/*
 * dcdetect.c
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 *
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <nodeos/an_event.h>

#include "peer.h"
#include "main.h"
#include "packets.h"
#include "tic.h"
#include "trace.h"

static an_event_t kickEvent;

/* XXX This has some odd network races when multiple clients are killed at
 * about the same time.  Different servers will blame the game lockup on
 * different clients and nothing will get done about it.
 *
 * Note that the setup protocol doesn't have a dead client shutdown mechanism,
 * and I cannot forsee a way to give it one due to the fundamental structure
 * of the network.
 */
static void
kickTrigger(an_event_t *event, void *arg)
{
	Client *client;
	TicCommand quitTic;
	Client *behind;
	TRACEONLY(char buf[128]);

	/* Go through all clients and increment their noRecv time */
	for (client = clientList; client; client = client->next)
	{
		if (!client->active ||
		    !client->local || 
		    client->quitTic != NOT_QUIT)
			continue;

		++client->noRecvTime;
		TRACE(TRACE_KICK, "clients%s->noRecvime = %d\n",
		      clientToString(client, buf, sizeof buf),
		      client->noRecvTime);
	}
	
	/* Find farthest behind, tic-wise, and see if it should be kicked */
	behind = NULL;
	for (client = clientList; client; client = client->next)
	{
		if (client->quitTic != NOT_QUIT)
			continue;

		if (!behind || (client->madeToTic < behind->madeToTic))
			behind = client;
	}

	if (behind && (behind->noRecvTime > options.kicktime))
	{
		TRACE(TRACE_KICK, "Client %s has gone over kick time\n",
		      clientToString(behind, buf, sizeof buf));
		
		/* Forge a quit tic */
		makeTic(behind->madeToTic, &quitTic, behind->identity);
		quitTic.flags |= TFLAG_QUIT;
		
		behind->quitTic = behind->madeToTic;
		ticListInsertOneTic(&behind->receivedPackets,
				    behind->madeToTic,
				    &quitTic);
		++behind->madeToTic;
		
		/* Flush with new information */
		sendConglomPackets();
	}
}

void
registerKickTrigger(void)
{
	an_timespec_t timeout;
	an_error_t err;
	
	an_event_create(&kickEvent);
	
	timeout.tv_nsec = 100*1000*1000;
	timeout.tv_sec = 0;

	err = an_event_schedule(&kickEvent, &timeout, AN_EVTYPE_PERIODIC,
				kickTrigger, NULL);
	assert(err == 0);
}
