/*
 * adoom/main.c
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 *
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <nodeos/an_event.h>

#include "main.h"
#include "options.h"
#include "peer.h"
#include "trace.h"
#include "error.h"
#include "stats.h"
#include "packets.h"
#include "send.h"
#include "dcdetect.h"

static void ee_init(void *arg);
an_domaininit_func_t InitEE = { ee_init };

static an_event_t timeoutEvent;

AdoomOptions options;

int ticfillsize;
OptionSpec optSpec[] = {
	{ "-myaddr", OT_STR, &options.myaddr, { strspec: { "127.0.0.1:32432" }}},
	{ "-protocol", OT_STR, &options.protocol, { strspec: { "udp" }}},
	{ "-timeout", OT_INT, &options.timeout, { intspec: { 20, 0, 1000*1000*1000 }}},
	{ "-statsprintgap", OT_INT, &options.statsprintgap, { intspec: { 1, 2, 120 }}},
	{ "-setuppoll", OT_INT, &options.setuppoll, { intspec: { 2, 2, 10 }}},
	{ "-tracemask", OT_STR, &options.tracemask, { strspec: { NULL }}},
	{ "-trace", OT_STR, &options.tracemask, { strspec: { NULL }}}, /* same as -tracemask */
	{ "-ticfillsize", OT_INT, &ticfillsize, { intspec: { 40, 0, 4096 }}},
	{ "-kicktime", OT_INT, &options.kicktime, { intspec: { 40, 10, 100 }}},
};

static void parsePeerArgs(an_initargs_t *initargs, int peerArgs, const char *myAddr);
static void registerTimeout(void);
static void gameOver(an_event_t *event, void *arg);

/*
 * TODO
 *
 * Deal with dead clients better
 *  Sometimes, when multiple clients die at about the same time, different nodes
 *  start blaming the problem on different clients, and the game locks
 *
 *  The setup protocol doesn't have a dead client mechanism and there is no way
 *  that I can forsee to give it one
 */

static void
ee_init(void *arg)
{
	an_initargs_t *initargs;
	int peerArgs;

	/* XXX configure time checks: */
	assert(sizeof(Uint64) == 8);
	assert(sizeof(Uint32) == 4);

	statsInit();

	/*
	 * Parse the initial arguments.
	 * Usage:
	 *   adoom [options] <adoom peers ...> -- <clients ...>
	 *
	 * Each <adoom peer> and <client> is an "addr[:port]" string
	 * where 'addr' is the IP address of the peer/client and
	 * the optional port is the UDP port for that peer/client.
	 */
	initargs = (an_initargs_t*)arg;
	peerArgs = parseArgs(initargs, optSpec, sizeof(optSpec)/sizeof(optSpec[0]));

	traceInit(options.tracemask);

	if (peerArgs == initargs->argc)
	{
		FATALERROR("No arguments provided.\n");
	}

	/*
	 * Start profiling
	 */
#	ifdef GPROF
	an_gprof_start(initargs->argv[0]);
#	endif

	/*
	 * Init peer miscellani
	 * This creates some needed mutex's
	 */
	initPeers();

	/*
	 * Create send buffers
	 * Note that this has to occur before parsePeerArgs because
	 * parsePeerArgs creates connections
	 */
	initSendBufs();

	parsePeerArgs(initargs, peerArgs, options.myaddr);

	/*
	 * Print stats at specified rate.
	 */
	registerStatsTrigger();

	/*
	 * Shutdown after a given timeout.
	 */
	registerTimeout();

	/*
	 * Run setup protocol
	 */
	registerSetupTrigger();

	/*
	 * Start client timeout system
	 */
	registerKickTrigger();
	
	// From here on out, eveything is handled by packet recv and triggers
}

static void 
parsePeerArgs(an_initargs_t *initargs, int peerArgs, const char *myAddr)
{
	int doPeers = 1;

	assert(initargs);
	assert(myAddr);

	/* <adoom peers ... > -- <clients ...> */

	TRACE(TRACE_SETUP, "Processing command line arguments.  Adoom peers first.\n");

	lockPeerList();
	lockClientList();
	
	myClientCount = 0;
	myPeerCount = 0;
	for (; peerArgs < initargs->argc; peerArgs++)
	{
		if (!strcmp("--", initargs->argv[peerArgs]))
		{
			TRACE(TRACE_SETUP, "command line arguments:  Switching to clients.\n");
			if (doPeers != 1)
				FATALERROR("Extra -- encountered in arguments");
			doPeers = 0;
		}
		else
		{
			if (doPeers)
			{
				TRACE(TRACE_SETUP,
				      "command line arguments: adding adoom peer `%s'.\n",
				      initargs->argv[peerArgs]);
				createADoomPeer(initargs->argv[peerArgs], myAddr);
				myPeerCount++;
			}
			else
			{
				TRACE(TRACE_SETUP,
				      "command line arguments: adding client `%s'.\n",
				      initargs->argv[peerArgs]);
				createClient(initargs->argv[peerArgs], myAddr);
				myClientCount++;
			}
		}
	}

	unlockClientList();
	unlockPeerList();
	
	TRACE(TRACE_SETUP,
	      "command line arguments: done [%d peers, %d clients].\n",
	      myPeerCount, myClientCount);
}

static void
registerTimeout(void)
{
	an_timespec_t timeout;
	an_error_t err = 0;
	
	an_event_create(&timeoutEvent);
	
	timeout.tv_nsec = 0;
	timeout.tv_sec = options.timeout;
	
	if (options.timeout > 0)
		err = an_event_schedule(&timeoutEvent, &timeout,
					AN_EVTYPE_ONESHOT_RELATIVE,
					gameOver, NULL);
	else
		/* run indefinetly */;

	assert(err == 0);
}


static void
gameOver(an_event_t *event, void *arg)
{
	// Clean up everything
	unregisterSetupTrigger();

	shutdownStats();
	
	printFinalStats();

	shutdownSendBufs();

	// My work here is done
	an_domain_destroy(an_domain_current());
}


void
shutdownGame(void)
{
	/* Just call gameOver with bogus params */
	gameOver(NULL,NULL);
}
