/*
 * packets.c
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 *
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <nodeos/an_event.h>

#include "packets.h"
#include "packetTypes.h"
#include "peer.h"
#include "main.h"
#include "trace.h"
#include "stats.h"
#include "error.h"
#include "send.h"
#include "ticlist.h"

static an_event_t setupEvent;
static Boolean setupRegistered;
static enum {
	STATE_SETUP,
	STATE_RUNNING
} state;

void
unregisterSetupTrigger(void)
{
	an_error_t err;

	if (!setupRegistered) return;
	
	err = an_event_deschedule(&setupEvent, NULL, NULL);
	assert(err == 0);
	err = an_event_destroy(&setupEvent, NULL, NULL);
	assert(err == 0);
	setupRegistered = false;
}

static void
setupPoll(an_event_t *event, void *arg)
{
	ADoomPeer *peer;
	TRACEONLY(char buf[128]);

	// Has everything already been setup?
	if (clientCount() >= myClientCount)
	{
		assert(clientCount() == myClientCount);
		TRACE(TRACE_SETUP, "Unregistering setup trigger; all clients connected\n");
		unregisterSetupTrigger();
		return;
	}
	if (peerCount() >= myPeerCount)
	{
		assert(peerCount() == myPeerCount);
		TRACE(TRACE_SETUP, "Unregistering setup trigger; all peers connected\n");
		unregisterSetupTrigger();
		return;
	}
	TRACE(TRACE_SETUP, "Polling other peers for setup\n");

	/*
	 * Iterate through peer adoom's and ping the ones that have not
	 * yet been determined alive
	 */
	lockPeerList();
	for (peer = adoomPeerList; peer; peer = peer->next)
	{
		if (!peer->active)
		{
			// Ping it
			TRACE(TRACE_SETUP|TRACE_SEND,
			      "Sending ping packet to peer %s\n",
			      adoomPeerToString(peer, buf, sizeof buf));
			pingToPeer(peer);
		}
	}
	unlockPeerList();
}

void
registerSetupTrigger(void)
{
	an_error_t err;
	an_timespec_t gap;
	
	an_event_create(&setupEvent);
	
	gap.tv_sec = options.setuppoll;
	gap.tv_nsec = 0;
	
	err = an_event_schedule(&setupEvent,
				&gap, AN_EVTYPE_PERIODIC,
				setupPoll, NULL);
	assert(err == 0);

	setupRegistered = true;
}

static void
advanceToRunning(void)
{
	Client *client;
	TRACEONLY(char buf[128]);

	if (state == STATE_RUNNING)
		return;

	assert(clientCount());

	initRunningBufs();

	state = STATE_RUNNING;
	
	for (client = clientList; client; client = client->next)
	{
		assert(client->active);
		if (!client->local) continue;
		
		TRACE(TRACE_SETUP|TRACE_SEND, "Sending recognize packet to client %s\n",
		      clientToString(client, buf, sizeof buf));
		recognizeToClient(client);
	}
}

static void
checkForAllClients(void)
{
	if (clientCount() == myClientCount)
	{
		// Hey, cool.  This game is ready to roll
		advanceToRunning();
	}
}

static void
sendAllClients(void)
{
	ADoomPeer *peer;
	Client *client;
	TRACEONLY(char buf1[128]);
	TRACEONLY(char buf2[128]);

	// Do I have all peer connections open?
	lockPeerList();
	if (peerCount() != myPeerCount)
	{
		unlockPeerList();
		return;
	}
	unlockPeerList();
	
	lockClientList();
	for (peer = adoomPeerList; peer; peer = peer->next)
	{
		assert(peer->active);
		if (peer->sentInitialClients) continue;
		
		for (client = clientList; client; client = client->next)
		{
			if (client->active && client->local)
			{
				TRACE(TRACE_SETUP|TRACE_SEND, "Sending client packet for client %s to peer %s seq %d\n",
				      clientToString(client, buf1, sizeof buf1),
				      adoomPeerToString(peer, buf2, sizeof buf2),
				      peer->clientSendSeq);
				clientToPeer(peer, client);
			}
		}

		peer->sentInitialClients = true;
	}	
	unlockClientList();
}

void
sendConglomPackets(void)
{
	ADoomPeer *peer;
	Client *client;
	Tic leastMadeToTicLocal,leastMadeToTicAll,mostMadeToTicLocal;
	TRACEONLY(char buf[128]);

	/* Find lowest madeToTic of local clients and all clients */
	leastMadeToTicLocal = (Uint64)(-1);
	leastMadeToTicAll = (Uint64)(-1);
	lockClientList();  /* This lock holds all the way through sending the
			    * client congloms */
	for (client = clientList; client; client = client->next)
	{
		if (client->quitTic != NOT_QUIT)
			continue;
		if (client->madeToTic < leastMadeToTicAll)
		{
			leastMadeToTicAll = client->madeToTic;
		}
		if (!client->local)
			continue;
		if (client->madeToTic < leastMadeToTicLocal)
		{
			leastMadeToTicLocal = client->madeToTic;
		}
	}

	/* If all local clients have been shutdown, there is nothing to send */
	if (leastMadeToTicLocal == (Uint64)(-1))
	{
		/* Send out this server's last breath.  Note that the last quit
		 * tic must be sent out at this point, because it will not have
		 * been sent out yet.  Find the highest madeToTic and flush
		 * everything up to there
		 */
		mostMadeToTicLocal = 0;
		for (client = clientList; client; client = client->next)
		{
			if (!client->local)
				continue;
			if (client->madeToTic > mostMadeToTicLocal)
				mostMadeToTicLocal = client->madeToTic;
		}
		/* Flush up to and including this tic to every peer */
		for (peer = adoomPeerList; peer; peer = peer->next)
		{
			while (peer->lastSentTic <= mostMadeToTicLocal)
			{
				TRACE(TRACE_CONGLOM|TRACE_SEND, "Flushing conglomerated tic %d to peer %s\n",
				      (int)(peer->lastSentTic),
				      adoomPeerToString(peer, buf, sizeof buf));
				conglomToPeer(peer, peer->lastSentTic);
				++peer->lastSentTic;
			}
		}
		/* Shutdown and hope nobody asks for a retransmit... */
		shutdownGame();
	}

	/* Do clients need congloms? */
	for (client = clientList; client; client = client->next)
	{
		if (client->quitTic != NOT_QUIT)
			continue;
		
		if (!client->local)
			continue;
		while (client->lastSentTic < leastMadeToTicAll)
		{
			/* XXX Will this cause obscure resends? */
			/* Have more tics for for client */
			TRACE(TRACE_CONGLOM|TRACE_SEND, "Sending conglomerated tic %d to client %s\n",
			      (int)(client->lastSentTic),
			      clientToString(client, buf, sizeof buf));
			conglomToClient(client, client->lastSentTic);
			++client->lastSentTic;
		}
	}
	unlockClientList();

	/* Do peers need congloms? */
	lockPeerList();
	for (peer = adoomPeerList; peer; peer = peer->next)
	{
		while (peer->lastSentTic < leastMadeToTicLocal)
		{
			/* All local clients have more than this peer */
			TRACE(TRACE_CONGLOM|TRACE_SEND, "Sending conglomerated tic %d to peer %s\n",
			      (int)(peer->lastSentTic),
			      adoomPeerToString(peer, buf, sizeof buf));
			conglomToPeer(peer, peer->lastSentTic);
			++peer->lastSentTic;
		}
	}
	unlockPeerList();
}

/*
 * This cannot be in the shared library because it needs to know about clients
 * and peers
 */
void
ticListInsertADoomRunningPacket(ADoomPeer *peer, RunningADoomConglomPacket *packet)
{
	Client *client;
	TicCommand *curcommand;
	int numSkip, numClients;

	curcommand = packet->ticCommand;
	numSkip = packet->startClient;
	numClients = packet->clientCount;

	for (client = clientList; client; client = client->next)
	{
		if (client->quitTic != NOT_QUIT && client->quitTic != packet->tic)
			continue;

		if (client->server == peer)
		{
			if (numSkip)
			{
				--numSkip;
				continue;
			} else {
				if (numClients-- == 0)
					break;
			}

			if (curcommand->flags & TFLAG_QUIT)
			{
				client->quitTic = packet->tic;
			}

			ticListInsertOneTic(&client->receivedPackets, packet->tic, curcommand);
			curcommand = (TicCommand*)(((void*)curcommand)+ticCommandSize());
		}
	}
}

void
handlePeerPingPacket(ADoomPingPacket *packet, ADoomPeer *peer)
{
	TRACEONLY(char buf[128]);
	
	TRACE(TRACE_SETUP, "Ping packet recv'd from peer %s\n",
	      adoomPeerToString(peer, buf, sizeof buf));

	peer->active = true;

	TRACE(TRACE_SETUP|TRACE_SEND, "Sending pong packet to peer %s\n",
	      adoomPeerToString(peer, buf, sizeof buf));
	pongToPeer(peer);

	sendAllClients();
}

void
handlePeerPongPacket(ADoomPongPacket *packet, ADoomPeer *peer)
{
	TRACEONLY(char buf[128]);
	// This peer has responded, it is now active
	TRACE(TRACE_SETUP, "Pong packet recv'd from peer %s\n",
	      adoomPeerToString(peer, buf, sizeof buf));

	peer->active = true;

	sendAllClients();
}

void
handlePeerClientPacket(ADoomClientPacket *packet, ADoomPeer *peer)
{
	Client *client;
	TRACEONLY(char buf1[128]);
	TRACEONLY(char buf2[128]);
	
	//assert(peer->active);
	if (!peer->active)
		TRACE(TRACE_SETUP, "*** Client packet from non-active peer\n");

	// Do I care?
	if (clientCount() == myClientCount)
		return;

	// Check the sequence number
	if (peer->nextClientSeq != packet->seqNum)
	{
		TRACE(TRACE_SETUP, "Out of order client packet recv'd from %s; expected %d got %d\n",
		      adoomPeerToString(peer, buf1, sizeof buf1),
		      peer->nextClientSeq, packet->seqNum);
		STAT_INC(outOfOrderClientPacket);

		// Sending a ping will cause this peer to resend all client
		// packets
		TRACE(TRACE_SETUP|TRACE_SEND, "Sending resend ping packet to peer\n");
		pingToPeer(peer);

		// Let this drop though, because it is still a valid client
	}

	for (client = clientList; client; client = client->next)
	{
		if (client->identity == packet->identity)
		{
			assert(client->active ? 1 : !client->local);
			TRACE(TRACE_SETUP, "Client packet recv'd from peer %s for client %s seq %d\n",
			      adoomPeerToString(peer, buf1, sizeof buf1),
			      clientToString(client, buf2, sizeof buf2),
			      packet->seqNum);
			client->local = false;
			client->server = peer;
			client->active = true;
			break;
		}
	}
	peer->nextClientSeq = packet->seqNum+1;

	checkForAllClients();
}

void
handlePeerConglomPacket(RunningADoomConglomPacket *packet, ADoomPeer *peer)
{
	Client *client;
	TRACEONLY(char buf[128]);

	assert(state == STATE_RUNNING);

	// Is it a retransmit request?
	if (packet->retransmitFrom != NO_RETRANSMITS)
	{
		TRACE(TRACE_RETRAN, "Retransmit request from peer %s from tic %d\n",
		      adoomPeerToString(peer, buf, sizeof buf),
		      (int)(packet->retransmitFrom));
		/* Bump lastSentTic, send routines will handle the rest */
		peer->lastSentTic = packet->retransmitFrom;
		return;
	}
	
	
	// Need to request retransmit?
	if (packet->tic > peer->madeToTic+1)
	{
		TRACE(TRACE_RETRAN, "Requesting retransmission from tic %d from peer %s\n",
		      (int)(peer->madeToTic+1),
		      adoomPeerToString(peer, buf, sizeof buf));
		retransmitToPeer(peer, peer->madeToTic+1);
		return;
	} else if (packet->tic < peer->madeToTic+1) {
		TRACE(TRACE_TIC, "Warning: duplicate tics recv'd from peer %s\n",
		      adoomPeerToString(peer, buf, sizeof buf));
		// Let this pass through because it is still valid
	}
	
	// Put tics in ticlists
	ticListInsertADoomRunningPacket(peer, packet);

	// Update peer's client's madeToTic's
	for (client = clientList; client; client = client->next)
	{
		if (client->server == peer)
			client->madeToTic = packet->tic;
	}
	
	peer->madeToTic = packet->tic;
	
	sendConglomPackets();
}

void
peerPacketRecv(an_pbuf_t *pbuf, void *arg, int offset)
{
	FdoomPacket *packet;
	ADoomPeer *peer;
	TRACEONLY(char buf[128]);
	
	assert(pbuf);
	assert(arg);
	assert(offset >= 0);
	
	packet = an_pbuf_data(pbuf)+offset;
	if (offset > 0)
		an_pbuf_setoffset(pbuf, an_pbuf_offset(pbuf)+offset, NULL);
	
	peer = (ADoomPeer*)arg;
	
	STAT_INC(peerPacketRecvCount);
	
	TRACE(TRACE_RECV, "Received packet (type=%d) from peer %s\n",
	      packet->packetType,
	      adoomPeerToString(peer, buf, sizeof buf));
	
	switch(packet->packetType)
	{
	case PT_ADOOM_PING:
		handlePeerPingPacket(&packet->pingPacket, peer);
		break;
	case PT_ADOOM_PONG:
		handlePeerPongPacket(&packet->pongPacket, peer);
		break;
	case PT_ADOOM_CLIENT:
		handlePeerClientPacket(&packet->clientPacket, peer);
		break;
	case PT_RUNNING_ADOOM:
		handlePeerConglomPacket(&packet->adoomconglomPacket, peer);
		break;
	default:
		FATALERROR("Unknown packet type received: %d\n",
			   packet->packetType);
	}

	// Recycle the pbuf
	an_inchan_addpbuf(&peer->recvChan, pbuf);
}

void
handleClientConnectPacket(ADoomConnectPacket *packet, Client *client)
{
	ADoomPeer *peer;
	TRACEONLY(char buf1[128]);
	TRACEONLY(char buf2[128]);
	
	if (client->active)
	{
		assert(client->local);
		// Ignore this packet, this is just a periodic resend
		STAT_INC(connectPacketFromActive);
		return;
	}

	TRACE(TRACE_SETUP, "Connect packet from new local client %s\n",
	      clientToString(client, buf1, sizeof buf1));

	lockClientList();

	// XXX Be more elegant about these checks
	assert(packet->peerCount == myClientCount-1);
	assert(packet->ticCommandSize == ticCommandSize());

	// Setup this client's state
	client->local=true;
	client->server=NULL;
	
	// Mark it as active
	client->active=true;

	unlockClientList();

	sendAllClients();

	// Notify peers that I have a live client
	lockPeerList();
	for (peer = adoomPeerList; peer; peer = peer->next)
	{
		if (peer->active && peer->sentInitialClients)
		{
			TRACE(TRACE_SETUP|TRACE_SEND, "Sending client packet for client %s to peer %s seq %d\n",
			      clientToString(client, buf1, sizeof buf1),
			      adoomPeerToString(peer, buf2, sizeof buf2),
			      peer->clientSendSeq);
			clientToPeer(peer, client);
		}
	}
	unlockPeerList();

	checkForAllClients();
}

void
handleClientRunningPacket(RunningPacket *packet, Client *client)
{
	char buf[128];

	// Is it a retransmit request?
	if (packet->retransmitFrom != NO_RETRANSMITS)
	{
		TRACE(TRACE_RETRAN, "Retransmit request from client %s from tic %d\n",
		      clientToString(client, buf, sizeof buf),
		      (int)(packet->retransmitFrom));
		/* Bump lastSentTic, send routines will handle the rest */
		client->lastSentTic = packet->retransmitFrom;
		if (!packet->ticCount)
			return;
	}
	
	// Need to request retransmit?
	if (packet->startTic > client->madeToTic+1)
	{
		TRACE(TRACE_RETRAN, "Requesting retransmission from tic %d from client %s\n",
		      (int)(client->madeToTic+1),
		      clientToString(client, buf, sizeof buf));
		retransmitToClient(client, client->madeToTic+1);
		return;
	} else if (packet->startTic < client->madeToTic+1) {
		TRACE(TRACE_TIC, "Warning: Duplicate tics recv'd from client %s\n",
		      clientToString(client, buf, sizeof buf));
		// Let this pass through because it is still valid
	}
	
	// Put tics in ticlist
	ticListInsertRunningPacket(&client->receivedPackets, packet);

	client->madeToTic = packet->startTic+packet->ticCount;

	sendConglomPackets();
}

void
clientPacketRecv(an_pbuf_t *pbuf, void *arg, int offset)
{
	FdoomPacket *packet;
	Client *client;
	TRACEONLY(char buf[128]);
	
	assert(pbuf);
	assert(arg);
	assert(offset >= 0);

	packet = an_pbuf_data(pbuf)+offset;
	if (offset > 0)
		an_pbuf_setoffset(pbuf, an_pbuf_offset(pbuf)+offset, NULL);

	client = (Client*)arg;
	assert(client);
	client->noRecvTime = 0;
	
	STAT_INC(clientPacketRecvCount);

	TRACE(TRACE_RECV, "Received packet (type=%d) from client %s\n",
	      packet->packetType,
	      clientToString(client, buf, sizeof buf));

	switch(packet->packetType)
	{
	case PT_ADOOM_CONNECT:
		handleClientConnectPacket(&packet->connectPacket, client);
		break;
	case PT_RUNNING:
		handleClientRunningPacket(&packet->runningPacket, client);
		break;
	default:
		FATALERROR("Unknown packet type received: %d\n",
			   packet->packetType);
	}
	
	// Recycle the pbuf
	an_inchan_addpbuf(&client->recvChan, pbuf);
}
