/*
 * adoom/peer.c
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 *
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <stdlib.h> // malloc
#include <stdio.h>  // snprintf

#include <nodeos/an_mutex.h>

#include "peer.h"
#include "network.h"
#include "identity.h"
#include "packets.h"
#include "trace.h"
#include "stats.h"

ADoomPeer *adoomPeerList;
Client *clientList;

int myClientCount, myPeerCount;  // What they're supposed to be

static an_mutex_t peerListLock;
static an_mutex_t clientListLock;

void
initPeers(void)
{
	an_mutex_t *rc;

	rc = an_mutex_create(&peerListLock);
	assert(rc == &peerListLock);

	rc = an_mutex_create(&clientListLock);
	assert(rc == &clientListLock);
}

void
createADoomPeer(const char *peerAddr, const char *myAddr)
{
	ADoomPeer* adp;
	int i;
	
	assert(peerAddr);
	assert(myAddr);
	
	adp = malloc(sizeof(ADoomPeer));
	adp->name=peerAddr;
	adp->identity = computeIdentity(/*XXX*/"udp", adp->name);

	adp->nextClientSeq = 0;
	adp->clientSendSeq = 0;
	
	adp->lastSentTic = 0;
	adp->madeToTic = 0;

	for (i = 0; i < BUFS_PER_PEER; i++)
	{
		adp->recvBufPtrs[i] = an_pbuf_create(&(adp->recvBufs[i]), AN_PBUF_DEFAULT, 0, 0);
		assert(adp->recvBufPtrs[i] == &(adp->recvBufs[i]));
	}
	adp->next=adoomPeerList;
	adoomPeerList = adp;
	
	// Open the connection
	createConnections(/*XXX*/"udp",
			  myAddr,
			  peerAddr,
			  &adp->recvChan,
			  peerPacketRecv,
			  (void*)adp,
			  &adp->recvDemux,
			  adp->recvBufPtrs,
			  BUFS_PER_PEER,
			  &adp->sendChan);
}

void
createClient(const char *clientAddr, const char *myAddr)
{
	Client *client;
	Client *listpos;
	int i;
	
	client = malloc(sizeof(Client));
	client->name = clientAddr;
	client->identity = computeIdentity(/*XXX*/"udp", client->name);
	client->active = false;
	client->local = false;
	client->server = NULL;

	ticListInit(&client->receivedPackets);
	client->madeToTic = 0;
	client->lastSentTic = 0;

	client->noRecvTime = 0;
	client->quitTic = NOT_QUIT;

	for (i = 0; i < BUFS_PER_PEER; i++)
	{
		client->recvBufPtrs[i] = an_pbuf_create(&(client->recvBufs[i]),
						       AN_PBUF_DEFAULT, 0, 0);
		assert(client->recvBufPtrs[i] == &(client->recvBufs[i]));
	}

	// Insert it into the proper place in the list
	if (!clientList)
	{
		client->next = NULL;
		clientList = client;
	} else if (client->identity < clientList->identity) {
		// Goes on front of list
		client->next = clientList;
		clientList = client;
	} else {
		for (listpos = clientList; listpos->next; listpos = listpos->next)
		{
			if (client->identity < listpos->next->identity)
			{
				client->next = listpos->next;
				listpos->next = client;
				break;
			}
		}
		if (!listpos->next)
		{
			client->next = NULL;
			listpos->next = client;
		}
	}
	
	createConnections(/*XXX*/"udp",
			  myAddr,
			  clientAddr,
			  &client->recvChan,
			  clientPacketRecv,
			  (void*)client,
			  &client->recvDemux,
			  client->recvBufPtrs,
			  BUFS_PER_PEER,
			  &client->sendChan);
}

int
peerCount(void)
{
	ADoomPeer *elem;
	int count;

	count = 0;
	for (elem = adoomPeerList; elem; elem = elem->next)
		if (elem->active) count++;
	
	return count;
}

int
clientCount(void)
{
	Client *elem;
	int count;
	
	count = 0;
	for (elem = clientList; elem; elem = elem->next)
		if (elem->active) count++;

	return count;
}

int
localClientCount(void)
{
	Client *elem;
	int count;
	
	count = 0;
	for (elem = clientList; elem; elem = elem->next)
		if (elem->active && elem->local) count++;

	return count;
}

void
lockPeerList(void)
{
	an_mutex_lock(&peerListLock);
	TRACE(TRACE_GSLOCK, "lockPeerList\n");
	STAT_INC(peerLockAcquire);
}

void
unlockPeerList(void)
{
	TRACE(TRACE_GSLOCK, "unlockPeerList\n");
	STAT_INC(peerLockRelease);
	an_mutex_unlock(&peerListLock);
}

void
lockClientList(void)
{
	an_mutex_lock(&clientListLock);
	TRACE(TRACE_GSLOCK, "lockClientList\n");
	STAT_INC(clientLockAcquire);
}

void
unlockClientList(void)
{
	TRACE(TRACE_GSLOCK, "unlockClientList\n");
	STAT_INC(clientLockRelease);
	an_mutex_unlock(&clientListLock);
}

const char *
adoomPeerToString(ADoomPeer *peer, char *buf, size_t bufsize)
{
	assert(peer);
	assert(buf);

	snprintf(buf, bufsize,
		 "[%s ...]", peer->name);

	return buf;
}

const char *
clientToString(Client *c, char *buf, size_t bufsize)
{
	assert(c);
	assert(buf);

	/*snprintf(buf, bufsize,
		 "[%s made=%lld, sent=%lld]",
		 c->name,
		 c->madeToTic,
		 c->lastSentTic);*/
	snprintf(buf, bufsize,
		 "[%s]",
		 c->name);

	return buf;
}
