/*
 * peer.h
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 *
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#ifndef PEER_HEADER
#define PEER_HEADER

#include <nodeos/an_chan.h>

#include "types.h"
#include "ticlist.h"

#define BUFS_PER_PEER 64
#define BUFS_PER_CLIENT 32

/**
 * This node's representation of the state of its ADoom peers.  The
 * peers are other nodes (like this one) that conglomerate traffic.
 *
 * XXX note that although each peer is responsible for a set of fdoom
 * clients, the current protocol only forwards consistent state from
 * one peer to another, so we don't need per-remote-client
 * lastSentTic/madeToTic information.  (Possibly a good design,
 * possibly we should track each remote client independently...)
 */
typedef struct ADoomPeer {
	const char *name;
	Uint64 identity; /* XXX debug only */

	/** Number of clients this ADoomPeer manages. */
	int clientCount;

	Boolean active;

	int nextClientSeq;  // Next expected sequence number of a recv'd client
	int clientSendSeq;  // Next seqNum of a client send
	Boolean sentInitialClients;
	
	Tic lastSentTic;	// Highest tic sent to peer
	Tic madeToTic;		// Highest tic recv'd from peer

	an_chan_t sendChan;
	an_chan_t recvChan;
	
	an_demuxkey_t recvDemux;

	an_pbuf_t recvBufs[BUFS_PER_PEER];
	an_pbuf_t* recvBufPtrs[BUFS_PER_PEER];

	struct ADoomPeer *next;
} ADoomPeer;

/**
 * This node's representation of its immediate clients.  These are
 * nodes that are "playing" the game (fdoom nodes).
 */
typedef struct Client {
	const char *name;
	Uint64 identity;

	Boolean active;  // If connection has been recv'ed
	Boolean local;   /* If connected to this adoom node (only valid if
			  * active is true) */

	int noRecvTime;  /* Number of 100 millis that nothing recv'd */
	Tic quitTic;

	TicList receivedPackets;
	Tic madeToTic;		// Highest tic recv'd from client
	Tic lastSentTic;	// Highest tic sent to client

	an_chan_t sendChan;
	an_chan_t recvChan;
	
	an_demuxkey_t recvDemux;

	an_pbuf_t recvBufs[BUFS_PER_PEER];
	an_pbuf_t* recvBufPtrs[BUFS_PER_PEER];

	ADoomPeer *server;

	struct Client *next;
} Client;

#define NOT_QUIT	(Uint64)(-1)

extern ADoomPeer *adoomPeerList;
extern Client *clientList;
extern int myClientCount, myPeerCount;

void initPeers(void);
void createADoomPeer(const char *peerAddr, const char *myAddr);
void createClient(const char *clientAddr, const char *myAddr);
int peerCount(void);
int clientCount(void);
int localClientCount(void);
void lockPeerList(void);
void unlockPeerList(void);
void lockClientList(void);
void unlockClientList(void);
const char *adoomPeerToString(ADoomPeer *peer, char *buf, size_t bufsize);
const char *clientToString(Client *c, char *buf, size_t bufsize);

#endif /* PEER_HEADER */
