/*
 * send.c
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 *
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <stdlib.h>  // malloc, free
#include <nodeos/an_chan.h>
#include <nodeos/an_mutex.h>

#include "send.h"
#include "packetTypes.h"
#include "main.h"
#include "conglom.h"
#include "stats.h"
#include "trace.h"

/*
 * Encapsulate a generic send buf pool mechanism.  Currently this is only a way
 * to wrap the structure needed by a send buf along with a mutex, but it
 * could be made more complex.  On small experiments, the mutex never blocks
 * anyways.
 */
typedef struct PooledSendBuf
{
	Boolean valid;
	an_pbuf_t pbuf;
	void *data;
	an_mutex_t mutex;
} PooledSendBuf;

/*
 * A single sendbuf sucked out of a theoretically pooled sendbuf
 */
typedef struct SingleSendBuf
{
	an_pbuf_t *pbuf;
	void *data;
	PooledSendBuf *srcpool;
} SingleSendBuf;

static PooledSendBuf pingPool = { false };
static PooledSendBuf pongPool = { false };
static PooledSendBuf clientPool = { false };
static PooledSendBuf recognizePool = { false };
static PooledSendBuf retransmitPool = { false };
static PooledSendBuf conglomPool = { false };
static PooledSendBuf adoomconglomPool = { false };
static PooledSendBuf retransmitADoomPool = { false };

static void
createPool(PooledSendBuf *pool, int size, void *initdata)
{
	an_pbuf_t *pbuf_rc;
	an_mutex_t *mutex_rc;
	
	pbuf_rc = an_pbuf_create(&(pool->pbuf), AN_PBUF_DEFAULT, 0, size);
	assert(pbuf_rc == &(pool->pbuf));

	mutex_rc = an_mutex_create(&(pool->mutex));
	assert(mutex_rc == &(pool->mutex));

	pool->data = an_pbuf_data(pbuf_rc);
	assert(pool->data);
	memcpy(pool->data, initdata, size);

	pool->valid = true;
}

static void
destroyPool(PooledSendBuf *pool)
{
	if (!pool->valid) return;
	
	an_mutex_lock(&(pool->mutex));  // Just in case
	pool->valid = false;
	an_pbuf_destroy(&(pool->pbuf));
	an_mutex_unlock(&(pool->mutex));
	an_mutex_destroy(&(pool->mutex));
}

void
initSendBufs(void)
{
	ADoomPingPacket pingpkt;
	ADoomPongPacket pongpkt;
	ADoomClientPacket clientpkt;
	ADoomRecognizePacket recognizepkt;
	RunningPacket retransmitpkt;
	RunningADoomConglomPacket retransmitadoompkt;

	// Fill default packets
	pingpkt.packetType = PT_ADOOM_PING;
	pongpkt.packetType = PT_ADOOM_PONG;
	clientpkt.packetType = PT_ADOOM_CLIENT;
	clientpkt.seqNum = 0;
	clientpkt.identity = 0;
	recognizepkt.packetType = PT_ADOOM_RECOGNIZE;
	retransmitpkt.packetType = PT_RUNNING;
	retransmitpkt.retransmitFrom = NO_RETRANSMITS;
	retransmitpkt.startTic = 0;
	retransmitpkt.ticCount = 0;
	retransmitadoompkt.packetType = PT_RUNNING_ADOOM;
	retransmitadoompkt.retransmitFrom = NO_RETRANSMITS;
	retransmitadoompkt.tic = 0;
	
	// Create the pools
	createPool(&pingPool, sizeof(ADoomPingPacket), &pingpkt);
	createPool(&pongPool, sizeof(ADoomPongPacket), &pongpkt);
	createPool(&clientPool, sizeof(ADoomClientPacket), &clientpkt);
	createPool(&recognizePool, sizeof(ADoomRecognizePacket), &recognizepkt);
	createPool(&retransmitPool, sizeof(RunningPacket), &retransmitpkt);
	createPool(&retransmitADoomPool, sizeof(RunningADoomConglomPacket), &retransmitadoompkt);
}

/*
 * Initializes buffers and pools that can only be created once the game
 * topology is known.  These need to know such things as the number of local
 * clients.
 */
void
initRunningBufs(void)
{
	RunningConglomPacket *conglompkt;
	int conglomsize;
	RunningADoomConglomPacket *adoomconglompkt;
	int adoomconglomsize;

	// Create packets with right sizes
	conglomsize = getConglomBufSize(sizeof(RunningConglomPacket),
					clientCount()-1);
	conglompkt = (RunningConglomPacket*)malloc(conglomsize);
	assert(conglompkt != NULL);

	adoomconglomsize = getConglomBufSize(sizeof(RunningADoomConglomPacket),
					     localClientCount());
	adoomconglompkt = (RunningADoomConglomPacket*)malloc(adoomconglomsize);
	assert(adoomconglompkt != NULL);

	TRACE(TRACE_CONGLOM, "MAX_CONGLOM_SIZE = %d\n", MAX_CONGLOM_SIZE);
	TRACE(TRACE_CONGLOM, "conglomsize = %d\n", conglomsize);
	TRACE(TRACE_CONGLOM, "adoomconglomsize = %d\n", adoomconglomsize);
	
	// Fill default packets
	conglompkt->packetType = PT_RUNNING_CONGLOM;
	conglompkt->retransmitFrom = NO_RETRANSMITS;
	conglompkt->tic = 0;
	conglompkt->startClient = 0;
	conglompkt->clientCount = 0;
	adoomconglompkt->packetType = PT_RUNNING_ADOOM;
	adoomconglompkt->retransmitFrom = NO_RETRANSMITS;
	adoomconglompkt->tic = 0;
	adoomconglompkt->startClient = 0;
	adoomconglompkt->clientCount = 0;

	// Create the pools
	createPool(&conglomPool, conglomsize, conglompkt);
	createPool(&adoomconglomPool, adoomconglomsize, adoomconglompkt);
	
	free(conglompkt);
	free(adoomconglompkt);
}

void
shutdownSendBufs(void)
{
	destroyPool(&pingPool);
	destroyPool(&pongPool);
	destroyPool(&clientPool);
	destroyPool(&recognizePool);
	destroyPool(&retransmitPool);	
	destroyPool(&conglomPool);
	destroyPool(&adoomconglomPool);
	destroyPool(&retransmitADoomPool);
}

static void
getSendbuf(PooledSendBuf *pool, SingleSendBuf *out)
{
	assert(pool != NULL);
	assert(out != NULL);
	assert(pool->valid);

	if (an_mutex_trylock(&(pool->mutex)) != 0)
	{
		STAT_INC(blockOnPoolMutex);
		an_mutex_lock(&(pool->mutex));
	}
	STAT_INC(poolLockAcquire);
	out->pbuf = &pool->pbuf;
	out->data = pool->data;
	out->srcpool = pool;

	if (an_pbuf_isbusy(out->pbuf))
	{
		STAT_INC(poolWaitOnPbuf);
		while (an_pbuf_isbusy(out->pbuf));
	}
}

static void
releaseSendbuf(SingleSendBuf *single)
{
	an_mutex_unlock(&(single->srcpool->mutex));
	STAT_INC(poolLockRelease);
}

void
pingToPeer(ADoomPeer *peer)
{
	an_error_t err;
	SingleSendBuf sendbuf;
	
	getSendbuf(&pingPool, &sendbuf);
	
	err = an_outchan_send(&peer->sendChan, sendbuf.pbuf);
	assert(err == 0);
	STAT_INC(peerPacketSentCount);
	
	releaseSendbuf(&sendbuf);
}

void
pongToPeer(ADoomPeer *peer)
{
	an_error_t err;
	SingleSendBuf sendbuf;
	
	getSendbuf(&pongPool, &sendbuf);
	
	err = an_outchan_send(&peer->sendChan, sendbuf.pbuf);
	assert(err == 0);
	STAT_INC(peerPacketSentCount);
	
	releaseSendbuf(&sendbuf);
}

void
clientToPeer(ADoomPeer *peer, Client *client)
{
	an_error_t err;
	SingleSendBuf sendbuf;
	
	getSendbuf(&clientPool, &sendbuf);
	
	((ADoomClientPacket*)sendbuf.data)->seqNum = (peer->clientSendSeq++);
	((ADoomClientPacket*)sendbuf.data)->identity = client->identity;

	err = an_outchan_send(&peer->sendChan, sendbuf.pbuf);
	assert(err == 0);
	STAT_INC(peerPacketSentCount);
	
	releaseSendbuf(&sendbuf);
}

void
retransmitToPeer(ADoomPeer *peer, Tic from)
{
	an_error_t err;
	SingleSendBuf sendbuf;

	getSendbuf(&retransmitADoomPool, &sendbuf);

	((RunningADoomConglomPacket*)sendbuf.data)->retransmitFrom = from;
	
	err = an_outchan_send(&peer->sendChan, sendbuf.pbuf);
	assert(err == 0);
	STAT_INC(peerPacketSentCount);
	
	releaseSendbuf(&sendbuf);
}

/*
 * The returns and args to this warrant some explanation:
 * nClient is a pointer to the index of the first client to claim the packet
 * is for.  When this function is done, the value of this is the index of the
 * client that should start the next conglom packet (or meaningless if it's
 * done).  client is a pointer to the first actual client structure to include
 * in the conglom packet.  The return is a pointer to the first client to
 * include in the next packet, or NULL if no more clients need to be conglom'd.
 *
 * XXX The last conglomerate in a split conglom sends the same size as the rest
 * of them, even if it doesn't fill the packet.  Resizing the pbuf just to
 * better fit the last packet would probably introduce more overhead than it's
 * worth
 */
static Client *
singleConglomToPeer(ADoomPeer *peer, Tic tic, int *nClient, Client *client)
{
	an_error_t err;
	SingleSendBuf sendbuf;
	TicCommand *ticcommand;
	int size;
	TRACEONLY(char buf[128]);
	
	getSendbuf(&adoomconglomPool, &sendbuf);
	
	((RunningADoomConglomPacket*)sendbuf.data)->tic = tic;
	((RunningADoomConglomPacket*)sendbuf.data)->startClient = *nClient;
	((RunningADoomConglomPacket*)sendbuf.data)->clientCount = 0;

	size = sizeof(RunningADoomConglomPacket);
	
	ticcommand = ((RunningADoomConglomPacket*)sendbuf.data)->ticCommand;
	for (; client; client = client->next)
	{
		if (client->local)
		{
			if (client->quitTic < tic)
			{
				++(*nClient);
				continue;
			}
			
			ticListCopyTicOut(&client->receivedPackets, tic, ticcommand);
			++((RunningADoomConglomPacket*)sendbuf.data)->clientCount;
			/* Advance by ticCommandSize() bytes */
			ticcommand = (TicCommand*)(((void*)ticcommand)+ticCommandSize());
			++(*nClient);
			/* Have I reached my maximum conglom size? */
			size += ticCommandSize();
			if (size+ticCommandSize() > MAX_CONGLOM_SIZE)
			{
				/* Make sure return values aren't bogus */
				client = client->next;
				while (client && !client->local)
				{
					client = client->next;
				}
				break;
			}
		}
	}
	
	TRACE(TRACE_SEND, "Sending single conglom tic %d to peer %s\n",
	      (int)(tic),
	      adoomPeerToString(peer, buf, sizeof buf));
	err = an_outchan_send(&peer->sendChan, sendbuf.pbuf);
	assert(err == 0);
	STAT_INC(peerPacketSentCount);
	
	releaseSendbuf(&sendbuf);

	return client;
}

void
conglomToPeer(ADoomPeer *peer, Tic tic)
{
	int nClient;
	Client *client;
	
	client = clientList;
	nClient = 0;
	
	while (client)
	{
		client = singleConglomToPeer(peer, tic, &nClient, client);
	}
}

void
recognizeToClient(Client *client)
{
	an_error_t err;
	SingleSendBuf sendbuf;
	
	getSendbuf(&recognizePool, &sendbuf);
	
	err = an_outchan_send(&client->sendChan, sendbuf.pbuf);
	assert(err == 0);
	STAT_INC(clientPacketSentCount);
	
	releaseSendbuf(&sendbuf);
}

void
retransmitToClient(Client *client, Tic from)
{
	an_error_t err;
	SingleSendBuf sendbuf;

	getSendbuf(&retransmitPool, &sendbuf);

	((RunningPacket*)sendbuf.data)->retransmitFrom = from;
	
	err = an_outchan_send(&client->sendChan, sendbuf.pbuf);
	assert(err == 0);
	STAT_INC(clientPacketSentCount);
	
	releaseSendbuf(&sendbuf);
}

/*
 * The return values and args for this are the same peculiar ones used for
 * singleConglomToPeer
 *
 * XXX And it has the same problems as singleConglomToPeer
 */
static Client *
singleConglomToClient(Client *dest, Tic tic, int *nClient, Client *client)
{
	an_error_t err;
	SingleSendBuf sendbuf;
	TicCommand *ticcommand;
	int size;
	TRACEONLY(char buf[128]);
	
	getSendbuf(&conglomPool, &sendbuf);
	
	((RunningConglomPacket*)sendbuf.data)->tic = tic;
	((RunningConglomPacket*)sendbuf.data)->startClient = *nClient;
	((RunningConglomPacket*)sendbuf.data)->clientCount = 0;
	
	size = sizeof(RunningConglomPacket);
	
	ticcommand = ((RunningConglomPacket*)sendbuf.data)->ticCommand;
	for (; client; client = client->next)
	{
		if (client != dest)
		{
			if (!(client->quitTic < tic))
			{
				ticListCopyTicOut(&client->receivedPackets, tic, ticcommand);
				/* Advance by ticCommandSize() bytes */
				ticcommand = (TicCommand*)(((void*)ticcommand)+ticCommandSize());
				size += ticCommandSize();
			}
			
			/* Increment counters */
			++((RunningConglomPacket*)sendbuf.data)->clientCount;
			++(*nClient);
			/* Have I reached my maximum conglom size? */
			if (size+ticCommandSize() > MAX_CONGLOM_SIZE)
			{
				/* Make sure return values aren't bogus */
				client = client->next;
				while (client && client == dest)
				{
					client = client->next;
				}
				break;
			}
		}
	}

	TRACE(TRACE_SEND, "Sending single conglom tic %d to client %s\n",
	      (int)(tic),
	      clientToString(dest, buf, sizeof buf));
	err = an_outchan_send(&dest->sendChan, sendbuf.pbuf);
	assert(err == 0);
	STAT_INC(clientPacketSentCount);
	
	releaseSendbuf(&sendbuf);
	
	return client;
}

void
conglomToClient(Client *client, Tic tic)
{
	int nClient;
	Client *client2;
	
	client2 = clientList;
	nClient = 0;
	
	while (client2)
	{
		client2 = singleConglomToClient(client, tic, &nClient, client2);
	}
}
