/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * Setup a pingreply by installing a Click graph using the click library.
 */
 
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <sys/types.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/in_systm.h>
#include <netinet/if_ether.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>

#include <nodeos/an_impl.h>
#include <nodeos/an_domain.h>
#include <nodeos/an_thread.h>
#include <nodeos/an_chan.h>
#include <nodeos/an_pbuf.h>
#include <nodeos/an_event.h>
#include <nodeos/an_time.h>
#include <nodeos/an_info.h>
#include <nodeos/impl/an_gprof.h>
#include <nodeos/lib/utils.h>

#include <nodeos/lib/click.hh>
#define RUNFOR	6000	/* run for this many seconds */
#define MAXBUFS	64

#define ETHIF	0

static void ee_init(void *arg);
an_domaininit_func_t InitEE = { ee_init };

static void ee_die(an_event_t *event, void *arg);
static void setup_channels(void);

static an_chan_t *ethinchan, ethinchan_data, *ethoutchan, ethoutchan_data;
static an_pbuf_t cutbufs_data[MAXBUFS];
anlib_clickgraphdummy_t graphmem;
anlib_clickgraph_t *graph;
an_event_t eventmem;
an_event_t *event_end;

an_ethaddrspec_t myaddr = {
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x00}, {0}, ETHERTYPE_IP
};

/*
 * A Click router graph. Mostly to ensure that
 * stuff is working. What is does? Take all incoming packets and
 * classify them into ARP requests, IP packets, or anything else.
 * Reply to ARP requests, so we caan hand out our own address.
 * Send IP packets into the pinger. Drop everything else.
 */
char click_config[] =
	"in :: FromInChan(in0);"
	"in -> Print(60) -> Discard; "
	;

int
ee_main(int argc, const char **argv)
{
	assert(argc >= 1);

	printf("%p/%p: %s started\n",
	       an_domain_current(), an_thread_current(), argv[0]);

	/*
	 * Setup channels
	 */
	
	setup_channels();

	printf("(%02x:%02x:%02x:%02x:%02x:%02x) ready\n",
	       myaddr.daddr[0], myaddr.daddr[1], myaddr.daddr[2],
	       myaddr.daddr[3], myaddr.daddr[4], myaddr.daddr[5]);

	an_gprof_start(argv[0]);
	return 0;
}

static void
setup_channels(void)
{
	an_error_t err;
	struct {
		an_chanaddr_t addr;
		an_chanproto_t proto;
		an_cpuspec_t cpu;
		an_pbuf_t *bufs[MAXBUFS];
	} attr;
	an_domain_t *domain;
	an_infoif_t info[8];
	an_netspec_t net;

	
	an_timespec_t etime;
	an_chanrecvfunc_t func;
	void *funcarg;
	
	int i, nif = 8;

	/* 
	 * Initialize the graph with the current config
	 */
	graph = anlib_clickgraph_create (&graphmem, click_config);
	assert (graph != 0);

	domain = an_domain_current();
	memset(&attr, 0, sizeof attr);

	/*
	 * Find my address.
	 * XXX assumes first interface is ethernet.
	 */
	err = an_info_getifs(info, &nif);
	assert(err == 0);
	assert(nif > ETHIF);
	assert(info[ETHIF].type == ANIF_ETH);
	memcpy(myaddr.daddr, info[ETHIF].addr.eth.daddr, sizeof myaddr.daddr);

	/*
	 * Setup a timer event for RUNFOR seconds
	 */
	event_end = an_event_create (&eventmem);
	assert (event_end);
	
	etime.tv_sec = RUNFOR;
	etime.tv_nsec = 0;
	err = an_event_schedule(event_end, &etime, AN_EVTYPE_ONESHOT_RELATIVE,
							ee_die, 0 );
	assert(err == 0);
	
	/*
	 * Setup in channel for ICMP/IP packets.
	 * No demux; all packets are handed off to the Click graph.
	 * No address; means a "raw" interface at the moment. Entire packet
	 * including the ethernet header is handed off.
	 */
	strncpy(attr.proto.data, info[ETHIF].pname, sizeof info[ETHIF].pname);
	attr.proto.len = strlen(attr.proto.data);
	memset(&attr.addr, 0, sizeof attr.addr);
	attr.cpu.ttype = AN_TH_IMPLICIT;
	attr.cpu.threads.im.maxthreads = 1;
	for (i = 0; i < MAXBUFS; i++) {
		attr.bufs[i] = an_pbuf_create(&cutbufs_data[i], AN_PBUF_ETH,
					      0, 0);
		assert(attr.bufs[i] != 0);
	}
	/*
   	 * before we create the channel we should lookup the click input we want to
     * connect to this channel 
     */
	err = anlib_clickgraph_lookupInput (graph, "in0", &func, &funcarg);
	assert(err == 0);		
	
	net.maxthreads = 1;
	net.buffers.npbufs = MAXBUFS;
	net.buffers.pbufs = attr.bufs;
	ethinchan = an_inchan_create(&ethinchan_data, domain, 0,
				     &attr.proto, &attr.addr,
				     &net, func, funcarg);
	assert(ethinchan != 0);

	printf("%p: inchannel for %02x:%02x:%02x:%02x:%02x:%02x, type=%x\n",
	       ethinchan,
	       myaddr.daddr[0], myaddr.daddr[1], myaddr.daddr[2],
	       myaddr.daddr[3], myaddr.daddr[4], myaddr.daddr[5],
	       ntohs(myaddr.type));

	/* are we done? */
}

static void
ee_die(an_event_t *event, void *arg)
{
	an_gprof_stop();

	printf("All done!\n");

	if (graph)
		anlib_clickgraph_destroy(graph);
	an_domain_destroy(an_domain_current());
	an_event_destroy (event_end, 0, 0);
}

static void
ee_init(void *arg)
{
	int status;
	int argc = 1;
	const char *argv[] = { "ClickPingreplyEE" }; 

	status = ee_main(argc, argv);
}

