/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * EE to test events, looks suspiciously like chantest...
 */

#include <stdio.h>
#include <string.h>
#include <assert.h>

#include <nodeos/an_impl.h>
#include <nodeos/an_domain.h>
#include <nodeos/an_thread.h>
#include <nodeos/an_event.h>
#include <nodeos/an_mempool.h>

#define RUNFOR	10

static void ee_init(void *arg);
an_domaininit_func_t InitEE = { ee_init };

static void misctests(void);
static void setup_events(void);
static void ee_event1(an_event_t *event, void *arg);
static void ee_event100(an_event_t *event, void *arg);
static unsigned long eventres, expected;

int
ee_main(int argc, const char **argv)
{
	printf("%p/%p: Channel test EE started\n",
	       an_domain_current(), an_thread_current());

	/*
	 * Setup events
	 */
	setup_events();

	/*
	 * No longer need this thread
	 */
	return 0;
}

static an_event_t ev1;
static an_event_t ev100;

static void
setup_events(void)
{
	an_error_t err;
	an_timespec_t gap;
	void *rc;

	eventres = an_event_granularity();
	assert(eventres < 1000*1000*1000);
	printf("Event granularity: %ld nsec (%ld per sec max)\n",
	       eventres, (1000*1000*1000)/eventres);

	expected = (1000*1000*1000 / eventres) * RUNFOR;

	rc = an_event_create(&ev1);
	assert(rc == &ev1);
	rc = an_event_create(&ev100);
	assert(rc == &ev100);
	
	/* max resolution */
	gap.tv_sec = 0;
	gap.tv_nsec = eventres;
	err = an_event_schedule(&ev1, &gap,
				AN_EVTYPE_PERIODIC,
				ee_event1, (void *)1);
	assert(err == 0);

	/* every second */
	gap.tv_sec = 1;
	gap.tv_nsec = 0;
	err = an_event_schedule(&ev100, &gap,
				AN_EVTYPE_PERIODIC,
				ee_event100, (void *)100);
	misctests();

	assert(err == 0);
}

static int immed_ev_ct = 0;

static an_event_t immed_ev;

static void immed_func(an_event_t *ev, void *arg)
{
	assert(ev == &immed_ev);
	assert(arg == NULL);

	immed_ev_ct++;
	if (immed_ev_ct < 10)
		an_event_fire(ev, immed_func, arg);
	else
		an_event_destroy(ev, NULL, NULL);
}


static void nevercalled(an_event_t *ev, void *arg)
{
	fprintf(stderr, "nevercalled() was called: ev=%p\n", ev);
	abort();
}

static void checkmisctests(void)
{
	printf("immed_ev_ct should be 10, is %d\n", immed_ev_ct);
	assert(immed_ev_ct == 10);
}

static void misctests(void)
{
	an_error_t errrc;
	an_event_t fooev;
	an_event_t *evrc;
	an_timespec_t evtime;
	an_eventtype_t etype;
	an_eventstate_t estate;

	/*
	 * Make sure create/destroy work and that destroy
	 * really destroys the event.
	 */
	evrc = an_event_create(&fooev);
	assert(evrc == &fooev);
	errrc = an_event_destroy(&fooev, NULL, NULL);
	assert(errrc == 0);
	evrc = an_event_create(&fooev);
	assert(evrc == &fooev);

	/*
	 * Try a double-schedule (should fail).
	 */
	evtime.tv_sec = 2 * 1000 * 1000 * 1000; /* Tue May 17 21:33:20 MDT 2033 */
	evtime.tv_nsec = 0;
	errrc = an_event_schedule(&fooev, &evtime,
				  AN_EVTYPE_ONESHOT_ABSOLUTE,
				  nevercalled, NULL);
	assert(errrc == 0);
	evtime.tv_sec -= 100;  /* a bit earlier, how about */
	errrc = an_event_schedule(&fooev, &evtime,
				  AN_EVTYPE_ONESHOT_ABSOLUTE,
				  nevercalled, NULL);
	assert(errrc != 0); /* XXX which error again? */

	/*
	 * Try getting the status of this far-off event.
	 */
	errrc = an_event_status(&fooev, &etype, &estate);
	assert(errrc == 0);
	assert(etype == AN_EVTYPE_ONESHOT_ABSOLUTE);
	assert(estate == AN_EVSTATE_SCHEDULED);

	/*
	 * Try firing the already scheduled event (should fail).
	 */
	errrc = an_event_fire(&fooev, nevercalled, NULL);
	assert(errrc != 0);

	/*
	 * Try descheduling the event (should succeed).
	 */
	errrc = an_event_deschedule(&fooev, &etype, &estate);
	assert(errrc == 0);
	assert(etype == AN_EVTYPE_ONESHOT_ABSOLUTE);
	assert(estate == AN_EVSTATE_SCHEDULED);

	/* destroy event before returning!  Its on the stack */
	errrc = an_event_destroy(&fooev, &etype, &estate);
	assert(errrc == 0);
	assert(etype == AN_EVTYPE_UNDEFINED);
	assert(estate == AN_EVSTATE_DESCHEDULED);

	/*
	 * Try firing an immediate event (should succeed).
	 * Handler will clean it up.
	 */
	evrc = an_event_create(&immed_ev);
	assert(evrc == &immed_ev);
	errrc = an_event_fire(&immed_ev, immed_func, NULL);
	assert(errrc == 0);
}

static int events;

static void
ee_event1(an_event_t *event, void *arg)
{
	an_eventstate_t state;
	an_eventtype_t type;

	assert(event);
	assert((int)arg == 1);

	an_event_status(event, &type, &state);
	assert(type == AN_EVTYPE_PERIODIC);
	assert(state == AN_EVSTATE_SCHEDULED);

	events++;
}

static void
ee_event100(an_event_t *event, void *arg)
{
	static int count;
	an_eventstate_t state;
	an_eventtype_t type;
	an_timespec_t tspec;

	assert(event);
	assert((int)arg == 100);

	an_event_status(event, &type, &state);
	assert(type == AN_EVTYPE_PERIODIC);
	assert(state == AN_EVSTATE_SCHEDULED);

	an_gettime(&tspec);
	printf("%p/%p: got %d events at %u seconds\n",
	       an_domain_current(), an_thread_current(), events,
	       (unsigned int)tspec.tv_sec);

	if (++count == RUNFOR) {
		/*
		 * Make sure various misctest handlers
		 * fired correctly.
		 */
		checkmisctests();

		/*
		 * Should be within 1% of the expected value
		 */
		if ((expected * 99) / 100 <= events &&
		    events <= ((expected * 101) / 100))
			printf("eventtest: PASS\n");
		else
			printf("eventtest: FAIL\nNOT ");
		printf("within +/-%lu of expected value %lu\n",
		       expected / 100, expected);		       

		an_domain_destroy(an_domain_current());
	}
}

static void
ee_init(void *arg)
{
	int status;
	int argc = 1;
	const char *argv[] = { "ChantestEE" }; 

	status = ee_main(argc, argv);
}

