/*
 * gamestate.c
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 *
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <assert.h>
#include <nodeos/an_mutex.h>
#include <nodeos/an_cond.h>
#include <nodeos/an_thread.h>
#include <nodeos/an_time.h>
#include <nodeos/an_error.h>
#include <nodeos/an_event.h>

#include <stdlib.h>  // malloc

#include "gamestate.h"
#include "tic.h"
#include "peer.h"
#include "packets.h"
#include "identity.h"
#include "trace.h"
#include "stats.h"
#include "error.h"

GameState gameState = {};
FdoomOptions options = {};

#if !defined(NDEBUG)
static int stateLocked = 0;
#endif

void lockGameState(void)
{
	an_mutex_lock(&gameState.lock);
	TRACE(TRACE_GSLOCK, "lockedGameState\n");
	DEBUGONLY(assert(stateLocked == 0));
	DEBUGONLY(stateLocked = 1);
	STAT_INC(lockAcquire);
}

void waitOnGameState(void)
{
	DEBUGONLY(assert_gameStateLocked());
	DEBUGONLY(stateLocked = 0);
	TRACE(TRACE_GSLOCK, "block on GameState\n");
	an_cond_wait(&gameState.cv, &gameState.lock);
	DEBUGONLY(stateLocked = 1);
}

void signalOnGameState(void)
{
	DEBUGONLY(assert_gameStateLocked());
	TRACE(TRACE_GSLOCK, "signal on GameState\n");
	an_cond_signal(&gameState.cv);
}

void unlockGameState(void)
{
	STAT_INC(lockRelease);
	DEBUGONLY(assert_gameStateLocked());
	DEBUGONLY(stateLocked = 0);
	TRACE(TRACE_GSLOCK, "unlockedGameState\n");
	an_mutex_unlock(&gameState.lock);
}

#if !defined(NDEBUG)
void assert_gameStateLocked(void)
{
	assert(stateLocked == 1);
}
#endif

static inline void delay(int ms)
{
	an_timespec_t delaytime;
	delaytime.tv_sec = 0;
	delaytime.tv_nsec = ms * 1000 * 1000;
	STAT_INCBY(sleepyTimeTicsMS, ms);
	an_thread_sleep(&delaytime);
}

/*
 * Fabricate N tics.  
 */
static void
makeTics(int ticCount)
{
	int i;

	assert(ticCount > 0);
	assert(ticCount <= MAX_TICS_AHEAD);

	lockGameState(); /* XXX move lock inside loop? */
	for (i = 0; i <ticCount; i++)
	{
		gameState.madeToTic++;
		makeTic(gameState.madeToTic, getLocalTic(gameState.madeToTic), gameState.identity);
	}
	unlockGameState();

	if (ticCount > 1)
		delay(options.millisPerTic/2);
}

static int initialized = 0;
static an_timespec_t lastUglyTime;

/*
 * Update the "tic count" which is the game-specific notion of time.
 *
 * Basically, just get the current real time, compare with the last
 * time we came into this function, and figure out the number of
 * tics that have elapsed since then.
 *
 * XXX does not deal with re-entrancy (i.e., multiple simultaneous
 * calls to updateTicker() will probably fail in odd ways).
 */
static void
updateTicker(an_event_t *event, void *arg)
{
	an_timespec_t uglyNow;
	Uint32 ticDiff;
	
	an_gettime(&uglyNow);
	
	/* Initialize if this is the first call. */
	if (!initialized)
	{
		initialized = 1;
		lastUglyTime = uglyNow;
		return;
	}

	/*
	 * Compute the difference in tics since this function
	 * was last called.
	 */
	ticDiff = (uglyNow.tv_sec*1000/options.millisPerTic) -
		(lastUglyTime.tv_sec*1000/options.millisPerTic) +
		((uglyNow.tv_nsec/(1000*1000))/options.millisPerTic) -
		((lastUglyTime.tv_nsec/(1000*1000))/options.millisPerTic);

	/* update lastUglyTime for the next entry into this function. */
	lastUglyTime = uglyNow;


	TRACE(TRACE_TIC, 
	      "uglyNow=%lu.%lu -> (%lu+%lu tic)\n",
	      (Uint32)(uglyNow.tv_sec),
	      (Uint32)(uglyNow.tv_nsec),
	      (Uint32)(uglyNow.tv_sec*1000/options.millisPerTic),
	      (Uint32)((uglyNow.tv_nsec/(1000*1000))/options.millisPerTic)
	      );

	TRACE(TRACE_TIC, "ticDiff=%lu\n", ticDiff);

	assert(ticDiff >= 0);

	if (ticDiff == 0)
	{
		STAT_INC(ticDiffDistribution[0]);
		return;
	}

	/*
	 * Cap the tic generation if we get too far ahead of the gamestate.
	 *
	 * XXX improve the adaptation code so this doesn't happen at run-time.
	 *
	 * XXX install a legit detector for games that have stalled for
	 * too long (and have the client exit).
	 */
	lockGameState();
	if ((gameState.madeToTic + ticDiff) >
	    (gameState.playedToTic + MAX_TICS_AHEAD))
	{
		STAT_INC(ticCappedCount);
		ticDiff = gameState.playedToTic + MAX_TICS_AHEAD - gameState.madeToTic;
	}
	
	// XXX bit of a hack:
	if (ticDiff == MAX_TICS_AHEAD)
		ticDiff--;

	unlockGameState();

	assert(ticDiff >= 0);
	assert(ticDiff < MAX_TICS_AHEAD);
	STAT_INC(ticDiffDistribution[ticDiff]);

	/*
	 * Fabricate some tics to match the number we're allowed
	 * to collect locally.
	 */
	if (ticDiff > 0)
		makeTics(ticDiff);

	/* Update state of the world, if possible */
	gameUpdate();
}

static an_event_t tickerEvent;

void
registerTicker(void)
{
	an_error_t err;
	an_timespec_t gap;

	gap.tv_sec = (options.millisPerTic / 1000);
	gap.tv_nsec = (options.millisPerTic % 1000) * 1000 * 1000;
		
	an_event_create(&tickerEvent);
	err = an_event_schedule(&tickerEvent, &gap, AN_EVTYPE_PERIODIC,
				updateTicker, NULL);
	assert(err == 0);
}

void
shutdownTicker(void)
{
	an_error_t err;
	err = an_event_deschedule(&tickerEvent, NULL, NULL);
	assert(err == 0); /* XXX */
	err = an_event_destroy(&tickerEvent, NULL, NULL);
	assert(err == 0); /* XXX */
}

/*
 * Initializes the game state and one-time allocates necessary memory.
 * This must be called after the arguments have been parsed, but before
 * peers are parsed.
 *
 * Sets up "gameState.identity" which is needed to parse the peer
 * options, based on the current node's addr.
 */
void
gameStateInit(void)
{
	int i;
	
	memset(&gameState, 0, sizeof gameState);
	gameState.gamePhase = GP_SETUP;
	an_mutex_create(&gameState.lock);
	an_cond_create(&gameState.cv);
	gameState.identity = computeIdentity(options.protocol, options.myaddr);
	gameState.controllerIndex = -1; /* assume this is controller */

	/* Allocate a contiguous chunk for the tics, then point to it */
	gameState.ticChunk = malloc(ticCommandSize()*TIC_HISTORY_LEN);
	assert(gameState.ticChunk);
	memset(gameState.ticChunk, 0, ticCommandSize()*TIC_HISTORY_LEN);
	for (i = 0; i < TIC_HISTORY_LEN; i++)
	{
		gameState.localTics[i] = gameState.ticChunk+ticCommandSize()*i;
	}
}

/*
 * Cleans up the game state and free's allocated memory
 */
void
gameStateShutdown(void)
{
	/* Free chunk (NOT individual tics) */
	if (gameState.ticChunk)
		free(gameState.ticChunk);
	gameState.ticChunk = NULL;
}

/*
 * Advance the game engine.  This is where the game is actually "played".
 * The game is "played" for each set of consistent tics from all players.
 *
 * The actual "playing" consists of just throwing out the used tics.
 */
static void
advanceEngine(void)
{
	Peer *peer;
	int advanceToTic = gameState.madeToTic;

	// Figure out the greatest consistent tic
	lockGameState(); // XXX long lock hold
	peer = peerList;
	while(peer != NULL)
	{
		checkPeerMagic(peer);
  		if (peer->quitTic != NOT_QUIT)
  		{
  			peer = peer->next;
  			continue;
  		}

		if (peer->madeToTic < advanceToTic)
			advanceToTic = peer->madeToTic;
		peer = peer->next;
	}

	// Advance the game to that consistent tic
	assert(gameState.playedToTic <= advanceToTic);
	if (gameState.playedToTic < advanceToTic)
	{
		assert((advanceToTic - gameState.playedToTic) >= 0);
		assert((advanceToTic - gameState.playedToTic) <= MAX_TICS_AHEAD);
		STAT_INC(ticAdvanceDistribution[advanceToTic - gameState.playedToTic]);

#ifdef DEBUG
		{
			Tic i;
			
			for (i = gameState.playedToTic; i < advanceToTic; i++)
			{
				TicCommand *t;
				// Sanity check remote ticcommands
				Peer *p = peerList;
				while(p != NULL)
				{
					if (p->quitTic != NOT_QUIT)
					{
						p = p->next;
						continue;
					}

					t = ticListGetTicForGet(&p->ticCommand, i);
					assert(t->identity == p->identity);

					p = p->next;
				}

				// Sanity check local ticcommand
				t = getLocalTic(i);
				assert(t->tic == i);
				assert(t->identity == gameState.identity);
			}
		}
#endif /*DEBUG*/

		/*
		 * "advance" the engine; since the tics are stored
		 * in a ring buffer they'll eventually get overwritten.
		 */
		gameState.playedToTic = advanceToTic;

	}
	else
	{
		STAT_INC(ticAdvanceDistribution[0]);
	}
	STAT_INC(ticsAheadDistribution[gameState.madeToTic - gameState.playedToTic]);
	unlockGameState();
}

/*
 * Move this game out of the setup phase and into the 'running' phase.
 */
void
advanceToRunning(void)
{
	an_timespec_t curtime;
	Uint64 millisstart,millisend;
	Uint32 timediff;
	int i;

	assert(gameState.gamePhase == GP_SETUP);
	gameState.gamePhase = GP_RUNNING;
	gameState.playedToTic = 0;
	gameState.madeToTic = 0;
	for (i = 0; i < TIC_HISTORY_LEN; i++)
		memset(gameState.localTics[i], 0, ticCommandSize());
	makeTic(0, getLocalTic(0), gameState.identity);

	/* Start the ticker going */
	registerTicker();

	/* Take care of some stats */
	an_gettime(&curtime);
	millisstart = (Uint64)(gameState.starttime.tv_sec)*(Uint64)(1000) +
		(Uint64)(gameState.starttime.tv_nsec)/(Uint64)(1000000);
	millisend = (Uint64)(curtime.tv_sec)*(Uint64)(1000) +
		(Uint64)(curtime.tv_nsec)/(Uint64)(1000000);
	timediff = (Uint32)(millisend - millisstart);
	STAT_SET(setupTimeMS,timediff);
}

/*
 * Update the game state.  Advance the engine as far as possible
 * and then send packets to anyone who needs packets.
 *
 * This is called once per tic, and once per packet received.  Its
 * probably called too often.
 */
void
gameUpdate(void)
{
	advanceEngine();
	sendRunningPackets();
}
