/*
 * gamestate.h
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 *
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#ifndef GAMESTATE_HEADER
#define GAMESTATE_HEADER

#include <nodeos/an_mutex.h>
#include <nodeos/an_cond.h>

#include "types.h"
#include "debug.h"
#include "tic.h"

typedef enum GamePhase {
	GP_SETUP=10, GP_RUNNING=11, GP_SHUTDOWN=12
} GamePhase;

typedef struct GameState {
	GamePhase gamePhase;
	int controllerIndex;	/* index of the controller peer */
	Uint64 identity;	/* this machine's identity */
	an_mutex_t lock;	/* lock protects Peer state */
	an_cond_t  cv;
	/*
	 * Time that the game started so that various events can be measured
	 */
	an_timespec_t starttime;
	/*
	 * Highest number tic that's been "played" (gotten from
	 * all peers and "executed" in the engine).
	 */
	Tic playedToTic;
	/*
	 * Highest number tic made for this node.
	 */
	Tic madeToTic;		/* Maxes out at (playedToTic + MAX_TICS_AHEAD) */
	/*
	 * The actual contiguous chunk of memory that localTics
	 * points off to.
	 */
	void *ticChunk;
	/*
	 * Local tics generated on this node.  Valid between
	 * (gameState.playedToTic%TIC_HISTORY_LEN) and 
	 * (gameState.madeToTic%TIC_HISTORY_LEN).
	 * Each element of this array points off to a TicCommand
	 * continguously allocated in ticChunk.
	 */
	TicCommand *localTics[TIC_HISTORY_LEN];
} GameState;

typedef struct FdoomOptions {
	int ticcmdsize;
	int statprintgap; /* in seconds */
	Boolean checksum;
	Boolean proxypeer;
	int gameid;
	const char *protocol;
	const char *tracemask;
	const char *myaddr;
	const char *serveraddr;
	int timeout;  /* in seconds */
	int millisPerTic;
	//int ticfillsize;  /* This becomes a global for convoluted reasons */
} FdoomOptions;

extern FdoomOptions options;

extern GameState gameState;

static inline TicCommand* getLocalTic(Tic t) __attribute__((__unused__));
static inline TicCommand* getLocalTic(Tic t)
{
	return gameState.localTics[t%TIC_HISTORY_LEN];
}

extern void lockGameState(void);
extern void unlockGameState(void);
void waitOnGameState(void);
void signalOnGameState(void);
#if !defined(NDEBUG)
extern void assert_gameStateLocked(void);
#endif

extern void gameStateInit(void);
extern void gameStateShutdown(void);

extern void advanceToRunning(void);

/*
 * Register the ticker that counts gametics.
 */
extern void registerTicker(void);

/*
 * Shutdown the gameticer.
 */
extern void shutdownTicker(void);

/* 
 * Advance the game engine and send out any update packets that can
 * be sent. 
 */
extern void gameUpdate(void);


#endif /* GAMESTATE_HEADER */
