/*
 * fdoom main.c
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 *
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <stdlib.h> // malloc

#include <nodeos/an_domain.h>
#include <nodeos/an_mutex.h>
#include <nodeos/an_cond.h>
#include <nodeos/an_event.h>

#include "peer.h"
#include "types.h"
#include "stats.h"
#include "trace.h"
#include "error.h"
#include "random.h"
#include "options.h"
#include "packets.h"
#include "identity.h"
#include "gamestate.h"
#include "server.h"

/*
 * TODO:
 *
 *  Fill the packets with variable length garbage.
 *
 *  Compute a checksum for each packet and verify on receipt.
 *
 *  Create and handle shutdown packets.
 */

static void ee_init(void *arg);
an_domaininit_func_t InitEE = { ee_init };

static void registerTimeout(void);
static an_event_t timeoutEvent;
static void gameOver(an_event_t* event, void* arg);

int ticfillsize;
static OptionSpec optSpec[] = {
	{ "-ticcmdsize", OT_INT, &options.ticcmdsize, {intspec: { 32, 0, 128 }}},
	{ "-statprintgap", OT_INT, &options.statprintgap, {intspec: { 1, 0, 120 }}},
	{ "-checksum", OT_BOOL, &options.checksum, {boolspec: { true }}},
	{ "-proxypeer", OT_BOOL, &options.proxypeer, {boolspec: { false }}},
	{ "-gameid", OT_INT, &options.gameid, {intspec: { 1, 1, 65535}}},
	{ "-protocol", OT_STR, &options.protocol, {strspec: { "udp" }}},
	{ "-timeout", OT_INT, &options.timeout, {intspec: { 60, 0, 100*1000*1000 }}},
	{ "-millispertic", OT_INT, &options.millisPerTic, {intspec: { (1000 / 5), 1, 1000 }}},
	{ "-tracemask", OT_STR, &options.tracemask, {strspec: { NULL }}},
	{ "-trace", OT_STR, &options.tracemask, {strspec: { NULL }}}, /* same as -tracemask */
	{ "-myaddr", OT_STR, &options.myaddr, {strspec: { "127.0.0.1:0" }}},
	{ "-adoomserver", OT_STR, &options.serveraddr, {strspec: { NULL }}},
	{ "-ticfillsize", OT_INT, &ticfillsize, {intspec: { 40, 0, 4096 }}},	  
};

static void
ee_init(void *arg)
{
	Boolean createRealPeers;
	an_initargs_t *initargs;
	int peerArgs;

	/* XXX configure time checks: */
	assert(sizeof(Uint64) == 8);
	assert(sizeof(Uint32) == 4);

	initargs = (an_initargs_t*)arg;

	/* Initialize random number generator */
	initRandom();

	/* Initialize stats. XXX statsInit() */
	memset(&fdoomStats, 0, sizeof fdoomStats);

	/*
	 * Usage: fdoom [options] <peers ...>
	 * Options:
	 *    See optSpec, above
	 *
	 * Each <peer> is an addr[:port] pair where 'addr' is the
	 * IP address of the peer and the option port specifies
	 * what port they're listening on.
	 */
	peerArgs = parseArgs(initargs, optSpec, sizeof(optSpec)/sizeof(optSpec[0])); 
	
	traceInit(options.tracemask);

	if (peerArgs == initargs->argc)
	{
		FATALERROR("No arguments provided.\n");
	}

	/*
	 * Start profiling
	 */
#	ifdef GPROF
	an_gprof_start(initargs->argv[0]);
#	endif

	/*
	 * Initialize gameState.
	 */
	gameStateInit();

	createRealPeers = true;
	if (options.serveraddr != NULL)
	{
		/*
		 * Setup connections only to our adoom server.
		 * XXX multiple adoom servers?
		 */
		createADoomServer(options.serveraddr,
				  options.myaddr,
				  options.protocol);

		/* Don't attach to remote peers. */
		createRealPeers = false;
	}
		
	/*
	 * Setup input and output channels to every peer.
	 */
	for (; peerArgs < initargs->argc; peerArgs++)
	{
		Peer* peer;
		TRACEONLY(char buf[128]);
		
		/*
		 * XXX hack if our own address appears in the peer
		 * list, just skip it.  This is a concession to startup
		 * scripts which put all nodes in a list and then fire
		 * off nodes one at a time passing the entire list
		 * (which includes itself).
		 *
		 * Note: I am assuming that the address string will be
		 * different (due to inclusion of port number) if multiple
		 * clients are running on the same physical host.
		 */
		if (strcmp(initargs->argv[peerArgs], options.myaddr) == 0)
		{
			TRACE(TRACE_SETUP,
			      "Myaddr (%s) skipped in peer list\n",
			      options.myaddr);
			continue;
		}

		peer = createPeer(initargs->argv[peerArgs],
				  options.myaddr,
				  options.protocol,
				  createRealPeers);
		
		TRACE(TRACE_SETUP, "Registered peer %s\n",
		      peertostring(peer, buf, sizeof buf));
		STAT_INC(peerCount);
		
		assert(peer->identity != gameState.identity);
		if (peer->identity < gameState.identity)
			gameState.controllerIndex = peer->index;
	}
	
	/*
	 * Print stats at specified rate.
	 */
	registerStatsTrigger();

	/*
	 * Shutdown after a given timeout.
	 */
	registerTimeout();

	/*
	 * Synchronize with peers to simulate joining the game and
	 * exchanging initial game state.
	 */
	synchronizeGameSetup();
	if (gameState.gamePhase == GP_SETUP)
	{
		TRACE(TRACE_SETUP, "synchronizeGameSetup failed\n");
		gameOver(NULL, NULL);
		return;
	}

	/*
	 * Threads handling the tic timer, the stats timer, the
	 * shutdown timer, and packet arrival will handle things from
	 * now on...
	 */
}

static void
registerTimeout(void)
{
	an_timespec_t timeout;
	an_error_t err = 0;
	
	an_event_create(&timeoutEvent);

	timeout.tv_nsec = 0;
	timeout.tv_sec = options.timeout;

	if (options.timeout > 0)
		err = an_event_schedule(&timeoutEvent, &timeout,
					AN_EVTYPE_ONESHOT_RELATIVE,
					gameOver, NULL);
	else
		/* run indefinetly */;

	assert(err == 0);

	an_gettime(&gameState.starttime);
}

static void 
gameOver(an_event_t* event, void* arg)
{
	/* XXX cleanup other components: */
 	shutdownTicker();

	gameStateShutdown();

	printFinalStats();

	// Clean up other an state.

	an_domain_destroy(an_domain_current());
}

