/*
 * Packets.c
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 *
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <nodeos/an_thread.h>
#include <nodeos/an_chan.h>
#include <nodeos/an_event.h>
#include <nodeos/an_mutex.h>
#include <nodeos/an_error.h>

#include "packets.h"
#include "options.h"
#include "peer.h"
#include "trace.h"
#include "stats.h"
#include "types.h"
#include "error.h"
#include "gamestate.h"
#include "server.h"

static an_pbuf_t setupSendBuf;
static an_pbuf_t connectSendBuf;
static an_mutex_t runningBufLock;
static an_pbuf_t runningSendBuf;

static void broadcastSetupPacket(void);
static void initSendBufs(void);
static void recvSetupPacket(SetupPacket* vaguePacket);
static void recvRunningPacket(Peer* peer, RunningPacket* packet);
static an_error_t emitSetupPacket(an_chan_t *chan);
//static void sendServerSetupPacket(Server *server);
static int sendConnectPacket(void);

static inline void delay(int ms)
{
	an_timespec_t delaytime;
	delaytime.tv_sec = ms / 1000;
	delaytime.tv_nsec = (ms % 1000) * 1000 * 1000;
	STAT_INCBY(sleepyTimeMS, ms);
	an_thread_sleep(&delaytime);
}

static void
initSendBufs(void)
{
	SetupPacket* sp;
	RunningPacket* rp;
	ADoomConnectPacket* cp;
	an_pbuf_t *rc;

	rc = an_pbuf_create(&setupSendBuf, AN_PBUF_DEFAULT, 0, sizeof(SetupPacket));
	assert(rc == &setupSendBuf);
	/* XXX The following is severely limiting to ticfillsize (max at 105-ish),
	 * because it makes it able to stick the whole history into one packet.
	 * It would be better if this could be more dynamic.
	 */
	rc = an_pbuf_create(&runningSendBuf, AN_PBUF_DEFAULT, 0,
			    sizeof(RunningPacket) + ticCommandSize()*TIC_HISTORY_LEN);
	assert(rc == &runningSendBuf);
	rc = an_pbuf_create(&connectSendBuf, AN_PBUF_DEFAULT, 0, sizeof(ADoomConnectPacket));
	assert(rc == &connectSendBuf);

	an_mutex_create(&runningBufLock);

	sp = an_pbuf_data(&setupSendBuf);
	sp->packetType = PT_SETUP;
	sp->senderIdentity = gameState.identity;
	sp->peerCount = peerCount();
	sp->ticCommandSize = ticCommandSize();
	an_pbuf_setsize(&setupSendBuf, sizeof(SetupPacket), NULL);

	rp = an_pbuf_data(&runningSendBuf);
	rp->packetType = PT_RUNNING;
	rp->senderIdentity = gameState.identity;
	rp->retransmitFrom = NO_RETRANSMITS;
	rp->startTic = 0;
	rp->ticCount = 0;
	memset(&rp->ticCommand, 0, sizeof(rp->ticCommand));
	an_pbuf_setsize(&runningSendBuf, runningPacketSize(rp), NULL);

	cp = an_pbuf_data(&connectSendBuf);
	cp->packetType = PT_ADOOM_CONNECT;
	cp->senderIdentity = gameState.identity;
	cp->peerCount = peerCount();
	cp->ticCommandSize = ticCommandSize();
	an_pbuf_setsize(&connectSendBuf, sizeof(ADoomConnectPacket), NULL);
}

int runningPacketSize(RunningPacket* rp)
{
	int baseSize = ((void*)(&rp->ticCommand[0])) - ((void*)rp); // XXX that's a constant
	return baseSize + (rp->ticCount * ticCommandSize());
}


void
synchronizeGameSetup(void)
{
	int foundPeerCt;

	foundPeerCt = 0;

	initSendBufs(); // XXX initBuffers, move call into main.c

	if (options.serveraddr != NULL)
	{
		/*
		 * I am behind an adoom server.  Send connection
		 * requests to it until it responds.
		 */
		int checkCount = 5; // XXX
		
		while (gameState.gamePhase == GP_SETUP
		       && (--checkCount > 0))
		{
			TRACE(TRACE_SETUP, "Sending connect packet\n");
			sendConnectPacket();
			delay(2*options.millisPerTic);
			checkCount++;
		}
		if (checkCount)
			TRACE(TRACE_SETUP, "Received connection response\n");
		else
			TRACE(TRACE_SETUP, "Connection response timeout\n");
		return;
	}

	if (gameState.controllerIndex != -1)
	{
		/*
		 * I am a listener.  Wait until a setup packet
		 * has come in.  Then, go.
		 */
		TRACE(TRACE_SETUP, "[Controllee] waiting on controller...\n");

		lockGameState();
		while (gameState.gamePhase == GP_SETUP)
			waitOnGameState();
		unlockGameState();

		TRACE(TRACE_SETUP, "[Controllee] synchronized with controller.\n");

		return;
	}

	/*
	 * I am the controller.  Send a Setup packet to
	 * everyone.  Wait until everyone has sent a regular
	 * packet before returning.
	 */

	while(foundPeerCt < peerCount())
	{
		int checkCount = 5; // XXX

		TRACE(TRACE_SETUP, "[Controller] Sending setup packet to all peers...\n");

		broadcastSetupPacket();

#if 1
		while ((foundPeerCt < peerCount())
			&& (--checkCount > 0))
		{
			Peer *peer;

			foundPeerCt = 0;
			lockGameState();
			peer = peerList;
			while(peer != NULL)
			{
				checkPeerMagic(peer);
				if (peer->joined)
					foundPeerCt++;
				peer = peer->next;
			}
			unlockGameState();

			STAT_INC(controllerSetupLoops);

			TRACE(TRACE_SETUP, "[Controller] %d peer(s) have accepted.\n",
				foundPeerCt);

			if (foundPeerCt < peerCount())
			{
				delay(500);
			}
		}
#else

		/*
		 * XXX THIS IS A BIG HACK.  I punted the setup protocol
		 * completely to avoid deadlock with adoom in the net.
		 */

		lockGameState();
		{
			Peer* peer;
			for (peer = peerList;
			     peer != NULL;
			     peer = peer->next)
			{
				peer->joined = true;
			}
			unlockGameState();
		}
		
		delay(2*options.millisPerTic);

		foundPeerCt = peerCount();
#endif
	}

	TRACE(TRACE_SETUP,
	      "[Controller] All peer setup packets received.  Moving to GP_RUNNING.\n");

	/*
	 * Move out of GP_SETUP into GP_RUNNING state.
	 */
	advanceToRunning();
}

static an_error_t
emitSetupPacket(an_chan_t *chan)
{
	an_error_t err;
	Boolean blocked;

	do 
	{
		blocked = false;
		/* XXX synchronous send.  What to do if async? */
		err = an_outchan_send(chan, &setupSendBuf);
		
		if (err == EBUSY)
		{
			TRACE(TRACE_SEND_ERR,
			      "an_outchan_send failed, busy; retrying.\n");
			delay(20);
			blocked = true;
			STAT_INC(packetSendBusyCount);
		}
	}
	while (blocked);

	return err;
}

void
broadcastSetupPacket(void)
{
	Peer *peer;
	TRACEONLY(char buf[129]);
	
	peer = peerList;
	while (peer != NULL)
	{
		an_error_t err;
		
		checkPeerMagic(peer);

		err = emitSetupPacket(&peer->sendChan);
		if (err != 0)
		{
			TRACE(TRACE_SEND_ERR,
			      "an_outchan_send failed %s: %s(%x)\n",
			      peertostring(peer, buf, sizeof buf),
			      an_error_message(), err);
			STAT_INC(packetSendErrCount);
		}
		else
		{
			TRACE(TRACE_SEND,
			      "Sent packet to %s\n",
			      peertostring(peer, buf, sizeof buf));
			STAT_INC(packetSentCount);
		}
		
		peer = peer->next;
	}
}

static int
sendConnectPacket(void)
{
	an_error_t err;
	Server *server;
	Boolean blocked;

	server=adoomServer();
	
	do
	{
		blocked = false;
		/* XXX synchronous send.  What to do if async? */
		err = an_outchan_send(&server->sendChan, &connectSendBuf);

		if (err == EBUSY)
		{
			TRACE(TRACE_SEND_ERR,
			      "an_outchan_send failed, busy: retrying.\n");
			delay(20);
			blocked = true;
			STAT_INC(packetSendBusyCount);
		}
	}
	while (blocked);

	return err;
}

/* returns number of tics sent to dest */
static int
setupRunningPacket(Boolean needRetransmit,
		   Tic madeToTic,
		   Tic lastSentTic,
		   RunningPacket *rp,
		   const char *traceName)
{
	int ticsInPacket = 0;

	TRACE(TRACE_SEND, "setupRunningPacket(retran=%d, madeTo=%lld, lastSent=%lld, name=%s)\n",
	      needRetransmit, madeToTic, lastSentTic, traceName);

	DEBUGONLY(assert_gameStateLocked());
	/* XXX assert rp pbuf locked. */
	
	assert(an_pbuf_offset(&runningSendBuf) == 0);
	assert(rp->packetType == PT_RUNNING);
	assert(rp->senderIdentity == gameState.identity);
	assert(gameState.madeToTic >= lastSentTic);
	
	/*
	 * If I need a retransmit from this peer, tell them by
	 * setting the 'retransmitFrom' field of the packet.
	 */
	if (needRetransmit)
	{
		TRACE(TRACE_RETRAN,
		      "Need retrans.  Asking for backups from %s.\n",
		      traceName);
		STAT_INC(requestRetransmitCount);
		assert((madeToTic >= 0) && (madeToTic <= gameState.madeToTic));
		rp->retransmitFrom = madeToTic;
	}
	else
	{
		rp->retransmitFrom = NO_RETRANSMITS;
	}
	
	/*
	 * Define the startTic and ticCount for the packet.
	 */
	rp->startTic = lastSentTic;
	assert(gameState.madeToTic >= lastSentTic);
	rp->ticCount = gameState.madeToTic - lastSentTic;
	assert(rp->ticCount < TIC_HISTORY_LEN);
	STAT_INC(ticSentDistribution[rp->ticCount]);

	/*
	 * Copy the tics into the packet
	 */
	if (rp->ticCount > 0)
	{
		TicCommand *src;
		TicCommand *dest;
		int i;

		TRACE(TRACE_SEND,
		      "Sending (%lld +%d /retran=%lld) to %s\n",
		      rp->startTic, rp->ticCount,
		      rp->retransmitFrom, traceName);
		
		/* Copy local command store into outgoing packet. */
			
		for (i = 0; i < rp->ticCount; i++)
		{
			src = getLocalTic(rp->startTic+i);
			dest = &(rp->ticCommand[i]);
			
			// sanity checks
			DEBUGONLY(assert(src->tic == (rp->startTic+i)));
			DEBUGONLY(assert(src->identity == gameState.identity));
			
			*dest = *src;
		}
		ticsInPacket = rp->ticCount;
	}
	else
	{
		TRACE(TRACE_SEND,
		      "Peer %s is up to date.\n", traceName);
		STAT_INC(packetNotNeededCount);
		ticsInPacket = 0;
	}

	DEBUGONLY(assert_gameStateLocked());
	return ticsInPacket;
}


static an_error_t
sendRunningPacketTo(RunningPacket* rp,
		    an_chan_t *sendChan,
		    const char *traceName)
{
	an_error_t err;

	/* XXX synchronous send.  What to do if async? */
	an_pbuf_setsize(&runningSendBuf, runningPacketSize(rp), NULL);
	err = an_outchan_send(sendChan, &runningSendBuf);
	if (err != 0)
	{
		TRACE(TRACE_SEND_ERR,
		      "an_outchan_send() failed %s: %s(%x)\n",
		      traceName, an_error_message(), err);
		STAT_INC(packetSendErrCount);
	}
	else
	{
		TRACE(TRACE_SEND,
		      "Sent packet to %s (%d tics)\n",
		      traceName, rp->ticCount);
		STAT_INC(packetSentCount);
	}

	return err;
}

/*
 * Send running packets to listeners.  Running packets include
 * any and all tics we've completed recently that a peer might
 * be interested in.
 */
void
sendRunningPackets(void)
{
	RunningPacket* rp = an_pbuf_data(&runningSendBuf);

	/*
	 * XXX this switch between server and legacy mode could
	 * be done more elegantly elsewhere, I think.
	 *
	 * XXX the code duplication between the two variations
	 * is pretty obscene.  Should right coalesce the Server
	 * and Peer objects a bit (define the common bits
	 * based on what's used here, actually!)
	 */
	if (options.serveraddr != NULL)
	{
		Server *server = adoomServer();
		int ticsInPacket;
		an_error_t err;

		TRACE(TRACE_SEND, "Sending running burst to adoom server.\n");
		
		lockGameState();
		an_mutex_lock(&runningBufLock);
		while(an_pbuf_isbusy(&runningSendBuf))
			an_thread_yield();
	
		ticsInPacket = setupRunningPacket(server->needRetransmit,
						  server->madeToTic,
						  server->lastSentTic,
						  rp,
						  "server");
		if ((ticsInPacket > 0)
		    || (server->needRetransmit))
		{
			err = sendRunningPacketTo(rp,
						  &server->sendChan,
						  "server");

			if (err == 0)
			{
				server->lastSentTic += ticsInPacket;
				assert(server->lastSentTic <= gameState.madeToTic);
				server->needRetransmit = false;
			}
		}
		an_mutex_unlock(&runningBufLock);
		unlockGameState();

	}
	else
	{
		Peer *peer;
		char traceName[128];
		int ticsInPacket;
		an_error_t err;

		TRACE(TRACE_SEND, "Sending running burst to all peers.\n");

		peer = peerList;
		while (peer != NULL)
		{
			peertostring(peer, traceName, sizeof traceName);
			checkPeerMagic(peer);

			lockGameState();
			an_mutex_lock(&runningBufLock);
			while(an_pbuf_isbusy(&runningSendBuf))
				an_thread_yield();
	
			ticsInPacket = setupRunningPacket(peer->needRetransmit,
							  peer->madeToTic,
							  peer->lastSentTic,
							  rp,
							  traceName);
			if ((ticsInPacket > 0)
			    || (peer->needRetransmit))
			{
				err = sendRunningPacketTo(rp, &peer->sendChan,
							  traceName);
				
				if (err == 0)
				{
					peer->lastSentTic += ticsInPacket;
					assert(peer->lastSentTic <= gameState.madeToTic);
					peer->needRetransmit = false;
				}
			}
			an_mutex_unlock(&runningBufLock);
			unlockGameState();

			peer = peer->next;
		}
	}
}

static void
recvSetupPacket(SetupPacket* packet)
{
	TRACE(TRACE_RECV|TRACE_SETUP,
	      "received SetupPacket[id=%#qx peer ct=%d]\n",
	      packet->senderIdentity, packet->peerCount);

	/* If I'm running, ignore setup packets */
	if (gameState.gamePhase == GP_RUNNING)
	{
		STAT_INC(setupPacketWhileRunningCount);
		return;
	}
	
	/*
	 * If I'm the controller, just ignore any setup packets.
	 *
	 * XXX use these as a resend to peer...?
	 */
	if (gameState.controllerIndex == -1)
	{
		STAT_INC(setupPacketWhileControllerCount);
		return;
	}
	
	/*
	 * Reject configuration if identies overlap, or configurations
	 * are different.
	 */
	lockGameState();
	if (packet->senderIdentity == gameState.identity)
	{
		FATALERROR("Bad setup: peer has my identity! (id=%#qx)\n",
			   gameState.identity);
	}
	else if (packet->peerCount != peerCount())
	{
		FATALERROR("Bad setup: wrong number of peers in setup packet!\n");
	}
	else
	{
		TRACE(TRACE_RECV|TRACE_SETUP,
		      "[Controllee] advancing to GP_RUNNING\n");
		/*
		 * Advance to running phase.
		 */
		advanceToRunning();
		signalOnGameState();
	}
	unlockGameState();
}

static void
addTicToPeer(Peer *peer,
	     Tic packetStartTic, int packetTicCount,
	     TicCommand *packetTics)
{
	int skipTics;
	TRACEONLY(char buf[129]);

	assert(peer);
	assert(packetTics);
	// assert(packetStartTic > 0);
	assert(packetTicCount > 0);
	DEBUGONLY(assert_gameStateLocked());

	skipTics = 0;

	if (packetStartTic > peer->madeToTic)
	{
		/*
		 * Some tics were skipped, ask for re-transmit.
		 * XXX toss the tics received, we'll get them
		 * back eventually, too.
		 */
		peer->needRetransmit = true;
		TRACE(TRACE_RETRAN,
		      "Need retrans.  I'm missing %d packet(s) from %s.\n",
		      (int)(packetStartTic - peer->madeToTic),
		      peertostring(peer, buf, sizeof buf));
		
	}
	else if (packetStartTic < peer->madeToTic)
	{
		/*
		 * Must be a retransmit we don't need?  Skip the tics
		 * that we don't need and take what we can from the
		 * the packet.
		 *
		 * XXX ADD A STAT() for this
		 */
		
		while ((packetStartTic < peer->madeToTic)
		       && (packetTicCount > 0))
		{
			packetStartTic++;
			packetTicCount--;
			skipTics++;
		}
		
		TRACE(TRACE_RECV,
		      "Received some old packets, skipped %d (%d left)\n",
		      skipTics, packetTicCount);
		
		/*
		 * Fall through and copy the packets.
		 */
	}
	
	if (packetStartTic == peer->madeToTic)
	{
		TicCommand *src;
		TicCommand *dest;
		int i;
		
		/* Got tics at exactly the tic offset we were expecting. */
		
		/* Copy tic commands over, increment progress for this peer */
		for (i = skipTics; i < packetTicCount; i++)
		{
			src = &(packetTics[i]);
			dest = ticListGetTicForPut(&peer->ticCommand, peer->madeToTic);
			
			DEBUGONLY(assert(src->tic == peer->madeToTic));
			DEBUGONLY(assert(src->identity == peer->identity));
			
			memcpy(dest, src, ticCommandSize());

			if (src->flags & TFLAG_QUIT)
			{
				TRACE(TRACE_KICK, "Recv'd quit flag for peer %s for tic %lld\n",
				      peertostring(peer, buf, sizeof buf),
				      peer->madeToTic);
				peer->quitTic = peer->madeToTic;
			}
			
			peer->madeToTic++;
		}
	}
}


static void
recvRunningPacket(Peer* peer, RunningPacket *packet)
{
	TRACEONLY(char buf[129]);

	TRACE(TRACE_RECV,
	      "received runningPacket[tic=%qu + %d (retrans=%d)]\n",
	      packet->startTic, packet->ticCount,
	      (packet->retransmitFrom == NO_RETRANSMITS)
	      ? 0 : (int)(peer->lastSentTic - packet->retransmitFrom));

	/* If I'm in the setup phase */
	switch (gameState.gamePhase)
	{
	case GP_SETUP:
		if (gameState.controllerIndex != -1)
		{
			/*
			 * I'm not a controller, and I haven't received
			 * a setup packet yet, so ignore this running
			 * packet.
			 */
			STAT_INC(runningPacketDuringSetupCount);
			return;
		}

		/* Otherwise, I'm the controller, record who sent this. */
		lockGameState();
		peer->joined = true;
		unlockGameState();
		/* Fall through and handle as regular running packet */
	case GP_RUNNING:
		/* Update the stored tics from the peer. */
		assert(packet->senderIdentity == peer->identity);

		/*
		 * Packet must either have tics or a request for
		 * retransmit (or both).  Never a useless packet.
		 */
		assert((packet->ticCount > 0)
		       || (packet->retransmitFrom != NO_RETRANSMITS));

		lockGameState();
		if (packet->retransmitFrom != NO_RETRANSMITS)
		{
			/* Peer needs some old packets.  Just whack
			 * lastSentTic back down to whatever they need
			 * it at.  Send routines will pick up the
			 * change.
			 */
			TRACE(TRACE_RETRAN,
			      "Peer %s needs retrans.  Skipping back %d packet(s).\n",
			      peertostring(peer, buf, sizeof buf),
			      (int)(peer->lastSentTic - packet->retransmitFrom));
			assert(peer->lastSentTic >= packet->retransmitFrom);
			assert(packet->retransmitFrom <= gameState.madeToTic);
			assert(gameState.madeToTic - packet->retransmitFrom <= TIC_HISTORY_LEN);
			assert(packet->retransmitFrom <= peer->lastSentTic); /* XXX handle gracefully */
			TRACE(TRACE_RETRAN,"peer->lastSentTic before=%d\n",
			      (int)(peer->lastSentTic));
			peer->lastSentTic = packet->retransmitFrom;
			TRACE(TRACE_RETRAN,"peer->lastSentTic after=%d\n",
			      (int)(peer->lastSentTic));
			STAT_INC(doRetransmitCount);
		}

		if (packet->ticCount > 0)
		{
			addTicToPeer(peer,
				     packet->startTic, packet->ticCount, 
				     &(packet->ticCommand[0]));
		}
		unlockGameState();
		
		break;
	default:
		FATALERROR("unknown gameState %d\n",
			gameState.gamePhase);
	}
}

void
recvPeerData(an_pbuf_t *pbuf, void *arg, int offset)
{
	Peer* peer = (Peer*)arg;
	FdoomPacket* packet = (FdoomPacket*)(an_pbuf_data(pbuf)+offset);

	STAT_INC(packetRecvCount);

	TRACE(TRACE_RECV,
	      "received packet from %s (type=%d)\n",
	      peer->name, packet->packetType);

	switch(packet->packetType)
	{
	case PT_SETUP:
		recvSetupPacket((SetupPacket*)packet);
		break;
	case PT_RUNNING:
		recvRunningPacket(peer, (RunningPacket*)packet);
		break;
	default:
		FATALERROR("Unknown packet type received: %d\n",
			   packet->packetType);
	}

	gameUpdate();
	
	/* Recycle the pbuf */
	an_inchan_addpbuf(&peer->recvChan, pbuf);
}

void
recvConglomPacket(Server* server, RunningConglomPacket* packet)
{
	TicCommand *tic;
	Peer *p;
	int numSkip, numClients;
	
	assert(server);
	assert(packet);
	assert(packet->packetType == PT_RUNNING_CONGLOM);

	/*
	 * XXX we're throwing away many of the possible optimizations
	 * here: by definition we have a consistent snapshot
	 * everyone's state.  Just pass this to the engine as a single
	 * chunk dammitp!
	 *
	 * Until then, just disperse the conglomerate into the
	 * different Peer structures.
	 *
	 * XXX This uses memcpy to get the tic bodies from the conglomerated
	 * packet into the full-fledged tics.  They're not exactly big, but
	 * this sort of evades the whole "no copy" idea.
	 */

	//assert(packet->ticCount == 1); /* XXX */
	assert(packet->retransmitFrom == NO_RETRANSMITS);

	tic = packet->ticCommand;

	numSkip = packet->startClient;
	numClients = packet->clientCount;

	TRACE(TRACE_CONGLOM,"Recv'd conglom for tic %d\n", (int)(packet->tic));

	lockGameState();
	for (p = peerList; p; p = p->next)
	{
		if (numSkip)
		{
			--numSkip;
			continue;
		} else {
			if (numClients-- == 0)
				break;
		}
		
		if (p->quitTic == NOT_QUIT)
		{
			addTicToPeer(p, packet->tic, 1, tic);
			tic = (TicCommand*)(((void*)tic)+ticCommandSize());
		}
	}
	unlockGameState();
}

static void
recvServRecognizePacket(ADoomRecognizePacket *packet)
{
	/*
	 * This packet does nothing but tell us that we're on the
	 * network now and the game is ready to go
	 */
	advanceToRunning();
}

void
recvServRetransmitRequst(Server *server, RunningPacket* packet)
{
	assert(server);
	assert(packet);
	assert(packet->packetType == PT_RUNNING);
	assert(packet->retransmitFrom != NO_RETRANSMITS);
	/* Hokey asserts: */
	assert(packet->startTic == 0);
	assert(packet->ticCount == 0);
	
	TRACE(TRACE_RETRAN,
	      "Server %p needs retrans.  Skipping back %d packet(s).\n",
	      server, (int)(server->lastSentTic - packet->retransmitFrom));

	assert(server->lastSentTic >= packet->retransmitFrom);
	assert(packet->retransmitFrom <= gameState.madeToTic);
	assert(gameState.madeToTic - packet->retransmitFrom <= TIC_HISTORY_LEN);
	assert(packet->retransmitFrom <= server->lastSentTic); /* XXX handle gracefully */
	server->lastSentTic = packet->retransmitFrom;
	STAT_INC(doServerRetransmitCount);
}

void
recvServerData(an_pbuf_t *pbuf, void *arg, int offset)
{
	Server* server = (Server*)arg;
	FdoomPacket* packet = (FdoomPacket*)(an_pbuf_data(pbuf)+offset);

	STAT_INC(serverPacketRecvCount);
	
	TRACE(TRACE_RECV,
	      "received packet from server %s (type=%d)\n",
	      server->name, packet->packetType);

	switch(packet->packetType)
	{
  	case PT_RUNNING:
  		/* XXX its a retransmit request... */
  		recvServRetransmitRequst(server, (RunningPacket*)packet);
  		break;
	case PT_ADOOM_RECOGNIZE:
		recvServRecognizePacket((ADoomRecognizePacket*)packet);
		break;
	case PT_RUNNING_CONGLOM:
		recvConglomPacket(server, (RunningConglomPacket*)packet);
		break;
	default:
		FATALERROR("Unknown packet type received: %d\n",
			   packet->packetType);
	}

	gameUpdate();
	
	/* Recycle the pbuf */
	an_inchan_addpbuf(&server->recvChan, pbuf);
}

