/*
 * peer.c
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 *
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <stdio.h> // snprintf
#include <stdlib.h> // malloc

#include "peer.h"
#include "error.h"
#include "packets.h"
#include "trace.h"
#include "gamestate.h"
#include "identity.h"
#include "network.h"

static int peerIdNum = 0;
Peer *peerList = NULL;

int
peerCount(void)
{
	return peerIdNum;
}



Peer*
createPeer(const char *peerAddr,
	   const char *myAddr,
	   const char *protocol,
	   Boolean createRealPeer)
{
	Peer *peer;
	Peer *listpos;
	int i;

	peer = malloc(sizeof(Peer));
	peer->name = peerAddr;
	peer->index = peerIdNum++;
	peer->identity = computeIdentity(options.protocol, peer->name);
	peer->joined = false;
	peer->needRetransmit = false;
	peer->madeToTic = 0;
	peer->lastSentTic = 0;
	peer->quitTic = NOT_QUIT;
	ticListInit(&peer->ticCommand);

DEBUGONLY(
	peer->magic1 = PEER_MAGIC_1;
	peer->magic2 = PEER_MAGIC_2;
	peer->magic3 = PEER_MAGIC_3;
	peer->magic4 = PEER_MAGIC_4;
	peer->magic5 = PEER_MAGIC_5;
	peer->magic6 = PEER_MAGIC_6;
	peer->magic7 = PEER_MAGIC_7;
)

DEBUGONLY(
	/* 0th packet is bogus (never sent), just fixup enough
	 * so debugging asserts will work. */
	ticListGetTicForPut(&peer->ticCommand, 0)->tic = 0;
	ticListGetTicForPut(&peer->ticCommand, 0)->identity = peer->identity;
	)

	/* Insert it into the proper place in the list */
	if (!peerList)
	{
		peer->next = NULL;
		peerList = peer;
	} else if (peer->identity < peerList->identity) {
		/* Goes on front of list */
		peer->next = peerList;
		peerList = peer;
	} else {
		for (listpos = peerList; listpos->next; listpos = listpos->next)
		{
			if (peer->identity < listpos->next->identity)
			{
				peer->next = listpos->next;
				listpos->next = peer;
				break;
			}
		}
		if (!listpos->next)
		{
			peer->next = NULL;
			listpos->next = peer;
		}
	}

	/*
	 * A "real" peer is one we're connected to and exchange data
	 * with.  We don't create real peers when they're hidden
	 * behind an adoom server.
	 */
	if (createRealPeer)
	{
		for (i = 0; i < BUFS_PER_PEER; i++)
		{
			peer->bufptrs[i] = an_pbuf_create(&(peer->bufs[i]),
							  AN_PBUF_DEFAULT, 0, 0);
			assert(peer->bufptrs[i] == &(peer->bufs[i]));
		}

		createConnections(protocol, myAddr, peerAddr,
				  &peer->recvChan,
				  recvPeerData,  (void*)peer,
				  &peer->recvDemux,
				  peer->bufptrs, BUFS_PER_PEER,
				  &peer->sendChan);
	}

	return peer;
}

const char *
peertostring(Peer* p, char *buf, int buflen)
{
	assert(p);
	assert(buf);
	assert(buflen > 0);
#if defined(DEBUG) && !defined(NDEBUG)
#if 0
	snprintf(buf, buflen,
		"[Peer %p; id=%s; next=%p]",
		p, p->name, p->next);

	snprintf(buf, buflen,
		"[Peer %s]",
		p->name);
#endif
	snprintf(buf, buflen,
		 "[Peer %s sent=%+d recv=%+d]",
		 p->name,
		 (int)(p->lastSentTic - gameState.playedToTic),
		 (int)(p->madeToTic - gameState.playedToTic)
		 );
#else
	buf[0] = '\0';
#endif
	return buf;
}

void 
checkPeerMagic(Peer *peer)
{
#if defined(DEBUG) && !defined(NDEBUG)
	assert(peer->magic1 == PEER_MAGIC_1);
	assert(peer->magic2 == PEER_MAGIC_2);
	assert(peer->magic3 == PEER_MAGIC_3);
	assert(peer->magic4 == PEER_MAGIC_4);
	assert(peer->magic5 == PEER_MAGIC_5);
	assert(peer->magic6 == PEER_MAGIC_6);
	assert(peer->magic7 == PEER_MAGIC_7);
#endif
}

