/*
 * stats.c
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 *
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
#include <nodeos/an_event.h>

#include "stats.h"
#include "types.h"
#include "gamestate.h"
#include "peer.h"

FdoomStats fdoomStats;

static an_event_t statsEvent;
static void printStats(an_event_t *event, void *arg);

/*
 * Use official nodeOS logging interface.
 * XXX this proto (and the implementation) should be elsewhere
 */
extern void an_printf(const char *fmt, ...);

void
registerStatsTrigger(void)
{
	an_error_t err;
	an_timespec_t gap;
	
	if (options.statprintgap == 0)
		return;

	an_event_create(&statsEvent);

	gap.tv_sec = options.statprintgap;
	gap.tv_nsec = 0;

	err = an_event_schedule(&statsEvent,
				&gap, AN_EVTYPE_PERIODIC,
				printStats, NULL);
	assert(err == 0);
}

static void
doPrintStats(void)
{

	an_printf("GS: tics=%d/%d\n",
		  (int)(gameState.playedToTic),
		  (int)(gameState.madeToTic));
	
#if defined(DEBUG) && !defined(NDEBUG)
	{
		Peer *peer;
		
		peer = peerList;
		while(peer != NULL)
		{
			char buf[128];
			checkPeerMagic(peer);
			an_printf("%s\n", peertostring(peer, buf, sizeof buf));
			peer = peer->next;
		}
	}
#endif

	an_printf("sent=%ld; recv=%ld; (senderr=%ld; recvTooLong=%ld); "
		  "reqRe=%ld; doRe=%ld; noSend=%ld;\n",
		  fdoomStats.packetSentCount,
		  fdoomStats.packetRecvCount,
		  fdoomStats.packetSendErrCount,
		  fdoomStats.recvTookTooLongCount,
		  fdoomStats.requestRetransmitCount,
		  fdoomStats.doRetransmitCount,
		  fdoomStats.packetNotNeededCount
		  );
}

static void
printStats(an_event_t *event, void *arg)
{
	doPrintStats();
}

static void
printDistribution(const char* name, Uint32 *array, int len)
{
	int i;
	Uint32 min, max;
	min = (Uint32)-1;
	max = 0;
	for (i = 0; i < len; i++)
	{
		if (array[i] < min)
			min = array[i];
		if (array[i] > max)
			max = array[i];
	}

	//an_printf("%s: %lu/%lu (min/max)\n", name, min, max);
	an_printf("%10s: ", name);
	for (i = 0; i < len; i++)
	{
		an_printf("%4lu ", array[i]);
	}
	an_printf("\n");
}

void
printFinalStats(void)
{
	/* Print some of the option info */
#define POPT_D(VAR) an_printf("%32s: %d\n", #VAR, options.VAR)
	POPT_D(ticcmdsize);
	an_printf("%32s: %d\n", "sizeof(TicCommand)", sizeof(TicCommand));
	POPT_D(statprintgap);
	POPT_D(millisPerTic);
	POPT_D(timeout);
	an_printf("%32s: %d\n", "ticfillsize", ticfillsize);
	
	/* Print some actual statistics */
#define PSTAT_U32(VAR) an_printf("%32s: %lu\n", #VAR, fdoomStats.VAR)
	if (gameState.gamePhase == GP_RUNNING)
	{
		STAT_SET(avgRunningTicRate, gameState.madeToTic /
			 (options.timeout - fdoomStats.setupTimeMS/1000));
	}

	PSTAT_U32(peerCount);
	PSTAT_U32(packetSentCount);
	PSTAT_U32(packetSendErrCount);
	PSTAT_U32(packetSendBusyCount);
	PSTAT_U32(packetRecvCount);
	PSTAT_U32(serverPacketRecvCount);
	PSTAT_U32(recvTookTooLongCount);
	PSTAT_U32(lockAcquire);
	PSTAT_U32(lockRelease);
	PSTAT_U32(sleepyTimeMS);
	PSTAT_U32(sleepyTimeTicsMS);
	PSTAT_U32(setupTimeMS);
	PSTAT_U32(requestRetransmitCount);
	PSTAT_U32(doRetransmitCount);
	PSTAT_U32(doServerRetransmitCount);
	PSTAT_U32(ticCappedCount);
	PSTAT_U32(controllerSetupLoops);
	PSTAT_U32(packetNotNeededCount);
	PSTAT_U32(setupPacketWhileRunningCount);
	PSTAT_U32(setupPacketWhileControllerCount);
	PSTAT_U32(runningPacketDuringSetupCount);
	PSTAT_U32(avgRunningTicRate);

	printDistribution("ticDiff", fdoomStats.ticDiffDistribution,
			  MAX_TICS_AHEAD);
	printDistribution("ticAdvance", fdoomStats.ticAdvanceDistribution,
			  MAX_TICS_AHEAD);
	printDistribution("ticSent", fdoomStats.ticSentDistribution,
			  MAX_TICS_AHEAD);
	printDistribution("ticsAhead", fdoomStats.ticsAheadDistribution,
			  MAX_TICS_AHEAD);
}

