/*
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * Forward "simple" (no options) IP packets using a special cutchan.
 *
 * This is the "null IPC" or "getpid" of the router world.
 */

#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <sys/types.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>

#include <nodeos/an_impl.h>
#include <nodeos/an_domain.h>
#include <nodeos/an_thread.h>
#include <nodeos/an_demuxkey.h>
#include <nodeos/an_chan.h>
#include <nodeos/an_mutex.h>
#include <nodeos/an_pbuf.h>
#include <nodeos/an_info.h>
#include <nodeos/an_event.h>

#include <nodeos/impl/an_gprof.h>	/* XXX */

#include "timeout.h"

/*
 * Use a cuthrough channel instead of passing up to EE
 */
#define USE_CUTCHAN

/*
 * Perform minimal IP processing (checksum, TTL, re-checksum)
 */
#define DO_IPFORWARD

/*
 * Do a simple, one-segment demux that will only match IP packets
 * with no options.
 */
#define SINGLEDEMUX

/*
 * Topology:
 *
 *	SRC:srcif ---> router_srcif\
 *	                            ROUTER:router_dstif --> dstif:DST
 *	SRC2:srcif -> router_src2if/
 */
//#define SRCIF_MAC		0x00,0xa0,0xc9,0xdb,0xcc,0x62 /* falcon:eth1 */
#define SRCIF_MAC		0			      /* anything */
#define SRC2IF_MAC		0			      /* anything */

#define ROUTER_SRCIF		1
#define ROUTER_DSTIF		0
#define OUT_BW			100000

//#define ROUTER_SRC2IF		2
//#define ROUTER_DST2IF		3
//#define OUT2_BW			100000

#define DSTIF_MAC		0x00,0xd0,0xb7,0x13,0xf3,0x99
//#define DST2IF_MAC		0x00,0xd0,0xb7,0x10,0x07,0xdc /* tbpc38:eth0 */

#define RUNFOR		(20)	/* run for this many seconds */
#define MAXTHREADS	1
#define MAXBUFS		64

/*
 * For some reason the headers define ETHER_MAX_LEN as
 * (ETH_FRAME_LEN + ETH_CRC_LEN), but then don't define ETH_CRC_LEN.
 */
#ifdef __linux__
#ifndef ETH_CRC_LEN
#define ETH_CRC_LEN ETHER_CRC_LEN
#endif
#endif

static void ee_init(void *arg);
an_domaininit_func_t InitEE = { ee_init };

static void ee_die(an_event_t *event, void *arg);
static void setup_channels(void);
static an_chan_t *n1in, n1in_data, *n2out, n2out_data;
static an_demuxkey_t demux_data, *demux;
static an_pbuf_t cutbufs_data[MAXBUFS];
#ifdef USE_CUTCHAN
static an_chan_t *n1cut, n1cut_data;
#else
static void ee_recv(an_pbuf_t *pbuf, void *arg, int msize);
#ifdef DO_IPFORWARD
unsigned int dropped;
#endif
#endif
unsigned int running;

#ifdef ROUTER_SRC2IF
static an_demuxkey_t demux2_data, *demux2;
static an_pbuf_t cutbufs2_data[MAXBUFS];
static an_chan_t *n2in, n2in_data, *n3out, n3out_data;
#ifdef USE_CUTCHAN
static an_chan_t *n2cut, n2cut_data;
#endif
#endif

an_ethaddrspec_t srcifaddr = {
	{0}, {SRCIF_MAC}, ETHERTYPE_IP
};
#ifdef ROUTER_SRC2IF
an_ethaddrspec_t src2ifaddr = {
	{0}, {SRC2IF_MAC}, ETHERTYPE_IP
};
#endif
an_ethaddrspec_t dstifaddr = {
	{DSTIF_MAC}, {0}, ETHERTYPE_IP
};
#ifdef ROUTER_DST2IF
an_ethaddrspec_t dst2ifaddr = {
	{DST2IF_MAC}, {0}, ETHERTYPE_IP
};
#endif

int
ee_main(int argc, char **argv)
{
	assert(argc >= 1);

	printf("%p/%p: %s started\n",
	       an_domain_current(), an_thread_current(), argv[0]);

	setup_channels();

	printf("Ready\n");
	running = 1;

	an_gprof_start(argv[0]);

	return 0;
}

static void
setup_channels(void)
{
	struct {
		an_chanaddr_t addr;
		an_chanproto_t proto;
		an_netspec_t net;
		an_pbuf_t *bufs[MAXBUFS];
	} attr;
	an_domain_t *domain;
	static an_infoif_t info[8];
	an_infoif_t *if1info, *if2info;
#ifdef ROUTER_SRC2IF
	an_infoif_t *if3info;
#endif
#ifdef ROUTER_DST2IF
	an_infoif_t *if4info;
#endif
	an_error_t err;
	int i, nif = 8;
	an_chanrecvfunc_t recvfunc;
	int bandwidth;

	domain = an_domain_current();
	memset(&attr, 0, sizeof attr);

	/*
	 * Find addresses for desired interfaces.
	 */
	err = an_info_getifs(info, &nif);
	assert(err == 0);
	assert(nif != 0);
	if1info = if2info = 0;
#ifdef ROUTER_SRC2IF
	if3info = 0;
#endif
#ifdef ROUTER_DST2IF
	if4info = 0;
#endif
	for (i = 0; i < nif; i++) {
		if (info[i].type != ANIF_ETH)
			continue;
		if (info[i].id == ROUTER_SRCIF)
			if1info = &info[i];
		if (info[i].id == ROUTER_DSTIF)
			if2info = &info[i];
#ifdef ROUTER_SRC2IF
		if (info[i].id == ROUTER_SRC2IF)
			if3info = &info[i];
#endif
#ifdef ROUTER_DST2IF
		if (info[i].id == ROUTER_DST2IF)
			if4info = &info[i];
#endif
	}
	assert(if1info != 0);
	assert(if2info != 0);
#ifdef ROUTER_SRC2IF
	assert(if3info != 0);
#endif
#ifdef ROUTER_DST2IF
	assert(if4info != 0);
#endif

	/*
	 * Setup an event to time us out in RUNFOR seconds
	 */
	set_die_timeout(RUNFOR, ee_die);
	printf("Timer started, running for %d seconds\n", RUNFOR);
	
	/*
	 * Create a demux to match no option IPv4 packets targetted to
	 * our MAC address on ROUTER_SRCIF.
	 */
	demux = an_demuxkey_create(&demux_data, an_domain_current());
	assert(demux != 0);
#ifdef SINGLEDEMUX
	{
		struct ethip {
			u_int8_t daddr[6];
			u_int8_t saddr[6];
			u_int16_t etype;
			u_int8_t ipvhl;
		} key, mask;

		memcpy(key.daddr, if1info->addr.eth.daddr, 6);
		memset(mask.daddr, ~0, 6);
		memset(key.saddr, 0, 6);
		memset(mask.saddr, 0, 6);
		key.etype = htons(ETHERTYPE_IP);
		mask.etype = ~0;
		key.ipvhl = 0x45;
		mask.ipvhl = ~0;

		err = an_demuxkey_add(demux, 15, &key, &mask);
		assert(err == 0);
	}
#else
	err = an_demuxkey_addeth(demux, AN_DEMUXKEY_ETH_ANYSADDR, 0,
				 if1info->addr.eth.daddr, htons(ETHERTYPE_IP));
	assert(err == 0);
	err = an_demuxkey_addipv4(demux, AN_DEMUXKEY_IPV4_ALLFLAGS,
				  0, 0, 0, 0, 0);
	assert(err == 0);
#endif
	/*
	 * Create an inchan with that demux.
	 */
	memset(&attr, 0, sizeof attr);
	strcpy(attr.proto.data, if1info->pname);
	attr.proto.len = strlen(attr.proto.data);
	for (i = 0; i < MAXBUFS; i++) {
		attr.bufs[i] = an_pbuf_create(&cutbufs_data[i], AN_PBUF_ETH,
					      0, 0);
		assert(attr.bufs[i] != 0);
	}
	attr.net.maxthreads = MAXTHREADS;
	attr.net.buffers.npbufs = MAXBUFS;
	attr.net.buffers.pbufs = attr.bufs;
#ifdef USE_CUTCHAN
	recvfunc = 0;
#else
	recvfunc = ee_recv;
#endif
	n1in = an_inchan_create(&n1in_data, domain, demux,
				&attr.proto, &attr.addr, &attr.net,
				recvfunc, 0);
	assert(n1in != 0);
	printf("IN: %s: "
	       "%02x:%02x:%02x:%02x:%02x:%02x -> "
	       "%02x:%02x:%02x:%02x:%02x:%02x\n",
	       if1info->pname, 
	       srcifaddr.saddr[0], srcifaddr.saddr[1],
	       srcifaddr.saddr[2], srcifaddr.saddr[3],
	       srcifaddr.saddr[4], srcifaddr.saddr[5],
	       if1info->addr.eth.daddr[0], if1info->addr.eth.daddr[1],
	       if1info->addr.eth.daddr[2], if1info->addr.eth.daddr[3],
	       if1info->addr.eth.daddr[4], if1info->addr.eth.daddr[5]);

#ifdef ROUTER_SRC2IF
	/*
	 * Create a demux to match no option IPv4 packets targetted to
	 * our MAC address on ROUTER_SRC2IF.
	 */
	demux2 = an_demuxkey_create(&demux2_data, an_domain_current());
	assert(demux2 != 0);
	err = an_demuxkey_addeth(demux2, AN_DEMUXKEY_ETH_ANYSADDR, 0,
				 if3info->addr.eth.daddr, htons(ETHERTYPE_IP));
	assert(err == 0);
	err = an_demuxkey_addipv4(demux, AN_DEMUXKEY_IPV4_ALLFLAGS,
				  0, 0, 0, 0, 0);
	assert(err == 0);

	/*
	 * Create an inchan with that demux.
	 */
	memset(&attr, 0, sizeof attr);
	strcpy(attr.proto.data, if3info->pname);
	attr.proto.len = strlen(attr.proto.data);
	for (i = 0; i < MAXBUFS; i++) {
		attr.bufs[i] = an_pbuf_create(&cutbufs2_data[i], AN_PBUF_ETH,
					      0, 0);
		assert(attr.bufs[i] != 0);
	}
	attr.net.maxthreads = MAXTHREADS;
	attr.net.buffers.npbufs = MAXBUFS;
	attr.net.buffers.pbufs = attr.bufs;
#ifdef USE_CUTCHAN
	recvfunc = 0;
#else
	recvfunc = ee_recv;
#endif
	n2in = an_inchan_create(&n2in_data, domain, demux2,
				&attr.proto, &attr.addr, &attr.net,
				recvfunc, (void *)1);
	assert(n2in != 0);
	printf("IN2: %s: "
	       "%02x:%02x:%02x:%02x:%02x:%02x -> "
	       "%02x:%02x:%02x:%02x:%02x:%02x\n",
	       if3info->pname, 
	       src2ifaddr.saddr[0], src2ifaddr.saddr[1],
	       src2ifaddr.saddr[2], src2ifaddr.saddr[3],
	       src2ifaddr.saddr[4], src2ifaddr.saddr[5],
	       if3info->addr.eth.daddr[0], if3info->addr.eth.daddr[1],
	       if3info->addr.eth.daddr[2], if3info->addr.eth.daddr[3],
	       if3info->addr.eth.daddr[4], if3info->addr.eth.daddr[5]);
#endif

	/*
	 * Determine total bandwidth of outchans
	 */
	bandwidth = OUT_BW;
#ifdef ROUTER_SRC2IF
	bandwidth += OUT2_BW;
#endif

	/*
	 * We preconstruct an ethernet header targetted to DESTIF_MAC
	 * via ROUTER_DSTIF.  For the cutchan case this must be provided
	 * for the nodeos to send packets.  For the non-cutchan case we
	 * use it to quickly write an outgoing packet header.
	 */
	/* daddr initialized statically */
	memcpy(dstifaddr.saddr, if2info->addr.eth.daddr, 6);
	dstifaddr.type = htons(ETHERTYPE_IP);

	/*
	 * Create an outchan sending packets.
	 */
	memset(&attr, 0, sizeof attr);
	strcpy(attr.proto.data, if2info->pname);
	attr.proto.len = strlen(attr.proto.data);
	attr.addr.eth = dstifaddr;
	attr.net.bandwidth = OUT_BW;
	n2out = an_outchan_create(&n2out_data, domain, &attr.proto, &attr.addr,
				  &attr.net);
	assert(n2out != 0);
	printf("OUT1 (%d%%): %s: "
	       "%02x:%02x:%02x:%02x:%02x:%02x -> "
	       "%02x:%02x:%02x:%02x:%02x:%02x\n",
	       OUT_BW*100/bandwidth, if2info->pname,
	       dstifaddr.saddr[0], dstifaddr.saddr[1],
	       dstifaddr.saddr[2], dstifaddr.saddr[3],
	       dstifaddr.saddr[4], dstifaddr.saddr[5],
	       dstifaddr.daddr[0], dstifaddr.daddr[1],
	       dstifaddr.daddr[2], dstifaddr.daddr[3],
	       dstifaddr.daddr[4], dstifaddr.daddr[5]);

#ifdef ROUTER_SRC2IF
#ifdef ROUTER_DST2IF
	/*
	 * We preconstruct an ethernet header targetted to DESTIF_MAC
	 * via ROUTER_DSTIF.  For the cutchan case this must be provided
	 * for the nodeos to send packets.  For the non-cutchan case we
	 * use it to quickly write an outgoing packet header.
	 */
	/* daddr initialized statically */
	memcpy(dst2ifaddr.saddr, if4info->addr.eth.daddr, 6);
	dst2ifaddr.type = htons(ETHERTYPE_IP);

	/*
	 * Create a second outchan sending packets.
	 */
	memset(&attr, 0, sizeof attr);
	strcpy(attr.proto.data, if4info->pname);
	attr.proto.len = strlen(attr.proto.data);
	attr.addr.eth = dst2ifaddr;
#endif
	attr.net.bandwidth = OUT2_BW;
	n3out = an_outchan_create(&n3out_data, domain, &attr.proto, &attr.addr,
				  &attr.net);
	assert(n3out != 0);
	printf("OUT2 (%d%%): %s: "
	       "%02x:%02x:%02x:%02x:%02x:%02x -> "
	       "%02x:%02x:%02x:%02x:%02x:%02x\n",
	       OUT2_BW*100/bandwidth,
#ifdef ROUTER_DST2IF
	       if4info->pname,
	       dst2ifaddr.saddr[0], dst2ifaddr.saddr[1],
	       dst2ifaddr.saddr[2], dst2ifaddr.saddr[3],
	       dst2ifaddr.saddr[4], dst2ifaddr.saddr[5],
	       dst2ifaddr.daddr[0], dst2ifaddr.daddr[1],
	       dst2ifaddr.daddr[2], dst2ifaddr.daddr[3],
	       dst2ifaddr.daddr[4], dst2ifaddr.daddr[5]);
#else
	       if2info->pname,
	       dstifaddr.saddr[0], dstifaddr.saddr[1],
	       dstifaddr.saddr[2], dstifaddr.saddr[3],
	       dstifaddr.saddr[4], dstifaddr.saddr[5],
	       dstifaddr.daddr[0], dstifaddr.daddr[1],
	       dstifaddr.daddr[2], dstifaddr.daddr[3],
	       dstifaddr.daddr[4], dstifaddr.daddr[5]);
#endif
#endif

#ifdef USE_CUTCHAN
	/*
	 * Create a cutchan tying the two together and performing minimal
	 * IP processing.
	 */
	memset(&attr, 0, sizeof attr);
#ifdef DO_IPFORWARD
	sprintf(attr.proto.data, "eth/ip|ip/eth");
	attr.proto.len = strlen(attr.proto.data);
#endif
	n1cut = an_cutchan_create(&n1cut_data, domain, &attr.proto, n1in, n2out);
	assert(n1cut != 0);
	printf("CUT1: IN1->OUT1 ");
	if (attr.proto.len)
		printf("(%s)\n", attr.proto.data);
	else
		printf("(eth|eth)\n");

#ifdef ROUTER_SRC2IF
	n2cut = an_cutchan_create(&n2cut_data, domain, &attr.proto, n2in, n3out);
	assert(n2cut != 0);
	printf("CUT2: IN2->OUT2 ");
	if (attr.proto.len)
		printf("(%s)\n", attr.proto.data);
	else
		printf("(eth|eth)\n");
#endif
#endif
}

#ifndef USE_CUTCHAN
#include <oskit/machine/proc_reg.h>
static unsigned long long cpuspeed = 600 * 1000000;
long long firstpacket[2], lastpacket[2];

static void
ee_recv(an_pbuf_t *pbuf, void *arg, int msize)
{
	int outif = (int)arg;
	an_error_t err;
	void *pdata;
#ifdef DO_IPFORWARD
	int psize;
	struct iph {
		unsigned char vhl, tos;
		unsigned short len, id, off;
		unsigned char ttl, proto;
		unsigned short sum;
		unsigned long saddr, daddr;
	} *iph;
#endif

	if (!running)
		goto done;

	lastpacket[outif] = get_tsc();
	if (firstpacket[outif] == 0)
		firstpacket[outif] = lastpacket[outif];

	pdata = an_pbuf_data(pbuf);
#ifdef DO_IPFORWARD
	psize = an_pbuf_size(pbuf);
	if (psize < sizeof dstifaddr + sizeof *iph) {
		dropped++;
		goto done;
	}

	iph = pdata + sizeof dstifaddr;
#if 0
	printf("%p: %d: vhl=%2x, len=%d, ttl=%d, s=%#lx, d=%#lx\n",
	       n1in, psize - sizeof(dstifaddr),
	       iph->vhl, ntohs(iph->len), iph->ttl,
	       ntohl(iph->saddr), ntohl(iph->daddr));
#endif

	/* Verify the checksum */
	{
		unsigned short *usp = (unsigned short *)iph;
		unsigned long cs;

		cs = *usp++;  cs += *usp++;
		cs += *usp++; cs += *usp++;
		cs += *usp++; cs += *usp++;
		cs += *usp++; cs += *usp++;
		cs += *usp++; cs += *usp;
		cs = (cs >> 16) + (cs & 0xffff);
		cs += (cs >> 16);
		if (cs != 0xffff) {
			dropped++;
			goto done;
		}
	}

	/* Decrement the TTL */
	if (--iph->ttl == 0) {
		dropped++;
		goto done;
	}

	/* Recompute the checksum (from Scout) */
	{
		unsigned long cs = iph->sum;

		cs += htons(0x0100);
		cs += (cs >> 16);
		iph->sum = cs;
	}
#endif

	/* Rewrite the ethernet header */
#ifdef ROUTER_DST2IF
	if (outif == 1)
		memcpy(pdata, &dst2ifaddr, sizeof dst2ifaddr);
	else
#endif
	memcpy(pdata, &dstifaddr, sizeof dstifaddr);

	/* Pump it back out */
#ifdef ROUTER_SRC2IF
	if (outif == 1)
		err = an_outchan_send(n3out, pbuf);
	else
#endif
	err = an_outchan_send(n2out, pbuf);
	assert(err == 0);

 done:
	/* Return the buffer to the channel */
#ifdef ROUTER_SRC2IF
	if (outif == 1)
		an_inchan_addpbuf(n2in, pbuf);
	else
#endif
	an_inchan_addpbuf(n1in, pbuf);
}
#endif

#undef CYCLECOUNT

#ifdef CYCLECOUNT
#include <oskit/machine/proc_reg.h>

unsigned long long _cycles_[64];
unsigned int _counts_[64];
static unsigned long long _last_, _cur_;
#define COUNT(ev) { \
	_cur_ = get_tsc(); \
	_cycles_[ev] += (_cur_ - _last_); \
	_counts_[ev]++; \
	_last_ = _cur_; \
}
#define FIRSTCOUNT { \
	_last_ = get_tsc(); \
} 
#else
#define COUNT(ev)
#define FIRSTCOUNT
#endif

static void
ee_die(an_event_t *event, void *arg)
{
	an_chanstats_t n1stats;
#ifdef ROUTER_SRC2IF
	an_chanstats_t n2stats;
#endif
	an_error_t err;
	unsigned prate, brate, seconds;

	an_gprof_stop();

	running = 0;

	/*
	 * Gather all the stats before we print anything since printf is
	 * expensive.
	 */
#ifdef USE_CUTCHAN
	err = an_chan_stats(n1cut, &n1stats);
	assert(err == 0);
	an_chan_destroy(n1cut);
#else
	err = an_chan_stats(n1in, &n1stats);
	assert(err == 0);
	{
		an_chanstats_t n3stats;
		err = an_chan_stats(n2out, &n3stats);
		assert(err == 0);
		n1stats.outbufcopies = n3stats.outbufcopies;
		n1stats.droppackets += n3stats.droppackets;
		n1stats.dropnobufs += n3stats.dropnobufs;
	}
#ifdef DO_IPFORWARD
	n1stats.dropbadpacket = dropped;
#endif
	an_chan_destroy(n1in);
#endif

#ifndef USE_CUTCHAN
	if ((lastpacket[0] -= firstpacket[0]) == 0)
		lastpacket[0] = cpuspeed;
	seconds = (lastpacket[0] / cpuspeed);
	prate = (unsigned)((n1stats.inpackets * cpuspeed) / lastpacket[0]);
	brate = (unsigned)((n1stats.inbytes * cpuspeed) / lastpacket[0]);
#else
	seconds = RUNFOR;
	prate = (unsigned)(n1stats.inpackets / seconds);
	brate = (unsigned)(n1stats.inbytes / seconds);
#endif

#ifdef ROUTER_SRC2IF
#ifdef USE_CUTCHAN
	err = an_chan_stats(n2cut, &n2stats);
	assert(err == 0);
	an_chan_destroy(n2cut);
#else
	err = an_chan_stats(n2in, &n2stats);
	assert(err == 0);
	{
		an_chanstats_t n3stats;
		err = an_chan_stats(n3out, &n3stats);
		assert(err == 0);
		n2stats.outbufcopies = n3stats.outbufcopies;
		n2stats.droppackets += n3stats.droppackets;
		n2stats.dropnobufs += n3stats.dropnobufs;
	}
#ifdef DO_IPFORWARD
	n2stats.dropbadpacket = dropped;
#endif
	an_chan_destroy(n2in);
#endif

#ifndef USE_CUTCHAN
	if ((lastpacket[1] -= firstpacket[1]) == 0)
		lastpacket[1] = 1;
	seconds = (lastpacket[1] / cpuspeed);
	prate = (unsigned)((n2stats.inpackets * cpuspeed) / lastpacket[1]);
	brate = (unsigned)((n2stats.inbytes * cpuspeed) / lastpacket[1]);
#else
	seconds = RUNFOR;
	prate = (unsigned)(n2stats.inpackets / seconds);
	brate = (unsigned)(n2stats.inbytes / seconds);
#endif
#endif

	printf("All done\n");
#ifdef USE_POLLING
ani_chan_pollstats();
#endif

#ifdef CYCLECOUNT
	{
	int i;
	for (i = 0; i < 64; i++) {
		if (_counts_[i] == 0)
			continue;
		printf("%d: %d events, %u cycles/event\n",
		       i, _counts_[i], (unsigned)(_cycles_[i]/_counts_[i]));
	}
	}
#endif

	printf("Domain1: %d seconds, %qu pkts (%u pkts/sec), "
	       "%qu bytes (%u bytes/sec)\n",
	       seconds, n1stats.inpackets, prate, n1stats.inbytes, brate);
	printf("  Packets: %qu queued, %qu incopied, %qu outcopied, %qu dropped\n",
	       n1stats.inbufqueued, n1stats.inbufcopies,
	       n1stats.outbufcopies, n1stats.droppackets);
	printf("  Drops: %qu no bufs, %qu failed buf setup, %qu bad IP packet\n",
	       n1stats.dropnobufs, n1stats.dropfailedbuf,
	       n1stats.dropbadpacket);

#ifdef ROUTER_SRC2IF
	printf("Domain2: %d seconds, %d pkts (%u pkts/sec), "
	       "%d bytes (%u bytes/sec)\n",
	       seconds, n2stats.inpackets, prate, n2stats.inbytes, brate);
	printf("  Packets: %d queued, %d incopied, %d outcopied, %d dropped\n",
	       n2stats.inbufqueued, n2stats.inbufcopies,
	       n2stats.outbufcopies, n2stats.droppackets);
	printf("  Drops: %d no bufs, %d failed buf setup, %d bad IP packet\n",
	       n2stats.dropnobufs, n2stats.dropfailedbuf,
	       n2stats.dropbadpacket);
#endif

	an_domain_destroy(an_domain_current());
}

static void
ee_init(void *arg)
{
	an_initargs_t *args = arg;
	int status;

	status = ee_main(args->argc, args->argv);
}
