/*
 * network.c
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 *
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <string.h>

/* These are just needed for parsing string addresses, and the INADDR_ANY, etc, constants: */
#include <stdlib.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <assert.h>

#include "types.h"
#include "network.h"
#include "trace.h"
#include "error.h"

#define DEFAULT_REMOTE_PORT	9666

#ifndef INADDR_NONE
#define INADDR_NONE	0xffffffff
#endif

void
strtoaddrport(const char *id, Uint32 *addr, Uint16 *port)
{
	const char *p;
	
	assert(id);
	assert(addr);
	assert(port);

	/* split id at the ':' */
	p = strchr(id, ':');
	if (p == NULL)
	{
		/* No port, default is zero. */
		*addr = inet_addr(id); 
		*port = 0;
	}
	else
	{
		char addrstr[p - id + 1];
		strncpy(addrstr, id, p - id);
		addrstr[p - id] = '\0';
		*addr = inet_addr(addrstr);
		*port = htons((Uint16)atoi(p+1));
	}

	assert(*addr != INADDR_NONE); // XXX
}

void
createConnections(const char* protocolId, const char* myAddr,
		  const char* peerAddr,
		  an_chan_t* recvChan,
		  an_chanrecvfunc_t recvFunc,
		  void *recvArg,
		  an_demuxkey_t* recvDemux,
		  an_pbuf_t* recvBufs[], int recvBufCount,
		  an_chan_t* sendChan)
{
	an_chan_t *rc;
	an_chanproto_t protocol;
	an_chanaddr_t recvAddr;
	an_chanaddr_t sendAddr;
	an_netspec_t inNetSpec;

	assert(protocolId);
	assert(myAddr);
	assert(peerAddr);
	assert(recvChan);
	assert(recvFunc);
	assert(recvDemux);
	assert(recvBufs);
	assert(recvBufCount > 0);
	assert(sendChan);

	memset(&protocol, 0, sizeof protocol);
	memset(&recvAddr, 0, sizeof recvAddr);
	memset(&sendAddr, 0, sizeof sendAddr);

	/*
	 * Setup the protocol and address specs based on the
	 * protocol, id, and various global bits of info.
	 */
	if (!strcasecmp(protocolId, "udp"))
	{
		long peerIpAddr, myIpAddr;
		short peerPort, myPort;

		/* Setup protocol for UDP connections */
		strcpy(protocol.data, "socket:if/ipv4/udp");
		protocol.len = strlen(protocol.data);

		strtoaddrport(peerAddr, &peerIpAddr, &peerPort);
		if (peerPort == 0)
			peerPort = htons(DEFAULT_REMOTE_PORT);

		strtoaddrport(myAddr, &myIpAddr, &myPort);
		if (myPort == 0)
			myPort = htons(DEFAULT_REMOTE_PORT);

		TRACE(TRACE_SETUP, "Peer inchan from %#lx:%d to %#lx:%d\n",
		      ntohl(myIpAddr), ntohs(myPort),
		      ntohl(peerIpAddr), ntohs(peerPort));

		/* Setup send addr for UDP connection to peer */
		sendAddr.udp4.daddr = peerIpAddr; // already in network order
		sendAddr.udp4.dport = peerPort;   //  "
		sendAddr.udp4.saddr = myIpAddr; // htonl(INADDR_ANY); // myIpAddr;   //  "
		sendAddr.udp4.sport = myPort;     //  "

		/* Setup recv addr for UDP connection from peer */
		recvAddr.udp4.daddr = sendAddr.udp4.saddr;
		recvAddr.udp4.dport = sendAddr.udp4.sport;
		recvAddr.udp4.saddr = sendAddr.udp4.daddr;
		recvAddr.udp4.sport = sendAddr.udp4.dport;
	}
	else
	{
		FATALERROR("unsupported protocol '%s'", protocolId);
	}

	/* Empty key is fine.  Don't expect any other sort of packet. */
	an_demuxkey_create(recvDemux, an_domain_current());

	inNetSpec.maxthreads = 1;  // XXX MAKE THIS 1 OR OUT-OF-ORDER CONST
	inNetSpec.buffers.npbufs = recvBufCount;
	inNetSpec.buffers.pbufs = recvBufs;
	// inCPUSpec.threads.im.stacksize = 32 * 1024;

	rc = an_inchan_create(recvChan,
			      an_domain_current(),
			      recvDemux,
			      &protocol, &recvAddr, &inNetSpec,
			      recvFunc, recvArg);
	if (rc != recvChan)
	{
		int code;
		code = an_error_code();
		FATALERROR("an_inchan_create from %s failed: %s (%#x: %s)\n",
			   peerAddr, an_error_message(),
			   code, strerror(code));
	}

	rc = an_outchan_create(sendChan,
			       an_domain_current(),
			       &protocol,
			       &sendAddr,
			       &inNetSpec);
			       //NULL /* netspec */);
	assert(rc == sendChan); // XXX
}
