/*
 * options.c
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 *
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <stdlib.h> // atoi
#include <string.h> // strcasecmp

#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>  // inet_addr()

#include "types.h"
#include "options.h"
#include "error.h"

static void
optionsInit(OptionSpec *argSpecs, int specSize)
{
	int i;

	//memset(&options, 0, sizeof options);
	// Initilize the default options from the option spec
	for (i = 0; i < specSize; i++)
	{
		switch (argSpecs[i].type)
		{
		case OT_INT:
			memcpy(argSpecs[i].addr, &argSpecs[i].spec.intspec.defaultValue, sizeof(int));
			break;
		case OT_BOOL:
			memcpy(argSpecs[i].addr, &argSpecs[i].spec.boolspec.defaultValue, sizeof(Boolean));
			break;
		case OT_STR:
			memcpy(argSpecs[i].addr, &argSpecs[i].spec.strspec.defaultValue, sizeof(char*));
			break;
		case OT_INETADDR:
			memcpy(argSpecs[i].addr, &argSpecs[i].spec.inetaddrspec.defaultValue, sizeof(char*));
			break;
		default:
			FATALERROR("Unknown argSpec type");
		}
	}
}

int
parseArgs(an_initargs_t* args, OptionSpec *argSpecs, int specSize)
{
	int i;

	optionsInit(argSpecs, specSize);

	i = 1; // skip args[0]
	while ((i < args->argc)
	       && (args->argv[i][0] == '-')
	       && (strcmp("--", args->argv[i])) // and is NOT '--'
	       )
	{
		const char* arg = args->argv[i];
		const char* opt = args->argv[i+1];
		Boolean found = false;
		int optid;
		
		if (i == args->argc-1) 
		{
			FATALERROR("error: no option for arg.\n");
		}
		
		for (optid=0; optid < specSize; optid++)
		{
			if (!strcasecmp(arg, argSpecs[optid].name)) {
				found = true;
				switch(argSpecs[optid].type)
				{
				case OT_INT:
					{
						int val = 0;
						val = atoi(opt);
						if ((val < argSpecs[optid].spec.intspec.min)
							|| (val > argSpecs[optid].spec.intspec.max))
						{
							FATALERROR("error: %s value out of range.\n",
								arg);
						}
						*((int*)(argSpecs[optid].addr))=val;
					}
					break;
				case OT_BOOL:
					{
						Boolean val = false;
						if (!strcasecmp(opt, "yes")
							|| !strcasecmp(opt, "true")
							|| !strcasecmp(opt, "on"))
							val = true;
						*((Boolean*)(argSpecs[optid].addr))=val;
					}
					break;
				case OT_STR:
					*((const char**)(argSpecs[optid].addr))=opt;
					break;
				case OT_INETADDR:
					*((int*)(argSpecs[optid].addr))=inet_addr(opt);
					break;
				default:
					FATALERROR("Unknown argSpec type: %d\n",
						   argSpecs[optid].type);
				}
			}
		}

		if (!found)
		{
			FATALERROR("Unknown option `%s' (with argument `%s')\n",
				   arg, opt);
		}

		i+=2;
	}

	return i;
}

