/*
 * packetTypes.h
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 *
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#ifndef PACKETTYPES_HEADER
#define PACKETTYPES_HEADER

#include "types.h"
#include "tic.h"

typedef enum PacketType {
	PT_SETUP=20, 
	PT_RUNNING=21, 
	PT_SHUTDOWN=22,
	/* 
	 * Following packet types are only used with 
	 * the adoom active network nodes.
	 */
	PT_ADOOM_CONNECT=30,
	PT_ADOOM_RECOGNIZE=31, /* XXX Same struct as PT_ADOOM_PONG */
	PT_ADOOM_PING=32,
	PT_ADOOM_PONG=33,/* Same struct as PT_ADOOM_PING */
	PT_ADOOM_CLIENT=34,
	PT_RUNNING_CONGLOM=35, /* Same struct as PT_RUNNING. */
	PT_RUNNING_ADOOM=36, /* Same struct as PT_RUNNING. */
	PT_ADOOM_SHUTDOWN=37,
} PacketType;

typedef struct SetupPacket {
	PacketType packetType; /* Must be first */
	Uint64 senderIdentity; /* XXX DEBUGONLY */
	int peerCount;
	int ticCommandSize;
	/* XXX "peerHash" hash of peer info to make sure all are on same page */
} SetupPacket;

/**
 * Packet sent and received by fdoom client (i.e., players).  
 * 
 * NOTE: there are three different types of running packets that use
 * this structure: PT_RUNNING, PT_RUNNING_CONGLOM, and PT_RUNNING_ADOOM.
 *
 * The PT_RUNNING version is used when fdoom clients communicate 
 * amongst themselves and contains only the sender's ticCommands.
 * The ticCommands are contiguous starting at the absolute "startTic"
 * and including "ticCount" commands.
 *
 * The PT_RUNNING_CONGLOM packets are for packets sent from an
 * adoom node to an fdoom client.  These packets contain tics
 * for all of the fdoom nodes hidden by the adoom node.  The commands
 * are ordered by client(**), then by tic for that client.  E.g.,
 * client1's packets, then client2's packets, etc.  Each client
 * has the same number of packets, for the same tic range.
 * (**) Note that the number of clients and their order is implicitly
 * known by each fdoom client.
 *
 * The PT_RUNNING_ADOOM packets are for packets sent between
 * adoom nodes.  They are structured just like PT_RUNNING_CONGLOM
 * packets.
 *
 * XXX Need to update the above comment, it is out of date
 */
typedef struct RunningPacket {
	/* XXX can combine following four fields into a single Uint32 */
	PacketType packetType;	/* Must be first, should be 8 bits */
	Tic    retransmitFrom;	// XXX Only need bottom 8 bits
	Tic    startTic;	// XXX Only need bottom 8 bits
	Uint8  ticCount;	// XXX Only need bottom 8 bits

	Uint64 senderIdentity;	// XXX debugging only

	/* Variable length array of commands sent/recvd. */
	TicCommand ticCommand[0]; 
} RunningPacket;

#define NO_RETRANSMITS	((Uint64)(-1))

/**
 * Conglomerated tics sent over adoom<->adoom and adoom<->fdoom
 *
 * These are similar to RunningPacket's, but they contain one one tic for
 * every client that the receiving side needs to know about.
 *
 * The identities of the clients referred to in the tic list are implicitly
 * known because each server and client keeps an ordered list of all other
 * clients.  XXX Well, idealy; note that there is still a TicBody.identity
 * and that it is used
 *
 * PT_RUNNING_CONGLOM packets are sent between adoom servers and fdoom clients.
 * These contain tics for _all_ other clients on the network.
 *
 * PT_RUNNING_ADOOM packets are sent between the adoom servers themselves.
 * These contain tics for only the clients that the originating adoom server
 * was directly connected to.  Otherwise, they have the same structure as
 * PT_RUNNING_CONGLOM packets.
 */
typedef struct RunningConglomPacket {
	PacketType packetType;
	Tic retransmitFrom;
	Tic tic;
	//Uint8 ticCount;  // XXX This _should_ be used
	Uint8 startClient;  /* Index of client to which the first tic body
			     * applies */
	Uint8 clientCount;  /* Number of clients with tics in this packet */

	TicCommand ticCommand[0];
} RunningConglomPacket;

typedef RunningConglomPacket RunningADoomConglomPacket;

/*
 * Fdoom to Adoom connection request packet.
 *
 * This is sent by each fdoom client to the adoom server they are
 * trying to connect to, repeatedly until the Adoom server responds
 * with a PT_ADOOM_RECOGNIZE packet.
 *
 * This just requests a connection to the adoom server and passes some
 * state info to make sure the config isn't bogus.
 *
 * Currently this is the same structure as PT_SETUP packets, but this
 * could change later, so this is none-the-less kept as a different
 * structure.
 */
typedef SetupPacket ADoomConnectPacket;

/*
 * Adoom to Adoom ping packet.
 *
 * Sent during network setup to determine when an adoom node has come
 * up.  Rather minimal packet, but, hey, it's a ping.  If the adoom
 * to which this is sent receives it, it will reply with a
 * PT_ADOOM_PONG packet and the two nodes will commence with sending
 * PT_ADOOM_CLIENT packets containing information about their connected
 * clients.  
 *
 * Note that, because the ping triggers the sending of all clients to
 * the other adoom server, it is also used as the way to re-request client
 * information if something is found amiss (ie out of order client packet).
 */
typedef struct ADoomPingPacket {
	PacketType packetType;	
} ADoomPingPacket;

typedef ADoomPingPacket ADoomPongPacket;

typedef ADoomPongPacket ADoomRecognizePacket;

/*
 * Adoom to Adoom client information packet.
 *
 * Sent during network setup to give other Adoom nodes information about
 * connected clients.  These are sent out for all connected clients to
 * other adoom nodes immediately after a PT_PONG_PACKET is sent.  If
 * the adoom connection is already established, these are sent out when
 * a new client connects.
 *
 * Once an adoom node knows that all of the clients are either connected
 * locally or to another adoom node, the game begins.
 *
 * XXX These are not conglomerated on initial connection because this would
 * be out of the scope of the pool system.  I'm not sure if the current pool
 * system (which never has to alloc anything after initial setup), or
 * conglomerating these is faster.
 */
typedef struct ADoomClientPacket {
	PacketType packetType;
	int seqNum;     /* Used to identify dropped client packets */
	Uint64 identity;
} ADoomClientPacket;

/*
 * Union of all the FdoomPacket types.  The only important thing is 
 * that the 'packetType' field be at the same offset (0) in all
 * of the types.  This just avoids some ugly casting.
 */
typedef union FdoomPacket {
	PacketType packetType;
	SetupPacket setupPacket;
	RunningPacket runningPacket;
	RunningConglomPacket conglomPacket;
	RunningADoomConglomPacket adoomconglomPacket;
	ADoomConnectPacket connectPacket;
	ADoomPingPacket pingPacket;
	ADoomPongPacket pongPacket;
	ADoomRecognizePacket recognizePacket;
	ADoomClientPacket clientPacket;  // Not actually an Fdoom packet...
} FdoomPacket;


#endif /* PACKETTYPES_HEADER */
