/*
 * trace.c
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 *
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "trace.h"

TraceMask traceMask = TRACE_NONE;

static struct {
	const char *name;
	TraceMask mask;
	const char *desc;
} traceTable[] = {
#define D(name, str) { #name, TRACE_ ## name, str }
	D(NONE,		"Nothing."),
	D(ALL,		"Everything."),
	D(SETUP,	"Setup."),
	D(SEND,		"Packet send."),
	D(SEND_DELAY,	"Delayed packet send."),
	D(SEND_ERR,	"Errors during packet send."),
	D(RECV,		"Packet receive."),
	D(TIC,		"Tic-related stuff."),
	D(RETRAN,	"Retransmits."),
	D(GSLOCK,	"GameState lock/unlock."),
	D(CONGLOM,	"Conglomeration stuff."),
	D(KICK,		"Automatic kick detection."),
};

#define NELEMS(ARR) (sizeof(ARR)/sizeof(ARR[0]))

void
trace(const char* func, int line, TraceMask mask, const char *fmt, ...)
{
	if (mask & traceMask)
	{
		va_list ap;
	
		fprintf(stderr, "%s: ", func);
		va_start(ap, fmt);
		vfprintf(stderr, fmt, ap);
		va_end(ap);
	}
}

const char*
traceInit(const char *mask_str)
{
	int i;
	int skipped;
	const char *separators = "|,";
	const char *curr_maskstr;

	/*
	 * Note that we don't use "strtok" because it modifies its argument
	 * and we *need* to be able to handle constant strings.  We also
	 * want to avoid allocating any memory this early in the game.
	 */

	if (mask_str == NULL)
		return NULL;

	curr_maskstr = mask_str;

	/* count leading separators */
	skipped = strspn(curr_maskstr, separators);
	if (skipped >= strlen(curr_maskstr))
	{
		/* String was nothing but separators, this is an error. */
		traceMask = 0;
		return "No valid options passed to traceInit";
	}

	/* skip leading separators */
	curr_maskstr+=skipped;
	
	/* Special target 'list' lists all the defined options, and returns 0 */
	if (!strcasecmp(curr_maskstr, "list")) {
		printf("Available debug opts: \n");
		printf("  %-15s\t%16s  %s\n",
		       "Option", "Mask (hex)", "Description");
		for (i = 0; i < NELEMS(traceTable); i++)
		{
			const char* name;
			const char* desc;
			int hi, lo;

			name = traceTable[i].name;
			desc = traceTable[i].desc;
			hi = (int)(traceTable[i].mask>>32);
			lo = (int)(traceTable[i].mask);

			if (hi)
				printf("  %-15s\t%8X%08X  %s\n",
					    name, hi, lo, desc);
			else
				printf("  %-15s\t        %8X  %s\n", 
					    name, lo, desc);
		}
		return "'list' option passed to traceInit()";
	}
	

	/*OPTSTRINGLOOP:*/
	while (curr_maskstr) {
		int invertedOpt = 0;
		size_t optlen;

		/* Find out how big this option is */
		optlen = strcspn(curr_maskstr, separators);
		assert(optlen > 0);
		assert(optlen <= strlen(curr_maskstr));

		
		/* Check for inverted opts */
		if (curr_maskstr[0] == '~') 
		{
			invertedOpt = 1;
			curr_maskstr++;
			
			/* ignore the degenerate case */
			if (curr_maskstr[0] == '\0')
				continue /*OPTSTRINGLOOP*/;
		}
		
		/*TABLECHECK:*/ 
		for (i = 0; i < NELEMS(traceTable); i++)
			if (!strncasecmp(curr_maskstr, traceTable[i].name, optlen))
			{
				if (invertedOpt)
					traceMask &= ~(traceTable[i].mask);
				else 
					traceMask |= traceTable[i].mask;
				break /*TABLECHECK*/;
			}
		
		/* Return bogon element if it wasn't found in table. */
		if (i == NELEMS(traceTable)) 
		{
			return curr_maskstr;
		}
		
		/* Skip the option just parsed. */
		curr_maskstr += optlen;

		/* End of options? */
		if (strlen(curr_maskstr) == 0)
			curr_maskstr = NULL;
		else
		{
			int sepct;
			/* Count the separators following the last option. */
			sepct = strspn(curr_maskstr, separators);
			assert(sepct > 0);
			assert(sepct <= strlen(curr_maskstr));

			/* Skip the separators */
			curr_maskstr+=sepct;

			/* Is that it? */
			if (strlen(curr_maskstr) == 0)
				curr_maskstr = NULL;
		}

		/*
		 * At this point curr_maskstr either points to the
		 * the next option name, or NULL (in which case the
		 * loop will end).
		 */
	}

	return NULL;
}

