/*
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * Test basic object operations
 */
#include <stdio.h>

#include <nodeos/an_mutex.h>
#include <nodeos/an_cond.h>
#include <nodeos/an_demuxkey.h>

#define MAXOBJS	10240

static void *objtab[MAXOBJS];
static int validtests(void);
static int invalidtests(void);

int
ee_main(int argc, const char **argv)
{
	int fails = 0;

	fails = validtests();
	fails += invalidtests();

	exit(fails);
}

static int
validtests(void)
{
	an_domain_t *domain;
	an_error_t err;
	int i;
	int fails = 0;

	domain = an_domain_current();
	assert(domain != 0);

	/* Create/test/destroy a bunch of small, harmless objects */
	for (i = 0; i < MAXOBJS; i++) {
		switch (i % 3) {
		case 0:
			objtab[i] = an_mutex_create(0);
			break;
		case 1:
			objtab[i] = an_cond_create(0);
			break;
		case 2:
			objtab[i] = an_demuxkey_create(0, domain);
			break;
		}
		if (objtab[i] == 0) {
			printf("* * * could not create object%d (%s)\n",
			       i, an_error_message());
			fails++;
		}
	}
	for (i = 0; i < MAXOBJS; i++) {
		int len = 0;

		if (objtab[i] == 0)
			continue;
		switch (i % 3) {
		case 0:
			err = an_mutex_trylock(objtab[i]);
			if (err) {
				printf("* * * mutex lock failed on object%d "
				       "(%s)\n",
				       i, an_error_message());
				fails++;
			} else
				an_mutex_unlock(objtab[i]);
			break;
		case 1:
			/* XXX no error returning operations here */
			an_cond_signal(objtab[i]);
			break;
		case 2:
			err = an_demuxkey_len(objtab[i], &len);
			if (err != 0 || len != 0) {
				printf("* * * demuxkey_len failed (%s) or "
				       "returned wrong len (%d) on object%d\n",
				       an_error_message(), len, i);
				fails++;
			}
			break;
		}
	}

	while (--i >= 0) {
		switch (i % 3) {
		case 0:
			err = an_mutex_destroy(objtab[i]);
			break;
		case 1:
			err = an_cond_destroy(objtab[i]);
			break;
		case 2:
			err = an_demuxkey_destroy(objtab[i]);
			break;
		default:
			err = 1;
			break;
		}
		if (err != 0) {
			printf("* * * could not destroy object%d (%s)\n",
			       i, an_error_message());
			fails++;
		}
	}

	return fails;
}

static int
invalidtests(void)
{
	return 0;
}

static void
ee_init(void *arg)
{
	int status;
	int argc = 1;
	const char *argv[] = { "objtest" }; 
	an_domain_t *domain = an_domain_current();

	status = ee_main(argc, argv);
	an_domain_destroy(domain);
}
an_domaininit_func_t InitEE = { ee_init };
