/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <sys/types.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <nodeos/an_impl.h>
#include <nodeos/an_info.h>
#include <nodeos/an_domain.h>
#include <nodeos/an_thread.h>
#include <nodeos/an_chan.h>
#include <nodeos/an_pbuf.h>
#include <nodeos/an_event.h>

#include <nodeos/impl/an_gprof.h>	/* XXX */

#undef BLOWHARD	/* enable multiple senders, see blowhosts below */

#define RUNFOR		10	/* run for this many seconds */
#define MAXBUFS		32

#define TTCPPORT	5001	/* default ttcp port */
#define SENDSIZE	1024

static void ee_init(void *arg);
an_domaininit_func_t InitEE = { ee_init };

static void ee_sendudp(an_chan_t *outchan);
static void setup_channels(int doSend, int doRecv, int blowhostip, int runfor);
static void ee_finish(an_event_t *event, void *arg);
static void ee_die(an_event_t *event, void *arg);

static an_chan_t *udpoutchan, udpoutchan_data;
static void ee_udp(an_pbuf_t *packet, void *arg, int msize);
static void ee_ttcp(an_pbuf_t *packet, void *arg, int msize);

static an_chan_t *udpinchan, udpinchan_data, *ttcpinchan, ttcpinchan_data;
static an_pbuf_t udpbuf_data, ttcpbufs_data[MAXBUFS];
static struct {
	int len;
	char data[8];
} ttcppattern = { 8, { 040, 041, 042, 043, 044, 045, 046, 047 } };
static an_demuxkey_t ttcpdemux_data, *ttcpdemux;

static int npackets, ttcppackets, ndrops;
static volatile int chandead;
static unsigned long nbytes;

int myip = 0xc0a80203;
#ifdef BLOWHARD
struct blowhost {
	int ip;
	an_chan_t *chan;
} blowhosts[] = {
	{ 0xc0a80202, 0 },
	{ 0xc0a80204, 0 },
};
int nblowhosts = sizeof(blowhosts) / sizeof(blowhosts[0]);
#endif
// int blowhostip = 0x9b63d658;	/* moab.cs.utah.edu */
int blowhostip = 0x7F000001;	/* localhost */

static void
ee_init(void *arg)
{
	an_initargs_t *initargs;
	int blow = 0;
	int doSend, doRecv;

	/*
	 * Parse command line arguments.
	 *    socktest [blow [blowhostip]]
	 */
	initargs = (an_initargs_t*)arg;
	assert(initargs);
#ifdef BLOWHARD
	blow = 1;
	blowhostip = blowhosts[0].ip;
#else
	if (initargs->argc > 2) 
	{
		blow = atoi(initargs->argv[1]);
		sscanf(initargs->argv[2], "%i", &blowhostip);
	}
	else if (initargs->argc == 2)
	{
		blow = atoi(initargs->argv[1]);
	}
#endif

	if (blow > 0) {
		doSend=1;
		doRecv=0;
	} else if (blow < 0) {
		doSend=1;
		doRecv=1;
	} else {
		doSend=0;
		doRecv=1;
	}

	{
		an_infoipv4_t info[8];
		int i, count = 8;
		struct in_addr tmp;

		an_info_getipv4(info, &count);
		printf("Available interfaces:\n");
		for (i = 0; i < count; i++) {
			printf("if%d: ", info[i].ifid);
			tmp.s_addr = info[i].addr;
			printf("addr=%s ", inet_ntoa(tmp));
			tmp.s_addr = info[i].mask;
			printf("mask=%s\n", inet_ntoa(tmp));
		}
	}

	/*
	 * Print a banner.
	 */
	if (doSend) {
		printf("%p/%p: socktest started (sending to %d.%d.%d.%d)\n",
		       an_domain_current(), an_thread_current(),
		       (blowhostip>>24)&0xFF, (blowhostip>>16)&0xFF,
		       (blowhostip>>8)&0xFF, blowhostip&0xFF);
	}

	if (doRecv)
	{
		printf("%p/%p: socktest started (recv)\n",
		       an_domain_current(), an_thread_current());
	}

	/*
	 * Setup channels
	 */
	setup_channels(doSend, doRecv, blowhostip, RUNFOR);

	if (doSend)
	{
#ifdef BLOWHARD
		int i;
		for (i = 0; i < nblowhosts; i++) {
			printf("sending to 0x%x\n", blowhosts[i].ip);
			an_domain_startthread(an_domain_current(),
					    ee_sendudp, blowhosts[i].chan, 0);
		}
#else
		printf("sending\n");
		ee_sendudp(udpoutchan);
#endif
	}

	if (doRecv && !doSend)
	{
		printf("waiting\n");
	}

	an_gprof_start(initargs->argv[0]);
	return;
}

/*
 * doSend/doRecv is 1 or 0.  blowhostip is the host to blow packets at, if blow
 * is 1.  Runfor is the timeout to set the clock channel at.  We'll
 * close down after the clock channel fires.
 */
static void
setup_channels(int doSend, int doRecv, int ip, int runfor)
{
	struct {
		an_chanaddr_t addr;
		an_chanproto_t proto;
		an_netspec_t net;
		an_pbuf_t *bufs[MAXBUFS];
	} attr;
	an_error_t err;
	an_event_t *ev;
	an_timespec_t fire;
	an_domain_t *domain = an_domain_current();
	int i;

	/*
	 * Setup a "shutdown coming" event at RUNFOR seconds and
	 * an actual shutdown event at RUNFOR+3 seconds. 
	 */
	ev = an_event_create(0);
	assert(ev != 0);
	fire.tv_sec = RUNFOR;
	fire.tv_nsec = 0;
	err = an_event_schedule(ev, &fire, AN_EVTYPE_ONESHOT_RELATIVE,
				ee_finish, 0);
	assert(err == 0);
	ev = an_event_create(0);
	assert(ev != 0);
	fire.tv_sec = RUNFOR+3;
	fire.tv_nsec = 0;
	err = an_event_schedule(ev, &fire, AN_EVTYPE_ONESHOT_RELATIVE,
				ee_die, 0);
	assert(err == 0);
	printf("Running for %d seconds\n", RUNFOR);

	memset(&attr, 0, sizeof attr);
#ifdef BLOWHARD
	for (i = 0; i < nblowhosts; i++) {
		/*
		 * Setup out channel for UDP
		 */
		strcpy(attr.proto.data, "socket:if/ipv4/udp");
		attr.proto.len = strlen(attr.proto.data);
		attr.addr.udp4.saddr = htonl(myip);
		attr.addr.udp4.daddr = htonl(blowhosts[i].ip);
		attr.addr.udp4.dport = htons(TTCPPORT);
		blowhosts[i].chan = an_outchan_create(0, domain, &attr.proto,
						      &attr.addr, 0);
		assert(blowhosts[i].chan != 0);
		printf("%p: outchannel for 0x%x:%d\n", blowhosts[i].chan,
		       blowhosts[i].ip, TTCPPORT);
		err = 0;
	}
#else
	if (doSend)
	{
		/*
		 * Setup out channel for UDP
		 */
		strcpy(attr.proto.data, "socket:if/ipv4/udp");
		attr.proto.len = strlen(attr.proto.data);
		attr.addr.udp4.daddr = htonl(ip);
		attr.addr.udp4.dport = htons(TTCPPORT);
		udpoutchan = an_outchan_create(&udpoutchan_data, domain,
					       &attr.proto, &attr.addr, 0);
		assert(udpoutchan != 0);
		printf("%p: outchannel for port=%d\n", udpoutchan, TTCPPORT);
	}

	if (doRecv)
	{
		/*
		 * Setup in channels for UDP and UDP-based TTCP
		 */
		strcpy(attr.proto.data, "socket:if/ipv4/udp");
		attr.proto.len = strlen(attr.proto.data);
		attr.addr.udp4.daddr = htonl(INADDR_ANY);
		attr.addr.udp4.dport = htons(TTCPPORT);

		attr.bufs[0] = an_pbuf_create(&udpbuf_data, AN_PBUF_SOCK, 0, 0);
		assert(attr.bufs[0] != 0);
		attr.net.maxthreads = 10;
		attr.net.buffers.npbufs = 1;
		attr.net.buffers.pbufs = attr.bufs;
		udpinchan = an_inchan_create(&udpinchan_data, domain, 0,
					     &attr.proto, &attr.addr,
					     &attr.net, ee_udp, 0);
		if (udpinchan == 0)
			printf("WARNING! Could not create UDP socket channel.\n"
			       "         Probably configured without "
			       "--enable-chan_socket.\n");
		assert(udpinchan != 0);
		printf("%p: inchannel for TTCP (port %d) traffic\n",
		       udpinchan, TTCPPORT);

		ttcpdemux = an_demuxkey_create(&ttcpdemux_data, an_domain_current());
		assert(ttcpdemux != 0);
		err = an_demuxkey_add(ttcpdemux, ttcppattern.len,
				      ttcppattern.data, 0);
		assert(err == 0);
		
		strcpy(attr.proto.data, "socket:if/ipv4/udp");
		attr.proto.len = strlen(attr.proto.data);
		attr.addr.udp4.daddr = htonl(INADDR_ANY);
		attr.addr.udp4.dport = htons(TTCPPORT);
		for (i = 0; i < MAXBUFS; i++) {
			attr.bufs[i] = an_pbuf_create(&ttcpbufs_data[i],
						      AN_PBUF_SOCK, 0, 0);
			assert(attr.bufs[i] != 0);
		}
		attr.net.maxthreads = 10;
		attr.net.buffers.npbufs = MAXBUFS;
		attr.net.buffers.pbufs = attr.bufs;
		ttcpinchan = an_inchan_create(&ttcpinchan_data, domain, ttcpdemux,
					      &attr.proto, &attr.addr,
					      &attr.net, ee_ttcp, 0);
		assert(ttcpinchan != 0);
		printf("%p: inchannel for TTCP (port %d) "
		       "default pattern traffic\n",
		       ttcpinchan, TTCPPORT);
	}
#endif
}

#include <nodeos/an_mutex.h>
#include <nodeos/an_cond.h>

#ifdef OSKIT
#include <oskit/machine/proc_reg.h>
#else
#define get_tsc() ({ \
	long long val; \
	an_timespec_t ts; \
	an_gettime(&ts); \
	val = ts.tv_sec * 1000000 + ts.tv_nsec / 1000; \
	val * tsc_freq; \
})
#endif

unsigned long long sstamp, estamp;
unsigned tsc_freq = 600;

#define PAUSE { \
	an_timespec_t ts; \
	ts.tv_sec = 0; \
	ts.tv_nsec = 10000000; \
	an_thread_sleep(&ts); \
}

static void
ee_sendudp(an_chan_t *outchan)
{
	an_pbuf_t *packet, packet_data;
	an_error_t err;
	int i;

	assert(SENDSIZE <= NODEOS_CHAN_BUFSIZE);

	packet = an_pbuf_create(&packet_data, AN_PBUF_SOCK,
				0, NODEOS_CHAN_BUFSIZE);
	assert(packet != 0);
	memset(an_pbuf_data(packet), 0x69, SENDSIZE);

	/* XXX so we can talk to ttcp */
	err = an_pbuf_setsize(packet, 4, 0);
	assert(err == 0);
	for (i = 0; i < 500; i++) {
		err = an_outchan_send(outchan, packet);
		if (err == 0)
			break;
	}
	err = an_pbuf_setsize(packet, SENDSIZE, 0);
	assert(err == 0);
	PAUSE;

	if (sstamp == 0)
		sstamp = estamp = get_tsc();
	while (chandead == 0) {
		err = an_outchan_send(outchan, packet);
		if (err == 0) {
			npackets++;
			nbytes += SENDSIZE;
			estamp = get_tsc();
		} else {
			ndrops++;
			PAUSE;
		}
	}
	PAUSE;

	/* XXX EOF for ttcp */
	err = an_pbuf_setsize(packet, 4, 0);
	assert(err == 0);
	for (i = 0; i < 500; i++) {
		err = an_outchan_send(outchan, packet);
		if (err == 0)
			break;
	}
}

static void
ee_udp(an_pbuf_t *pbuf, void *arg, int msize)
{
	int psize = an_pbuf_size(pbuf);

	if (chandead)
		return;

	if (npackets == 0)
		sstamp = get_tsc();

	npackets++;
	nbytes += psize;
	an_inchan_addpbuf(udpinchan, pbuf);

	estamp = get_tsc();
}

static void
ee_ttcp(an_pbuf_t *pbuf, void *arg, int msize)
{
	int psize = an_pbuf_size(pbuf);

	if (chandead)
		return;

	if (npackets == 0)
		sstamp = get_tsc();

	npackets++;
	ttcppackets++;
	nbytes += psize;
	an_inchan_addpbuf(ttcpinchan, pbuf);

	estamp = get_tsc();
}

static void
ee_finish(an_event_t *event, void *arg)
{
	chandead = 1;
#if 0
	an_gprof_stop();
#endif
}

static void
ee_die(an_event_t *event, void *arg)
{
	if (sstamp == estamp)
		printf("Done.  Probably didn't get/send any packets at all.\n");
	else {
		unsigned long long usec = (estamp - sstamp) / tsc_freq;
		printf("Done: %lu bytes, %d packets "
		       "in %u.%06u sec (%u bytes/sec)\n",
		       nbytes, npackets,
		       (unsigned)(usec / 1000000), (unsigned)(usec % 1000000),
		       (unsigned)(((unsigned long long)nbytes*1000000) / usec));
		if (ndrops)
			printf(" * * * %d packets dropped\n", ndrops);
	}
	if (ttcppackets)
		printf("      %d default TTCP packets\n", ttcppackets);
	printf("socktest: PASS\n");

	an_domain_destroy(an_domain_current());
}

