/*
 * stat-listen.c
 *
 * Copyright (c) 1999-2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 * 
 * This program is not built by the Moab install/build files.  It is not
 * to be compiled against Moab, but against standard unix header files.
 * Build with:
 *    gcc -o stat-listen stat-listen.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "../nodeos/impl/an_stats.h"
#include "stat.h"

//#define NOGRAPH

#define FATAL_ERROR(msg) fatal_error(__FILE__, __FUNCTION__, __LINE__, msg)

#define GRAPHPROG	"./rtGraph"	/* name of the graphing binary */
#define GRAPHDPS	2		/* datapoints/sec to graph */

int verbose = 0;
unsigned cpuspeed = 600000000;
unsigned netspeed = 100000000;
unsigned memsize  = (8*1024*1024);

#define cpucap(v) (unsigned)((v) > cpuspeed ? cpuspeed : (v))
#define netcap(v) (unsigned)((v) > netspeed ? netspeed : (v))
#define memcap(v) (unsigned)((v) > memsize ? memsize : (v))

#define cpuscale(v) (cpucap(v) / 1000000)
#define netscale(v) (netcap(v) / 1000000)
#define memscale(v) (memcap(v) / 1024)

FILE *cpufd, *netfd, *memfd;

struct StatPort
{
	int fd;
};

static void setup_recieve_port(struct StatPort* sp, struct in_addr myAddr, short myPort);
static void setup_stat_output(void);
static void handle_stat_packets(struct StatPort* sp);
static void fatal_error(const char* file, const char* func, int line, const char* msg);
static void print_stat(char* buf, int size);
static void diff_time_stats(an_timespec_t *new, an_timespec_t *old,
			    an_timespec_t *diff);
static void add_time_stats(an_timespec_t *stamp, an_timespec_t *totstamp);
static void diff_domain_stats(an_domainstats_t *new, an_domainstats_t *old,
			    an_domainstats_t *diff);
static void add_domain_stats(an_domainstats_t *stats, an_domainstats_t *totstats);
static void print_domain_stats(an_domainstats_t *stats, int verbose,
			     const char *str, int interval);
static void diff_chan_stats(an_chanstats_t *new, an_chanstats_t *old,
			    an_chanstats_t *diff);
static void add_chan_stats(an_chanstats_t *stats, an_chanstats_t *totstats);
static void print_chan_stats(an_chanstats_t *stats, int verbose,
			     const char *str, int interval);
static void print_mpool_stats(an_mempoolstats_t *stats, int verbose,
			      const char *str);

int
main(int argc, char **argv)
{
	struct StatPort statPort;
	struct in_addr myAddr; 
	int rc;
	char *myAddrStr = "155.99.214.74"; /* XXX PLASTIC */
	short myPort = htons(STAT_RECV_PORT);

	if (argc == 2)
	{
		myAddrStr = argv[1];
	}
	else if (argc == 3)
	{
		myAddrStr = argv[1];
		myPort = htons(atoi(argv[2]));
	}
	
	rc = inet_aton(myAddrStr, &myAddr); /* XXX PLASTIC */
	assert(rc == 1);

	printf("Opening UDP port %s/%d\n", myAddrStr, ntohs(myPort));
	setup_recieve_port(&statPort, myAddr, myPort);
	setup_stat_output();
	handle_stat_packets(&statPort);

	return 0;
}

static void
setup_recieve_port(struct StatPort *sp, struct in_addr myaddr, short myport)
{
	struct sockaddr_in sender;
	int rc;

	sp->fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (sp->fd < 0)
		FATAL_ERROR("creating socket");

	sender.sin_len = sizeof(struct sockaddr_in);
	sender.sin_family = AF_INET;
	sender.sin_addr.s_addr = htonl(INADDR_ANY);
	sender.sin_port = myport;

	rc = 1;
	rc = setsockopt(sp->fd, SOL_SOCKET, SO_REUSEADDR, &rc, sizeof rc);
	if (rc != 0)
		FATAL_ERROR("SO_REUSEADDR");

	rc = bind(sp->fd, (struct sockaddr*)&sender, sizeof(struct sockaddr_in));
	if (rc != 0)
		FATAL_ERROR("binding socket");
}

static void
setup_stat_output(void)
{
	char cmdbuf[1024];
		
	assert(GRAPHDPS <= 1000);

#ifdef NOGRAPH
	cpufd = netfd = memfd = stdout;
#else
	snprintf(cmdbuf, sizeof(cmdbuf),
		 "%s -title 'CPU Usage (Mcycles/s)'"
		 " -legend0 'max'     -color0 black"
		 " -legend1 forwarder -color1 blue"
		 " -legend2 'net hog' -color2 red"
		 " -legend3 'cpu hog' -color3 deeppink"
		 " -style blob"
		 " -max %d",
		 GRAPHPROG, (int)(cpuscale(cpuspeed)*1.10));

	cpufd = popen(cmdbuf, "w");
	if (cpufd == NULL)
		FATAL_ERROR("CPU graph start failed");
	setlinebuf(cpufd);

	snprintf(cmdbuf, sizeof(cmdbuf),
		 "%s -title 'Network Bandwidth Usage (Mbits/s)'"
		 " -legend0 'max'     -color0 black"
		 " -legend1 forwarder -color1 blue"
 		 " -legend2 'net hog' -color2 red"
		 " -style blob"
		 " -max %d",
		 GRAPHPROG, (int)(netscale(netspeed)*1.10));

	netfd = popen(cmdbuf, "w");
	if (netfd == NULL)
		FATAL_ERROR("Net graph start failed");
	setlinebuf(netfd);
	
	snprintf(cmdbuf, sizeof(cmdbuf),
		 "%s -title 'Memory Usage (KBytes)'"
		 " -legend0 'max'     -color0 black"
		 " -legend1 forwarder -color1 blue"
 		 " -legend2 'mem hog' -color2 magenta4"
		 " -style blob"
		 " -max %d",
		 GRAPHPROG, (int)(memscale(memsize)*1.10));

	memfd = popen(cmdbuf, "w");
	if (memfd == NULL)
		FATAL_ERROR("Net graph start failed");
	setlinebuf(memfd);
#endif

	printf("Formatted stat output is initalized.\n");
}

static void
handle_stat_packets(struct StatPort *sp)
{
	char buf[4096];
	struct sockaddr_in from;
	int fromlen;

	fromlen = sizeof(struct sockaddr_in);

	while(1)
	{
		int rc;
		
		rc = recvfrom(sp->fd, buf, sizeof(buf), MSG_WAITALL,
			      (struct sockaddr*)&from, &fromlen);
		if (rc == -1)
			FATAL_ERROR("-1 returned from recvfrom");
		
		assert(rc > 0);
		assert(rc < sizeof(buf));
		print_stat(buf, rc);
	}
}

/*
 * Compute a <whatever>/sec rate based on the given value, elapsed interval
 * and cap value.  Assumes no interval is greater than 47 days.
 */
static inline unsigned int
compute_rate(an_statval_t val, unsigned int ms, unsigned int capval)
{
	unsigned int rate;

	if (ms == 0)
		ms = 1;
	rate = (unsigned int)((val * 1000) / ms);
	return (rate > capval) ? capval : rate;
}

static void
print_stat(char *buf, int size)
{
	stat_packet_t *statpkt = (stat_packet_t*)buf;
	an_domainstats_t cpustats, cputotal;
	an_chanstats_t netstats, nettotal;
	an_timespec_t stamp, timetotal;
	static an_domainstats_t lastdatacpu, lastnhcpu, lastchcpu;
	static an_chanstats_t lastdatanet, lastnhnet;
	static an_timespec_t laststamp;
	static int called;

	if (size < sizeof(stat_packet_t)) {
		printf("Malformed packet.  Size is %d (expecting at least %d)\n",
		       size, sizeof(stat_packet_t));
		return;
	}

	if (statpkt->timestamp.tv_sec == -1) {
		printf("Done!\n");
#ifndef NOGRAPH
		pclose(cpufd);
		pclose(netfd);
		pclose(memfd);
#endif
		exit(0);
	}

	statpkt->version = ntohl(statpkt->version);
	statpkt->timestamp.tv_sec = ntohl(statpkt->timestamp.tv_sec);
	statpkt->timestamp.tv_nsec = ntohl(statpkt->timestamp.tv_nsec);
	if (!called) {
		laststamp = statpkt->timestamp;
		lastdatacpu = statpkt->datadomain;
		lastnhcpu = statpkt->nethogdomain;
		lastchcpu = statpkt->cpuhogdomain;
		lastdatanet = statpkt->datachan;
		lastnhnet = statpkt->nethogchan;
		called = 1;
	}

	if (verbose) {
		if (statpkt->version == STATS_VERSION_MOAB)
			printf("Moab stats: ");
		else if (statpkt->version == STATS_VERSION_JANOS)
			printf("Jnodeos stats: ");

#ifdef RAW
		printf("\n\t");
		printf("ts=%d.%03d ", statpkt->timestamp.tv_sec,
		       statpkt->timestamp.tv_nsec / 1000000);
		printf("\n\t");
		print_domain_stats(&statpkt->datadomain, 0, "Forward", 0);
		printf("\n\t");
		print_domain_stats(&statpkt->nethogdomain, 0, "NetHog", 0);
		printf("\n\t");
		print_domain_stats(&statpkt->cpuhogdomain, 0, "CpuHog", 0);
		printf("\n");

		printf("\t");
		print_chan_stats(&statpkt->datachan, 0, "Forward", 0);
		printf("\n\t");
		print_chan_stats(&statpkt->nethogchan, 0, "NetHog", 0);
		printf("\n");

		printf("\t");
		print_mpool_stats(&statpkt->mempool, 0, "Mempool", 0);
		printf("\n");
#else
		printf("\n");
		diff_time_stats(&statpkt->timestamp, &laststamp, &stamp);
		printf("ts=%d.%03d ", (int)stamp.tv_sec, (int)(stamp.tv_nsec / 1000000));
		laststamp = statpkt->timestamp;
		printf("\n");
		diff_domain_stats(&statpkt->datadomain, &lastdatacpu, &cpustats);
		print_domain_stats(&cpustats, 0, "Forward", 0);
		lastdatacpu = statpkt->datadomain;
		diff_domain_stats(&statpkt->nethogdomain, &lastnhcpu, &cpustats);
		print_domain_stats(&cpustats, 0, "NetHog", 0);
		lastnhcpu = statpkt->nethogdomain;
		diff_domain_stats(&statpkt->cpuhogdomain, &lastchcpu, &cpustats);
		print_domain_stats(&cpustats, 0, "CpuHog", 0);
		lastchcpu = statpkt->cpuhogdomain;

		diff_chan_stats(&statpkt->datachan, &lastdatanet, &netstats);
		print_chan_stats(&netstats, 0, "Forward", 0);
		lastdatanet = statpkt->datachan;
		diff_chan_stats(&statpkt->nethogchan, &lastnhnet, &netstats);
		print_chan_stats(&netstats, 0, "NetHog", 0);
		lastnhnet = statpkt->nethogchan;

		print_mpool_stats(&statpkt->mempool, 0, "Mempool");
		printf("\n");
#endif
	} else {
		int ms, i, intervals;
		unsigned int cpud, cpunh, cpuch, netd, netnh;
		unsigned int memd, memmh;

		memset(&cputotal, 0, sizeof cputotal);

		/*
		 * Diff and sum the stats
		 */
		diff_time_stats(&statpkt->timestamp, &laststamp, &stamp);
		add_time_stats(&stamp, &timetotal);
		laststamp = statpkt->timestamp;
		statpkt->timestamp = stamp;

		diff_domain_stats(&statpkt->datadomain, &lastdatacpu, &cpustats);
		add_domain_stats(&cpustats, &cputotal);
		lastdatacpu = statpkt->datadomain;
		statpkt->datadomain = cpustats;

		diff_domain_stats(&statpkt->nethogdomain, &lastnhcpu, &cpustats);
		add_domain_stats(&cpustats, &cputotal);
		lastnhcpu = statpkt->nethogdomain;
		statpkt->nethogdomain = cpustats;

		diff_domain_stats(&statpkt->cpuhogdomain, &lastchcpu, &cpustats);
		add_domain_stats(&cpustats, &cputotal);
		lastchcpu = statpkt->cpuhogdomain;
		statpkt->cpuhogdomain = cpustats;

		diff_chan_stats(&statpkt->datachan, &lastdatanet, &netstats);
		add_chan_stats(&netstats, &nettotal);
		lastdatanet = statpkt->datachan;
		statpkt->datachan = netstats;

		diff_chan_stats(&statpkt->nethogchan, &lastnhnet, &netstats);
		add_chan_stats(&netstats, &nettotal);
		lastnhnet = statpkt->nethogchan;
		statpkt->nethogchan = netstats;

		/*
		 * Display the averaged data as many times as necessary
		 * to fill covered graphing intervals.
		 *
		 * XXX this is a hack to "hide" delayed stats packets.
		 */
		assert(statpkt->timestamp.tv_sec < 4000000);
		ms = statpkt->timestamp.tv_sec * 1000;
		ms += statpkt->timestamp.tv_nsec / 1000000;
		intervals = ms / (1000 / GRAPHDPS);

		cpud = compute_rate(statpkt->datadomain.cpuused, ms, cpuspeed);
		cpunh = compute_rate(statpkt->nethogdomain.cpuused, ms, cpuspeed);
		cpuch = compute_rate(statpkt->cpuhogdomain.cpuused, ms, cpuspeed);
		netd = compute_rate(statpkt->datachan.outbytes*8, ms, netspeed);
		netnh = compute_rate(statpkt->nethogchan.outbytes*8, ms, netspeed);
		if (statpkt->version == STATS_VERSION_JANOS) {
			memd = compute_rate(statpkt->datadomain.memused, 1000,
					    memsize);
			memmh = compute_rate(statpkt->memhogdomain.memused,
					     1000, memsize);
		} else {
			memd = compute_rate(statpkt->mempool.exchunks +
					    statpkt->mempool.imchunks, 1000,
					    memsize);
			memmh = 0;
		}

		for (i = 0; i < intervals; i++) {
#ifdef NOGRAPH
			printf("Time: %d\n", i * (1000 / GRAPHDPS));
			printf("CPU: ");
#endif
			fprintf(cpufd, "%u %u %u %u\n",
				cpuscale(cpuspeed),
				cpuscale(cpud), cpuscale(cpunh),
				cpuscale(cpuch));
#ifdef NOGRAPH
			printf("Net: ");
#endif
			fprintf(netfd, "%u %u %u\n",
				netscale(netspeed),
				netscale(netd), netscale(netnh));
#ifdef NOGRAPH
			printf("Mem: ");
#endif
			fprintf(memfd, "%u %u %u\n",
				memscale(memsize),
				memscale(memd), memscale(memmh));
		}
	}
}

static void
fatal_error(const char* file, const char* func, int line, const char* msg)
{
	int e = errno;
	fprintf(stderr, "%s:%s:%d: %s (%d: %s)\n",
		file, func, line, msg, e, strerror(e));
	abort();
}

static void
diff_time_stats(an_timespec_t *new, an_timespec_t *old, an_timespec_t *diff)
{
	diff->tv_sec = new->tv_sec - old->tv_sec;
	if (new->tv_nsec < old->tv_nsec) {
		diff->tv_sec--;
		diff->tv_nsec += 1000000000;
	}
	diff->tv_nsec = new->tv_nsec - old->tv_nsec;
}

static void
add_time_stats(an_timespec_t *stamp, an_timespec_t *totstamp)
{
	totstamp->tv_sec += stamp->tv_sec;
	totstamp->tv_nsec += stamp->tv_nsec;
	if (totstamp->tv_nsec >= 1000000000) {
		totstamp->tv_sec++;
		totstamp->tv_nsec -= 1000000000;
	}
}

/*
 * XXX ripped from lib/{domain,chan}stats.c
 */
static void
diff_domain_stats(an_domainstats_t *new, an_domainstats_t *old, an_domainstats_t *diff)
{
	diff->curthreads = new->curthreads;
	diff->maxthreads = new->maxthreads;
	diff->cpuunits = new->cpuunits;
	diff->activations = new->activations - old->activations;
	diff->cpuused = new->cpuused - old->cpuused;
	diff->memused = new->memused - old->memused;
}

static void
add_domain_stats(an_domainstats_t *stats, an_domainstats_t *totstats)
{
	totstats->curthreads = stats->curthreads;
	totstats->maxthreads = stats->maxthreads;
	totstats->cpuunits = stats->cpuunits;
	totstats->activations += stats->activations;
	totstats->cpuused += stats->cpuused;
	totstats->memused += stats->memused;
}

static void
print_domain_stats(an_domainstats_t *stats, int verbose, const char *str,
		 int interval)
{
	if (stats->cpuused == 0) {
		if (verbose)
			printf("%s: idle\n", str);
		return;
	}

	if (stats->cpuunits == NODEOS_CPU_CYCLES)
		printf("%s: %qu cycles of CPU", str, stats->cpuused);
	else
		printf("%s: %qu microseconds of CPU", str, stats->cpuused);
	if (interval > 0)
		printf(" (%u cycles/sec)",
		       (unsigned)(stats->cpuused / interval));
	printf("\n");

	if (verbose)
		printf("  %d thread activations (%d current, %d max)\n",
		       (int)stats->activations, (int)stats->curthreads,
		       (int)stats->maxthreads);
}

static void
diff_chan_stats(an_chanstats_t *new, an_chanstats_t *old, an_chanstats_t *diff)
{
	diff->inpackets = new->inpackets - old->inpackets;
	diff->inbytes = new->inbytes - old->inbytes;
	diff->outpackets = new->outpackets - old->outpackets;
	diff->outbytes = new->outbytes - old->outbytes;
	diff->droppackets = new->droppackets - old->droppackets;
	diff->dropnobufs = new->dropnobufs - old->dropnobufs;
	diff->dropfailedbuf = new->dropfailedbuf - old->dropfailedbuf;
	diff->dropbadpacket = new->dropbadpacket - old->dropbadpacket;
	diff->dropbytes = new->dropbytes - old->dropbytes;
	diff->inbufqueued = new->inbufqueued - old->inbufqueued;
	diff->inbufcopies = new->inbufcopies - old->inbufcopies;
	diff->outbufcopies = new->outbufcopies - old->outbufcopies;
}

static void
add_chan_stats(an_chanstats_t *stats, an_chanstats_t *totstats)
{
	totstats->inpackets += stats->inpackets;
	totstats->inbytes += stats->inbytes;
	totstats->outpackets += stats->outpackets;
	totstats->outbytes += stats->outbytes;
	totstats->droppackets += stats->droppackets;
	totstats->dropnobufs += stats->dropnobufs;
	totstats->dropfailedbuf += stats->dropfailedbuf;
	totstats->dropbadpacket += stats->dropbadpacket;
	totstats->dropbytes += stats->dropbytes;
	totstats->inbufqueued += stats->inbufqueued;
	totstats->inbufcopies += stats->inbufcopies;
	totstats->outbufcopies += stats->outbufcopies;
}

static void
print_chan_stats(an_chanstats_t *stats, int verbose, const char *str,
		 int interval)
{
	if (stats->inpackets == 0 && stats->outpackets == 0 &&
	    stats->droppackets == 0) {
		if (verbose)
			printf("%s: no packets\n", str);
		return;
	}

	if (stats->inpackets > 0) {
		printf("%s: %qu pkts", str, stats->inpackets);
		if (interval > 0)
			printf(" (%u pkts/s)",
			       (unsigned)(stats->inpackets/interval));
		printf(", %qu bytes", stats->inbytes);
		if (interval > 0)
			printf("(%u bytes/s)",
			       (unsigned)(stats->inbytes/interval));
		printf(" received\n");
		if (verbose)
			printf("  %qu pkts copied\n", stats->inbufcopies);
	}

	if (stats->outpackets > 0) {
		printf("%s: %qu pkts", str, stats->outpackets);
		if (interval > 0)
			printf(" (%u pkts/s)",
			       (unsigned)(stats->outpackets/interval));
		printf(", %qu bytes", stats->outbytes);
		if (interval > 0)
			printf("(%u bytes/s)",
			       (unsigned)(stats->outbytes/interval));
		printf(" sent\n");
		if (verbose)
			printf("  %qu pkts copied\n", stats->outbufcopies);
	}

	if (stats->droppackets > 0) {
		printf("%s: %qu pkts", str, stats->droppackets);
		if (interval > 0)
			printf(" (%u pkts/s)",
			       (unsigned)(stats->droppackets/interval));
		printf(", %qu bytes", stats->dropbytes);
		if (interval > 0)
			printf("(%u bytes/s)",
			       (unsigned)(stats->dropbytes/interval));
		printf(" dropped\n");
		if (verbose)
			printf("  %qu no bufs, %qu failed buffer setup\n",
			       stats->dropnobufs, stats->dropfailedbuf);
	}
}

static void
print_mpool_stats(an_mempoolstats_t *stats, int verbose, const char *str)
{
	if (stats->exchunks + stats->imchunks == 0) {
		if (verbose)
			printf("%s: no memory\n", str);
		return;
	}

	printf("%s: %qu domains using %qu bytes of memory\n",
	       str, stats->numdomains,
	       (stats->exchunks+stats->imchunks) * stats->chunksize);

	if (verbose)
		printf("  %qu explicit, %qu implicit\n",
		       stats->exchunks, stats->imchunks);
}
