/*
 * Copyright (c) 1999-2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * Test various thread functions.
 */

/*
 * Thread interrupt test option TRYSETJMP.
 *
 * Define to make sure longjmp'ing out of thread_interrupt handler works.
 * This is a dubious test since the NodeOS spec says nothing about
 * setjmp/longjmp.  However, currently it should work in Moab.
 */
#define TRYSETJMP

/*
 * Thread interrupt test option CHECKFAULTS.
 *
 * Define to tet interrupt handler invocation on exceptions.
 * Undefine if for some reason you cannot come up with a FAULT_ME
 * macro (below) that works.
 */
#define CHECKFAULTS

/*
 * Macro to cause a fault (signal)
 */
#ifdef CHECKFAULTS
#define FAULT_ME	*(unsigned int *)0xFFFFFFFF = 1
#endif

#include <assert.h>
#include <stdio.h>
#ifdef TRYSETJMP
#include <setjmp.h>
#endif

#include <nodeos/an_thread.h>
#include <nodeos/an_mempool.h>
#include <nodeos/an_domain.h>
#include <nodeos/an_mutex.h>
#include <nodeos/an_cond.h>

static int errors;

/*
 * We need assert to always be defined and to note an error without dying.
 */
void failed(int ln, const char *str)
{
	printf("* * * threadtest: failed assertion on line %d: %s\n", ln, str);
	errors++;
}
#undef assert
#define	assert(e)	((e) ? (void)0 : failed(__LINE__, #e))

static void ee_init(void *arg);
static void dotests(void);

an_domaininit_func_t InitEE = { ee_init };

int
ee_main(int argc, const char **argv)
{
#if defined(OSKIT) && defined(GPROF)
	oskit_start_gprof();
#endif
	dotests();
#if defined(OSKIT) && defined(GPROF)
	oskit_stop_gprof();
#endif
	printf("threadtest: %s\n", errors ? "FAIL" : "PASS");
	return 0;
}

static void
ee_init(void *arg)
{
	int status;
	int argc = 1;
	const char *argv[] = { "NullEE" }; 
	an_domain_t *domain = an_domain_current();

	status = ee_main(argc, argv);
	an_domain_destroy(domain);
}

static void testcurrent(void);
static void testcreatedestroy(void);
static void teststackinfo(void);
static void testdata(void);
static void testyield(void);
static void testsleep(void);
static void testinterrupt(void);

static void dotests(void)
{
	printf("test an_thread_current()\n");
	testcurrent();
	printf("test an_thread_create/destroy()\n");
	testcreatedestroy();
	printf("test an_thread_stackinfo()\n");
	teststackinfo();
	printf("test an_thread_getdata/setdata()\n");
	testdata();
	printf("test an_thread_yield()\n");
	testyield();
	printf("test an_thread_interrupts\n");
	testinterrupt();
	printf("test an_thread_sleep()\n");
	testsleep();

	printf("All tests complete.\n");
}

static void testcurrent(void)
{
	an_thread_t *th;

	th = an_thread_current();
	assert(th);
}

static void testcreatedestroy(void)
{
	an_thread_t newth;
	int stacksize = 20 * 1024; // OSKit pthread min stack size
	void* stack;
	an_thread_t *th;
	an_error_t err;

	stack = an_mempool_alloc(an_mempool_current(), stacksize);
	th = an_thread_create(&newth, stack, stacksize);
	assert(th);
	assert(th == &newth);

	err = an_thread_destroy(th);
	assert(err == 0);
	
	an_mempool_free(an_mempool_current(), stack, stacksize);
}

static void teststackinfo(void)
{
	void *stackbase;
	int stacksize;
	void *stackptr;
	an_error_t err;
	void *fp = __builtin_frame_address(0);

	err = an_thread_stackinfo(an_thread_current(),
				  &stackbase,
				  &stacksize,
				  &stackptr);
	if (err == ENODEV) {
		/* Running in unix mode, there is no support for stackinfo(). */
		return;
	}

	assert(err == 0); /* XXX */

	/*     |***********|--ACTIVE-STACK---|          */
	/* stackbase    stackptr    stackbase+stacksize */

	assert(fp > stackbase);
	assert(fp < (stackbase + stacksize));
	assert(fp >= stackptr);
}

static void testdata(void)
{
	int data;
	int data2;
	void *d;

	d = an_thread_getdata();
	assert(d == NULL);

	an_thread_setdata(&data);
	d = an_thread_getdata();
	assert(d = &data);

	an_thread_setdata(&data2);
	d = an_thread_getdata();
	assert(d = &data2);
}

static void testyield(void)
{
	an_thread_yield();
}

static void testsleep(void)
{
	int sleepdelay = 2;

	an_timespec_t start, end, wait;

	an_gettime(&start);

	wait.tv_sec = sleepdelay;
	wait.tv_nsec = 0;
	an_thread_sleep(&wait);

	an_gettime(&end);

	assert(end.tv_sec > start.tv_sec);
	assert((end.tv_sec - start.tv_sec) >= sleepdelay);
}

/*
 * Thread interrupt testing stuff.
 * Mostly taken from domaintest.
 */
static an_mutex_t *threadlock;
static an_cond_t *threadcond;
static an_cond_t *forever;
volatile int threadstarted;

static void
ihandler(int itype, void *info)
{
	an_mutex_lock(threadlock);
	printf("  ");
	switch (threadstarted) {
	case 1:
	case 2:
		printf("SPIN"); break;
	case 3:
	case 4:
		printf("SYSCALL"); break;
	case 7:
	case 8:
		printf("FAULT"); break;
	default:
		printf("????"); break;
	}
	printf(" thread %p interrupted (%s), ",
	       an_thread_current(),
	       itype == AN_THREAD_INTR ? "INTR" : "TRAP");
#ifdef TRYSETJMP
	{
		sigjmp_buf *sjb = an_thread_getdata();

		if (sjb) {
			printf("throw...");
			an_thread_unblockintr();
			siglongjmp(*sjb, 1);
		}
	}
#endif
	printf("exiting\n");

	threadstarted = 0;
	an_cond_signal(threadcond);
	an_mutex_unlock(threadlock);
	an_thread_exit();
}

static void
spin(void *arg)
{
#ifdef TRYSETJMP
	sigjmp_buf sjb;
#endif

	an_mutex_lock(threadlock);
	printf("  SPIN thread %p started...\n", an_thread_current());
	an_thread_sethandler(ihandler);
	threadstarted = 1;
#ifdef TRYSETJMP
	if (sigsetjmp(sjb, 0) != 0) {
		printf("catch, looping again...\n");
		an_thread_setdata(0);
		threadstarted = 2;
	} else {
		an_thread_setdata(&sjb);
	}
#endif
	an_cond_signal(threadcond);
	an_mutex_unlock(threadlock);

	while (1)
		;
}

static void
syscall(void *arg)
{
#ifdef TRYSETJMP
	sigjmp_buf sjb;
#endif

	an_mutex_lock(threadlock);
	printf("  SYSCALL thread %p started...\n", an_thread_current());
	an_thread_sethandler(ihandler);
	threadstarted = 3;
#ifdef TRYSETJMP
	if (sigsetjmp(sjb, 0) != 0) {
		printf("catch, looping again...\n");
		an_thread_setdata(0);
		threadstarted = 4;
	} else {
		an_thread_setdata(&sjb);
	}
#endif
	an_cond_signal(threadcond);
	an_mutex_unlock(threadlock);

	while (1)
		(void) an_domain_current();
}

static void
fault(void *arg)
{
#ifdef TRYSETJMP
	sigjmp_buf sjb;
#endif

	an_mutex_lock(threadlock);
	printf("  FAULT thread %p started...\n", an_thread_current());
	an_thread_sethandler(ihandler);
	threadstarted = 7;
#ifdef TRYSETJMP
	if (sigsetjmp(sjb, 0) != 0) {
		printf("catch, fault again...\n");
		an_thread_setdata(0);
		threadstarted = 8;
	} else {
		an_thread_setdata(&sjb);
	}
#endif
	an_cond_signal(threadcond);
	an_mutex_unlock(threadlock);

	FAULT_ME;
	printf("  ERROR: FAULT thread after fault instruction!\n");
}

static void
ihandler2(int itype, void *info)
{
	switch (threadstarted) {
	case 5:
	case 6:
		printf("  BLOCKED"); break;
	default:
		printf("  ????"); break;
	}
	printf(" thread %p interrupted (%s), ",
	       an_thread_current(),
	       itype == AN_THREAD_INTR ? "INTR" : "TRAP");
	if (an_mutex_trylock(threadlock) == 0)
		printf("  ERROR: "
		       "BLOCKED thread inthandler, mutex not locked!\n");
#ifdef TRYSETJMP
	{
		sigjmp_buf *sjb = an_thread_getdata();

		if (sjb) {
			printf("throw...");
			an_thread_unblockintr();
			siglongjmp(*sjb, 1);
		}
	}
#endif
	printf("exiting\n");

	threadstarted = 0;
	an_cond_signal(threadcond);
	an_mutex_unlock(threadlock);
	an_thread_exit();
}

static void
block(void *arg)
{
#ifdef TRYSETJMP
	sigjmp_buf sjb;
#endif

	an_mutex_lock(threadlock);
	printf("  BLOCKING thread %p started...\n", an_thread_current());
	an_thread_sethandler(ihandler2);
	threadstarted = 5;
#ifdef TRYSETJMP
	if (sigsetjmp(sjb, 0) != 0) {
		printf("catch, blocking again...\n");
		an_thread_setdata(0);
		threadstarted = 6;
	} else {
		an_thread_setdata(&sjb);
	}
#endif
	an_cond_signal(threadcond);
	while (1)
		an_cond_wait(forever, threadlock);
}

#if defined(__FreeBSD__) && !defined(__FreeBSD_version)
#include <sys/param.h>
#ifndef __FreeBSD_version
#define __FreeBSD_version 0
#endif
#endif

static void testinterrupt(void)
{
	an_error_t err;
	an_thread_t *th;

#if !defined(OSKIT) && defined(__FreeBSD__) && (__FreeBSD_version < 440000)
	/*
	 * XXX thread interrupt doesn't work right now for older (pre-4.4)
	 * version of FreeBSD due to a probable bug in the pthreads package,
	 * see http://www.freebsd.org/cgi/query-pr.cgi?pr=21943.
	 */
	printf(" does not work in FreeBSD version %d, skipping...\n",
	       __FreeBSD_version);
	return;
#endif

	threadlock = an_mutex_create(NULL);
	assert(threadlock != NULL);
	threadcond = an_cond_create(NULL);
	assert(threadcond != NULL);

	forever = an_cond_create(NULL);
	assert(forever != NULL);

	an_mutex_lock(threadlock);
	printf(" interrupting thread in infinite user loop...\n");
	err = an_domain_startthread(an_domain_current(), spin, 0, &th);
	assert(err == 0);
	while (threadstarted != 1)
		an_cond_wait(threadcond, threadlock);
	an_thread_intr(th);
#ifdef TRYSETJMP
	while (threadstarted != 2)
		an_cond_wait(threadcond, threadlock);
	an_thread_intr(th);
#endif
	while (threadstarted != 0)
		an_cond_wait(threadcond, threadlock);

	printf(" interrupting thread in infinite syscall loop...\n");
	err = an_domain_startthread(an_domain_current(), syscall, 0, &th);
	assert(err == 0);
	while (threadstarted != 3)
		an_cond_wait(threadcond, threadlock);
	an_thread_intr(th);
#ifdef TRYSETJMP
	while (threadstarted != 4)
		an_cond_wait(threadcond, threadlock);
	an_thread_intr(th);
#endif
	while (threadstarted != 0)
		an_cond_wait(threadcond, threadlock);

	printf(" interrupting thread in cond_wait...\n");
	err = an_domain_startthread(an_domain_current(), block, 0, &th);
	assert(err == 0);
	while (threadstarted != 5)
		an_cond_wait(threadcond, threadlock);
	an_thread_intr(th);
#ifdef TRYSETJMP
	while (threadstarted != 6)
		an_cond_wait(threadcond, threadlock);
	an_thread_intr(th);
#endif
	while (threadstarted != 0)
		an_cond_wait(threadcond, threadlock);

#ifdef CHECKFAULTS
	printf(" testing thread fault catching...\n");
	err = an_domain_startthread(an_domain_current(), fault, 0, &th);
	assert(err == 0);
	threadstarted = 7;
	while (threadstarted != 0)
		an_cond_wait(threadcond, threadlock);
#endif

	an_mutex_unlock(threadlock);
}
