/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */


/*
 * ARP implementation: There are 2 APIs implemented here.
 * 1. ARP Querier - Basically uses an outchan's raw channel to send out ARP queries
 *         and uses a shared "inchan" to receive ARP responses.
 * 2. ARP Responder - ARP responder creates shared "inchan" and "outchan" to receive 
 *         ARP queries and to send out responses respectively. The API implemented here
 *         is a simple one and doesn't let you create ARP responders for a subnet etc.
 */
 
/*
 * Abbreviations:-
 * arpr - arp responder
 * arpq - arp querier
 */

#include <nodeos/an_error.h>
#include <nodeos/an_chan.h>
#include <nodeos/an_pbuf.h>
#include <nodeos/an_demuxkey.h>
#include <nodeos/lib/utils.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <limits.h>

#include <nodeos/lib/click.hh>
#include <nodeos/lib/anlib_chan.h>
#include "anlib_chan_in.h"
#include "anlib_chanconfig.h"

typedef struct anlib_arpq_data {
	anlib_clickgraph_t *graph;
	an_chan_t rawchan_data;
	an_chan_t *raw_inchan;
	struct anlib_arpq_data *next;
	char ifname[16];
	int ref_count;
	an_mempool_t *mempool; /* mempool for this datastructure */ 
} anlib_arpq_data_t;

typedef struct anlib_arpr_data {
	anlib_clickgraph_t *graph;
	an_chan_t inchan_data, outchan_data;
	an_chan_t *inchan, *outchan;
	char ifname[16];
	int ref_count;
	struct anlib_arpr_data *next;
	an_mempool_t *mempool; 
} anlib_arpr_data_t;

static an_chan_t *anlib_arp_inchan_create(int q, an_domain_t *domain, 
        an_netspec_t *net, an_chan_t *data, void *dk, an_chanproto_t *proto,
		an_chanrecvfunc_t recv_func, void* recv_funcarg);
static anlib_arpq_data_t * arpq_find(an_chanproto_t *proto);
static anlib_arpr_data_t * arpr_find(an_chanproto_t *proto);

/*
 * These statics would create a problem for the multiple EE case (May be).
 * But until then, there's no problem.
 */
static anlib_arpq_data_t *arpq_inchandata_head = 0;
static anlib_arpr_data_t *arpr_data_head = 0;
static an_demuxkey_t arpr_demux_data;
static an_demuxkey_t arpq_demux_data;

int 
anlib_arpQuerier_exists(an_chanproto_t *proto)
{
    if (arpq_find(proto))
		return true;
	return false;
}

int 
anlib_outchan_arpCreateGraph(an_domain_t *domain, anlib_chanaddr_t *addr, 
					pcg_t *pcg, an_chanproto_t *proto, an_netspec_t *net)
{
	char *config = 0;
	anlib_clickgraph_t *graph = 0;
	anlib_arpq_data_t *arpdata = 0;
	an_chanrecvfunc_t recv_func;
	void *recv_arg;
	int retval = false;
	an_mempool_t *mpool = 0;
	int ret;
	
	/* First lookup if we already have a graph for this interface */
    if ((arpdata = arpq_find(proto))) {
        pcg->graph = arpdata->graph;
	    arpdata->ref_count++;
        return true;
    }
	
	/* Get the mempool to which all memory should be charged */
	mpool = an_mempool_domain(domain);
	if (!mpool)
		return false;
	
	/* create a graph as well as an "inchan" on which to listen for
	   arp responses */
	arpdata = (anlib_arpq_data_t *)an_mempool_alloc( mpool, 
						sizeof(anlib_arpq_data_t));
	if (0 == arpdata) 
		goto error;
	arpdata->mempool = mpool;
	
	graph = (anlib_clickgraph_t *)
		an_mempool_alloc(mpool, sizeof(anlib_clickgraphdummy_t));
	if (0 == graph)
		return false;

	config = (char *)an_mempool_alloc(mpool, ARPOUT_CONFIG_LEN);

	if (0 == config)
		goto error;

	sprintf(config, anlib_outchan_arpConfig, addr->arp.ip[0], addr->arp.ip[1],
			addr->arp.ip[2], addr->arp.ip[3], 
			addr->arp.eth[0], addr->arp.eth[1], addr->arp.eth[2], 
			addr->arp.eth[3], addr->arp.eth[4], addr->arp.eth[5]);

	pcg->graph = anlib_clickgraph_create(graph, config);
	if (0 == pcg->graph)
		goto error;
	
	/* Lookup for arp response input, which we can connect to the input channel
	 for this interface */
 	ret = anlib_clickgraph_lookupInput(pcg->graph, "arpr", &recv_func, 
																	&recv_arg);
	if (0 != ret)
		goto error;

	arpdata->raw_inchan = anlib_arp_inchan_create(false, domain, net, 
			&arpdata->rawchan_data, &arpq_demux_data, proto, 
			recv_func, recv_arg);
	if (0 == arpdata->raw_inchan)
		goto error;
	arpdata->graph = pcg->graph;
	arpdata->ref_count = 1;
	strcpy(arpdata->ifname, proto->data);
	arpdata->next = arpq_inchandata_head;
	arpq_inchandata_head = arpdata;
	retval = true;
	goto cleanup;
error:
	if (graph)
		an_mempool_free(mpool, graph, sizeof graph);
	if (arpdata)
		an_mempool_free(mpool, arpdata, sizeof arpdata);
cleanup:
	if (config)
		an_mempool_free(mpool, config, sizeof config);
	return retval;
}

/* The corresponding cleanup routine */
int 
anlib_outchan_arpDestroy(anlib_clickgraph_t *graph)
{
	anlib_arpq_data_t **pn = &arpq_inchandata_head;
	anlib_arpq_data_t *n = 0;

	for (; *pn; pn = &(*pn)->next) {
		if ((*pn)->graph == graph) {
			if (--(*pn)->ref_count == 0) {
				n = *pn;	
				*pn = (*pn)->next;
				anlib_clickgraph_destroy(graph);
				an_mempool_free(n->mempool, graph, sizeof graph);
				an_chan_destroy(n->raw_inchan);
				an_mempool_free(n->mempool, n, sizeof *n);
			}
			return true;
		}
	}
	return false;
}
 
/* ARP responder functions */

int 
anlib_arpResponder_exists(an_chanproto_t *proto)
{
	if (arpr_find(proto))
		return true;
	return false;
}

int 
anlib_arpResponder_create(an_domain_t *domain, anlib_chanaddr_t *addr, 
					an_chanproto_t *proto, an_netspec_t *net)
{
	anlib_arpr_data_t *arpr_data = 0;
	anlib_chanaddr_t laddr;
	addr_data_t taddr;
	an_chanaddr_t naddr;
	an_error_t err;
	anlib_clickgraph_t *graph = 0;
	char *config = 0;
	int retval = false;
	int flags;
	int ret;
	an_chanrecvfunc_t recv_func;
	void *recv_arg;
	an_mempool_t *mpool;
	
	/* First lookup if we already have a graph for this interface */
	if (0 != (arpr_data = arpr_find(proto))) {
		arpr_data->ref_count++;
		return true;
    }

    taddr.cur = addr->raw.data;
    taddr.len = addr->raw.len;
	/* TBD: This should be replaced */
	mpool = an_mempool_domain(domain);
	if (!mpool)
		return false;
	
	/* Create the data structure that would hold all the state for this
	   interface */
	arpr_data = (anlib_arpr_data_t *)an_mempool_alloc(mpool, sizeof
							(anlib_arpr_data_t)); 
	if (0 == arpr_data) 
		goto error;
	memset(arpr_data, 0, sizeof arpr_data);
	arpr_data->mempool = mpool;
	
	/* First create the Graph and lookup input in that so that we can create
	   inchannel */
	graph = (anlib_clickgraph_t *)
		an_mempool_alloc(mpool, sizeof(anlib_clickgraphdummy_t));
	if (0 == graph)
		return false;

	config = (char *)an_mempool_alloc(mpool, ARPIN_CONFIG_LEN);
	if (0 == config)
		goto error;

	ret = get_next_address(&taddr, &laddr, CHANPROTO_IN_ARP, &flags);
	if (!ret)
		goto error;
	
	sprintf(config, anlib_inchan_arpConfig, laddr.arp.ip[0], laddr.arp.ip[1],
			laddr.arp.ip[2], laddr.arp.ip[3], 
			laddr.arp.eth[0], laddr.arp.eth[1], laddr.arp.eth[2], 
			laddr.arp.eth[3], laddr.arp.eth[4], laddr.arp.eth[5]);

	graph = anlib_clickgraph_create(graph, config);
	if (0 == graph)
		goto error;
	
 	ret = anlib_clickgraph_lookupInput(graph, "in", &recv_func, &recv_arg);
	if (0 != ret)
		goto error;

	/* Create an inchan to receive ARP queries */
	arpr_data->inchan = anlib_arp_inchan_create(true, domain, net,
			&arpr_data->inchan_data, &arpr_demux_data, proto, 
			recv_func, recv_arg);
	if (0 == arpr_data->inchan)
		goto error;

	memset(&naddr, 0, sizeof naddr);
	/* Create an outchan on which we would respond to ARP queries */
	arpr_data->outchan = an_outchan_create(&arpr_data->outchan_data, domain,
							proto, &naddr, net);
	if (0 == arpr_data->outchan)
		goto error;

	/* attach outchan to ARP graph and we are done */
	err = anlib_clickgraph_attachOutput(graph, "out", an_outchan_send,
			arpr_data->outchan);
	if (err)
		goto error;
	
	arpr_data->graph = graph;
	arpr_data->ref_count = 1;
	arpr_data->next = arpr_data_head;
	arpr_data_head = arpr_data;
	strcpy(arpr_data->ifname, proto->data);
	retval = true;
	goto cleanup;
error:
	if (graph)
		an_mempool_free(mpool, graph, sizeof graph);
	if (arpr_data) {
		if (arpr_data->inchan)
			an_chan_destroy(arpr_data->inchan);
		if (arpr_data->outchan)
			an_chan_destroy(arpr_data->outchan);
		an_mempool_free(mpool, arpr_data, sizeof arpr_data);
	}
cleanup:
	if (config)
		an_mempool_free(mpool, config, sizeof config);
	return retval;
}

int 
anlib_arpResponder_destroy(an_chanproto_t *proto)
{
	anlib_arpr_data_t **pn= &arpr_data_head;
	anlib_arpr_data_t *n = 0;
	
	for (; *pn; pn = &(*pn)->next) {
		if (!strcmp((*pn)->ifname, proto->data)) {
			if (--(*pn)->ref_count == 0) {
				n = *pn;
				*pn = (*pn)->next;
				anlib_clickgraph_destroy(n->graph);
				an_mempool_free(n->mempool, n->graph, sizeof n->graph);
				an_chan_destroy(n->inchan);
				an_chan_destroy(n->outchan);
				an_mempool_free(n->mempool, n, sizeof *n);
				return true;
			}
		}
	}
	return false;
}

/* Static helper functions */

static anlib_arpq_data_t *
arpq_find(an_chanproto_t *proto)
{
	anlib_arpq_data_t *arpdata = arpq_inchandata_head;
	
	for (; arpdata; arpdata = arpdata->next) {
		if (!strcmp(arpdata->ifname, proto->data)) {
			break;
		}
	}
	return arpdata;
}

static anlib_arpr_data_t * 
arpr_find(an_chanproto_t *proto)
{
	anlib_arpr_data_t *arpr_data = arpr_data_head;    
	for (; arpr_data; arpr_data = arpr_data->next) {
		if (!strcmp(arpr_data->ifname, proto->data)) {
			break;
		}
	}
	return arpr_data;
}

static an_chan_t *
anlib_arp_inchan_create(int q, an_domain_t *domain, an_netspec_t *net, 
		an_chan_t *data, void *demux_mem, an_chanproto_t *proto, 
		an_chanrecvfunc_t recv_func, void* recv_funcarg)
{
	an_chanaddr_t addr;
	an_demuxkey_t *arp_demux = 0;
	an_chan_t *arpr_chan = 0;
	unsigned short pkt[11], pktmask[11];
	
	/* Create demuxkey to select queries or responses as requested */
	memset(&addr, 0, sizeof addr);
	arp_demux = an_demuxkey_create(demux_mem, domain);
	if (0 == arp_demux)
		return 0;

	/* Select IP/ARP */
	memset(pkt, 0, sizeof pkt);
	memset(pktmask, 0, sizeof pktmask);
	pkt[6] = htons(0x0806); /* ETHERTYPE_ARP */
	pktmask[6] = ~0;
	if (q)
		pkt[10] = htons(1); /* ARPOP_QUERY */
	else
		pkt[10] = htons(2); /* ARPOP_REPLY */
	pktmask[10] = ~0;
	an_demuxkey_add(arp_demux, sizeof pkt, pkt, pktmask);

	/* In channel to select packets */
	arpr_chan = an_inchan_create(data, domain, arp_demux, proto, &addr, net, 
				     recv_func, recv_funcarg);
	return arpr_chan;	
}

void 
print_arp_stats(void)
{	
	moab_chan_printstats(arpq_inchandata_head->raw_inchan, 1, "arp stats", 
			120, 0);
}

