/*
 * Copyright (c) 1999-2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * The Channel API implemented as click graphs. It uses the 
 * click library (anlib_clickgraph_*).
 */
 
#include <nodeos/an_error.h>
#include <nodeos/an_chan.h>
#include <nodeos/an_pbuf.h>
#include <nodeos/an_demuxkey.h>
#include <nodeos/ani/ani_log.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <limits.h>

#include <nodeos/lib/click.hh>
#include <nodeos/lib/anlib_chan.h>
#include "anlib_chanconfig.h"
#include "anlib_chan_in.h"

pcg_ops_t anlib_inchan_ethOps = {
	CHANPROTO_IN_ETH, anlib_chanprefix_matches, (char *)"eth", 0, 0, 0, anlib_clickgraph_destroy 
};

pcg_ops_t anlib_strip_ethOps = {
	CHANPROTO_NONE, 0, 0, 0,
	anlib_inchan_createGraph, anlib_strip_ethConfig, anlib_clickgraph_destroy
};

pcg_ops_t anlib_inchan_ifOps = {
	CHANPROTO_IN_IF, anlib_chanprefix_matches, (char *)"if", 0, 0, 0, anlib_clickgraph_destroy 
};

pcg_ops_t anlib_inchan_ipOps = {
	CHANPROTO_IN_IP, anlib_chan_matches, (char *)"ip", anlib_ipUpdateDemux,
	anlib_inchan_createGraph, anlib_inchan_ipConfig, anlib_clickgraph_destroy
};

pcg_ops_t anlib_inchan_udpOps = {
	CHANPROTO_IN_UDP, anlib_chan_matches, (char *)"udp", anlib_udpUpdateDemux,
	anlib_inchan_createGraph, anlib_inchan_udpConfig, anlib_clickgraph_destroy
};

pcg_ops_t *anlib_inchan_ops[] = {
	&anlib_inchan_ethOps,
	&anlib_inchan_ifOps,
	&anlib_inchan_ipOps,
	&anlib_inchan_udpOps,
		 0
};

int anlib_outchan_ipCreateGraph(anlib_chanaddr_t *addr, an_mempool_t *mpool, anlib_clickgraph_t **pg, 
	char *config);
int anlib_outchan_udpipCreateGraph(anlib_chanaddr_t *addr, an_mempool_t *mpool, anlib_clickgraph_t **pg, 
	char *config);
int anlib_outchan_arpDestroy(anlib_clickgraph_t *graph);

pcg_ops_t anlib_outchan_ethOps = {
	CHANPROTO_OUT_ETH, anlib_chanprefix_matches, (char *)"eth", 0, 0, 0, anlib_clickgraph_destroy
};

pcg_ops_t anlib_outchan_ifOps = {
	CHANPROTO_OUT_IF, anlib_chanprefix_matches, (char *)"if", 0, 0, 0, anlib_clickgraph_destroy
};

pcg_ops_t anlib_outchan_ipOps = {
	CHANPROTO_OUT_IP, anlib_chan_matches, (char *)"ip", 0, anlib_outchan_ipCreateGraph, 
	anlib_outchan_ipConfig, anlib_clickgraph_destroy
};

pcg_ops_t anlib_outchan_udpOps = {
	CHANPROTO_OUT_UDP, anlib_chan_matches, (char *)"udp", 0, 0, 0, anlib_clickgraph_destroy
};

pcg_ops_t anlib_outchan_udpipOps = {
	CHANPROTO_OUT_UDPIP, anlib_chan_matches, (char *)"udpip", 0,
	anlib_outchan_udpipCreateGraph, anlib_outchan_udpipConfig, anlib_clickgraph_destroy
};

pcg_ops_t anlib_outchan_arpOps = {
	CHANPROTO_OUT_ARP, anlib_chan_matches, (char *)"arp", 0,
	0, anlib_outchan_arpConfig, anlib_outchan_arpDestroy
};

pcg_ops_t *anlib_outchan_ops[] = {
	&anlib_outchan_ethOps,
	&anlib_outchan_ifOps,
	&anlib_outchan_ipOps,
	&anlib_outchan_udpOps,
	&anlib_outchan_udpipOps,
	&anlib_outchan_arpOps,
	0
};


/*
 * @allocation (optional): sizeof(an_domain_t) bytes for channel
 * @allocation (temporary): sizeof(ani_pbuf_t *) * npbufs bytes for pbuf ptrs
 */

	/* 
	 * Address specification format is:- e.g. for "ip/ip/udp"
	 * <ip address>/<ip address>/<udp address>
	 * For now My convention would be:-
	 * eth:- <saddr>:<daddr>:<proto>
	 * IP address:- <src>:<dst>:proto
	 * udp/tcp:- <sport>:<dport>:proto
	 */

	/* 
	 * I am also assuming that for each protocol component that the user
	 * specifies there would be a corresponding address component, or else 
	 * the address string would be of 0 length. 
	 */

anlib_chan_t *
anlib_inchan_create(void *mem, an_domain_t *domain, an_demuxkey_t *key,
		 an_chanproto_t *proto, anlib_chanaddr_t *addr,
		 an_netspec_t *net, an_chanrecvfunc_t func, void *farg)
{
	anlib_chan_t *xchan = 0;
	anlib_chanaddr_t native_libaddr;
	an_chanaddr_t native_addr;
	an_chanproto_t native_proto;
	int data_offset = 0; 
	void *demuxkey_data = 0;
	pcg_t *pcg_head = 0, *next_pcg = 0;
	an_demuxkey_t *demuxkey = 0;
	void *proto_last = 0;
	addr_data_t addr_last;
	anlib_chanproto_t next_proto;
	an_mempool_t *mpool = 0;

	mpool = an_mempool_domain(domain);
	xchan = (anlib_chan_t *) an_mempool_alloc(mpool, sizeof(anlib_chan_t));
	if (xchan == 0)
		return 0;
	memset(xchan, 0, sizeof *xchan);
    	xchan->mempool = mpool;
	
	if (0 == validate_proto(proto)) {
		goto try_native;		
	}
	
	/* 
	 * Strip off the "eth" like part in the proto string 
	 * This should be passed to NodeOS native APIs.
	 */

	next_proto = get_next_proto(proto, 1, mpool, &next_pcg, &pcg_head, &proto_last);

	addr_last.cur = addr->raw.data;
	addr_last.len = addr->raw.len;
	memset(&native_addr, 0, sizeof native_addr);
	
	if (CHANPROTO_IN_ETH == next_proto || CHANPROTO_IN_IF == next_proto) {
		char *cur = proto->data;
		int flags;
		/* Go to the end of current protocol */
		for (native_proto.len = 0; (*cur != '/') && (*cur); 
		                            native_proto.data[native_proto.len++] = *(cur++) )
	        ; /* no body */ 
			
		native_proto.data[native_proto.len] = 0;
		if (get_next_address(&addr_last, &native_libaddr, next_proto, &flags))
			native_addr.eth = native_libaddr.eth;
		
		if (next_proto == CHANPROTO_IN_ETH) {
			/* We receive MAC header in the case of "eth"ernet channels */
			/* Insert a click graph to strip off the eth header */
			pcg_t *p = alloc_pcg(&pcg_head, mpool);
			if (!p) goto cleanup;
			p->ops = &anlib_strip_ethOps;
		    if (p->ops->create_graph(addr, mpool, &(p->graph), 
		                              p->ops->create_arg) == false)
		      goto cleanup;	
			data_offset = 14; 
	    }		
	    /* Get the next protocol */
	    next_proto = get_next_proto(0, 1, mpool, &next_pcg, &pcg_head, &proto_last);
	}
	else {
	    ani_log_printf(ANI_LOG_DEBUG, "No interface protocol specified\n");
        goto cleanup;
	}
	/* Create a new demux key */
	/*
	 * There is a problem here. If the user specifies a demuxkey, then we should append 
	 * that key to the key we create for protocol demuxing and not the other-way-round. 
	 * But to do that I need a function that would let me append one key to another.
	 * All that function would need to do is append the segments from one key to another.
	 * Until then, user specified keys won't work, they are just ignored.
	 */
	demuxkey_data = an_mempool_alloc(mpool, sizeof(an_demuxkey_t));
	if (!demuxkey_data)
	    goto cleanup;
	demuxkey = an_demuxkey_create(demuxkey_data, domain);
	if (!demuxkey)
	    goto cleanup;

	/* 
	 * For each next proto do the following:-
	 * 1. Get the address
	 * 2. Append to the demux keys. I am assuming the order in the address spec
	 *  	is correct. The user has to make sure to provide valid proto and 
	 *		address strings.
	 * 3. Instantiate the corresponding graph.
	 */
	while ((CHANPROTO_NONE != next_proto) && (CHANPROTO_INVALID != next_proto)) {
		if (! next_pcg) {
			/* We don't know this protocol, so pass all parameters to native
			   channel creation API and see if it can handle it */
			break;
		}
		if (next_pcg->ops->update_demuxkey(&addr_last, data_offset, demuxkey) == false) 
		  goto cleanup;
		data_offset = 0;
		if (next_pcg->ops->create_graph(addr, mpool, &(next_pcg->graph), 
		      next_pcg->ops->create_arg) == false)
		      goto cleanup;		
		next_proto = get_next_proto(0, 1, mpool, &next_pcg, &pcg_head, 
		                        &proto_last);
	}

	/* CHANPROTO_NONE implies that we have finished parsing the input, and we
	   have the correct demuxkey in demuxkey */
	if (CHANPROTO_NONE == next_proto) {
		an_error_t err;
	    an_chanrecvfunc_t recv_func = 0;
		void *recv_farg;
		pcg_t *last_pcg = pcg_head;
		if (0 == connect_pcgs(pcg_head))
			goto cleanup;
		/* should now update the parameters to native channel
		   creation API. */
		err = anlib_clickgraph_lookupInput(pcg_head->graph, "in", &recv_func, &recv_farg);
		if (err != 0)
		    goto cleanup;
		/* Create the native channel */
		xchan->raw_chan = an_inchan_create(mem, domain, demuxkey, &native_proto, 
								&native_addr, net, recv_func, recv_farg);
		if (!xchan->raw_chan)
		    goto cleanup;
		if (last_pcg && last_pcg->next)
			last_pcg = last_pcg->next;
		if (last_pcg) {
			err = anlib_clickgraph_attachOutputA(last_pcg->graph, "out", func,
					farg);
			if (err != 0)
			    goto cleanup;
		}
	}
	else {
		/* We came out of the loop because we dint know the protocol or one of
		   the components of the protocol */
		/* Freeup all stuff */
		/* Make demuxkey equal to the user specified key */
cleanup:		
		anlib_chan_cleanpcgs(pcg_head, mpool);
		if (demuxkey)
			an_demuxkey_destroy(demuxkey);
try_native:
		xchan->raw_chan = an_inchan_create(mem, domain, key, proto, 
									(an_chanaddr_t *)addr, net, func, farg);
		if (!xchan->raw_chan) {
		    anlib_chan_destroy(xchan);
		    return 0;
		}
	}
	xchan->pcg_head = pcg_head;
	return xchan;
}

anlib_chan_t *
anlib_outchan_create(void *mem, an_domain_t *domain, an_chanproto_t *proto,
		  anlib_chanaddr_t *addr, an_netspec_t *net)
{
	anlib_chan_t *xchan = 0;
	pcg_t *pcg_head = 0;
	pcg_t *next_pcg = 0;
	void *proto_last = 0;
	addr_data_t addr_last;
	an_chanproto_t native_proto;
	an_chanaddr_t naddr;
	anlib_chanproto_t next_proto;
	int flags;
	anlib_chanaddr_t addr_data, arp_addr;
	pcg_t *arp_pcg = 0;
	int ret;
	an_mempool_t *mpool = 0;
	
	mpool = an_mempool_domain(domain);
	xchan = (anlib_chan_t *) an_mempool_alloc(mpool, sizeof(anlib_chan_t));
	if (xchan == 0)
		return 0;
	memset(xchan, 0, sizeof *xchan);
	xchan->mempool = mpool;
	memset(&naddr, 0, sizeof naddr);
	
	if (!validate_proto(proto)) {
		goto try_native;		
	}

    addr_last.cur = addr->raw.data;
	addr_last.len = addr->raw.len;
	next_proto = get_next_proto(proto, 2, mpool, &next_pcg, &pcg_head, &proto_last);
	ret = get_next_address(&addr_last, &addr_data, next_proto, &flags);
	
	while ((CHANPROTO_NONE != next_proto) && (CHANPROTO_INVALID != next_proto)
					&& (CHANPROTO_OUT_ETH != next_proto) 
					&& (CHANPROTO_OUT_IF != next_proto)) {
		if (NULL == next_pcg) {
			/* We don't know this protocol, so pass all parameters to native
			   channel creation API and see if it can handle it */
			break;
		}
		if (CHANPROTO_OUT_ARP == next_proto) {
			arp_pcg = next_pcg;
			arp_addr = addr_data;
		}
		else {
			if (next_pcg->ops->create_graph(&addr_data, mpool, &next_pcg->graph,
									   next_pcg->ops->create_arg) == 0)
			    goto cleanup;
		}
			
		next_proto = get_next_proto(0, 2, mpool, &next_pcg, 
							&pcg_head, &proto_last);
		ret = get_next_address(&addr_last, &addr_data, next_proto, &flags);
	}

	if (CHANPROTO_OUT_ETH == next_proto || CHANPROTO_OUT_IF == next_proto) {
		char *cur = proto->data + proto->len -1;
		int proto_len;
		/* Go to the end of current protocol */
		for (proto_len = 0; *cur != '/' && cur != proto->data; cur--, proto_len++);
		if (cur == proto->data)
			proto_len++;
		else
			cur++;
		strncpy(native_proto.data, cur, proto_len);
		native_proto.data[proto_len] = 0;
		native_proto.len = proto_len;
		if (ret) 
			naddr.eth = addr_data.eth;
		next_proto = CHANPROTO_NONE;
		/* If there was arp proto, its graph wouldn't have been created,
		   create it now */
		if (arp_pcg) {
			if (anlib_outchan_arpCreateGraph(domain, &arp_addr, arp_pcg, &native_proto, net)
			        == 0)
				goto cleanup;
		}
	}
	else {
        goto cleanup;
	}

	/* CHANPROTO_NONE means that we have finished parsing the input */
	if (CHANPROTO_NONE == next_proto) {
		pcg_t *last_pcg = 0;
		an_error_t err;
		if (0 == connect_pcgs(pcg_head)) 
			goto cleanup;
		/* Create the native channel */
		xchan->raw_chan = an_outchan_create(mem, domain, &native_proto, &naddr, net);
		if (! xchan->raw_chan)
		    goto cleanup;
		/* update the parameters to native channel creation API. */
		if ((last_pcg = pcg_head)) {
			if (pcg_head->next)
				last_pcg = pcg_head->next;
		}

		if (last_pcg) {
			err = anlib_clickgraph_attachOutput(last_pcg->graph, "out", 
						an_outchan_send, xchan->raw_chan);
			if (err != 0)
			    goto cleanup;
		
			err = anlib_clickgraph_lookupInput(pcg_head->graph, "in", &xchan->send, &xchan->sendarg);
			if (err != 0)
			    goto cleanup;			
		}
	}
	else {
cleanup:
		/* Freeup all stuff */
		anlib_chan_cleanpcgs(pcg_head, mpool);
try_native:
		xchan->raw_chan = an_outchan_create(mem, domain, proto, (an_chanaddr_t*)addr, net);
		if (!xchan->raw_chan) {
		    anlib_chan_destroy(xchan);
		    return 0;
		}
		xchan->native = true;
	}
	xchan->pcg_head = pcg_head;
	return xchan;
}

/* CutChan: Simple implementation. */
anlib_chan_t *
anlib_cutchan_create(void *mem, an_domain_t *domain, anlib_chan_t *inchan, anlib_chan_t *outchan)
{
	int ret;	
	/* For now, I am not going to connect if any of the channels was just native */
	if (inchan->native || outchan->native)
		return 0;
		
	ret = connect_pcgpair(inchan->pcg_head, outchan->pcg_head);
	if (!ret)
		return 0;
	/* Inchan is the new cutchan */	
	/* Free memory from the outchan. Or may be I should let the user do that???? */
	an_mempool_free(outchan->mempool, outchan, sizeof outchan);	

	return inchan;
}

/* Inorder for lib to be able to destroy channels, I need to have a different
   channel type than an_chan_t */
an_error_t
anlib_chan_destroy(anlib_chan_t *chan)
{
	anlib_chan_cleanpcgs(chan->pcg_head, chan->mempool);
	if (chan->raw_chan)
	    an_chan_destroy(chan->raw_chan);
	an_mempool_free(chan->mempool, chan, sizeof(anlib_chan_t));
	return 0;	
}

an_error_t
anlib_outchan_send(anlib_chan_t *chan, an_pbuf_t *pbuf)
{
	if (chan->native) 
		return an_outchan_send(chan->raw_chan, pbuf);
	else if (chan->send)
		chan->send(pbuf, chan->sendarg, 0);
	else
	    return AN_ERR_INVALIDARG;
	return 0;
}


int
validate_proto(an_chanproto_t *proto)
{
	/* We need to probably make a table of which protocol can follow a protocol and then
	check against that table in this function. But the table should not deny any valid
	combinations. This could be tough except for simple things like there should not be 
	an etherenet protocol component in the middle of the protocol but then why not? */
	
	/* For now, all combinations that user asks for are valid */
	return true;
}

int 
anlib_chan_cleanpcgs(pcg_t *pcg, an_mempool_t *mpool)
{
	pcg_t *next;
	for (; pcg; pcg = next) {
		next = pcg->next; 
		pcg->ops->destroy(pcg->graph);
		an_mempool_free(mpool, pcg, sizeof(pcg_t));
	}
	return true;
}


int 
anlib_chanprefix_matches(char *proto, int len, char *proto_str)
{
	/* Match "eth" and "eth<N>" */
	int slen = strlen(proto_str);
	if (len < slen)
		return 0;
	if (strncmp(proto, proto_str, slen))
		return 0;
	/* For now assuming that the user would specify only valid N's in "eth<N>",
	 if not the user would get an error from the native channel creation API. */
	return 1; 
}

int 
anlib_outchan_ipCreateGraph(anlib_chanaddr_t *addr, an_mempool_t *mpool, 
	 anlib_clickgraph_t **pg, char *config_str)
{
	char *config;
	anlib_clickgraph_t *graph = (anlib_clickgraph_t*) an_mempool_alloc( mpool, 
									sizeof(anlib_clickgraphdummy_t));
	if (0 == graph)
		return false;
	
	config = (char *)an_mempool_alloc(mpool, IPOUT_CONFIG_LEN);
	if (0 == config)
		return false;
		
	sprintf(config, config_str, addr->ip.type, addr->ip.saddr[0],
			addr->ip.saddr[1], addr->ip.saddr[2], addr->ip.saddr[3],
			addr->ip.daddr[0], addr->ip.daddr[1], addr->ip.daddr[2], 
			addr->ip.daddr[3]);
	
	*pg = anlib_clickgraph_create(graph, config);
	an_mempool_free(mpool, config, sizeof IPOUT_CONFIG_LEN);
	if (0 == *pg)
		return false;
	return true;
}

int 
anlib_outchan_udpipCreateGraph(anlib_chanaddr_t *addr, an_mempool_t *mpool, 
			anlib_clickgraph_t **pg, char *config_str)
{
	char *config;
	anlib_clickgraph_t *graph;
	graph = (anlib_clickgraph_t *) an_mempool_alloc(mpool, sizeof(anlib_clickgraphdummy_t));
	if (0 == graph)
		return false;

	config = (char *)an_mempool_alloc(mpool, UDPIPOUT_CONFIG_LEN);
	if (0 == config)
		return false;
		
	sprintf(config, config_str, addr->udpip.saddr[0], 
			addr->udpip.saddr[1], addr->udpip.saddr[2], addr->udpip.saddr[3],
			addr->udpip.sport,
			addr->udpip.daddr[0], addr->udpip.daddr[1], addr->udpip.daddr[2],
			addr->udpip.daddr[3], 
			addr->udpip.dport);
	
	*pg = anlib_clickgraph_create(graph, config);
	an_mempool_free(mpool, config, sizeof IPOUT_CONFIG_LEN);
	if (0 == *pg)
		return false;
	return true;
}

int 
anlib_outchan_udpCreateGraph(anlib_chanaddr_t *addr, an_mempool_t *mpool,
				anlib_clickgraph_t **pg, char *config_str)
{
	char *config;
	anlib_clickgraph_t *graph = (anlib_clickgraph_t *)
		an_mempool_alloc(mpool, sizeof(anlib_clickgraphdummy_t));
	if (0 == graph)
		return false;

	config = (char *)an_mempool_alloc(mpool, UDPIPOUT_CONFIG_LEN);
	if (0 == config)
		return false;

	sprintf(config, config_str, addr->udp.sport, 
												addr->udp.dport);

	*pg = anlib_clickgraph_create(graph, config);
	an_mempool_free(mpool, config, sizeof IPOUT_CONFIG_LEN);
	if (0 == *pg)
		return false;
	return true;
}

int 
anlib_ipUpdateDemux(addr_data_t *addr, int offset, 
					an_demuxkey_t *key)
{
	int flags;
	an_error_t err;
	anlib_chanaddr_t ip;
	u_int32_t saddr = 0, daddr = 0;
	get_next_address(addr, &ip, CHANPROTO_IN_IP, &flags);
	if (flags != AN_DEMUXKEY_IPV4_ALLFLAGS) {
	  /* Convert between unsigned char* and ulong here, Does this work ok?? */
	  saddr = ip.ip.saddr[0] + 16 * ip.ip.saddr[1] + 256 * ip.ip.saddr[2] +
					4096 * ip.ip.saddr[3];
	  saddr = htonl(saddr);
	  daddr = ip.ip.daddr[0] + 16 * ip.ip.daddr[1] + 256 * ip.ip.daddr[2] +
					4096 * ip.ip.daddr[3];
	  daddr = htonl(daddr);	
	}  

	/* XXX gak, change in NodeOS spec: no more offset */
	if (offset != 0)
		panic("%s: FIX: demuxkey_add no longer supports offset",
		      __FUNCTION__);

	err = an_demuxkey_addipv4(key, flags, saddr, 0, daddr, 0, ip.ip.type);
	if (0 != err)
		return false;
	return true;	
}

int 
anlib_udpUpdateDemux(addr_data_t *addr, int offset, 
					an_demuxkey_t *key)
{
	int flags;
	an_error_t err;
	anlib_chanaddr_t udp;
	if (false == get_next_address(addr, &udp, CHANPROTO_IN_UDP, &flags))
	    return false;
	if (AN_DEMUXKEY_UDP_ALLFLAGS != flags) {

		/* XXX gak, change in NodeOS spec: no more offset */
		if (offset != 0)
			panic("%s: FIX: demuxkey_add no longer supports offset",
			      __FUNCTION__);

		err = an_demuxkey_addudp(key, flags, htons(udp.udp.sport),
					 htons(udp.udp.dport));
		if (0 != err) 
			return false;
	}
	return true;
}

int 
anlib_inchan_createGraph(anlib_chanaddr_t *addr, an_mempool_t *mpool, 
		anlib_clickgraph_t **pg, char *config)
{
	anlib_clickgraph_t *graph = (anlib_clickgraph_t*) an_mempool_alloc(
						mpool, sizeof(anlib_clickgraphdummy_t));
	if (0 == graph)
		return false;
	
	*pg = anlib_clickgraph_create(graph, config);
	if (0 == *pg)
		return false;
	return true;
}

int 
connect_pcgpair(pcg_t *p1, pcg_t *p2)
{
	an_error_t err;
	anlib_recvPCGfunc_t recv_func;
	void *recv_farg;
	err = anlib_clickgraph_lookupInputPCG(p2->graph, "in", &recv_func,
				&recv_farg);
	if (0 != err) 
		return false;
	err = anlib_clickgraph_attachOutputPCG(p1->graph, "out", 
				recv_func, recv_farg);
	if (0 != err)
		return false;
	return true;
}

int 
connect_pcgs(pcg_t *pcg_head)
{
	pcg_t *last_pcg = 0;
	pcg_t *next_pcg = 0;
	int ret;

	if (0 == pcg_head) {
		return false;
	}

	/* Connect all the pcgs from the last to the 2nd */
	if ((last_pcg = pcg_head->next)) {
		for (next_pcg = last_pcg->next; last_pcg && next_pcg; 
						last_pcg = next_pcg, next_pcg = next_pcg->next) {
			ret = connect_pcgpair(next_pcg, last_pcg);
			if (!ret)
				return false;
		}
	}
	/* Connect the 2nd pcg and the first pcg */
	if ((last_pcg) && (last_pcg != pcg_head)) {
		ret = connect_pcgpair(pcg_head, last_pcg);
		if (!ret)
			return false;
	}
	return true;
}


int 
anlib_chan_matches(char *proto, int len, char *str)
{
	int str_len = strlen(str);
	if (len != str_len)
		return false;
	return !strncmp(proto, str, str_len);
}

anlib_chanproto_t 
get_next_proto(an_chanproto_t *proto, int chantype, 
		an_mempool_t *mpool, pcg_t **pcg, pcg_t **ppcg_head, void **last)
{
	char *current = 0;
	int proto_len = 0;
	char *clast = 0;
	pcg_ops_t **ops = 0;
	pcg_t *new_pcg = 0;

	if (proto) 
		current = proto->data;		
	else {
		if (!last)
		    return CHANPROTO_INVALID;
		current = *(char **)last;	
	}

	/* proto is a null terminated string for now */
	if (*current == 0)
		return CHANPROTO_NONE;

	/* Go to the end of current protocol */
	clast = current;
	for (proto_len = 0; (*clast != '/') && (*clast); clast++, proto_len++);
	
	if (1 == chantype) 
		ops = anlib_inchan_ops;
	else
		ops = anlib_outchan_ops;

	for (; *ops; ops++)
		if ((*ops)->matches(current, proto_len, (*ops)->matches_arg))
			break;
	
	if (!*ops) 
		return CHANPROTO_INVALID;

	/* Create a new pcg in mempool of domain */
	if ((*ops)->id != CHANPROTO_IN_ETH && (*ops)->id != CHANPROTO_IN_IF &&
			(*ops)->id != CHANPROTO_OUT_IF && (*ops)->id != CHANPROTO_OUT_ETH)
	{
		new_pcg = alloc_pcg(ppcg_head, mpool);
		if (!new_pcg)
		    return CHANPROTO_INVALID;
		new_pcg->ops = (*ops);
	}
	
	/* Make the last point to one past the current string, to ignore next "/" */
	if (last) {
		if (*clast)
			*last = clast+1;
		else
			*last = clast;
	}
	
	*pcg = new_pcg;
	return (*ops)->id;
}

pcg_t *
alloc_pcg(pcg_t **ppcg_head, an_mempool_t *mpool)
{
	pcg_t *pcg = (pcg_t *)an_mempool_alloc(mpool, sizeof(pcg_t));
	if (!pcg)
	    return 0;
	memset(pcg, 0, sizeof *pcg);

	if (0 == ppcg_head)
		return 0;

	if (*ppcg_head) {
		pcg->next = (*ppcg_head)->next;
		(*ppcg_head)->next = pcg; 
	}
	else
		*ppcg_head = pcg;
		
    return pcg;
}

static int
cp_ip_address(char *s, int len, unsigned char *return_value)
{
  int pos = 0, part, d;
  unsigned char value[4];

  for (d = 0; d < 4; d++) {
    if (d && pos < len && s[pos] == '.')
      pos++;
    if (pos >= len || !isdigit(s[pos]))
      goto bad;
    for (part = 0; pos < len && isdigit(s[pos]) && part <= 255; pos++)
      part = part*10 + s[pos] - '0';
    if (part > 255)
      goto bad;
    value[d] = part;
  }

  if (pos == len) {
    memcpy(return_value, value, 4);
    return true;
  }

 bad:
  return false;
}

static int
xvalue(int x)
{
  if (x >= '0' && x <= '9')
    return x - '0';
  else if (x >= 'A' && x <= 'F')
    return x - 'A' + 10;
  else if (x >= 'a' && x <= 'f')
    return x - 'a' + 10;
  else
    return -1;
}

static int
cp_ethernet_address(char *s, int len, unsigned char *return_value)
{
  int i = 0, d;
  unsigned char value[6];

  if (s[0] == '0' && (s[1] == 'x' || s[1] == 'X')) 
	  i = 2;

  for (d = 0; d < 6; d++) {
    if (i < len - 1 && isxdigit(s[i]) && isxdigit(s[i+1])) {
      value[d] = xvalue(s[i])*16 + xvalue(s[i+1]);
      i += 2;
    } else if (i < len && isxdigit(s[i])) {
      value[d] = xvalue(s[i]);
      i += 1;
    } else
      goto bad;
    if (d == 5) break;
    if (i >= len - 1)
      goto bad;
  }

  if (i == len) {
    memcpy(return_value, value, 6);
    return true;
  }

 bad:
  return false;
}

int 
get_ethernet_address(char **pstr, int *plen, unsigned char *retval)
{
	char *s = *pstr;
	int com_len = get_component_len(pstr, plen);

	if ((com_len) && (cp_ethernet_address(s, com_len, retval))) 
		return com_len;
	return 0;
}

int 
get_ip_address(char **pstr, int *plen, unsigned char *retval)
{
	char *s = *pstr;
	int com_len = get_component_len(pstr, plen);

	if ((com_len) && (cp_ip_address(s, com_len, retval))) 
		return com_len;
	return 0;
}

int 
get_component_len(char **pstr, int *plen)
{
	int com_len;
	char *s = *pstr;
	int len = *plen;
	for (com_len = 0; *s != ':' && com_len < len; com_len++, s++)
		;
	*pstr = s;
	*plen -= com_len;
	if ((1 == com_len) && (**pstr == '*')) 
		com_len = 0; /* '*' is equivalent to no address */

	if (**pstr == ':') {
		(*pstr)++;
		(*plen)--;
	}
	return com_len;
}

int 
get_hexshort_address(char **pstr, int *plen, unsigned short *retval)
{
	char *s = *pstr;
	int com_len;
	int i;
	unsigned short value = 0;
	/* Strip off any leading 0x or 0X */
	if (s[0] == '0' && (s[1] == 'x' || s[1] == 'X')) {
		s += 2; *plen -= 2; *pstr = s;
	}
	com_len = get_component_len(pstr, plen);
	if (com_len > 4) 
		return false;
	for (i = 0; i < com_len-1; i += 2) {
		if (isxdigit(s[i]) && isxdigit(s[i+1])) {
      		value = 256 * value + xvalue(s[i])*16 + xvalue(s[i+1]);
		}
	}
	if (i == com_len-1)
		value = 16 * value + xvalue(s[i]);
	*retval = value;
	return com_len;
}

int 
get_short_address(char **pstr, int *plen, unsigned short *retval)
{
	char *s = *pstr;
	int com_len;
	char c;

	com_len = get_component_len(pstr, plen);
	if (com_len) {
		c = s[com_len];
		s[com_len] = 0;
		*retval = strtoul(s, 0, 10);
		s[com_len] = c;
		if (*retval)
			return true; 
	}
	return false;
}

int 
get_byte_address(char **pstr, int *plen, unsigned char *retval)
{
	char *s = *pstr;
	int com_len;
	char c;
	com_len = get_component_len(pstr, plen);
	if (com_len) {
		c = s[com_len];
		s[com_len] = 0;
		*retval = strtoul(s, 0, 10);
		s[com_len] = c;
		if (*retval)
			return true; 
	}
	return false;
}

int 
get_next_address(addr_data_t *addr, void *out_addr, int proto, 
		int *flags) 
{
	char *current;
	int len;
	char *clast;
	int proto_len;
	anlib_chanaddr_t *chan_addr;
	int ret;
	int rv = true;
	
	if (addr) {
		current = addr->cur;
		len = addr->len;
	}
	else
        return false;
		
	for (proto_len = 0, clast = current; *clast != '/' && proto_len < len; 
									proto_len++, clast++)
		; /* No body */
	
	/* Make clast point to the next component address */
	if (proto_len < len)
		clast++;

    addr->cur = clast; /* Return value, Next time we would point to the next address, 
					  if there is one, past the '/' */
	addr->len = len - (clast - current);
	
	chan_addr = (anlib_chanaddr_t *)out_addr;			
	memset(chan_addr, 0, sizeof *chan_addr);
	switch (proto) {
		case CHANPROTO_IN_ETH:			
		case CHANPROTO_OUT_ETH:			
		case CHANPROTO_IN_IF:			
		case CHANPROTO_OUT_IF:			
			if (0 == proto_len) {
				*flags = AN_DEMUXKEY_ETH_ALLFLAGS;
				return false;	
			}
			*flags = 0;
			/* src address */
			ret = get_ethernet_address(&current, &proto_len, chan_addr->eth.saddr);
			if (!ret)
				*flags |= AN_DEMUXKEY_ETH_ANYSADDR;	
			/* dst address */
			ret = get_ethernet_address(&current, &proto_len, chan_addr->eth.daddr);
			if (!ret)
				*flags |= AN_DEMUXKEY_ETH_ANYDADDR;	
			/* type */
			ret = get_hexshort_address(&current, &proto_len, &chan_addr->eth.type);
			if (ret)
				chan_addr->eth.type = htons(chan_addr->eth.type);
			else {
				*flags |= AN_DEMUXKEY_ETH_ANYETYPE;	
			}
			break;
		case CHANPROTO_IN_IP:
		case CHANPROTO_OUT_IP:
			if (0 == proto_len) {
				*flags = AN_DEMUXKEY_IPV4_ALLFLAGS;
				return true;
			}
			*flags = 0;
			/* saddr */
			ret = get_ip_address(&current, &proto_len, chan_addr->ip.saddr);
			if (!ret)
				*flags |= AN_DEMUXKEY_IPV4_ANYSADDR;	
			/* daddr */
			ret = get_ip_address(&current, &proto_len, chan_addr->ip.daddr);
			if (!ret)
				*flags |= AN_DEMUXKEY_IPV4_ANYDADDR;	
			/* type */
			ret = get_byte_address(&current, &proto_len, &chan_addr->ip.type);
			if (!ret)
				*flags |= AN_DEMUXKEY_IPV4_ANYPROTO;	
			break;
		case CHANPROTO_IN_UDP:
		case CHANPROTO_OUT_UDP:
			if (0 == len) {
				*flags = AN_DEMUXKEY_UDP_ALLFLAGS;
				return true;	
			}
			*flags = 0;
			/* sport */
			ret = get_short_address(&current, &proto_len,
					&chan_addr->udp.sport);
			if (!ret) 
				*flags |= AN_DEMUXKEY_UDP_ANYSPORT;
			/* dport */
			ret = get_short_address(&current, &proto_len,
					&chan_addr->udp.dport);
			if (!ret) 
				*flags |= AN_DEMUXKEY_UDP_ANYDPORT;
			break;
		case CHANPROTO_OUT_UDPIP:
			if (0 == len) {
				*flags = 0;
				return false;
			}
			*flags = 0;
			memset(chan_addr, 0, sizeof(*chan_addr));
			get_ip_address(&current, &proto_len, chan_addr->udpip.saddr);
			get_short_address(&current, &proto_len, &chan_addr->udpip.sport);
			get_ip_address(&current, &proto_len, chan_addr->udpip.daddr);
			get_short_address(&current, &proto_len, &chan_addr->udpip.dport);
			break;
		case CHANPROTO_IN_ARP:
		case CHANPROTO_OUT_ARP:
			if (0 == len) {
				*flags = 0;
				return false;
			}
			rv = get_ip_address(&current, &proto_len, chan_addr->arp.ip);
			if (rv)
				rv = get_ethernet_address(&current, &proto_len, chan_addr->arp.eth);
			break;
		default:
		    rv = false;
			ani_log_printf(ANI_LOG_DEBUG, 
					"Unknown protocol in get_next_address\n");
			break;
	}
	return rv;
}

