/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
#include <stdio.h>
#include <string.h> /* XXX memset */

#include <nodeos/an_chan.h>

static an_chanstats_t totstats;
static void print_stats(an_chanstats_t *stats,
			int verbose, const char *str, int interval);
static void diff_stats(an_chanstats_t *new, an_chanstats_t *old,
		       an_chanstats_t *diff);
static void add_stats(an_chanstats_t *stats);

an_error_t
moab_chan_printstats(an_chan_t *chan, int verbose,
		     const char *str, int interval, int accum)
{
	an_chanstats_t stats;
	an_error_t err;

	
	if (chan == 0) {
		print_stats(&totstats, verbose, str, interval);
		memset(&totstats, 0, sizeof totstats);
		return 0;
	}

	err = an_chan_stats(chan, &stats);
	if (err == 0) {
		print_stats(&stats, verbose, str, interval);
		if (accum)
			add_stats(&stats);
	}

	return err;
}

an_error_t
moab_chan_printstatsdelta(an_chan_t *chan, int verbose, const char *str,
			  int interval, void **cookie)
{
	an_chanstats_t *statp, stats, dstats;

	if (cookie == 0)
		return -1;

	statp = *cookie;
	if (statp == 0) {
		statp = an_mempool_alloc(an_mempool_current(), sizeof *statp);
		if (statp == 0)
			return -1;
		memset(statp, 0, sizeof *statp);
		*cookie = statp;
	}
	if (an_chan_stats(chan, &stats) != 0)
		return -1;
	diff_stats(&stats, statp, &dstats);
	print_stats(&dstats, verbose, str, interval);
	*statp = stats;

	return 0;
}

static void
print_stats(an_chanstats_t *stats, int verbose, const char *str, int interval)
{
	if (stats->inpackets == 0 && stats->outpackets == 0 &&
	    stats->droppackets == 0) {
		if (verbose)
			printf("%s: no packets\n", str);
		return;
	}

	if (stats->inpackets > 0) {
		printf("%s: %qu pkts", str, stats->inpackets);
		if (interval > 0)
			printf(" (%u pkts/s)",
			       (unsigned)(stats->inpackets/interval));
		printf(", %qu bytes", stats->inbytes);
		if (interval > 0)
			printf("(%u bytes/s)",
			       (unsigned)(stats->inbytes/interval));
		printf(" received\n");
		if (verbose)
			printf("  %qu pkts copied, %qu queued\n",
			       stats->inbufcopies, stats->inbufqueued);
	}

	if (stats->outpackets > 0) {
		printf("%s: %qu pkts", str, stats->outpackets);
		if (interval > 0)
			printf(" (%u pkts/s)",
			       (unsigned)(stats->outpackets/interval));
		printf(", %qu bytes", stats->outbytes);
		if (interval > 0)
			printf("(%u bytes/s)",
			       (unsigned)(stats->outbytes/interval));
		printf(" sent\n");
		if (verbose)
			printf("  %qu pkts copied\n", stats->outbufcopies);
	}

	if (stats->droppackets > 0) {
		printf("%s: %qu pkts", str, stats->droppackets);
		if (interval > 0)
			printf(" (%u pkts/s)",
			       (unsigned)(stats->droppackets/interval));
		printf(", %qu bytes", stats->dropbytes);
		if (interval > 0)
			printf("(%u bytes/s)",
			       (unsigned)(stats->dropbytes/interval));
		printf(" dropped\n");
		if (verbose)
			printf("  %qu no bufs, %qu failed buffer setup\n",
			       stats->dropnobufs, stats->dropfailedbuf);
	}
}

static void
diff_stats(an_chanstats_t *new, an_chanstats_t *old, an_chanstats_t *diff)
{
	diff->inpackets = new->inpackets - old->inpackets;
	diff->inbytes = new->inbytes - old->inbytes;
	diff->outpackets = new->outpackets - old->outpackets;
	diff->outbytes = new->outbytes - old->outbytes;
	diff->droppackets = new->droppackets - old->droppackets;
	diff->dropnobufs = new->dropnobufs - old->dropnobufs;
	diff->dropfailedbuf = new->dropfailedbuf - old->dropfailedbuf;
	diff->dropbadpacket = new->dropbadpacket - old->dropbadpacket;
	diff->dropbytes = new->dropbytes - old->dropbytes;
	diff->inbufqueued = new->inbufqueued - old->inbufqueued;
	diff->inbufcopies = new->inbufcopies - old->inbufcopies;
	diff->outbufcopies = new->outbufcopies - old->outbufcopies;
}

static void
add_stats(an_chanstats_t *stats)
{
	totstats.inpackets += stats->inpackets;
	totstats.inbytes += stats->inbytes;
	totstats.outpackets += stats->outpackets;
	totstats.outbytes += stats->outbytes;
	totstats.droppackets += stats->droppackets;
	totstats.dropnobufs += stats->dropnobufs;
	totstats.dropfailedbuf += stats->dropfailedbuf;
	totstats.dropbadpacket += stats->dropbadpacket;
	totstats.dropbytes += stats->dropbytes;
	totstats.inbufqueued += stats->inbufqueued;
	totstats.inbufcopies += stats->inbufcopies;
	totstats.outbufcopies += stats->outbufcopies;
}
