/*
 * Copyright (c) 1999-2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * NodeOS Demuxkey API
 */
#ifndef _NODEOS_DEMUXKEY_H_
#define _NODEOS_DEMUXKEY_H_

#ifdef NODEOS_SPEC_COMPAT
#define an_Demuxkey an_demuxkey_t
#define an_demuxkeyCreate an_demuxkey_create
#define an_demuxkeyDestroy an_demuxkey_destroy
#define an_demuxkeyAdd an_demuxkey_add
#define an_demuxkeyAddIPv4 an_demuxkey_addipv4
#define an_demuxkeyAddIPv4UDP an_demuxkey_addipv4udp
#define an_demuxkeyAddIPv4ANEP an_demuxkey_addipv4anep
#endif

#include <nodeos/an_impl.h>
#include <nodeos/an_domain.h>
#include <nodeos/impl/an_demuxkey.h>

/*
 * Create an empty demux key (packet filter) object.
 * Returns a pointer to the new object or zero on error.
 */
an_demuxkey_t *an_demuxkey_create(void *mem, an_domain_t *domain);

/*
 * Destroy a demux key.
 * Frees any resources allocated by an_demuxkey_match operations.
 * XXX make sure demux key is not "in use".
 */
an_error_t an_demuxkey_destroy(an_demuxkey_t *key);

/*
 * Clone a demux key.  New demux key is created in the provided memory.
 * Returns a pointer to the new object or zero on error.
 */
an_demuxkey_t *an_demuxkey_clone(void *mem, an_demuxkey_t *orig);

/*
 * Store the length of the demux key in len.
 */
an_error_t an_demuxkey_len(an_demuxkey_t *key, int *len);

/*
 * Adds a <length, data, mask> tuple segment to the indicated demux.
 * The offset at which the tuple is matched against a packet is implicit.
 * The first segment is matched starting at offset 0 of the packet payload.
 * Successive segments are matched at the offset immediately following
 * the range matched by the previous segment.
 *
 * If there is a range of the packet that needs to be skipped, that must
 * be encoded in a segment using a "don't care" value/mask (0/0).
 * XXX this strikes me as a bad idea.
 *
 * The match data buffer is currently copied and hence may be overwritten
 * upon return from this call.
 */
an_error_t an_demuxkey_add(an_demuxkey_t *key, int len, void *data, void *mask);

/*
 * Adds an ethernet header match segment to the indicated demuxkey.
 */
an_error_t an_demuxkey_addeth(an_demuxkey_t *key, int flags,
			      u_int8_t src[6], u_int8_t dst[6],
			      u_int16_t etype);

/* flags */
#define AN_DEMUXKEY_ETH_ANYSADDR	1
#define AN_DEMUXKEY_ETH_ANYDADDR	2
#define AN_DEMUXKEY_ETH_ANYETYPE	4
#define AN_DEMUXKEY_ETH_ALLFLAGS	7

/*
 * Adds an IPv4 header match segment to the indicated demux key.
 * This segment matches a particular IPv4 header (with the required
 * source, dest, and protocol fields).  This key implicitly 
 * matches variable length IP headers.  Thus following segments
 * will be matched with a packet offset following the entire IP header.
 * This function does NOT perform an IP header checksum comparison,
 * that is, this segment can match a corrupted IP header.
 */
an_error_t an_demuxkey_addipv4(an_demuxkey_t *key, int flags,
			       u_int32_t src, u_int32_t srcmask,
			       u_int32_t dst, u_int32_t dstmask,
			       u_int8_t proto);

/* flags */
#define AN_DEMUXKEY_IPV4_ANYSADDR	1
#define AN_DEMUXKEY_IPV4_ANYDADDR	2
#define AN_DEMUXKEY_IPV4_ANYPROTO	4
#define AN_DEMUXKEY_IPV4_ALLFLAGS	7

/*
 * Add a segment that matches a UDP header.
 * This function only matches the header, a packet matching this segment
 * is not necessarily a valid UDP packet (i.e., the length and checksum
 * are not verified).
 *
 * Usually used in conjunction with an_demuxkey_addipv4.
 */
an_error_t an_demuxkey_addudp(an_demuxkey_t *key, int flags,
			      u_int16_t src, u_int16_t dst);

/* flags */
#define AN_DEMUXKEY_UDP_ANYSPORT	1
#define AN_DEMUXKEY_UDP_ANYDPORT	2
#define AN_DEMUXKEY_UDP_ALLFLAGS	3

/*
 * Add a segment to the given demux key to match a TCP header.
 * This function only matches the header, a packet matching this segment
 * is not necessarily a valid TCP packet (i.e., the checksum is not verified).
 *
 * Usually used in conjunction with an_demuxkey_addipv4.
 */
an_error_t an_demuxkey_addtcp(an_demuxkey_t *key, int flags,
			      u_int16_t src, u_int16_t dst);

/* flags */
#define AN_DEMUXKEY_TCP_ANYSPORT	1
#define AN_DEMUXKEY_TCP_ANYDPORT	2
#define AN_DEMUXKEY_TCP_ALLFLAGS	3

/*
 * Add a segment to the given demux key to match an ANEP header.
 * This segment implicitly matches variable length ANEP headers.
 * Thus following segments will be matched with a packet offset following
 * the entire ANEP header.
 */
an_error_t an_demuxkey_addanep(an_demuxkey_t *key, int flags,
			       u_int8_t version, u_int16_t protoid);

/* flags */
#define AN_DEMUXKEY_ANEP_ANYVERSION	1
#define AN_DEMUXKEY_ANEP_ANYPROTOID	2
#define AN_DEMUXKEY_ANEP_ALLFLAGS	3

/* 
 * Dump a representation of the key into the given character buffer.
 */
an_error_t an_demuxkey_tostring(an_demuxkey_t *key, char *buf, int buflen);

/*
 * Test a data buffer against the given key.
 * Returns -1 if there was an error, otherwise the index
 * of the first byte that was not part of the match.  (Returns
 * 0 if there was no match.)
 * 'data' is the chunk of data to compare against.
 * 'datalength' is the length of data.
 *
 * XXX doesn't work with "null" keys.  Should banish null keys?
 */
int an_demuxkey_matches(an_demuxkey_t *key, void *data, int datalength);

#endif /* _NODEOS_DEMUXKEY_H_ */
