/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * NodeOS Event API
 */
#ifndef _NODEOS_EVENT_H_
#define _NODEOS_EVENT_H_

#ifdef NODEOS_SPEC_COMPAT
#define an_Event_t	   an_event_t
#define an_EventFunc_t	   an_eventfunc_t
#define an_EventType_t	   an_eventtype_t
#define an_EventState_t	   an_eventstate_t
#define an_eventCreate     an_event_create
#define an_eventDestroy    an_event_destroy
#define an_eventSchedule   an_event_schedule
#define an_eventFire	   an_event_fire
#define an_eventStatus	   an_event_status
#define an_eventDeschedule an_event_deschedule
#define an_eventRedefine   an_event_redefine
#endif

#include <nodeos/an_error.h>
#include <nodeos/an_impl.h>
#include <nodeos/an_time.h>
#include <nodeos/impl/an_event.h>


/*
 * Create an event object.  The object is unscheduled.  If no mem
 * is provided, an object is implicitly allocated and returned.
 * 
 * Errors: ENOMEM (returns NULL)
 * 	   mem is invalid
 */
an_event_t *an_event_create(void *mem);

/*
 * Destroy the event.  Deschedule if currently scheduled.  This can
 * be called at any time on any event.  If non-null, *eventType and
 * *eventState will hold the status of the event at the time it was
 * destroyed.
 *
 * Errors: not a valid event.
 *         eventState/eventType is neither a valid pointer nor NULL.
 */
an_error_t an_event_destroy(an_event_t *event, an_eventtype_t *eventType,
			    an_eventstate_t *eventState);


/*
 * Schedule an event.  When the event triggers, the given
 * callback will be executed and passed the event and the
 * given arg.  The trigger time depends on the type of event
 * (AN_EVTYPE_PERIODIC, AN_EVTYPE_ONESHOT_RELATIVE, or
 * AN_EVTYPE_ONESHOT_ABSOLUTE).  For the first two types,
 * the given time is a delta from the current time, for the
 * last it is an absolute time.  A periodic timer will first
 * fire after one delta has elapsed.
 *
 * One-shot events are implicitly "descheduled" immediately before the
 * callback is invoked.  Event must not already be scheduled.
 *
 * Periodic events may overlap if the callback takes longer to
 * execute than the period.
 *   
 * Errors: one-shot+absolute time that has already passed
 *	 delta time is too small (smaller than clock granularity)
 *	 not a valid event
 *	 event is already scheduled 
 *	 {time|callback|type} are invalid
 */
an_error_t an_event_schedule(an_event_t *event, an_timespec_t *etime,
			     an_eventtype_t type, an_eventfunc_t callback,
			     void *arg);

/*
 *   Fire an event off immediately.  Exactly equivalent to
 *   an_event_schedule of a one-shot with a delta time of zero.
 *
 *   Errors: not a valid event
 *	 event is already scheduled
 *	 callback is invalid
 */
an_error_t an_event_fire(an_event_t *event, an_eventfunc_t callback,
			 void *arg);

/*
 * Return the status of the given event.  Status includes
 * the type of event:
 *        one-shot|periodic|descheduled
 * plus the state:
 *        scheduled|running(periodic only)|descheduled
 * For perioidic events, the state may not be correct if
 * the handler takes longer than the period to run  (the
 * state will be running|scheduled, but if its 'scheduled'
 * there may be a handler still running).
 *
 * Errors: not a valid event
 *     eventStatus is neither a valid pointer nor NULL.
 */
an_error_t an_event_status(an_event_t *event, an_eventtype_t *eventType,
			   an_eventstate_t *eventState);

/*
 * Deschedule an event.  If eventState/eventType is non-null the
 * status of the event at time of deschedule is put in eventState/eventType.
 * Same values as an_event_status() reports.  No error if the
 * event isn't scheduled.  
 *
 * Errors: not a valid event
 *	  eventState/eventType is not a valid pointer
 */
an_error_t an_event_deschedule(an_event_t *event, an_eventtype_t *eventType,
			       an_eventstate_t *eventState);

    	     
/*
 *   Redefine the given event to have the given new parameters.
 *   NULL parameters are left unchanged (cannot legally have a NULL
 *   callback or NULL time).  Flags are same as for schedule.
 *   Status of event just before reschedule is returned in
 *   *eventStatus if that pointer isn't null.
 *   Provides useful additional atomicity wrt to events firing.
 *
 *   Errors: not a valid event
 *	 not a valid time
 *	 not valid flags
 *	 time is too small
 *	 {eventStatus|oldtime|oldFlags|oldCallback} are 
 *		 neither valid nor NULL.
 */
an_error_t an_event_redefine(an_event_t *event,
			     an_timespec_t *oldtime, an_eventtype_t *oldtype,
			     an_eventstate_t *oldstate,
			     an_eventfunc_t *oldfunc, void **oldfarg,
			     an_timespec_t *newtime, an_eventtype_t newtype,
			     an_eventfunc_t newfunc, void *newfarg);

/*
 * Return the granularity (in nanoseconds) of the NodeOS
 * event timer.
 */
unsigned long an_event_granularity(void);

#endif /* _NODEOS_EVENT_H_ */
