/*
 * Copyright (c) 1999-2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * NodeOS internal representation of a channel.
 */

#ifndef _NODEOS_ANI_CHANNEL_H_
#define _NODEOS_ANI_CHANNEL_H_

#define ANI_CHANNEL_STATS

#include <nodeos/an_impl.h>
#include <nodeos/an_resource.h>
#include <nodeos/impl/decls.h>

#include <nodeos/ani/ani_error.h>
#include <nodeos/ani/ani_ilock.h>
#include <nodeos/ani/ani_pbuf.h>
#include <nodeos/ani/ani_domain.h>

struct chanops;
struct ani_demuxkey;
struct ani_threadpool;

struct ani_chan {
	an_chanaddr_t	 addr;	   /* peer addressing info */
	an_chanproto_t	 proto;	   /* protocol processing info */
	struct ani_chan	 *next;	   /* link of all channels for a domain */
	struct ani_domain *domain;    /* domain we are associated with */
	struct ani_threadpool *tpool; /* thread pool for our CPU resources */
	int		 curthreads;/* threads processing packets */
	int		 maxthreads;/* max threads allowed */
	struct chanops	 *ops;	   /* operations */
	int		 flags;	   /* channel type and assorted flags */
	union {
		struct {
			struct ani_demuxkey *demux;/* demux information */
			ani_queue_head_t    rbufs; /* free, !busy pbufs */
			ani_queue_head_t    bbufs; /* free, busy pbufs */
			ani_queue_head_t    qbufs; /* queued, full pbufs */
			an_chanrecvfunc_t   func;  /* delivery function */
			void		    *farg; /* delivery function arg */
		} in;
		struct {
			unsigned int	mbs;	   /* mbits/sec of bandwidth */
		} out;
		struct {
			struct ani_chan	*inchan;   /* component in-chan */
			struct ani_chan	*outchan;  /* component out-chan */
		} cut;
	}		 tdata;	   /* channel type specific data */
	/*
	 * For Click, but would also be useful for cut. There will
	 * eventually be multiple component channels, but for now a
	 * click composite channel consists of just a single in or out.
	 */
	struct {
		struct ani_chan	*rawchan;  /* component in-chan */
		void		*data;	   /* Another odd bit of data */
	}		 component;
	void		 *data;	   /* channel-protocol specific data */
	int		 refs;
	ani_ilock_t	 lock;
	an_chanstats_t	 stats;	   /* channel statistics */	
};
typedef struct ani_chan ani_chan_t;

/* channel types */
#define CHAN_IN		1
#define CHAN_OUT	2
#define CHAN_CUT	3
#define CHAN_TMASK	3

/* channel flags */
#define CHAN_DYING	0x04	/* channel has been destroyed */
#define CHAN_WITHCUT	0x08	/* associated with a cut-through channel */
#define CHAN_ISRAW	0x10	/* provide protocol headers (if possible) */
#define CHAN_ISCLICK	0x20	/* Is a Click based channel */
#define CHAN_WITHCLICK  0x40	/* Input/output to/from a Click channel */

#ifdef ANI_CHANNEL_STATS
#define CHANSTATS(s)	s
#else
#define CHANSTATS(s)
#endif

struct an_infoif;

struct chanops {
	void (*init)(void);
	void (*deinit)(void);
	int (*matches)(an_chanproto_t *);
	ani_error_t (*create)(ani_chan_t *);
	void (*destroy)(ani_chan_t *);
	ani_error_t (*send)(ani_chan_t *, ani_pbuf_t *);
	ani_error_t (*enable)(ani_chan_t *);
	void (*disable)(ani_chan_t *);
	int (*poll)(void);
	void (*txflush)(ani_chan_t *);
};

void ani_chan_setup(struct ani_domain *rootdomain);
void ani_chan_shutdown(void);
ani_error_t ani_chan_init(ani_chan_t *chan, int ctype,
			  struct ani_domain *domain,
			  struct ani_threadpool *tpool,
			  an_chanproto_t *proto, an_chanaddr_t *addr,
			  struct ani_demuxkey *demux, int nthreads,
			  ani_pbuf_t **pbufs, int npbufs, unsigned int outbw,
			  an_chanrecvfunc_t func, void *farg);
ani_error_t ani_chan_deinit(ani_chan_t *chan);
void ani_chan_cleanup(ani_chan_t *chan);
ani_error_t ani_chan_enable(ani_chan_t *chan);
void ani_chan_disable(ani_chan_t *chan);
ani_error_t ani_chan_send(ani_chan_t *chan, ani_pbuf_t *pbuf);
ani_error_t ani_chan_handoff(ani_chan_t *chan, ani_pbuf_t *pbuf);
an_chanstats_t *ani_chan_stats(ani_chan_t *chan);
void ani_chan_thread_start(ani_chan_t *chan, ani_pbuf_t *pbuf, int msize);
void ani_chan_thread_startintr(ani_chan_t *chan, ani_pbuf_t *pbuf, int msize);
void ani_chan_thread_exit(ani_chan_t *chan);
ani_error_t ani_chan_pbuf_alloc(ani_chan_t *chan, size_t size,
				ani_pbuf_t **out_pbuf);
void ani_chan_pbuf_free(ani_chan_t *chan, ani_pbuf_t *pbuf);
ani_error_t ani_chan_pbuf_add(ani_chan_t *chan,
			      ani_pbuf_t *pbufs[], int npbufs);
void ani_chan_pbuf_removeall(ani_chan_t *chan);

ani_error_t ani_cutchan_init(ani_chan_t *chan, struct ani_domain *domain,
			     struct ani_threadpool *tpool, void *cdata,
			     an_chanproto_t *proto,
			     ani_chan_t *inchan, ani_chan_t *outchan,
			     an_chanrecvfunc_t rfunc, void *rfuncarg);
void ani_cutchan_receive(struct an_pbuf *pbuf, void *arg, int msize);
#ifdef ANI_CHANNEL_STATS
void ani_chan_stats_dump(ani_chan_t *chan);
#endif
#ifdef USE_POLLING
void ani_chan_poll(void *foo, void *arg);
void ani_chan_txflush(ani_chan_t *chan);
#endif

/*
 * Click Stuff.
 */
ani_error_t ani_clickchan_init(ani_chan_t *chan, int ctype,
			       ani_domain_t *domain,
			       ani_threadpool_t *tpool,
			       an_chanproto_t *proto, an_chanaddr_t *addr,
			       struct ani_demuxkey *demux, int nthreads,
			       ani_pbuf_t **pbufs, int npbufs,
			       unsigned int outbw,
			       an_chanrecvfunc_t func, void *farg,
			       /* Click stuff */
			       ani_chan_t *rawchan);

ani_error_t ani_clickchan_instantiate(char *config,
				ani_chan_t *iinchan, ani_chan_t *ioutchan,
				ani_threadpool_t *click_tpool,
				void (*chan_intr)(ani_chan_t *, ani_pbuf_t *),
				an_chanrecvfunc_t *rfunc, void **rfuncarg,
				void **router);
		   
#include <nodeos/ani/ani_chan_inline.h>

#endif /* _NODEOS_ANI_CHANNEL_H_ */
