/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#ifndef _NODEOS_ANI_DEMUXKEY_H_
#define _NODEOS_ANI_DEMUXKEY_H_

/*
 * Moab demuxkey segments come in two flavors:
 *
 * 1. A <off, len, key, mask> tuple.  A segment matches if "len" bytes of the
 *    packet data at offset "off" are equal to "key" according to the formula:
 *
 *	(data[byte] & mask[byte]) == key[byte]
 *
 *    If a segment key == NULL, the segment never matches.
 *    If a segment mask == NULL, the key must match exactly (i.e., mask == -1)
 *
 * 2. A packet-type specific function.  Specialize routines for matching
 *    (potentially variable length) headers for specific protocols
 *    (e.g., ANTS).
 *
 * The first in handled in the general context of the latter.
 */

/*
 * Demuxkey segment "match" function.
 *
 * Determines if the given packet data (pdata/plen) meets the critera of the
 * indicated demux data (ddata).  Returns the offset of the first byte
 * in the packet AFTER the part that matches.  Returns -1 if there is no
 * match.
 */
typedef int (*demuxkeyseg_func_t)(void *ddata, void *pdata, int plen);

/*
 * Demuxkey segment.
 *
 * Describes a function to apply to a range of packet data.  The indicated
 * function (``func'') is called with a pointer to packet data and the
 * segment-specific data (``data'').  The packet data pointer passed to the
 * function is ``offset'' bytes beyond any previous matched segment
 * (i.e., the value returned by the previous segment's function).
 * ``mlen'' is the amount of data covered by this segment.  For functions
 * which deal with variable length matches, it is the minimum amount matched.
 * This field is used for optimization (determining if a packet is even long
 * enough to possibly be matched by a demuxkey) and for prioritization of a
 * demuxkey (determining the order in which demuxkeys are tried).
 * ``dlen'' is the length of the segment-specific data and is used only
 * for generic segment deallocation.
 */
struct ani_demuxkeyseg {
	struct ani_demuxkeyseg	*next;
	int			offset;
	demuxkeyseg_func_t	func;
	int			mlen;
	int			dlen;
	unsigned char		data[0];
};
typedef struct ani_demuxkeyseg ani_demuxkeyseg_t;

/*
 * Demuxkey.
 *
 * A list of matching functions to be applied to a packet's data.
 * A demuxkey consists of a list of segments which must all be matched.
 *
 * The ``demuxkey list'' fields are items which need to be moved elsewhere
 * if we decide to allow keys to be associated with multiple channels.
 * We don't do this currently, in part because the chan_eth code adds
 * segments to user provided demuxkeys during channel creation.
 */
struct ani_chan;
struct ani_demuxkey {
	/* demuxkey fields */
	ani_demuxkeyseg_t	*segs;	 /* segments associated with key */
	int			minpsize;/* min len packet this could match */
	int			prio;	 /* priority (for list ordering) */

	/* demuxkey list fields */
	struct ani_chan		*chan;	/* associated channel */
	struct ani_demuxkey	*next;	/* next demuxkey in a list */
};
typedef struct ani_demuxkey ani_demuxkey_t;

struct ani_demuxkeylist {
	ani_demuxkey_t		*head;	/* first demuxkey on list */
};
typedef struct ani_demuxkeylist ani_demuxkeylist_t;

/* demuxkey segment ops */
ani_demuxkeyseg_t *ani_demuxkeyseg_tuple(int off, int len, void *key,
					 void *mask);
ani_demuxkeyseg_t *ani_demuxkeyseg_eth(int off, int flags, u_int8_t src[6],
				       u_int8_t dst[6], u_int16_t etype);
ani_demuxkeyseg_t *ani_demuxkeyseg_ipv4(int off, int flags,
					u_int32_t src, u_int32_t srcmask,
					u_int32_t dst, u_int32_t dstmask,
					u_int8_t proto);
ani_demuxkeyseg_t *ani_demuxkeyseg_udp(int off, int flags,
				       u_int16_t sport, u_int16_t dport);
ani_demuxkeyseg_t *ani_demuxkeyseg_tcp(int off, int flags,
				       u_int16_t sport, u_int16_t dport);
ani_demuxkeyseg_t *ani_demuxkeyseg_anep(int off, int flags,
					u_int8_t version, u_int16_t protoid);

void ani_demuxkeyseg_destroy(ani_demuxkeyseg_t *seg);

/* demuxkey ops */
ani_error_t ani_demuxkey_init(ani_demuxkey_t *dkey);
ani_error_t ani_demuxkey_deinit(ani_demuxkey_t *dkey);
ani_error_t ani_demuxkey_clone(ani_demuxkey_t *ndkey, ani_demuxkey_t *odkey);
ani_error_t ani_demuxkey_addchannel(ani_demuxkey_t *dkey,
				    struct ani_chan *chan);
void ani_demuxkey_removechannel(ani_demuxkey_t *dkey, struct ani_chan *chan);
ani_error_t ani_demuxkey_adjust(ani_demuxkey_t *dkey, int offset);
ani_error_t ani_demuxkey_prepend(ani_demuxkey_t *dkey, ani_demuxkeyseg_t *seg);
ani_error_t ani_demuxkey_append(ani_demuxkey_t *dkey, ani_demuxkeyseg_t *seg);
ani_error_t ani_demuxkey_len(ani_demuxkey_t *dkey, int *len);
int ani_demuxkey_match(ani_demuxkey_t *dkey, void *pdata, int plen);
int ani_demuxkey_match_with_header(ani_demuxkey_t *dkey, void *hdr, int hlen,
				   void *pdata, int plen);
ani_error_t ani_demuxkey_tostring(ani_demuxkey_t* dkey, char* buf, int buflen);
#ifdef DEBUG
void ani_demuxkey_dump(ani_demuxkey_t *dkey);
void ani_demuxkeylist_dump(ani_demuxkeylist_t *list);
#endif

/* demuxkey list ops */
void ani_demuxkeylist_init(ani_demuxkeylist_t *list);
ani_error_t ani_demuxkeylist_add(ani_demuxkeylist_t *list,
				 ani_demuxkey_t *dkey);
void ani_demuxkeylist_remove(ani_demuxkeylist_t *list, ani_demuxkey_t *dkey);
struct ani_chan *ani_demuxkeylist_apply(ani_demuxkeylist_t *list,
					void *data, int *lenp);
struct ani_chan *ani_demuxkeylist_apply_with_header(ani_demuxkeylist_t *list,
						    void *hdr, int hlen,
						    void *data, int *lenp);
#endif /* _NODEOS_ANI_DEMUXKEY_H_ */
