/*
 * Copyright (c) 1999-2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#ifndef _NODEOS_ANI_DOMAIN_H_
#define _NODEOS_ANI_DOMAIN_H_

#include <nodeos/an_resource.h>
#include <nodeos/impl/an_impl.h>

#include <nodeos/ani/ani_hnode.h>
#include <nodeos/ani/ani_resource.h>
#include <nodeos/ani/ani_ilock.h>
#include <nodeos/ani/ani_threadpool.h>

struct ani_cred;
struct ani_chan;
struct ani_mempool;

/*
 * The domain: rendezvous point for threads and resources.
 *
 * Every domain is part of the domain hierarchy as well as possibly associated
 * with a mempool.
 */

struct ani_domain {
	struct hnode		me;	/* domain hier link (must be first) */
	struct ani_cred		*cred;		/* credentials */
	struct ani_mempool	*mpool;		/* mempool */
	struct ani_chan		*chanlist;	/* list of channels */
	struct ani_threadpool	tpool;		/* thread pool */
	void			*apistate;	/* API boundary related state */
	int			flags;		/* various flags */
	an_domainterm_func_t	termfunc;	/* termination function */
	ani_resource_t		res;		/* reserved resources */
	int			refs;		/* reference count */
	ani_ilock_t		lock;		/* mutual exclusion */
	an_domainstats_t	stats;		/* statistics */
};
typedef struct ani_domain ani_domain_t;

#define DOMAIN_INUPCALL	1
#define DOMAIN_DYING	2

/*
 * Boot-time initialization of the domain component.
 */
void ani_domain_setup(ani_domain_t *rdomain);

ani_error_t ani_rootdomain_init(ani_domain_t *rdomain, struct ani_cred *rcred,
				struct ani_mempool *rmempool,
				int nthreads, int stacksize,
				void **root_threads, void **root_threadstacks,
				an_domaininit_func_t ifunc, void *initarg);

void ani_rootdomain_deinit_children(ani_domain_t *domain);

void ani_rootdomain_deinit(ani_domain_t *rdomain);

ani_error_t ani_rootdomain_startthread(ani_thread_t *thread,
				       ani_thread_func_t func,
				       void *farg1, void *farg2, void *farg3);

/*
 * Create a new domain.
 * The new domain is initialized with
 * the indicated parent domain, memory pool and thread scheduler.
 */
ani_error_t ani_domain_init(ani_domain_t *domain, struct ani_cred *cred,
			    an_cpuspec_t *cspec, an_memspec_t *mspec,
			    an_domaininit_func_t initfunc, void *initarg,
			    an_domainterm_func_t termfunc);

/*
 * Destroy the indicated domain.
 * Detach from the domain hierarchy, destroy all child domains and channels.
 * Domain resources are not freed til all domain threads are gone.
 */
ani_error_t ani_domain_deinit(ani_domain_t *domain);

/*
 * Final domain destruction.
 * Removes its memory resources from the associated memory pool
 * and frees the data structures.
 */
void ani_domain_cleanup(ani_domain_t *domain);

void ani_domain_addchannel(ani_domain_t *domain, struct ani_chan *chan);
void ani_domain_removechannel(ani_domain_t *domain, struct ani_chan *chan);

/*
 * Fire up a thread from the domain's threadpool at the specified function.
 */
ani_error_t ani_domain_startthread(ani_domain_t *domain, int inkern,
				   void (*startfunc)(void *, void *),
				   void *farg1, void *farg2);

an_domainstats_t *ani_domain_stats(ani_domain_t *domain);

#include <nodeos/ani/ani_domain_inline.h>

#endif /* _NODEOS_ANI_DOMAIN_H_ */

