/*
 * Copyright (c) 1999-2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * Macros and inline functions for the NodeOS system call
 * entry/exit layer and for the intern/extern of various types.
 */
#ifndef _NODEOS_ANI_NODEOS_H_
#define _NODEOS_ANI_NODEOS_H_

#include <string.h>
#include <assert.h>
#include <signal.h>

#include <nodeos/ani/ani_error.h>
#include <nodeos/ani/ani_thread.h>

#ifdef USE_POLLING
extern void ani_chan_poll(void *, void *);
#else
#define ani_chan_poll(x,y)
#endif

/*
 * We disallow both thread interrupts (signals) and thread death (cancelation)
 * while in the NodeOS.  Both are checked on exit from the NodeOS.
 */
#define NODEOS_ENTER { \
	ani_thread_t *th = ani_thread_current(); \
	int ostate; \
	assert_ani_thread_valid(th); \
	pthread_sigmask(SIG_BLOCK, &ani_thread_intrmask, 0); \
	pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, &ostate); \
	assert(ostate == PTHREAD_CANCEL_ASYNCHRONOUS); \
	assert((th->flags & THREAD_INKERNEL) == 0); \
	th->flags |= THREAD_INKERNEL; \
	ani_chan_poll(0, 0); \
}

#define NODEOS_LEAVE { \
	ani_thread_t *th = ani_thread_current(); \
	int ostate; \
	assert_ani_thread_valid(th); \
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &ostate); \
	assert(ostate == PTHREAD_CANCEL_DEFERRED); \
	pthread_testcancel(); \
	assert((th->flags & THREAD_INKERNEL) != 0); \
	th->flags &= ~THREAD_INKERNEL; \
	if (th->flags & THREAD_INTSON) \
		pthread_sigmask(SIG_UNBLOCK, &ani_thread_intrmask, 0); \
}

/*
 * For many API calls, it doesn't matter if a thread is interrupted or
 * cancelled as there are no dangerous side-effects (e.g., taking a lock)
 * and the calls cannot block.  For these calls, we don't even bother to
 * record that we are in the NodeOS.
 */
#define LW_NODEOS_ENTER
#define LW_NODEOS_LEAVE

/*
 * Copy memory into and out of NodeOS space.
 * Even if memory is directly accessible between NodeOS and EE space
 * (as in our implementation), these calls must copy the memory.
 * These calls must also always set the error return value.
 */
#define NODEOS_COPYIN(ndst, usrc, len, errp) { \
	memcpy((ndst), (usrc), (len)); \
	*errp = ani_error_clear(); \
}

#define NODEOS_COPYOUT(udst, nsrc, len, errp) { \
	memcpy((udst), (nsrc), (len)); \
	*errp = ani_error_clear(); \
}

/* Prototypes for object specific create/destroy/intern/extern functions */

struct an_domain;
struct ani_domain;

struct ani_domain *
NODEOS_CREATE_DOMAIN(struct ani_domain *domain, struct an_domain *xdomain);

void
NODEOS_DESTROY_DOMAIN(struct ani_domain *domain, struct ani_domain *idomain);

struct ani_domain *
NODEOS_INTERN_DOMAIN(struct ani_domain *domain, struct an_domain *xdomain);

struct an_domain *
NODEOS_EXTERN_DOMAIN(struct ani_domain *domain, struct ani_domain *idomain);


struct an_cred;
struct ani_cred;

struct ani_cred *
NODEOS_CREATE_CRED(struct ani_domain *domain, struct an_cred *xcred);

void
NODEOS_DESTROY_CRED(struct ani_domain *domain, struct ani_cred *icred);

struct ani_cred *
NODEOS_INTERN_CRED(struct ani_domain *domain, struct an_cred *xcred);

struct an_cred *
NODEOS_EXTERN_CRED(struct ani_domain *domain, struct ani_cred *icred);


struct an_thread;
struct ani_thread;

struct ani_thread *
NODEOS_CREATE_THREAD(struct ani_domain *domain, struct an_thread *xthread);

void
NODEOS_DESTROY_THREAD(struct ani_domain *domain, struct ani_thread *ithread);

struct ani_thread *
NODEOS_INTERN_THREAD(struct ani_domain *domain, struct an_thread *xthread);

struct an_thread *
NODEOS_EXTERN_THREAD(struct ani_domain *domain, struct ani_thread *ithread);


struct an_mempool;
struct ani_mempool;

struct ani_mempool *
NODEOS_CREATE_MEMPOOL(struct ani_domain *domain, struct an_mempool *xmempool);

void
NODEOS_DESTROY_MEMPOOL(struct ani_domain *domain, struct ani_mempool *imempool);

struct ani_mempool *
NODEOS_INTERN_MEMPOOL(struct ani_domain *domain, struct an_mempool *xmempool);

struct an_mempool *
NODEOS_EXTERN_MEMPOOL(struct ani_domain *domain, struct ani_mempool *imempool);


struct an_chan;
struct ani_chan;

struct ani_chan *
NODEOS_CREATE_CHAN(struct ani_domain *domain, struct an_chan *xchan);

void
NODEOS_DESTROY_CHAN(struct ani_domain *domain, struct ani_chan *ichan);

struct ani_chan *
NODEOS_INTERN_CHAN(struct ani_domain *domain, struct an_chan *xchan);

struct an_chan *
NODEOS_EXTERN_CHAN(struct ani_domain *domain, struct ani_chan *ichan);


struct an_pbuf;
struct ani_pbuf;

struct ani_pbuf *
NODEOS_CREATE_PBUF(struct ani_domain *domain, struct an_pbuf *xpbuf);

void
NODEOS_DESTROY_PBUF(struct ani_domain *domain, struct ani_pbuf *ipbuf);

struct ani_pbuf *
NODEOS_INTERN_PBUF(struct ani_domain *domain, struct an_pbuf *xpbuf);

struct an_pbuf *
NODEOS_EXTERN_PBUF(struct ani_domain *domain, struct ani_pbuf *ipbuf);


struct an_demuxkey;
struct ani_demuxkey;

struct ani_demuxkey *
NODEOS_CREATE_DEMUXKEY(struct ani_domain *domain, struct an_demuxkey *xdemuxkey);

void
NODEOS_DESTROY_DEMUXKEY(struct ani_domain *domain, struct ani_demuxkey *idemuxkey);

struct ani_demuxkey *
NODEOS_INTERN_DEMUXKEY(struct ani_domain *domain, struct an_demuxkey *xdemuxkey);

struct an_demuxkey *
NODEOS_EXTERN_DEMUXKEY(struct ani_domain *domain, struct ani_demuxkey *idemuxkey);


struct an_mutex;
struct ani_mutex;

struct ani_mutex *
NODEOS_CREATE_MUTEX(struct ani_domain *domain, struct an_mutex *xmutex);

void
NODEOS_DESTROY_MUTEX(struct ani_domain *domain, struct ani_mutex *imutex);

struct ani_mutex *
NODEOS_INTERN_MUTEX(struct ani_domain *domain, struct an_mutex *xmutex);

struct an_mutex *
NODEOS_EXTERN_MUTEX(struct ani_domain *domain, struct ani_mutex *imutex);


struct an_cond;
struct ani_cond;

struct ani_cond *
NODEOS_CREATE_COND(struct ani_domain *domain, struct an_cond *xcond);

void
NODEOS_DESTROY_COND(struct ani_domain *domain, struct ani_cond *icond);

struct ani_cond *
NODEOS_INTERN_COND(struct ani_domain *domain, struct an_cond *xcond);

struct an_cond *
NODEOS_EXTERN_COND(struct ani_domain *domain, struct ani_cond *icond);


struct an_event;
struct ani_event;

struct ani_event *
NODEOS_CREATE_EVENT(struct ani_domain *domain, struct an_event *xevent);

void
NODEOS_DESTROY_EVENT(struct ani_domain *domain, struct ani_event *ievent);

struct ani_event *
NODEOS_INTERN_EVENT(struct ani_domain *domain, struct an_event *xevent);

struct an_event *
NODEOS_EXTERN_EVENT(struct ani_domain *domain, struct ani_event *ievent);

#endif /* _NODEOS_ANI_NODEOS_H_ */
