/*
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * Doubly-linked lists.  Basically the Mach queue.h code with ani_ prefixes.
 */

#ifndef _NODEOS_ANI_QUEUE_H_
#define _NODEOS_ANI_QUEUE_H_

/*
 *
 * Mach Operating System
 * Copyright (c) 1991,1990,1989,1988,1987 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon rights
 * to redistribute these changes.
 */
/*
 *	File:	queue.h
 *	Author:	Avadis Tevanian, Jr.
 *	Date:	1985
 *
 *	Type definitions for generic queues.
 */

/*
 *	Queue of abstract objects.  Queue is maintained
 *	within that object.
 *
 *	Supports fast removal from within the queue.
 *
 *	How to declare a queue of elements of type "foo_t":
 *		In the "*foo_t" type, you must have a field of
 *		type "ani_queue_chain_t" to hold together this queue.
 *		There may be more than one chain through a
 *		"foo_t", for use by different queues.
 *
 *		Declare the queue as a "ani_queue_t" type.
 *
 *		Elements of the queue (of type "foo_t", that is)
 *		are referred to by reference, and cast to type
 *		"ani_queue_entry_t" within this module.
 */

/*
 *	A generic doubly-linked list (queue).
 */

struct ani_queue_entry {
	struct ani_queue_entry	*next;		/* next element */
	struct ani_queue_entry	*prev;		/* previous element */
};

typedef struct ani_queue_entry	*ani_queue_t;
typedef	struct ani_queue_entry	ani_queue_head_t;
typedef	struct ani_queue_entry	ani_queue_chain_t;
typedef	struct ani_queue_entry	*ani_queue_entry_t;

/*
 *	Macro:		ani_queue_init
 *	Function:
 *		Initialize the given queue.
 *	Header:
 *		void ani_queue_init(q)
 *			ani_queue_t		q;	*MODIFIED*
 */
#define	ani_queue_init(q)	((q)->next = (q)->prev = q)

/*
 *	Macro:		ani_queue_first
 *	Function:
 *		Returns the first entry in the queue,
 *	Header:
 *		ani_queue_entry_t ani_queue_first(q)
 *			ani_queue_t	q;		*IN*
 */
#define	ani_queue_first(q)	((q)->next)

/*
 *	Macro:		ani_queue_next
 *	Function:
 *		Returns the entry after an item in the queue.
 *	Header:
 *		ani_queue_entry_t ani_queue_next(qc)
 *			ani_queue_t qc;
 */
#define	ani_queue_next(qc)	((qc)->next)

/*
 *	Macro:		ani_queue_last
 *	Function:
 *		Returns the last entry in the queue.
 *	Header:
 *		ani_queue_entry_t ani_queue_last(q)
 *			ani_queue_t	q;		 *IN*
 */
#define	ani_queue_last(q)	((q)->prev)

/*
 *	Macro:		ani_queue_prev
 *	Function:
 *		Returns the entry before an item in the queue.
 *	Header:
 *		ani_queue_entry_t ani_queue_prev(qc)
 *			ani_queue_t qc;
 */
#define	ani_queue_prev(qc)	((qc)->prev)

/*
 *	Macro:		ani_queue_end
 *	Function:
 *		Tests whether a new entry is really the end of
 *		the queue.
 *	Header:
 *		boolean_t ani_queue_end(q, qe)
 *			ani_queue_t q;
 *			ani_queue_entry_t qe;
 */
#define	ani_queue_end(q, qe)	((q) == (qe))

/*
 *	Macro:		ani_queue_empty
 *	Function:
 *		Tests whether a queue is empty.
 *	Header:
 *		boolean_t ani_queue_empty(q)
 *			ani_queue_t q;
 */
#define	ani_queue_empty(q)		ani_queue_end((q), ani_queue_first(q))


/*----------------------------------------------------------------*/
/*
 * Macros that operate on generic structures.  The queue
 * chain may be at any location within the structure, and there
 * may be more than one chain.
 */

/*
 *	Macro:		ani_queue_enter
 *	Function:
 *		Insert a new element at the tail of the queue.
 *	Header:
 *		void ani_queue_enter(q, elt, type, field)
 *			ani_queue_t q;
 *			<type> elt;
 *			<type> is what's in our queue
 *			<field> is the chain field in (*<type>)
 */
#define ani_queue_enter(head, elt, type, field)			\
{ 								\
	ani_queue_entry_t prev;					\
								\
	prev = (head)->prev;					\
	if ((head) == prev) {					\
		(head)->next = (ani_queue_entry_t) (elt);	\
	} else {						\
		((type)prev)->field.next = (ani_queue_entry_t)(elt);\
	}							\
	(elt)->field.prev = prev;				\
	(elt)->field.next = head;				\
	(head)->prev = (ani_queue_entry_t) elt;			\
}

/*
 *	Macro:		ani_queue_enter_first
 *	Function:
 *		Insert a new element at the head of the queue.
 *	Header:
 *		void ani_queue_enter_first(q, elt, type, field)
 *			ani_queue_t q;
 *			<type> elt;
 *			<type> is what's in our queue
 *			<field> is the chain field in (*<type>)
 */
#define ani_queue_enter_first(head, elt, type, field)		\
{ 								\
	ani_queue_entry_t next;					\
								\
	next = (head)->next;					\
	if ((head) == next) {					\
		(head)->prev = (ani_queue_entry_t) (elt);	\
	} else {						\
		((type)next)->field.prev = (ani_queue_entry_t)(elt);\
	}							\
	(elt)->field.next = next;				\
	(elt)->field.prev = head;				\
	(head)->next = (ani_queue_entry_t) elt;			\
}

/*
 *	Macro:		ani_queue_enter_before
 *	Function:
 *		Insert a new element before the indicated element.
 *	Header:
 *		void ani_queue_enter_before(q, nelt, elt, type, field)
 *			ani_queue_t q;
 *			<type> nelt, elt;
 *			<type> is what's in our queue
 *			<field> is the chain field in (*<type>)
 */
#define ani_queue_enter_before(head, nelt, elt, type, field)		\
{									\
	ani_queue_entry_t prev;						\
									\
	(elt)->field.next = (ani_queue_entry_t)(nelt);			\
	if ((head) == (ani_queue_entry_t)(nelt)) {			\
		(elt)->field.prev = (head)->prev;			\
		prev = (head)->prev;					\
		(head)->prev = (ani_queue_entry_t)(elt);		\
	} else {							\
		(elt)->field.prev = (nelt)->field.prev;			\
		prev = (nelt)->field.prev;				\
		(nelt)->field.prev = (ani_queue_entry_t)(elt);		\
	}								\
	if ((head) == prev)						\
		(head)->next = (ani_queue_entry_t)(elt);		\
	else								\
		((type)prev)->field.next = (ani_queue_entry_t)(elt);	\
}

/*
 *	Macro:		ani_queue_enter_after
 *	Function:
 *		Insert a new element after the indicated element.
 *	Header:
 *		void ani_queue_enter_after(q, pelt, elt, type, field)
 *			ani_queue_t q;
 *			<type> pelt, elt;
 *			<type> is what's in our queue
 *			<field> is the chain field in (*<type>)
 */
#define ani_queue_enter_after(head, pelt, elt, type, field)		\
{									\
	ani_queue_entry_t next;						\
									\
	(elt)->field.prev = (ani_queue_entry_t)(pelt);			\
	if ((head) == (ani_queue_entry_t)(pelt)) {			\
		(elt)->field.next = (head)->next;			\
		next = (head)->next;					\
		(head)->next = (ani_queue_entry_t)(elt);		\
	} else {							\
		(elt)->field.next = (pelt)->field.next;			\
		next = (pelt)->field.next;				\
		(pelt)->field.next = (ani_queue_entry_t)(elt);		\
	}								\
	if ((head) == next)						\
		(head)->prev = (ani_queue_entry_t)(elt);		\
	else								\
		((type)next)->field.prev = (ani_queue_entry_t)(elt);	\
}

/*
 *	Macro:		ani_queue_remove
 *	Function:
 *		Remove an arbitrary item from the queue.
 *	Header:
 *		void ani_queue_remove(q, qe, type, field)
 *			arguments as in ani_queue_enter
 */
#define	ani_queue_remove(head, elt, type, field)		\
{								\
	ani_queue_entry_t next, prev;				\
								\
	next = (elt)->field.next;				\
	prev = (elt)->field.prev;				\
								\
	if ((head) == next)					\
		(head)->prev = prev;				\
	else							\
		((type)next)->field.prev = prev;		\
								\
	if ((head) == prev)					\
		(head)->next = next;				\
	else							\
		((type)prev)->field.next = next;		\
}

/*
 *	Macro:		ani_queue_remove_first
 *	Function:
 *		Remove and return the entry at the head of
 *		the queue.
 *	Header:
 *		ani_queue_remove_first(head, entry, type, field)
 *		entry is returned by reference
 */
#define	ani_queue_remove_first(head, entry, type, field)	\
{								\
	ani_queue_entry_t next;					\
								\
	(entry) = (type) ((head)->next);			\
	next = (entry)->field.next;				\
								\
	if ((head) == next)					\
		(head)->prev = (head);				\
	else							\
		((type)(next))->field.prev = (head);		\
	(head)->next = next;					\
}

/*
 *	Macro:		ani_queue_remove_last
 *	Function:
 *		Remove and return the entry at the tail of
 *		the queue.
 *	Header:
 *		ani_queue_remove_last(head, entry, type, field)
 *		entry is returned by reference
 */
#define	ani_queue_remove_last(head, entry, type, field)		\
{								\
	ani_queue_entry_t prev;					\
								\
	(entry) = (type) ((head)->prev);			\
	prev = (entry)->field.prev;				\
								\
	if ((head) == prev)					\
		(head)->next = (head);				\
	else							\
		((type)(prev))->field.next = (head);		\
	(head)->prev = prev;					\
}

/*
 *	Macro:		ani_queue_assign
 */
#define	ani_queue_assign(to, from, type, field)			\
{								\
	if (!ani_queue_empty(from)) {				\
		((type)((from)->prev))->field.next = (to);	\
		((type)((from)->next))->field.prev = (to);	\
	}							\
	*to = *from;						\
}

/*
 *	Macro:		ani_queue_iterate
 *	Function:
 *		iterate over each item in the queue.
 *		Generates a 'for' loop, setting elt to
 *		each item in turn (by reference).
 *	Header:
 *		ani_queue_iterate(q, elt, type, field)
 *			ani_queue_t q;
 *			<type> elt;
 *			<type> is what's in our queue
 *			<field> is the chain field in (*<type>)
 */
#define ani_queue_iterate(head, elt, type, field)			\
	for ((elt) = (type) ani_queue_first(head);			\
	     !ani_queue_end((head), (ani_queue_entry_t)(elt));		\
	     (elt) = (type) ani_queue_next(&(elt)->field))

/*
 *	Macro:		ani_queue_merge
 *	Function:
 *		append q2 to the end of q1.
 *		q2 is left empty.
 */
#define ani_queue_merge(q1, q2, type, field)				\
{									\
	if (!ani_queue_empty(q2)) {					\
		if (ani_queue_empty(q1)) {				\
			ani_queue_assign(q1, q2, type, field);		\
		} else {						\
			((type)((q1)->prev))->field.next = (q2)->next;	\
			((type)((q2)->next))->field.prev = (q1)->prev;	\
			((type)((q2)->prev))->field.next = (q1);	\
			(q1)->prev = (q2)->prev;			\
		}							\
		ani_queue_init(q2);					\
	}								\
}

#endif /* _NODEOS_ANI_QUEUE_H_ */
