/*
 * Copyright (c) 1999-2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#ifndef _NODEOS_ANI_THREAD_H_
#define _NODEOS_ANI_THREAD_H_

#include <pthread.h>
#include <sched.h>
#include <setjmp.h>
#include <signal.h>
#include <nodeos/an_impl.h>
#include <nodeos/ani/ani_ilock.h>
#include <nodeos/ani/ani_cond.h>
#include <nodeos/ani/ani_threadwork.h>

/*
 * Stuff That Doesn't Belong
 * XXX hack around pthread deficiencies:
 *	OSKit doesn't support attribute scope stuff currently,
 *		the default (only!) scope is system.
 *	FreeBSD 3.x doesn't support cancelation or thread scheduling
 *		the EE had really better cooperate!
 * XXX hack around missing functions in the OSKit/ARM C library:
 *	No sig{set,long}jmp functions, use _{set,long}jmp instead.
 */
#ifndef PTHREAD_CANCEL_ASYNCHRONOUS
#define THREAD_CANCEL_BROKEN
#define PTHREAD_CANCEL_ASYNCHRONOUS	0
#define PTHREAD_CANCEL_DEFERRED		0
#define pthread_setcanceltype(t, sp)	(*(sp) = 0)
#define pthread_cancel(t)		(0)
#define pthread_testcancel()		((void)0)
#undef PTHREAD_SCOPE_SYSTEM /* XXX freebsd defines this but doesn't impl */
#endif

#if !defined(_POSIX_THREAD_PRIORITY_SCHEDULING)
#define pthread_attr_setschedparam(ignore_attr, ignore_prio) (0)
#define pthread_setschedparam(ignore_tid, ignore_policy, ignore_prio) (0)
#endif

#ifndef PTHREAD_SCOPE_SYSTEM
#define pthread_attr_setscope(a,s) 	(0)
#endif

#ifdef OSKIT_ARM32_SHARK
#define sigsetjmp(b, m)		_setjmp(b)
#define siglongjmp(b, v)	_longjmp(b, v)
#endif

struct ani_threadpool;
struct ani_chan;
struct an_cpuspec;

typedef struct ani_thread {
	struct ani_thread *next;	/* threadpool link */
	struct ani_threadpool *pool;	/* our threadpool */
	struct ani_chan *chan;	 /* channel we are processing for */
	pthread_t id;		 /* pthread ID */
	pthread_attr_t attr;	 /* pthread attributes (stackptr, stacksize) */
	struct sched_param pprio;/* pthread scheduling priority */
	int nprio;		 /* NodeOS scheduling priority */
	ani_icond_t waitedon;	 /* place for others to wait for this thread */
	struct ani_threadwork work;	/* work targetted to this thread */
	an_threadintr_func_t handler;	/* interrupt/trap handler */
	ani_ilock_t lock;	 /* lock to protect interrupt state */
	ani_cond_t *cond;	 /* non-zero if in ani_cond_wait */
	int flags;		 /* various flags */
	ani_ilock_t *exitlock;	 /* lock to release on thread exit */
	sigjmp_buf jmpbuf;	 /* warp the thread back to its origin */

	/*
	 * The following may be manipulated at processor interrupt time.
	 * They are protected by the threadpool lock and by disabling
	 * processor interrupts.
	 */
	struct ani_thread *freenext;	/* threadpool freelist link */
	ani_icond_t waiting;	 	/* place for this thread to wait */
} ani_thread_t;

#define THREAD_INKERNEL	0x01	/* thread is in the NodeOS */
#define THREAD_WAITEDON	0x02	/* someone is waiting for thread */
#define THREAD_DONE	0x04	/* waiter is finished with the thread */
#define THREAD_INTSON	0x08	/* interrupts should be allowed */
#define THREAD_ALLOCED	0x10	/* thread was nodeos allocated */
#define THREAD_ONESHOT	0x20	/* thread runs a specific function and dies */
#define THREAD_KILLED	0x40	/* thread has been killed */
#define THREAD_BEDEAD	0x80	/* thread is not attached to a threadpool */
#define THREAD_DOUNLOCK	0x100	/* unlock exitlock on exit */

struct an_cpuspec;

/* variables */

extern sigset_t ani_thread_intrmask;
extern pthread_key_t ani_nodeos_thread;

/* prototypes */

void ani_thread_setup(void);
ani_error_t ani_thread_init(ani_thread_t *thread,
			    void *stackptr, size_t stacksize);
ani_error_t ani_thread_deinit(ani_thread_t *thread);
ani_thread_t *ani_thread_current(void);
ani_error_t ani_thread_spawn(struct ani_threadpool *tpool,
			     ani_thread_t **out_thread);
ani_error_t ani_thread_start(ani_thread_t *thread,
			     struct ani_threadpool *tpool);
void ani_thread_kill(ani_thread_t *thread, int dowait);
void ani_thread_kill_locked(ani_thread_t *thread, int dowait);
void ani_thread_unlock_exit(ani_ilock_t *exitlock);
void ani_thread_exit(void);
void ani_thread_yield(void);
void ani_thread_sleep(an_timespec_t *delay);
void *ani_thread_getdata(void);
void ani_thread_setdata(void *);
void ani_thread_interrupt(ani_thread_t *thread);
void ani_thread_block_interrupt(void);
void ani_thread_unblock_interrupt(void);
void ani_thread_sethandler(an_threadintr_func_t handler);
void ani_thread_condtimedwait(ani_cond_t *cond, an_timespec_t* abstime);
int ani_thread_getprio(ani_thread_t *thread);
ani_error_t ani_thread_setprio(ani_thread_t *thread, int prio);
ani_error_t ani_thread_stackinfo(ani_thread_t *thread,
				 void **base, size_t *size, void **ptr);

#include <nodeos/ani/ani_thread_inline.h>

#endif /* _NODEOS_THREAD_H_ */
