/*
 * Copyright (c) 1999-2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#ifndef _ANI_THREADPOOL_H_
#define _ANI_THREADPOOL_H_

#define ANI_THREADPOOL_CTHREADS	4	/* number of cached threads */
#define ANI_THREADPOOL_STATS

#include <nodeos/an_resource.h>

#include <nodeos/ani/ani_ilock.h>
#include <nodeos/ani/ani_thread.h>
#include <nodeos/ani/ani_threadwork.h>
#include <nodeos/ani/ani_resource.h>

/*
 * The threadpool: a group of threads sharing the same scheduling params
 */

struct ani_domain;
struct ani_chan;

struct ani_threadcpu {
	enum { ANI_CPU_USECS, ANI_CPU_CYCLES } units; /* units of amount */
	long long amount;			      /* amount of CPU */
};
typedef struct ani_threadcpu ani_threadcpu_t;

struct ani_threadstats {
	/* general */
	int active;		/* currently active threads */
	int starts;		/* total thread activations */
	int queued;		/* delayed thread activations (work queued) */
	int failed;		/* failed thread activations */
	int maxactive;		/* max simultaneous active threads */
	int maxqueued;		/* max work queue size */

	/* scheduling */
	ani_threadcpu_t pastcpu;/* cpu used by past threads */
	ani_threadcpu_t curcpu;	/* cpu used by current threads */

	/* caching */
	int creates;		/* actual threads created */
	int maxcached;		/* max threads in the thread cache */
};
typedef struct ani_threadstats ani_threadstats_t;

struct ani_threadpool {
	struct ani_domain *domain;	/* associated domain */
	int flags;			/* various flags */
	int curthreads;			/* current # of threads in pool */
	ani_thread_t *threads;		/* list of current threads */
	int reserve;			/* desired number of free threads */
	int freethreads;		/* current # of free threads */
	ani_thread_t *freelist;		/* list of free threads */
	int freeexitcount;		/* freelist threads that need to exit */
	int curosthreads;		/* number of one-shot threads */
	ani_thread_t *osthreads;	/* list of one-shot threads */
	ani_icond_t waiting;		/* place to wait for this pool */
	int worklen;			/* length of work queue */
	ani_threadwork_t *workhead;	/* head of work queue */
	ani_threadwork_t *worktail;	/* tail of work queue */
	an_cpuspec_t params;		/* user supplied cpu spec */
	ani_resource_t res;		/* normalized resources */
	int refs;			/* reference count */
	ani_ilock_t lock;		/* lock for entire pool */
	ani_threadstats_t stats;	/* stats */
};
typedef struct ani_threadpool ani_threadpool_t;

#define TPOOL_DYING	1
#define TPOOL_WAITING	2

ani_error_t ani_threadpool_init(ani_threadpool_t *tpool,
				struct ani_domain *domain,
				an_cpuspec_t *cspec, ani_resource_t *res);
void ani_threadpool_deinit(ani_threadpool_t *tpool, int waitfor);
void ani_threadpool_disable(ani_threadpool_t *tpool, int waitfor);
void ani_threadpool_cleanup(ani_threadpool_t *tpool);

#ifdef ANI_THREADPOOL_STATS
#define TPOOLSTATS(s)	s
#else
#define TPOOLSTATS(s)
#endif

/*
 * Stop scheduling new threads.
 * Used when the system is shutting down to make life easier.
 */
void ani_threadpool_shutdown(void);

/*
 * Wait for the indicated thread to terminate or until the given time has
 * elapsed.  Returns zero on success, an error code otherwise.
 */
ani_error_t ani_threadpool_thread_wait(ani_threadpool_t *tpool,
				       ani_thread_t *thread,
				       an_timespec_t *timo);

/*
 * Start a thread from the given pool running to do the indicated work.
 * If there are no threads available and the pool is at the thread limit,
 * returns an error.  If the pool is under limit but there are no cached
 * threads available, the call is delayed while a new thread is started.
 * That is, upon return from this call, an actual thread has been commited
 * to run the function.  If out_threadp is non-zero, the call returns a
 * pointer to the new thread.  The threadwork record may be reused upon
 * return from this call.  Returns zero if a thread was successfully
 * started, an error otherwise.
 */
ani_error_t ani_threadpool_thread_start(ani_threadpool_t *tpool,
					ani_threadwork_t *work,
					ani_thread_t **out_thread);

/*
 * Start a thread from the given pool to do the indcated work and wait
 * til it terminates or the timeout expires.  If the timeout expires,
 * the thread is killed and a distinguished error code is returned.
 * The threadwork recrod may be reused upon return from this call.
 * Returns zero on success, an error code otherwise.
 */
ani_error_t ani_threadpool_thread_run(ani_threadpool_t *tpool,
				      ani_threadwork_t *work,
				      an_timespec_t *timo);

/*
 * Schedule work to be done in a thread from the indicated pool.
 * If no threads are currently available or queueonly is non-zero, the
 * request will be queued.  Returns zero on success, an error code otherwise.
 *
 * XXX odd semantics: work record must remain valid after return,
 * until a thread is started.
 */
ani_error_t ani_threadpool_thread_schedule(ani_threadpool_t *tpool,
					   ani_threadwork_t *work,
					   int queueonly);

/*
 * Lightweight version of thread_schedule called at interrupt level.
 * Puts the given work record at the end of the pool work queue and does
 * a wakeup on any available thread.
 */
void ani_threadpool_thread_startintr(ani_threadpool_t *tpool,
				     ani_threadwork_t *work);

/*
 * Start the given thread as a ``one-shot'' thread in the context of the
 * given pool running the indicated function.  Returns zero if a thread was
 * successfully started, an error otherwise.
 */
ani_error_t ani_threadpool_osthread_start(ani_threadpool_t *tpool,
					  ani_thread_t *thread, int inkern,
					  ani_thread_func_t func,
					  void *farg1, void *farg2, void *farg3);

void ani_threadpool_thread_exit(ani_threadpool_t *tpool) NODEOS_FUNC_NORETURN;

ani_error_t ani_threadpool_thread_setreserve(ani_threadpool_t *tpool,
					     int nthreads, int wait);

void ani_threadpool_stats(ani_threadpool_t *tpool, ani_threadstats_t *stats);

#include <nodeos/ani/ani_threadpool_inline.h>

#endif /* _ANI_THREADPOOL_H_ */
