/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#ifndef __ANLIB_CHAN_H__
#define __ANLIB_CHAN_H__

#include <nodeos/lib/click.hh>

#define true 1
#define false 0

#define RAW_ADDR_LEN 256

struct pcg;

/*
 * New Channel structure to be used by chanlib.
 * This keeps track of memory allocated by the chan lib.
 */
typedef struct anlib_chan {
	an_chan_t *raw_chan;
	int native;
	int arp; 
	an_chanrecvfunc_t send;
	void *sendarg;
	struct pcg *pcg_head;
	an_mempool_t *mempool;
} anlib_chan_t;

/*
 * Address specs
 * The user should always use rawaddr to provide addresses and use the
 * format as indicated below for each protocol component.
 */

/*
 * Address specification conventions:-
 * Each protocol component has its own address format.
 * Address spec for multiple protocol components are combined using
 *  "/".
 */ 

typedef struct anlib_rawaddr {
        int len;
        char data[RAW_ADDR_LEN];
} anlib_rawaddr_t;

/*
 * Address spec for protocol "ip" is
 * "<src ip addr>:<dst ip addr>:<next proto>"
 */

typedef struct {
        unsigned char saddr[4];
        unsigned char daddr[4];
        unsigned char type;
} anlib_ipaddrspec_t;

/*
 * Address spec for protocol "udp" is
 * "<src port>:<dst port>"
 */
typedef struct {
        unsigned short sport;
        unsigned short dport;
} anlib_udpaddrspec_t;

/*
 * Address spec for protocol "udpip" is
 * "<src ip addr>:<src port>:<dst ip addr>:<dst port>"
 */

typedef struct {
        unsigned char saddr[4];
        unsigned char daddr[4];
        unsigned short sport;
        unsigned short dport;
} anlib_udpipaddrspec_t;

/*
 * Address spec for protocol "arp" is
 * "<ip addr of this interface>:<eth addr of this interface>"
 */

typedef struct {
	unsigned char ip[4];
	unsigned char eth[6];	
} anlib_arpaddrspec_t;

typedef union anlib_chanaddr {
        anlib_rawaddr_t raw;
        anlib_ipaddrspec_t ip;
        anlib_udpaddrspec_t udp;
        an_ethaddrspec_t eth;
	anlib_udpipaddrspec_t udpip;
	anlib_arpaddrspec_t arp;
} anlib_chanaddr_t;

/* Generic channel functions */

/*
 * Create an in-channel as described by the given parameters.  Every packet
 * matching the given protocol, addressing and demultiplex criteria is
 * handed off to the specified delivery function.  Returns a pointer to the
 * new channel object or zero on an error.
 *
 * Note - The demultiplexing key is ignored for now.
 *
 * ``domain'' is the domain with which the channel should be associated,
 * ``key'' contains the filtering criteria to be applied to every packet,
 * ``proto'' describes the protocol processing to apply to every packet,
 * ``addr'' contains the addressing info appropriate for the protocols,
 * ``net'' specifies the network (cpu, packet buffer) resources to use,
 * ``func'' is the packet delivery function,
 * ``farg'' is an additional argument provided to the delivery function.
 */

anlib_chan_t *
anlib_inchan_create(void *mem, an_domain_t *domain, an_demuxkey_t *key,
		 an_chanproto_t *proto, anlib_chanaddr_t *addr,
		 an_netspec_t *net, an_chanrecvfunc_t func, void *farg);


/*
 * Create an out-channel as described by the given parameters.  Every packet
 * sent on the channel undergoes the specified protocol processing and is
 * routed based on the addressing info.  Returns a pointer to the new channel
 * object or zero on an error.
 *
 * Note - If the proto spec contains as "arp" component in it. The user
 * should - a. check to see if an arpQuerier for this interface
 * already exists. b. and if not, the user is responsible for
 * providing buffers for the arpQuerier in "net" parameter.
 *
 * ``domain'' is the domain with which the channel should be associated,
 * ``proto'' describes the protocol processing to apply to every packet,
 * ``addr'' contains the addressing info appropriate for the protocols,
 * ``net'' specifies the network (bandwidth) resources to use, also
 *         includes packet buffer and cpu if need to create an arp querier
 */

anlib_chan_t *
anlib_outchan_create(void *mem, an_domain_t *domain, an_chanproto_t *proto,
		  anlib_chanaddr_t *addr, an_netspec_t *net);

/*
 * Create a cut-channel by composing the provided in- and out- channels.
 * The in-chan and out-chan can no longer be used once they are composed
 * into a cut-channel.
 *
 * ``domain'' is the domain with which the channel should be associated,
 * ``inchan'' is the in-channel created for receiving packets,
 * ``outchan'' is the out-channel created for sending the packets
 */

anlib_chan_t *
anlib_cutchan_create(void *mem, an_domain_t *domain, anlib_chan_t *inchan, 
                anlib_chan_t *outchan);

/*
 * Send the indicated packet to an out-channel.
 * No indication is provided as
 * to whether the packet was successfully transmitted.  Returns an error
 * if the channel has been destroyed.
 *
 * ``chan'' is the channel to send on.
 * ``packet'' is the packet to send.
 */

an_error_t
anlib_outchan_send(anlib_chan_t *chan, an_pbuf_t *pbuf);

/*
 * Cleanly shutdown a channel.
 * Further operations on the channel should return an error.  For
 * in-channels, no further packets will be received, though processing
 * of currently active packets continues.
 *
 * ``chan'' is the channel to destroy.
 */

an_error_t
anlib_chan_destroy(anlib_chan_t *chan);

/* ARP functions */

/*
 * Returns whether or not an arp querier exists for an interface.
 *
 * "proto" is the name of the interface, e.g. "if0"
 */
 
int 
anlib_arpQuerier_exists(an_chanproto_t *proto);

/*
 * Returns whether or not an arp responder exists for an interface.
 *
 * "proto" is the name of the interface, e.g. "if0"
 */

int 
anlib_arpResponder_exists(an_chanproto_t *proto);

/*
 * Create an arp responder for given interface.
 *
 * "domain" identifies the domain that owns this arp responder,
 * "addr" specifies ip and ethernet address for this interface,
 * "proto" is the name of the interface, e.g. "if0",
 * "net" specifies the network resources (cpu, buffer) to use
 */

int 
anlib_arpResponder_create(an_domain_t *domain, anlib_chanaddr_t *addr,
			  an_chanproto_t *proto, an_netspec_t *net);
/*
 * Cleanly shuts down an arp responder for an interface. The memory is
 * returned to the domain that created the arp responder.
 *
 * "proto" is the name of the interface, e.g. "if0",
 */

int 
anlib_arpResponder_destroy(an_chanproto_t *proto);

#endif /* __ANLIB_CHAN_H__ */
