/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * NodeOS cond entrypoints.
 */

#include <nodeos/an_cond.h>
#include <nodeos/an_mutex.h>
#include <nodeos/an_error.h>

#include <nodeos/ani/ani_nodeos.h>
#include <nodeos/ani/ani_mutex.h>
#include <nodeos/ani/ani_cond.h>
#include <nodeos/ani/ani_domain.h>

#include "nodeos.h"
#include "an_error_inline.h"
#include "an_cond_inline.h"
#include "an_mutex_inline.h"

/*
 * Initalize an existing piece of memory to be a cond object.
 */
an_cond_t *
an_cond_create(void *mem)
{
	an_cond_t *xcond;
	ani_cond_t *icond;
	ani_domain_t *hdomain;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	/*
	 * Create a new object
	 */
	icond = NODEOS_CREATE_COND(hdomain, mem);
	if (icond == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "cannot create new an_cond");
		goto done;
	}

	err = ani_cond_init(icond);

 done:
	if (ani_error_occured(err)) {
		if (icond)
			NODEOS_DESTROY_COND(hdomain, icond);
		xcond = 0;
	} else
		xcond = NODEOS_EXTERN_COND(hdomain, icond);

	NODEOS_LEAVE;
	return xcond;
}

an_error_t
an_cond_destroy(an_cond_t *cond)
{
	ani_cond_t *icond;
	ani_domain_t *hdomain;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	icond = NODEOS_INTERN_COND(hdomain, cond);
	if (icond == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_cond argument");
		goto done;
	}

	err = ani_cond_deinit(icond);

	if (!ani_error_occured(err))
		NODEOS_DESTROY_COND(hdomain, icond);

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

void 
an_cond_wait(an_cond_t *cond, an_mutex_t *mux)
{
	ani_error_t err;
	ani_domain_t *hdomain;
	ani_cond_t *icond;
	ani_mutex_t *imux;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	icond = NODEOS_INTERN_COND(hdomain, cond);
	if (icond == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_cond argument");
		goto done;
	}
	imux = NODEOS_INTERN_MUTEX(hdomain, mux);
	if (imux == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_mutex argument");
		goto done;
	}

	ani_cond_timedwait(icond, imux, 0);

 done:
	NODEOS_LEAVE;
	/* XXX should return error */
}

void
an_cond_timedwait(an_cond_t *cond, an_mutex_t *mux, an_timespec_t *abstime)
{
	ani_error_t err;
	ani_domain_t *hdomain;
	ani_cond_t *icond;
	ani_mutex_t *imux;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	icond = NODEOS_INTERN_COND(hdomain, cond);
	if (icond == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_cond argument");
		goto done;
	}
	imux = NODEOS_INTERN_MUTEX(hdomain, mux);
	if (imux == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_mutex argument");
		goto done;
	}

	/*
	 * N.B. no copyin of abstime necessary in our implementation
	 */
	ani_cond_timedwait(icond, imux, abstime);

 done:
	NODEOS_LEAVE;
	/* XXX should return error */
}

void
an_cond_signal(an_cond_t *cond)
{
	ani_error_t err;
	ani_domain_t *hdomain;
	ani_cond_t *icond;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	icond = NODEOS_INTERN_COND(hdomain, cond);
	if (icond == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_cond argument");
		goto done;
	}

	ani_cond_signal(icond);

 done:
	NODEOS_LEAVE;
	/* XXX should return error */
}

void
an_cond_broadcast(an_cond_t *cond)
{
	ani_error_t err;
	ani_domain_t *hdomain;
	ani_cond_t *icond;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	icond = NODEOS_INTERN_COND(hdomain, cond);
	if (icond == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_cond argument");
		goto done;
	}

	ani_cond_broadcast(icond);

 done:
	NODEOS_LEAVE;
}

/*
 * Include the inlined functions in non-inlined form
 */
#undef NODEOS_FUNC_INLINE
#define NODEOS_FUNC_INLINE

#include "an_cond_inline.h"
