/*
 * Copyright (c) 1999-2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * The Domain API
 */
#include <nodeos/an_domain.h>
#include <nodeos/an_cred.h>
#include <nodeos/an_thread.h>

#include <nodeos/ani/ani_domain.h>
#include <nodeos/ani/ani_cred.h>
#include <nodeos/ani/ani_mempool.h>
#include <nodeos/ani/ani_nodeos.h>

#include "nodeos.h"
#include "an_error_inline.h"
#include "an_domain_inline.h"
#include "an_thread_inline.h"
#include "an_cred_inline.h"

an_domain_t *
an_domain_create(void *mem, an_cred_t *cred,
		 an_cpuspec_t *cspec, an_memspec_t *mspec,
		 an_domaininit_func_t init, void *initarg,
		 an_domainterm_func_t term)
{
	an_domain_t *xdomain;
	ani_domain_t *idomain, *hdomain;
	ani_cred_t *icred;
	ani_error_t err;
	void *tab;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	/*
	 * Create a new object
	 */
	idomain = NODEOS_CREATE_DOMAIN(hdomain, mem);
	if (idomain == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "cannot create new domain");
		goto done;
	}

	/*
	 * Validate arguments.
	 *
	 * N.B.: no copyin of cpu/mem specs as copyin/out are not
	 * needed in our implementation.  We do need to copy the specs
	 * before returning, but that is done in domain_init.
	 */
	icred = NODEOS_INTERN_CRED(hdomain, cred);
	if (icred == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_cred argument");
		goto done;
	}

	/*
	 * Initialize the internal object.
	 */
	err = ani_domain_init(idomain, icred, cspec, mspec, init, initarg, term);
	if (ani_error_occured(err))
		goto done;

	/*
	 * Create the domain's descriptor table
	 */
	tab = nodeos_desc_alloctab(idomain,
				   NODEOS_DESC_GROWSIZE, NODEOS_DESC_MAXSIZE);
	if (tab == 0) {
		ani_domain_deinit(idomain);
		err = ani_error_set(AN_ERR_NOMEMORY,
				    "cannot allocate descriptor table");
		goto done;
	}
	ani_domain_setapistate(idomain, tab);

 done:
	if (ani_error_occured(err)) {
		if (idomain)
			NODEOS_DESTROY_DOMAIN(hdomain, idomain);
		xdomain = 0;
	} else
		xdomain = NODEOS_EXTERN_DOMAIN(hdomain, idomain);

	NODEOS_LEAVE;
	return xdomain;
}

an_error_t
an_domain_destroy(an_domain_t *domain)
{
	ani_domain_t *idomain, *hdomain;
	ani_error_t err;
	void *tab;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	idomain = NODEOS_INTERN_DOMAIN(hdomain, domain);
	if (idomain == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_domain argument");
		goto done;
	}

	/*
	 * Grab the domain's descriptor table pointer so we can use it
	 * after the domain is gone
	 */
	tab = ani_domain_getapistate(hdomain);

	err = ani_domain_deinit(idomain);
	if (ani_error_occured(err))
		goto done;

	nodeos_desc_freetab(tab);
	NODEOS_DESTROY_DOMAIN(hdomain, idomain);

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_domain_t *
an_domain_current(void)
{
	an_domain_t *xdomain;
	ani_domain_t *hdomain;

	NODEOS_ENTER;
	hdomain = ani_domain_current();
	xdomain = NODEOS_EXTERN_DOMAIN(hdomain, hdomain);
	NODEOS_LEAVE;
	return xdomain;
}

an_error_t
an_domain_startthread(an_domain_t *domain, void (*func)(void *arg), void *arg,
		      an_thread_t **out_thread)
{
	ani_domain_t *idomain, *hdomain;
	ani_thread_t *ithread;
	ani_threadwork_t work;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	idomain = NODEOS_INTERN_DOMAIN(hdomain, domain);
	if (idomain == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_domain argument");
		goto done;
	}

	if (func == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid thread function argument");
		goto done;
	}

	ani_threadwork_init(&work);
	ani_threadwork_set(&work, 0, 0, (ani_thread_func_t)func, arg, 0, 0);
	err = ani_threadpool_thread_start(&idomain->tpool, &work, &ithread);
	if (!ani_error_occured(err) && out_thread)
		*out_thread = NODEOS_EXTERN_THREAD(hdomain, ithread);

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_domain_stats(an_domain_t *domain, an_domainstats_t *stats)
{
	ani_domain_t *idomain, *hdomain;
	an_domainstats_t *istats;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	idomain = NODEOS_INTERN_DOMAIN(hdomain, domain);
	if (idomain == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_domain argument");
		goto done;
	}

	/*
	 * Get the stats and copy them out to the user's memory.
	 *
	 * N.B. we can just do a direct copy in our implementation.
	 */
	istats = ani_domain_stats(idomain);
	memcpy(stats, istats, sizeof *stats);
	err = ani_error_clear();

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

/*
 * Include the inlined functions in non-inlined form
 */
#undef NODEOS_FUNC_INLINE
#define NODEOS_FUNC_INLINE

#include "an_domain_inline.h"
