/*
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <nodeos/an_event.h>

#include <nodeos/ani/ani_nodeos.h>
#include <nodeos/ani/ani_event.h>
#include <nodeos/ani/ani_domain.h>

#include "nodeos.h"
#include "an_error_inline.h"
#include "an_event_inline.h"

an_event_t *
an_event_create(void *mem)
{
	an_event_t *xevent;
	ani_event_t *ievent;
	ani_domain_t *hdomain;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	/*
	 * Create a new object
	 */
	ievent = NODEOS_CREATE_EVENT(hdomain, mem);
	if (ievent == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "cannot create new event");
		goto done;
	}

	err = ani_event_init(ievent, ani_domain_threadpool(hdomain));

 done:
	if (ani_error_occured(err)) {
		if (ievent)
			NODEOS_DESTROY_EVENT(hdomain, ievent);
		xevent = 0;
	} else
		xevent = NODEOS_EXTERN_EVENT(hdomain, ievent);

	NODEOS_LEAVE;
	return xevent;
}

an_error_t
an_event_destroy(an_event_t *event, an_eventtype_t *type,
		 an_eventstate_t *state)
{
	ani_error_t err;
	ani_domain_t *hdomain;
	ani_event_t *ievent;
	int iflags;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	ievent = NODEOS_INTERN_EVENT(hdomain, event);
	if (ievent == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_event argument");
		goto done;
	}

	err = ani_event_deinit(ievent, &iflags);
	if (!ani_error_occured(err)) {
		NODEOS_DESTROY_EVENT(hdomain, ievent);
		if (type)
			*type = ani_event_flagstotype(iflags);
		if (state)
			*state = ani_event_flagstostate(iflags);
	}

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t 
an_event_schedule(an_event_t *event, an_timespec_t *etime,
		  an_eventtype_t type, an_eventfunc_t func, void *farg)
{
	ani_event_t *ievent;
	ani_error_t err;

	NODEOS_ENTER;
	ievent = NODEOS_INTERN_EVENT(ani_domain_current(), event);
	if (ievent == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_event argument");
		goto done;
	}

	/*
	 * N.B. no copyin of arguments needed in our implementation
	 */
	err = ani_event_schedule(ievent, etime, type, ANI_EV_USER, func, farg);

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}


an_error_t 
an_event_fire(an_event_t *event, an_eventfunc_t func, void *arg)
{
	ani_event_t *ievent;
	ani_error_t err;

	NODEOS_ENTER;
	ievent = NODEOS_INTERN_EVENT(ani_domain_current(), event);
	if (ievent == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_event argument");
		goto done;
	}

	err = ani_event_fire(ievent, func, arg);

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}


an_error_t 
an_event_status(an_event_t *event, an_eventtype_t *type,
		an_eventstate_t *state)
{
	int iflags = 0;
	ani_event_t *ievent;
	ani_error_t err;

	LW_NODEOS_ENTER;
	ievent = NODEOS_INTERN_EVENT(ani_domain_current(), event);
	if (ievent == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_event argument");
		goto done;
	}

	iflags = ani_event_flags(ievent);
	err = ani_error_clear();

 done:
	LW_NODEOS_LEAVE;

	if (!ani_error_occured(err)) {
		if (type)
			*type = ani_event_flagstotype(iflags);
		if (state)
			*state = ani_event_flagstostate(iflags);
	}

	return NODEOS_EXTERN_ERROR(err);
}

an_error_t 
an_event_deschedule(an_event_t *event, an_eventtype_t *type,
		    an_eventstate_t *state)
{
	ani_event_t *ievent;
	ani_error_t err;
	int iflags;

	NODEOS_ENTER;
	ievent = NODEOS_INTERN_EVENT(ani_domain_current(), event);
	if (ievent == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_event argument");
		goto done;
	}

	err = ani_event_deschedule(ievent, &iflags);

 done:
	NODEOS_LEAVE;

	if (!ani_error_occured(err)) {
		if (type)
			*type = ani_event_flagstotype(iflags);
		if (state)
			*state = ani_event_flagstostate(iflags);
	}
	
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_event_redefine(an_event_t *event, an_timespec_t *oldtime,
		  an_eventtype_t *oldtype, an_eventstate_t *oldstate,
		  an_eventfunc_t *oldfunc, void **oldfarg,
		  an_timespec_t *newtime, an_eventtype_t newtype,
		  an_eventfunc_t newfunc, void *newfarg)
{
	ani_event_t *ievent;
	ani_error_t err;
	int oflags;

	NODEOS_ENTER;
	ievent = NODEOS_INTERN_EVENT(ani_domain_current(), event);
	if (ievent == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_event argument");
		goto done;
	}

	err = ani_event_redefine(ievent, oldtime, &oflags, oldfunc, oldfarg,
				 newtime, newtype, newfunc, newfarg);

 done:
	NODEOS_LEAVE;

	if (!ani_error_occured(err)) {
		if (oldtype)
			*oldtype = ani_event_flagstotype(oflags);
		if (oldstate)
			*oldstate = ani_event_flagstostate(oflags);
	}

	return NODEOS_EXTERN_ERROR(err);
}

unsigned long
an_event_granularity(void)
{
	unsigned long res;

	LW_NODEOS_ENTER;
	ani_event_info(&res);
	LW_NODEOS_LEAVE;

	return res;
}

/*
 * Include the inlined functions in non-inlined form
 */
#undef NODEOS_FUNC_INLINE
#define NODEOS_FUNC_INLINE

#include "an_event_inline.h"

