/*
 * Copyright (c) 2000-2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * File system interface
 * XXX a file is not currently a first-class object.
 */
#include <nodeos/an_fs.h>
#include <nodeos/an_error.h>

#include <nodeos/ani/ani_nodeos.h>
#include <nodeos/ani/ani_domain.h>
#include <nodeos/ani/ani_fs.h>

#include "nodeos.h"
#include "an_error_inline.h"
#include "an_fs_inline.h"

an_error_t 
an_file_close(an_file_t *file) 
{
	ani_error_t err;
	ani_file_t *ifile;
	ani_domain_t *hdomain;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	/*
	 * Validate arguments.
	 */
	ifile = NODEOS_INTERN_FILE(hdomain, file);
	if (ifile == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_file argument");
		goto done;
	}

	err = ani_file_close(ifile);

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t 
an_file_fsync(an_file_t *file)
{
	ani_error_t err;
	ani_file_t *ifile;
	ani_domain_t *hdomain;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	/*
	 * Validate arguments.
	 */
	ifile = NODEOS_INTERN_FILE(hdomain, file);
	if (ifile == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_file argument");
		goto done;
	}

	err = ani_file_fsync(ifile);

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t 
an_file_truncate(an_file_t *file, an_offset_t offset)
{
	ani_error_t err;
	ani_file_t *ifile;
	ani_domain_t *hdomain;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	/*
	 * Validate arguments.
	 */
	ifile = NODEOS_INTERN_FILE(hdomain, file);
	if (ifile == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_file argument");
		goto done;
	}

	err = ani_file_truncate(ifile, offset);

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t 
an_file_lseek(an_file_t *file, an_offset_t length, an_whence_t whence,
	      an_offset_t *retoff)
{
	ani_error_t err;
	ani_file_t *ifile;
	ani_domain_t *hdomain;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	/*
	 * Validate arguments.
	 */
	ifile = NODEOS_INTERN_FILE(hdomain, file);
	if (ifile == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_file argument");
		goto done;
	}

	err = ani_file_lseek(ifile, length, whence, retoff);

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t 
an_file_mkdir(const char *path, an_acd_t *acd) 
{
	ani_error_t err;

	NODEOS_ENTER;

	err = ani_file_mkdir(path, acd);

	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t 
an_file_read(an_file_t *file, an_size_t nbytes, void *buf, an_size_t *retlen)
{
	ani_error_t err;
	ani_file_t *ifile;
	ani_domain_t *hdomain;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	/*
	 * Validate arguments.
	 */
	ifile = NODEOS_INTERN_FILE(hdomain, file);
	if (ifile == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_file argument");
		goto done;
	}

	err = ani_file_read(ifile, nbytes, buf, retlen);

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t 
an_file_rename(const char *from, const char *to)
{
	ani_error_t err;

	NODEOS_ENTER;

	err = ani_file_rename(from, to);

	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t 
an_file_write(an_file_t *file, an_size_t length, void *buf, an_size_t *retval)
{
	ani_error_t err;
	ani_file_t *ifile;
	ani_domain_t *hdomain;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	/*
	 * Validate arguments.
	 */
	ifile = NODEOS_INTERN_FILE(hdomain, file);
	if (ifile == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_file argument");
		goto done;
	}

	err = ani_file_write(ifile, length, buf, retval);

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t 
an_file_unlink(const char *path)
{
	ani_error_t err;

	NODEOS_ENTER;

	err = ani_file_unlink(path);

	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t 
an_file_open(const char *path, int flags, an_acd_t *acd, an_file_t **retval)
{
	ani_error_t err;
	ani_file_t *ifile;
	ani_domain_t *hdomain;
	void *mem;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	/*
	 * Create a new object
	 */
	mem = *retval;
	ifile = NODEOS_CREATE_FILE(hdomain, mem);
	if (ifile == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "cannot create new an_file");
		goto done;
	}

	/* N.B. no copyin of info needed in our implementation */
	err = ani_file_open(path, flags, acd, ifile);

 done:
	if (ani_error_occured(err)) {
		if (ifile)
			NODEOS_DESTROY_FILE(hdomain, ifile);
	} else
		*retval = NODEOS_EXTERN_FILE(hdomain, ifile);

	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t 
an_file_stat(const char *path, an_stat_t *retval)
{
	ani_error_t err;

	NODEOS_ENTER;

	err = ani_file_stat(path, retval);

	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t 
an_file_fstat(an_file_t *file, an_stat_t *retval)
{
	ani_error_t err;
	ani_file_t *ifile;
	ani_domain_t *hdomain;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	/*
	 * Validate arguments.
	 */
	ifile = NODEOS_INTERN_FILE(hdomain, file);
	if (ifile == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_file argument");
		goto done;
	}


	NODEOS_ENTER;
	err = ani_file_fstat(ifile, retval);

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

/*
 * Include the inlined functions in non-inlined form
 */
#undef NODEOS_FUNC_INLINE
#define NODEOS_FUNC_INLINE

#include "an_fs_inline.h"
