/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * The Mempool API
 */
#include <nodeos/an_impl.h>
#include <nodeos/an_mempool.h>
#include <nodeos/an_error.h>
#include <nodeos/an_domain.h>

#include <nodeos/ani/ani_nodeos.h>
#include <nodeos/ani/ani_mempool.h>
#include <nodeos/ani/ani_domain.h>

#include "nodeos.h"
#include "an_error_inline.h"
#include "an_mempool_inline.h"

an_mempool_t *
an_mempool_create(void *mem, an_mempool_t *parent, an_mempoolfull_func_t func)
{
	an_mempool_t *xmpool;
	ani_mempool_t *impool, *iparent;
	ani_domain_t *hdomain;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	/*
	 * Create a new object
	 */
	impool = NODEOS_CREATE_MEMPOOL(hdomain, mem);
	if (impool == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "cannot create new an_mempool");
		goto done;
	}

	if (parent) {
		iparent = NODEOS_INTERN_MEMPOOL(hdomain, parent);
		if (iparent == 0) {
			err = ani_error_set(AN_ERR_INVALIDARG,
					    "invalid an_mempool argument");
			goto done;
		}
	} else
		iparent = 0;

	err = ani_mempool_init(impool, iparent, func);

 done:
	if (ani_error_occured(err)) {
		if (impool)
			NODEOS_DESTROY_MEMPOOL(hdomain, impool);
		xmpool = 0;
	} else
		xmpool = NODEOS_EXTERN_MEMPOOL(hdomain, impool);

	NODEOS_LEAVE;
	return xmpool;
}

an_error_t
an_mempool_destroy(an_mempool_t *mpool)
{
	ani_mempool_t *impool;
	ani_domain_t *hdomain;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	impool = NODEOS_INTERN_MEMPOOL(hdomain, mpool);
	if (impool == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_mempool argument");
		goto done;
	}

	err = ani_mempool_deinit(impool);

	if (!ani_error_occured(err))
		NODEOS_DESTROY_MEMPOOL(hdomain, impool);

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_mempool_t *
an_mempool_current(void)
{
	an_mempool_t *xmempool;

	NODEOS_ENTER;
	xmempool = NODEOS_EXTERN_MEMPOOL(ani_domain_current(),
					 ani_mempool_current());
	NODEOS_LEAVE;
	return xmempool;
}

an_mempool_t *
an_mempool_domain(struct an_domain *domain)
{
	an_mempool_t *xmempool = 0;
	ani_domain_t *idomain;
	
	NODEOS_ENTER;
	
	idomain = NODEOS_INTERN_DOMAIN(0, domain);
	if (idomain == 0) {
		ani_error_set(AN_ERR_INVALIDARG, 
						"invalid domain argument");
		goto done;
	}
		
	xmempool = NODEOS_EXTERN_MEMPOOL(idomain, ani_domain_mempool(idomain));
 
 done:
	NODEOS_LEAVE;
	return xmempool;
}

void *
an_mempool_alloc(an_mempool_t *mpool, size_t size)
{
	ani_mempool_t *impool;
	ani_domain_t *hdomain;
	ani_error_t err;
	void *ptr = 0;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	impool = NODEOS_INTERN_MEMPOOL(hdomain, mpool);
	if (impool == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_mempool argument");
		goto done;
	}

	ptr = ani_mempool_alloc(impool, size);

 done:
	NODEOS_LEAVE;
	return ptr;
}

void
an_mempool_free(an_mempool_t *mpool, void *ptr, size_t size)
{
	ani_mempool_t *impool;
	ani_domain_t *hdomain;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	impool = NODEOS_INTERN_MEMPOOL(hdomain, mpool);
	if (impool == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_mempool argument");
		goto done;
	}

	ani_mempool_free(impool, ptr, size);

 done:
	NODEOS_LEAVE;
	/* XXX should return error */
}

/*
 * Return statistics for a mempool.
 * Return memory use statistics for the given mempool in the provided buffer.
 * Returns zero on success or an error code otherwise.
 */
an_error_t
an_mempool_stats(an_mempool_t *mpool, an_mempoolstats_t *stats)
{
	ani_mempool_t *impool;
	ani_domain_t *hdomain;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	impool = NODEOS_INTERN_MEMPOOL(hdomain, mpool);
	if (impool == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_mempool argument");
		goto done;
	}

	/*
	 * Get the stats and copy them out to the user's memory.
	 *
	 * N.B. we can just copy stats directly into user buffer
	 * in our implementation.
	 */
	err = ani_mempool_stats(impool, stats);

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

/*
 * Include the inlined functions in non-inlined form
 */
#undef NODEOS_FUNC_INLINE
#define NODEOS_FUNC_INLINE

#include "an_mempool_inline.h"
