/*
 * Copyright (c) 1999-2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
#include <nodeos/an_moab_setup.h>
#include <nodeos/an_thread.h>
#include <nodeos/an_error.h>

#include <nodeos/ani/ani_nodeos.h>
#include <nodeos/ani/ani_moab_setup.h>
#include <nodeos/ani/ani_domain.h>

#include "nodeos.h"

void
an_moab_setup(void)
{
	/* No context, cannot call NODEOS_ENTER or NODEOS_EXIT */

	ani_moab_setup();
}

int
an_moab_version(void)
{
	int ver;

	LW_NODEOS_ENTER;
	ver = ani_moab_version();
	LW_NODEOS_LEAVE;
	
	return ver;
}

void
an_moab_start(an_domain_t *root_domain, an_cred_t *root_cred,
	      an_mempool_t *root_mempool, int nthreads, int stacksize,
	      void **root_threads, void **root_threadstacks,
	      void (*initee_func)(void *), void *initee_arg)
{
	static ani_domain_t _rootdomain;
	static ani_cred_t _rootcred;
	static ani_mempool_t _rootpool;
	nodeos_dtab_t *_roottab;

	/*
	 * No context, cannot call NODEOS_ENTER or NODEOS_EXIT
	 */

	/*
	 * XXX gotta fire up the root mempool right away as we are going
	 * to need to allocate memory for objects.
	 */
	ani_mempool_setup(&_rootpool);

	/*
	 * Create the root domain's descriptor table and enter the static
	 * objects that we don't create normally.
	 */
	_roottab = nodeos_desc_allocroottab(&_rootpool,
					    NODEOS_DESC_GROWSIZE,
					    NODEOS_DESC_MAXSIZE);
	assert(_roottab != 0);

	nodeos_desc_inserttab(_roottab, root_domain, &_rootdomain,
			      NODEOS_TYPE_DOMAIN, &root_domain->index);
	nodeos_desc_inserttab(_roottab, root_cred, &_rootcred,
			      NODEOS_TYPE_CRED, &root_cred->index);
	nodeos_desc_inserttab(_roottab, root_mempool, &_rootpool,
			      NODEOS_TYPE_MEMPOOL, &root_mempool->index);


	/*
	 * Do the equivalent of NODEOS_CREATE_THREAD for root threads.
	 * This involves allocating memory for the internal objects
	 * and registering them in the root descriptor table where they
	 * can be found by ani_rootdomain_init which does NODEOS_INTERN_THREAD.
	 */
	if (root_threads != 0) {
		int i;
		an_thread_t *xthread;
		ani_thread_t *ithread;

		for (i = 0; i < nthreads; i++) {
			xthread = root_threads[i];
			assert(xthread != 0);
			ithread = ani_mempool_alloc(&_rootpool,
						    sizeof(ani_thread_t));
			assert(ithread != 0);

			nodeos_desc_inserttab(_roottab, xthread,
					      ithread, NODEOS_TYPE_THREAD,
					      &xthread->index);
		}
	}

	/*
	 * Point the as-of-yet uninitialized root domain at the descriptor
	 * table.  This has to happen before moab_start since NODEOS
	 * INTERN/EXTERN are used.
	 */
	ani_domain_setapistate(&_rootdomain, _roottab);

	/*
	 * Likewise, cannot use NODEOS_CREATE_* since there is no domain
	 * descriptor table to record the object in.  We use our statically
	 * allocated objects for moab_start and then record them in the root
	 * domain's descriptor table after the fact.
	 *
	 * N.B. No copyin of thread/stack arrays necessary in our impl.
	 */
	ani_moab_start(&_rootdomain, &_rootcred, &_rootpool,
		       nthreads, stacksize,
		       root_threads, root_threadstacks,
		       initee_func, initee_arg);

	/*
	 * XXX reset the root domain descriptor table pointer in case the
	 * "official" domain creation in moab_start cleared it.
	 */
	ani_domain_setapistate(&_rootdomain, _roottab);
}

void
an_moab_shutdown(void)
{
	/*
	 * We may or may not have a context here, but it doesn't matter,
	 * the world is about to end so we don't bother with ENTER/LEAVE
	 */
	ani_moab_shutdown();
}
