/*
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * The packet buffer (pbuf) API
 */
#include <nodeos/an_pbuf.h>
#include <nodeos/an_error.h>

#include <nodeos/ani/ani_nodeos.h>
#include <nodeos/ani/ani_domain.h>
#include <nodeos/ani/ani_pbuf.h>

#include "ani_os.h"

#include "nodeos.h"
#include "an_error_inline.h"
#include "an_pbuf_inline.h"

an_pbuf_t *
an_pbuf_create(void *mem, int flags, int offset, int size)
{
	an_pbuf_t *xpbuf;
	ani_pbuf_t *ipbuf = 0;
	ani_domain_t *hdomain;
	size_t bsize;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	/*
	 * Translate size-related flags.
	 * XXX we only have one buffer size so this is easy!
	 */
	switch (flags & AN_PBUF_SIZEMASK) {
	case AN_PBUF_DEFAULT:
	case AN_PBUF_ETH:
	case AN_PBUF_SOCK:
	case AN_PBUF_CLOCK:
	case AN_PBUF_STDIO:
		bsize = NODEOS_CHAN_BUFSIZE;
		break;
	default:
		err = ani_error_set(AN_ERR_INVALIDARG, "invalid flags");
		goto done;
	}

	/*
	 * Constrain the active region
	 */
	if (offset < 0 || size < 0 || offset + size > bsize) {
		err = ani_error_set(AN_ERR_INVALIDARG, "invalid offset/size");
		goto done;
	}

	/*
	 * Create a new object
	 */
	ipbuf = NODEOS_CREATE_PBUF(hdomain, mem);
	if (ipbuf == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "cannot create new an_pbuf");
		goto done;
	}

	err = ani_pbuf_init(ipbuf, hdomain, 0, bsize, offset, size);

 done:
	if (ani_error_occured(err)) {
		if (ipbuf)
			NODEOS_DESTROY_PBUF(hdomain, ipbuf);
		xpbuf = 0;
	} else
		xpbuf = NODEOS_EXTERN_PBUF(hdomain, ipbuf);

	NODEOS_LEAVE;
	return xpbuf;
}

an_error_t
an_pbuf_destroy(an_pbuf_t *pbuf)
{
	ani_error_t err;
	ani_domain_t *hdomain;
	ani_pbuf_t *ipbuf;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	ipbuf = NODEOS_INTERN_PBUF(hdomain, pbuf);
	if (ipbuf == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_pbuf argument");
		goto done;
	}

	ani_pbuf_deinit(ipbuf);
	NODEOS_DESTROY_PBUF(hdomain, ipbuf);
	err = ani_error_clear();

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

void *
an_pbuf_data(an_pbuf_t *pbuf)
{
	void *data;
	ani_domain_t *hdomain;
	ani_pbuf_t *ipbuf;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	ipbuf = NODEOS_INTERN_PBUF(hdomain, pbuf);
	if (ipbuf == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_pbuf argument");
		data = 0; /* XXX */
		goto done;
	}

	data = ani_pbuf_data(ipbuf);

 done:
	NODEOS_LEAVE;
	return data;
}

int
an_pbuf_size(an_pbuf_t *pbuf)
{
	int size;
	ani_domain_t *hdomain;
	ani_pbuf_t *ipbuf;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	ipbuf = NODEOS_INTERN_PBUF(hdomain, pbuf);
	if (ipbuf == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_pbuf argument");
		size = -1; /* XXX */
		goto done;
	}

	size = (int)ani_pbuf_size(ipbuf);

 done:
	NODEOS_LEAVE;
	return size;
}

int
an_pbuf_offset(an_pbuf_t *pbuf)
{
	int offset;
	ani_domain_t *hdomain;
	ani_pbuf_t *ipbuf;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	ipbuf = NODEOS_INTERN_PBUF(hdomain, pbuf);
	if (ipbuf == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_pbuf argument");
		offset = -1; /* XXX */
		goto done;
	}

	offset = (int)ani_pbuf_offset(ipbuf);

 done:
	NODEOS_LEAVE;
	return offset;
}

int
an_pbuf_bufsize(an_pbuf_t *pbuf)
{
	int size;
	ani_domain_t *hdomain;
	ani_pbuf_t *ipbuf;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	ipbuf = NODEOS_INTERN_PBUF(hdomain, pbuf);
	if (ipbuf == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_pbuf argument");
		size = -1; /* XXX */
		goto done;
	}

	size = (int)ani_pbuf_bufsize(ipbuf);

 done:
	NODEOS_LEAVE;
	return size;
}

/*
 * Determine if a pbuf is currently busy.
 *
 * XXX the busy state of a pbuffer can be cleared at interrupt time so there
 * is an obvious race here.  This shouldn't be a problem, but if it is,
 * ani_pbuf_isbusy should be responsible for ensuring the desired semantics.
 */
int
an_pbuf_isbusy(an_pbuf_t *pbuf)
{
	int isbusy;
	ani_domain_t *hdomain;
	ani_pbuf_t *ipbuf;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	ipbuf = NODEOS_INTERN_PBUF(hdomain, pbuf);
	if (ipbuf == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_pbuf argument");
		isbusy = 0; /* XXX */
		goto done;
	}

	isbusy = ani_pbuf_isbusy(ipbuf);

 done:
	NODEOS_LEAVE;
	return isbusy;
}

an_error_t
an_pbuf_return(an_pbuf_t *pbuf)
{
	ani_domain_t *hdomain;
	ani_pbuf_t *ipbuf;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	ipbuf = NODEOS_INTERN_PBUF(hdomain, pbuf);
	if (ipbuf == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_pbuf argument");
		goto done;
	}

	ani_pbuf_return(ipbuf);
	err = ani_error_clear();

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_pbuf_t *
an_pbuf_clone(an_pbuf_t *pbuf, void *newpbufmem)
{
	an_pbuf_t *xpbuf;
	ani_domain_t *hdomain;
	ani_pbuf_t *ipbuf, *inewpbuf = 0;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	ipbuf = NODEOS_INTERN_PBUF(hdomain, pbuf);
	if (ipbuf == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_pbuf argument");
		goto done;
	}

	/*
	 * Create a new object
	 */
	inewpbuf = NODEOS_CREATE_PBUF(hdomain, newpbufmem);
	if (inewpbuf == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "cannot create new an_pbuf");
		goto done;
	}

	err = ani_pbuf_clone(ipbuf, inewpbuf);

 done:
	if (ani_error_occured(err)) {
		if (inewpbuf)
			NODEOS_DESTROY_PBUF(hdomain, inewpbuf);
		xpbuf = 0;
	} else
		xpbuf = NODEOS_EXTERN_PBUF(hdomain, inewpbuf);

	NODEOS_LEAVE;
	return xpbuf;
}

an_error_t
an_pbuf_setoffset(an_pbuf_t *pbuf, int noffset, void **ndata)
{
	ani_domain_t *hdomain;
	void *data;
	ani_pbuf_t *ipbuf;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	ipbuf = NODEOS_INTERN_PBUF(hdomain, pbuf);
	if (ipbuf == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_pbuf argument");
		goto done;
	}

	err = ani_pbuf_setoffset(ipbuf, noffset, &data);

	/*
	 * Copy out the new data pointer
	 *
	 * N.B. can do a direct copy in our implementation.
	 */
	if (!ani_error_occured(err) && ndata != 0)
		*ndata = data;

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_pbuf_setsize(an_pbuf_t *pbuf, int nsize, void **ndata)
{
	ani_domain_t *hdomain;
	void *data;
	ani_pbuf_t *ipbuf;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	ipbuf = NODEOS_INTERN_PBUF(hdomain, pbuf);
	if (ipbuf == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_pbuf argument");
		goto done;
	}

	err = ani_pbuf_setsize(ipbuf, nsize, &data);

	/*
	 * Copy out the new data pointer
	 *
	 * N.B. can do a direct copy in our implementation.
	 */
	if (!ani_error_occured(err) && ndata != 0)
		*ndata = data;

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

/*
 * Include the inlined functions in non-inlined form
 */
#undef NODEOS_FUNC_INLINE
#define NODEOS_FUNC_INLINE

#include "an_pbuf_inline.h"
