/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * The Thread API
 */
#include <nodeos/an_thread.h>
#include <nodeos/an_error.h>

#include <nodeos/ani/ani_nodeos.h>
#include <nodeos/ani/ani_domain.h>
#include <nodeos/ani/ani_threadpool.h>

#include "nodeos.h"
#include "an_error_inline.h"
#include "an_thread_inline.h"
#include "an_domain_inline.h"

an_thread_t *
an_thread_create(void *mem, void *stackptr, size_t stacksize)
{
	an_thread_t *xthread;
	ani_thread_t *ithread;
	ani_domain_t *hdomain;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	/*
	 * Create a new object
	 */
	ithread = NODEOS_CREATE_THREAD(hdomain, mem);
	if (ithread == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "cannot create new an_thread");
		goto done;
	}

	err = ani_thread_init(ithread, stackptr, stacksize);

 done:
	if (ani_error_occured(err)) {
		if (ithread)
			NODEOS_DESTROY_THREAD(hdomain, ithread);
		xthread = 0;
	} else
		xthread = NODEOS_EXTERN_THREAD(hdomain, ithread);

	NODEOS_LEAVE;
	return xthread;
}

int
an_thread_destroy(an_thread_t *thread)
{
	ani_error_t err;
	ani_domain_t *hdomain;
	ani_thread_t *ithread;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	ithread = NODEOS_INTERN_THREAD(hdomain, thread);
	if (ithread == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_thread argument");
		goto done;
	}

	err = ani_thread_deinit(ithread);

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_thread_t *
an_thread_current(void)
{
	an_thread_t *xthread;

	NODEOS_ENTER;
	xthread = NODEOS_EXTERN_THREAD(ani_domain_current(),
				       ani_thread_current());
	NODEOS_LEAVE;
	return xthread;
}

an_error_t
an_thread_start(an_thread_t *thread, an_domain_t *domain,
		void (*func)(void *arg), void *arg)
{
	ani_thread_t *ithread;
	ani_domain_t *idomain;
	ani_domain_t *hdomain;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	idomain = NODEOS_INTERN_DOMAIN(hdomain, domain);
	if (idomain == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_domain argument");
		goto done;
	}

	/* XXX thread should be looked up in idomain context instead? */
	ithread = NODEOS_INTERN_THREAD(hdomain, thread);
	if (ithread == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_thread argument");
		goto done;
	}

	if (func == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid thread function argument");
		goto done;
	}

	err = ani_threadpool_osthread_start(&idomain->tpool, ithread, 0,
					    (ani_thread_func_t)func,
					    arg, 0, 0);

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

void
an_thread_exit(void)
{
	ani_thread_t *cur;
	
	NODEOS_ENTER;
	cur = ani_thread_current();
	assert_ani_thread_valid(cur);
	assert(cur->pool);
	ani_threadpool_thread_exit(cur->pool);
	NODEOS_LEAVE;
}

void
an_thread_yield(void)
{
	NODEOS_ENTER;
	ani_thread_yield();
	NODEOS_LEAVE;
}

void
an_thread_sleep(an_timespec_t *delay)
{
	NODEOS_ENTER;
	/*
	 * N.B. no copyin of delay needed in our implementation.
	 */
	ani_thread_sleep(delay);
	NODEOS_LEAVE;
}

void
an_thread_setdata(void *data)
{
	NODEOS_ENTER;
	ani_thread_setdata(data);
	NODEOS_LEAVE;
}

void *
an_thread_getdata(void)
{
	void *data;

	NODEOS_ENTER;
	data = ani_thread_getdata();
	NODEOS_LEAVE;

	return data;
}

void
an_thread_intr(an_thread_t *thread)
{
	ani_domain_t *hdomain;
	ani_thread_t *ithread;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	ithread = NODEOS_INTERN_THREAD(hdomain, thread);
	if (ithread == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_thread argument");
		goto done;
	}

	ani_thread_interrupt(ithread);

 done:
	NODEOS_LEAVE;
	/* XXX should return error */
}

void
an_thread_blockintr(void)
{
	NODEOS_ENTER;
	ani_thread_block_interrupt();
	NODEOS_LEAVE;
}

void
an_thread_unblockintr(void)
{
	NODEOS_ENTER;
	ani_thread_unblock_interrupt();
	NODEOS_LEAVE;
}

an_error_t
an_thread_setprio(an_thread_t *thread, int prio)
{
	ani_domain_t *hdomain;
	ani_thread_t *ithread;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	ithread = NODEOS_INTERN_THREAD(hdomain, thread);
	if (ithread == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_thread argument");
		goto done;
	}

	err = ani_thread_setprio(ithread, prio);

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_thread_getprio(an_thread_t *thread, int *priop)
{
	ani_domain_t *hdomain;
	ani_thread_t *ithread;
	ani_error_t err;
	int iprio;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	ithread = NODEOS_INTERN_THREAD(hdomain, thread);
	if (ithread == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_thread argument");
		goto done;
	}

	/*
	 * Return the current priority.
	 *
	 * N.B. no copyout needed in our implemenatation
	 */
	iprio = ani_thread_getprio(ithread);
	*priop = iprio;
	err = ani_error_clear();

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

/*
 * Implementation specific API calls
 */

/*
 * Establish a thread_interrupt handler.
 */
void
an_thread_sethandler(an_threadintr_func_t handler)
{
	NODEOS_ENTER;
	ani_thread_sethandler(handler);
	NODEOS_LEAVE;
}

/*
 * Get various bits of stack information on a specific
 * thread.  The thread may be the current thread.
 * If the target thread is running (on an SMP machine)
 * the results are undefined.
 *
 * XXX this call may evolve into a more generic thread_info call.
 */
an_error_t 
an_thread_stackinfo(an_thread_t *thread, void **base, size_t *size, void **ptr)
{
	ani_domain_t *hdomain;
	ani_thread_t *ithread;
	void *ibase, *iptr;
	size_t isize;
	ani_error_t err;
	
	NODEOS_ENTER;

	hdomain = ani_domain_current();

	ithread = NODEOS_INTERN_THREAD(hdomain, thread);
	if (ithread == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid an_thread argument");
		goto done;
	}

	/*
	 * Copy out parameters to user.
	 *
	 * N.B. we can just do a direct copy in our implementation.
	 */
	err = ani_thread_stackinfo(ithread, &ibase, &isize, &iptr);
	if (!ani_error_occured(err)) {
		*base = ibase;
		*size = isize;
		*ptr = iptr;
	}

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

/*
 * Include the inlined functions in non-inlined form
 */
#undef NODEOS_FUNC_INLINE
#define NODEOS_FUNC_INLINE

#include "an_thread_inline.h"
