/*
 * Copyright (c) 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * Implement descriptor tables providing small integer names for objects.
 * Note that 0 is an invalid index.
 *
 * The theory is that we implement a per-domain mapping of object descriptors
 * to the objects themselves.  But that doesn't really work with the API as
 * defined.
 *
 * Recall that the API is defined to return pointers to objects rather than
 * objects themselves.  For example, an_thread_current() must return a pointer
 * to a pre-existing thread object since we don't pass in a pointer to an
 * object that could be "filled in".
 *
 * This works fine when there is a 1-to-1 relationship between external and
 * internal objects.  For example, in the ``direct'' API version of Moab,
 * multiple domains could share the same mempool and no matter what thread in
 * what domain does an an_mempool_current(), they will get back the same
 * pointer to the One True Piece of Memory representing that mempool.
 * That is, names are not context sensitive in this world.
 *
 * But if we want to implement a descriptor-based naming scheme where the
 * descriptors are per-domain, then suddenly every domain has a distinct name
 * for any object.  For example, the same mempool might be descriptor 2 in
 * one domain and descriptor 8 in another.
 *
 * This is where we hit the wall because of calls like an_thread_current()
 * which return pointers to pre-existing memory.  Say domain 1 creates a bunch
 * of threads.  Its external objects get "filled in" with descriptors
 * representing those thread objects, where the values are chosen from this
 * domain's descriptor name space.  Now domain 1 creates a new domain 2 and
 * passes those threads to be associated with the newly created domain.
 * One of those threads starts executing in domain 2 and does an
 * an_thread_current().  The NodeOS must return a pointer.  The only pointer
 * it has is to the an_thread_t object memory in domain 1.  But the descriptor
 * embedded there is an index in domain 1's table and not meaningful in domain
 * 2.  The NodeOS could rewrite the stored descriptor value, but then the
 * an_thread_t object is no longer valid in domain 1.  What's a poor NodeOS to
 * do?
 *
 * Punt! Punt!  For now we have a single, global descriptor table instead of
 * doing the per-domain thing.  This has to be rethought if/when we enter the
 * multiple address space world
 */
#define GLOBAL_DTAB

#include <string.h>

#include <nodeos/ani/ani_error.h>
#include <nodeos/ani/ani_domain.h>
#include <nodeos/ani/ani_mempool.h>

#include "nodeos.h"

static nodeos_dent_t *nodeos_desc_alloc(nodeos_dtab_t *dtab, int *indexp);
static nodeos_dent_t *nodeos_desc_lookup(ani_domain_t *domain, void *iobj,
					 int ntype, int *indexp);
static nodeos_dtab_t *root_dtab;
static int bootstrapdone;

nodeos_dtab_t *
nodeos_desc_allocroottab(ani_mempool_t *mpool, int startlen, int maxlen)
{
	struct nodeos_dtab *tab;

	tab = ani_mempool_alloc(mpool, sizeof(nodeos_dtab_t));
	if (tab == 0)
		return 0;

	tab->ents = ani_mempool_alloc(mpool, startlen * sizeof(nodeos_dent_t));
	if (tab->ents == 0) {
		ani_mempool_free(mpool, tab, sizeof(nodeos_dtab_t));
		return 0;
	}

	memset(tab->ents, 0, startlen * sizeof(nodeos_dent_t));
	tab->len = startlen;
	tab->maxlen = maxlen;
	tab->pool = mpool;
	root_dtab = tab;

	return tab;
}

nodeos_dtab_t *
nodeos_desc_alloctab(ani_domain_t *domain, int startlen, int maxlen)
{
	struct nodeos_dtab *tab;

	/* XXX woeful hack */
	bootstrapdone = 1;

#ifdef GLOBAL_DTAB
	tab = root_dtab;
#else
	ani_mempool_t *mpool;

	assert_ani_domain_valid(domain);
	mpool = ani_domain_mempool(domain);

	tab = ani_mempool_alloc(mpool, sizeof(nodeos_dtab_t));
	if (tab == 0)
		return 0;

	tab->ents = ani_mempool_alloc(mpool, startlen * sizeof(nodeos_dent_t));
	if (tab->ents == 0) {
		ani_mempool_free(mpool, tab, sizeof(nodeos_dtab_t));
		return 0;
	}

	memset(tab->ents, 0, startlen * sizeof(nodeos_dent_t));
	tab->len = startlen;
	tab->maxlen = maxlen;
	tab->pool = mpool;
#endif

	return tab;
}

/*
 * Free a descriptor table.
 */
void
nodeos_desc_freetab(nodeos_dtab_t *dtab)
{
#ifdef GLOBAL_DTAB
	assert(dtab == root_dtab);
#else
	if (dtab != 0) {
		assert(dtab->ents != 0);
		assert(dtab->mpool != 0);

		ani_mempool_free(dtab->mpool, dtab->ents,
				 dtab->len * sizeof(nodeos_dent_t));
		ani_mempool_free(dtab->mpool, dtab, sizeof(nodeos_dtab_t));
	}
#endif
}

/*
 * Insert static objects into a descriptor table.
 * Used at boot time to allocate descriptors for the various "root" objects.
 */
void
nodeos_desc_inserttab(nodeos_dtab_t *dtab, void *xobj, void *iobj, int ntype,
		      int *indexp)
{
	nodeos_dent_t *dent;

	dent = nodeos_desc_alloc(dtab, indexp);
	assert(dent != 0);

	dent->type = ntype;
	dent->flags = 0;
	dent->iptr = iobj;
	dent->xptr = xobj;
}

ani_error_t
nodeos_obj_alloc(ani_domain_t *domain, void *mem, int ntype, int xsize, int isize,
		 void **iobjp)
{
	void *iobj, *xobj;
	nodeos_dent_t *dent;
	nodeos_dtab_t *dtab;

	assert_ani_domain_valid(domain);

	iobj = ani_mempool_alloc(ani_domain_mempool(domain),
				 isize + (mem ? 0 : xsize));
	if (iobj == 0)
		return ani_error_set(ENOMEM, "no memory for NodeOS object");

	if (mem == 0)
		xobj = iobj + isize;
	else
		xobj = mem;

#ifdef GLOBAL_DTAB
	dtab = root_dtab;
#else
	dtab = ani_domain_getapistate(domain);
#endif
	assert(dtab != 0);

	dent = nodeos_desc_alloc(dtab, 0);
	if (dent != 0) {
		dent->type = ntype;
		dent->flags = NODEOS_FLAG_ALLOCINT;
		if (mem == 0)
			dent->flags |= NODEOS_FLAG_ALLOCEXT;
		dent->iptr = iobj;
		dent->xptr = xobj;
	} else {
		ani_mempool_free(ani_domain_mempool(domain), iobj,
				 isize + (mem ? 0 : xsize));
		return ani_error_set(ENOSPC, "domain descriptor table full");
	}

	*iobjp = iobj;
	return ani_error_clear();
}

void
nodeos_obj_free(ani_domain_t *domain, void *iobj, int ntype, int xsize, int isize)
{
	nodeos_dent_t *dent;

	assert_ani_domain_valid(domain);

	dent = nodeos_desc_lookup(domain, iobj, ntype, 0);
	assert(dent != 0);

	if (dent->flags & NODEOS_FLAG_ALLOCINT) {
		int len = isize;

		if (dent->flags & NODEOS_FLAG_ALLOCEXT)
			len += xsize;
		ani_mempool_free(ani_domain_mempool(domain), dent->iptr, len);
	}

	dent->type = NODEOS_TYPE_NONE;
}

/*
 * Return the object corresponding to the given descriptor table index.
 */
ani_error_t
nodeos_obj_intern(ani_domain_t *domain, void *xobj, int ix, int ntype,
		  void **iobjp)
{
	struct nodeos_dtab *tab;

	if (!bootstrapdone && domain == 0)
		tab = root_dtab;
	else {
		assert_ani_domain_valid(domain);
#ifdef GLOBAL_DTAB
		tab = root_dtab;
#else
		tab = ani_domain_getapistate(domain);
#endif
		assert(tab != 0);
	}

	if (ix < 1 || ix >= tab->len)
		return ani_error_set(EINVAL, "object descriptor out of range");

	if (tab->ents[ix].type != ntype)
		return ani_error_set(EINVAL, "object descriptor of wrong type");

	if (tab->ents[ix].xptr != xobj)
		return ani_error_set(EINVAL, "object descriptor corrupted");

	*iobjp = tab->ents[ix].iptr;
	return ani_error_clear();
}

/*
 * Lookup an object in a descriptor table and return the corresponding index.
 */
ani_error_t
nodeos_obj_extern(ani_domain_t *domain, void *iobj, int ntype,
		  void **xobjp, int *ixp)
{
	nodeos_dent_t *dent;

	assert_ani_domain_valid(domain);
	assert(xobjp != 0);
	assert(ixp != 0);

	dent = nodeos_desc_lookup(domain, iobj, ntype, ixp);
	if (dent == 0)
		return ani_error_set(EINVAL,
				     "object not found in descriptor table");

	*xobjp = dent->xptr;
	return ani_error_clear();
}

static nodeos_dent_t *
nodeos_desc_alloc(nodeos_dtab_t *dtab, int *ixp)
{
	int i;

	for (i = 1; i < dtab->len; i++)
		if (dtab->ents[i].type == NODEOS_TYPE_NONE) {
			if (ixp)
				*ixp = i;
			return &dtab->ents[i];
		}

	/*
	 * If the table is full but can grow, try to grow it.
	 */
	if (dtab->len < dtab->maxlen) {
		nodeos_dent_t *ntab, *otab;
		int nlen, olen;

		assert(dtab->pool != 0);
		nlen = dtab->len + NODEOS_DESC_GROWSIZE;
		if (nlen > dtab->maxlen)
			nlen = dtab->maxlen;

		ntab = ani_mempool_alloc(dtab->pool,
					 nlen * sizeof(nodeos_dent_t));
		if (ntab != 0) {
			/*
			 * Do the realloc thing
			 */
			olen = dtab->len;
			otab = dtab->ents;
			memcpy(ntab, otab, olen * sizeof(nodeos_dent_t));
			memset(&ntab[olen], 0,
			       (nlen-olen) * sizeof(nodeos_dent_t));
			dtab->ents = ntab;
			dtab->len = nlen;
			ani_mempool_free(dtab->pool, otab,
					 olen * sizeof(nodeos_dent_t));

			assert(dtab->ents[olen].type == NODEOS_TYPE_NONE);
			if (ixp)
				*ixp = olen;
			return &dtab->ents[olen];
		}
	}

	return 0;
}

static nodeos_dent_t *
nodeos_desc_lookup(ani_domain_t *domain, void *iobj, int ntype, int *ixp)
{
	nodeos_dtab_t *tab;
	int i;

#ifdef GLOBAL_DTAB
	tab = root_dtab;
#else
	tab = ani_domain_getapistate(domain);
#endif
	assert(tab != 0);

	for (i = 1; i < tab->len; i++)
		if (tab->ents[i].type == ntype && tab->ents[i].iptr == iobj) {
			if (ixp)
				*ixp = i;
			return &tab->ents[i];
		}

	return 0;
}
