/*
 * Copyright (c) 1999-2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * The Channel API
 */
#include <nodeos/an_error.h>
#include <nodeos/an_chan.h>
#include <nodeos/an_pbuf.h>
#include <nodeos/an_demuxkey.h>

#include <nodeos/ani/ani_nodeos.h>
#include <nodeos/ani/ani_domain.h>
#include <nodeos/ani/ani_chan.h>
#include <nodeos/ani/ani_mempool.h>
#include <nodeos/config.h>

#include "ani_os.h"

ani_chan_t *
NODEOS_CREATE_CHAN(ani_domain_t *domain, an_chan_t *xchan)
{
	assert_ani_domain_valid(domain);

	if (xchan == 0) {
		xchan = ani_mempool_alloc(ani_domain_mempool(domain),
					  sizeof(an_chan_t));
		if (xchan == 0)
			return 0;
		xchan->nodeosobj = xchan;
	} else
		xchan->nodeosobj = 0;

	return &xchan->chan;
}

void
NODEOS_DESTROY_CHAN(ani_domain_t *domain, ani_chan_t *ichan)
{
	an_chan_t *xchan = (an_chan_t *)ichan;

	assert_ani_domain_valid(domain);

	assert(xchan != 0);
	if (xchan->nodeosobj != 0) {
		assert(xchan->nodeosobj == xchan);
		ani_mempool_free(ani_domain_mempool(domain), xchan,
				 sizeof(an_chan_t));
	}
}

/*
 * @allocation (optional): sizeof(an_domain_t) bytes for channel
 * @allocation (temporary): sizeof(ani_pbuf_t *) * npbufs bytes for pbuf ptrs
 */
an_chan_t *
an_inchan_create(void *mem, an_domain_t *domain, an_demuxkey_t *key,
		 an_chanproto_t *proto, an_chanaddr_t *addr,
		 an_netspec_t *net, an_chanrecvfunc_t func, void *farg)
{
	an_chan_t *xchan;
	ani_chan_t *ichan;
	ani_pbuf_t **ipbufs = 0;
	ani_domain_t *idomain, *hdomain;
	ani_demuxkey_t *ikey = 0;
	ani_mempool_t *mpool = 0;
	int npbufs = 0;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	/*
	 * Create a new object
	 */
	ichan = NODEOS_CREATE_CHAN(hdomain, mem);
	if (ichan == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "cannot create new channel");
		goto done;
	}

	/*
	 * Validate arguments.
	 *
	 * N.B.: no copyin of proto/addr/cpu specs as copyin/out are not
	 * needed in our implementation.  We do need to copy the specs
	 * before returning, but that is done in chan_init.
	 */
	idomain = NODEOS_INTERN_DOMAIN(hdomain, domain);
	assert(idomain != 0);
	if (proto == 0 || addr == 0 || net->maxthreads <= 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid proto/addr/maxthreads argument");
		goto done;
	}
	if (key != 0) {
		ikey = NODEOS_INTERN_DEMUXKEY(hdomain, key);
		assert(ikey != 0);
	}
	npbufs = net->buffers.npbufs;
	if (npbufs != 0) {
		an_pbuf_t **pbufs;
		int i;

		pbufs = net->buffers.pbufs;
		if (pbufs == 0) {
			err = ani_error_set(AN_ERR_INVALIDARG,
					    "invalid pbuf list");
			goto done;
		}
		mpool = ani_domain_mempool(hdomain);
		ipbufs = ani_mempool_alloc(mpool, sizeof(*ipbufs) * npbufs);
		if (ipbufs == 0) {
			err = ani_error_set(AN_ERR_NOMEMORY,
					    "could not allocate pbufs");
			goto done;
		}
		for (i = 0; i < npbufs; i++) {
			ipbufs[i] = NODEOS_INTERN_PBUF(hdomain, pbufs[i]);
			if (ipbufs[i] == 0) {
				err = ani_error_set(AN_ERR_INVALIDARG,
						    "invalid pbuf (%d)", i);
				goto done;
			}
		}
	}

	/*
	 * Create and enable the channel.
	 */
	err = ani_chan_init(ichan, CHAN_IN, idomain,
			    ani_domain_threadpool(idomain),
			    proto, addr, ikey, net->maxthreads,
			    ipbufs, npbufs, 0,
			    func, farg);
	if (ani_error_occured(err))
		goto done;

	if (func != 0) {
		err = ani_chan_enable(ichan);
		if (ani_error_occured(err)) {
			ani_chan_deinit(ichan);
			goto done;
		}
	}

 done:
	if (ipbufs)
		ani_mempool_free(mpool, ipbufs, sizeof(*ipbufs) * npbufs);

	if (ani_error_occured(err)) {
		if (ichan)
			NODEOS_DESTROY_CHAN(hdomain, ichan);
		xchan = 0;
	} else
		xchan = NODEOS_EXTERN_CHAN(hdomain, ichan);

	NODEOS_LEAVE;
	return xchan;
}

an_chan_t *
an_outchan_create(void *mem, an_domain_t *domain, an_chanproto_t *proto,
		  an_chanaddr_t *addr, an_netspec_t *net)
{
	an_chan_t *xchan;
	ani_chan_t *ichan;
	ani_domain_t *idomain, *hdomain;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	/*
	 * Create a new object
	 */
	ichan = NODEOS_CREATE_CHAN(hdomain, mem);
	if (ichan == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "cannot create new channel");
		goto done;
	}

	/*
	 * Validate arguments.
	 *
	 * N.B.: no copyin of proto/addr/net specs as copyin/out are not
	 * needed in our implementation.  We do need to copy the specs
	 * before returning, but that is done in chan_init.
	 */
	idomain = NODEOS_INTERN_DOMAIN(hdomain, domain);
	assert(idomain != 0);
	if (proto == 0 || addr == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid proto/addr argument");
		goto done;
	}

	/*
	 * Create the channel.
	 */
	err = ani_chan_init(ichan, CHAN_OUT, idomain,
			    ani_domain_threadpool(idomain),
			    proto, addr, 0, 0, 0, 0, net->bandwidth, 0, 0);

 done:
	if (ani_error_occured(err)) {
		if (ichan)
			NODEOS_DESTROY_CHAN(hdomain, ichan);
		xchan = 0;
	} else
		xchan = NODEOS_EXTERN_CHAN(hdomain, ichan);

	NODEOS_LEAVE;
	return xchan;
}

an_chan_t *
an_cutchan_create(void *mem, an_domain_t *domain, an_chanproto_t *proto,
		  an_chan_t *inchan, an_chan_t *outchan)
{
	an_chan_t *xchan;
	ani_chan_t *ichan, *iinchan, *ioutchan;
	ani_domain_t *idomain, *hdomain;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	/*
	 * Create a new object
	 */
	ichan = NODEOS_CREATE_CHAN(hdomain, mem);
	if (ichan == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "cannot create new channel");
		goto done;
	}

	/*
	 * Validate arguments.
	 */
	idomain = NODEOS_INTERN_DOMAIN(hdomain, domain);
	assert(idomain != 0);
	iinchan = NODEOS_INTERN_CHAN(hdomain, inchan);
	assert(iinchan != 0);
	ioutchan = NODEOS_INTERN_CHAN(hdomain, outchan);
	assert(ioutchan != 0);

	/*
	 * Create the channel.
	 */
	err = ani_cutchan_init(ichan, idomain, ani_domain_threadpool(idomain),
			       0, proto, iinchan, ioutchan,
			       ani_cutchan_receive, ichan);
	if (ani_error_occured(err))
		goto done;

	/*
	 * Enable the inchan
	 */
	err = ani_chan_enable(iinchan);
	if (ani_error_occured(err)) {
		ani_chan_deinit(ichan);
		goto done;
	}

 done:
	if (ani_error_occured(err)) {
		if (ichan)
			NODEOS_DESTROY_CHAN(hdomain, ichan);
		xchan = 0;
	} else
		xchan = NODEOS_EXTERN_CHAN(hdomain, ichan);

	NODEOS_LEAVE;
	return xchan;
}

#ifdef CHAN_CLICK
/*
 *
 */
an_chan_t *
an_clickinchan_create(void *mem, an_domain_t *domain, an_demuxkey_t *key,
		      an_chanproto_t *proto, an_chanaddr_t *addr,
		      an_netspec_t *net, an_chanrecvfunc_t func, void *farg,
		      an_chan_t *inchan)
{
	an_chan_t *xchan;
	ani_chan_t *ichan, *iinchan;
	ani_pbuf_t **ipbufs = 0;
	ani_domain_t *idomain, *hdomain;
	ani_demuxkey_t *ikey = 0;
	ani_mempool_t *mpool = 0;
	int npbufs = 0;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	/*
	 * Create a new object
	 */
	ichan = NODEOS_CREATE_CHAN(hdomain, mem);
	if (ichan == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "cannot create new channel");
		goto done;
	}

	/*
	 * Validate arguments.
	 *
	 * N.B.: no copyin of proto/addr/cpu specs as copyin/out are not
	 * needed in our implementation.  We do need to copy the specs
	 * before returning, but that is done in chan_init.
	 */
	idomain = NODEOS_INTERN_DOMAIN(hdomain, domain);
	assert(idomain != 0);
	iinchan = NODEOS_INTERN_CHAN(hdomain, inchan);
	assert(iinchan != 0);
	if (ani_chan_type(iinchan) != CHAN_IN || iinchan->tdata.in.func != 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "channel type/function not correct");
		goto done;
	}

	if (key != 0) {
		ikey = NODEOS_INTERN_DEMUXKEY(hdomain, key);
		assert(ikey != 0);
	}
	npbufs = net->buffers.npbufs;
	if (npbufs != 0) {
		an_pbuf_t **pbufs;
		int i;

		pbufs = net->buffers.pbufs;
		if (pbufs == 0) {
			err = ani_error_set(AN_ERR_INVALIDARG,
					    "invalid pbuf list");
			goto done;
		}
		mpool = ani_domain_mempool(hdomain);
		ipbufs = ani_mempool_alloc(mpool, sizeof(*ipbufs) * npbufs);
		if (ipbufs == 0) {
			err = ani_error_set(AN_ERR_NOMEMORY,
					    "could not allocate pbufs");
			goto done;
		}
		for (i = 0; i < npbufs; i++) {
			ipbufs[i] = NODEOS_INTERN_PBUF(hdomain, pbufs[i]);
			if (ipbufs[i] == 0) {
				err = ani_error_set(AN_ERR_INVALIDARG,
						    "invalid pbuf (%d)", i);
				goto done;
			}
		}
	}

	/*
	 * Create and enable the Nodeos Click channel.
	 */
	err = ani_clickchan_init(ichan, CHAN_IN, idomain,
				 ani_domain_threadpool(idomain),
				 proto, addr, ikey, net->maxthreads,
				 ipbufs, npbufs, 0,
				 func, farg,
				 iinchan);
	if (ani_error_occured(err))
		goto done;

	if (func != 0) {
		err = ani_chan_enable(ichan);
		if (ani_error_occured(err)) {
			ani_chan_deinit(ichan);
			goto done;
		}
	}

 done:
	if (ipbufs)
		ani_mempool_free(mpool, ipbufs, sizeof(*ipbufs) * npbufs);

	if (ani_error_occured(err)) {
		if (ichan)
			NODEOS_DESTROY_CHAN(hdomain, ichan);
		xchan = 0;
	} else
		xchan = NODEOS_EXTERN_CHAN(hdomain, ichan);

	NODEOS_LEAVE;
	return xchan;
}

/*
 *
 */
an_chan_t *
an_clickoutchan_create(void *mem, an_domain_t *domain,
		       an_chanproto_t *proto, an_chanaddr_t *addr,
		       an_netspec_t *net,
		       an_chan_t *outchan)
{
	an_chan_t *xchan;
	ani_chan_t *ichan, *ioutchan;
	ani_domain_t *idomain, *hdomain;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	/*
	 * Create a new object
	 */
	ichan = NODEOS_CREATE_CHAN(hdomain, mem);
	if (ichan == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "cannot create new channel");
		goto done;
	}

	/*
	 * Validate arguments.
	 *
	 * N.B.: no copyin of proto/addr/cpu specs as copyin/out are not
	 * needed in our implementation.  We do need to copy the specs
	 * before returning, but that is done in chan_init.
	 */
	idomain = NODEOS_INTERN_DOMAIN(hdomain, domain);
	assert(idomain != 0);
	ioutchan = NODEOS_INTERN_CHAN(hdomain, outchan);
	assert(ioutchan != 0);
	if (ani_chan_type(ioutchan) != CHAN_OUT) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid channel type");
		goto done;
	}

	/*
	 * Create and enable the Nodeos Click channel.
	 */
	err = ani_clickchan_init(ichan, CHAN_OUT, idomain,
				 ani_domain_threadpool(idomain),
				 proto, addr, 0, 0, 0, 0, net->bandwidth,
				 0, 0, ioutchan);

 done:
	if (ani_error_occured(err)) {
		if (ichan)
			NODEOS_DESTROY_CHAN(hdomain, ichan);
		xchan = 0;
	} else
		xchan = NODEOS_EXTERN_CHAN(hdomain, ichan);

	NODEOS_LEAVE;
	return xchan;
}
#endif

an_error_t
an_chan_destroy(an_chan_t *chan)
{
	ani_chan_t *ichan;
	ani_domain_t *hdomain;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	ichan = NODEOS_INTERN_CHAN(hdomain, chan);
	assert(ichan != 0);

	err = ani_chan_deinit(ichan);

	if (!ani_error_occured(err))
		NODEOS_DESTROY_CHAN(hdomain, ichan);

	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_outchan_send(an_chan_t *chan, an_pbuf_t *pbuf)
{
	ani_domain_t *hdomain;
	ani_error_t err;

	NODEOS_ENTER;
	hdomain = ani_domain_current();
	err = ani_chan_send(NODEOS_INTERN_CHAN(hdomain, chan),
			    NODEOS_INTERN_PBUF(hdomain, pbuf));
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

#if 0
an_error_t
an_outchan_sendv(an_chan_t *chan, an_pbuf_t *pbufs[], int npbufs)
{
	ani_os_panic("ani_chan_sendv called");
	return AN_ERR_INVALIDARG;
}
#endif

an_error_t
an_inchan_handoff(an_chan_t *chan, an_pbuf_t *pbuf)
{
	ani_domain_t *hdomain;
	ani_error_t err;

	NODEOS_ENTER;
	hdomain = ani_domain_current();
	err = ani_chan_handoff(NODEOS_INTERN_CHAN(hdomain, chan),
			       NODEOS_INTERN_PBUF(hdomain, pbuf));
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_inchan_addpbuf(an_chan_t *chan, an_pbuf_t *pbuf)
{
	ani_domain_t *hdomain;
	ani_chan_t *ichan;
	ani_pbuf_t *ipbuf[1];
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	ichan = NODEOS_INTERN_CHAN(hdomain, chan);
	assert(ichan != 0);
	ipbuf[0] = NODEOS_INTERN_PBUF(hdomain, pbuf);
	assert(ipbuf[0] != 0);

	ani_chan_pbuf_add(ichan, ipbuf, 1);
	err = ani_error_clear();

	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_inchan_rempbuf(an_chan_t *chan, int flags, an_pbuf_t **pbufp)
{
	ani_os_panic("ani_chan_rempbuf called");
	return AN_ERR_INVALIDARG;
}

an_error_t
an_chan_stats(an_chan_t *chan, an_chanstats_t *stats)
{
	ani_domain_t *hdomain;
	ani_chan_t *ichan;
	an_chanstats_t *istats;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();
	ichan = NODEOS_INTERN_CHAN(hdomain, chan);
	assert(ichan != 0);

	/*
	 * Get the stats and copy them out to the user's memory.
	 *
	 * N.B. we can just do a direct copy in our implementation.
	 */
	istats = ani_chan_stats(ichan);
	memcpy(stats, istats, sizeof *stats);
	err = ani_error_clear();

	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

/*
 * Include the inlined functions in non-inlined form
 * XXX must be last in the file
 */
#undef NODEOS_FUNC_INLINE
#define NODEOS_FUNC_INLINE

#include <nodeos/impl/an_chan_inline.h>
