/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * NodeOS cond entrypoints.
 */

#include <nodeos/an_cond.h>
#include <nodeos/an_mutex.h>
#include <nodeos/an_error.h>
#include <nodeos/ani/ani_mutex.h>
#include <nodeos/ani/ani_cond.h>
#include <nodeos/ani/ani_domain.h>
#include <nodeos/ani/ani_mempool.h>

ani_cond_t *
NODEOS_CREATE_COND(ani_domain_t *domain, an_cond_t *xcond)
{
	assert_ani_domain_valid(domain);

	if (xcond == 0) {
		xcond = ani_mempool_alloc(ani_domain_mempool(domain),
					  sizeof(an_cond_t));
		if (xcond == 0)
			return 0;
		xcond->nodeosobj = xcond;
	} else
		xcond->nodeosobj = 0;

	return &xcond->cond;
}

void
NODEOS_DESTROY_COND(ani_domain_t *domain, ani_cond_t *icond)
{
	an_cond_t *xcond = (an_cond_t *)icond;

	assert_ani_domain_valid(domain);

	assert(xcond != 0);
	if (xcond->nodeosobj != 0) {
		assert(xcond->nodeosobj == xcond);
		ani_mempool_free(ani_domain_mempool(domain), xcond,
				 sizeof(an_cond_t));
	}
}

/*
 * Initalize an existing piece of memory to be a cond object.
 */
an_cond_t *
an_cond_create(void *mem)
{
	an_cond_t *xcond;
	ani_cond_t *icond;
	ani_domain_t *hdomain;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	/*
	 * Create a new object
	 */
	icond = NODEOS_CREATE_COND(hdomain, mem);
	if (icond == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "cannot create new an_cond");
		goto done;
	}

	err = ani_cond_init(icond);

 done:
	if (ani_error_occured(err)) {
		if (icond)
			NODEOS_DESTROY_COND(hdomain, icond);
		xcond = 0;
	} else
		xcond = NODEOS_EXTERN_COND(hdomain, icond);

	NODEOS_LEAVE;
	return xcond;
}

an_error_t
an_cond_destroy(an_cond_t *cond)
{
	ani_cond_t *icond;
	ani_error_t err;

	NODEOS_ENTER;
	icond = NODEOS_INTERN_COND(ani_domain_current(), cond);
	assert(icond != 0);
	err = ani_cond_deinit(icond);
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

/* an_cond_signal is inlined from <nodeos/impl/an_cond_inline.h> */
/* an_cond_broadcast is inlined from <nodeos/impl/an_cond_inline.h> */
/* an_cond_wait is inlined from <nodeos/impl/an_cond_inline.h> */
/* an_cond_timedwait is inlined from <nodeos/impl/an_cond_inline.h> */

/*
 * Include the inlined functions in non-inlined form
 * XXX must be last in the file
 */
#undef NODEOS_FUNC_INLINE
#define NODEOS_FUNC_INLINE

#include <nodeos/impl/an_cond_inline.h>
