/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * The Credential API
 * XXX largely unspecified
 */
#include <nodeos/an_cred.h>
#include <nodeos/an_impl.h>

#include <nodeos/ani/ani_nodeos.h>
#include <nodeos/ani/ani_cred.h>
#include <nodeos/ani/ani_domain.h>

ani_cred_t *
NODEOS_CREATE_CRED(ani_domain_t *domain, an_cred_t *xcred)
{
	if (xcred == 0) {
		xcred = ani_mempool_alloc(ani_domain_mempool(domain),
					  sizeof(an_cred_t));
		if (xcred == 0)
			return 0;
		xcred->nodeosobj = xcred;
	} else
		xcred->nodeosobj = 0;

	return &xcred->cred;
}

void
NODEOS_DESTROY_CRED(ani_domain_t *domain, ani_cred_t *icred)
{
	an_cred_t *xcred = (an_cred_t *)icred;

	assert(xcred != 0);
	if (xcred->nodeosobj != 0) {
		assert(xcred->nodeosobj == xcred);
		ani_mempool_free(ani_domain_mempool(domain), xcred,
				 sizeof(an_cred_t));
	}
}

an_cred_t *
an_cred_create(void *mem)
{
	an_cred_t *xcred;
	ani_cred_t *icred;
	ani_domain_t *hdomain;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	/*
	 * Create a new object
	 */
	icred = NODEOS_CREATE_CRED(hdomain, mem);
	if (icred == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "cannot create new an_cred");
		goto done;
	}

	err = ani_cred_init(icred, 0, 0);

 done:
	if (ani_error_occured(err)) {
		if (icred)
			NODEOS_DESTROY_CRED(hdomain, icred);
		xcred = 0;
	} else
		xcred = NODEOS_EXTERN_CRED(hdomain, icred);

	NODEOS_LEAVE;
	return xcred;
}

an_error_t
an_cred_resource_get(an_cred_t *cred,
		     unsigned int *nthreads, unsigned int *cpurate,
		     unsigned int *memtotal, unsigned int *netrate)
{
	ani_error_t err;
	ani_cred_t *icred;

	NODEOS_ENTER;
	icred = NODEOS_INTERN_CRED(ani_domain_current(), cred);
	assert(icred != 0);

	/*
	 * N.B. no copyout needed in our implementation
	 */
	err = ani_cred_resource_get(icred, nthreads, cpurate, memtotal,
				    netrate);

	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_cred_resource_xfer(an_cred_t *srccred, an_cred_t *dstcred,
		      unsigned int nthreads, unsigned int cpurate,
		      unsigned int memtotal, unsigned int netrate)
{
	ani_error_t err;
	ani_cred_t *isrccred, *idstcred;
	ani_domain_t *hdomain;

	NODEOS_ENTER;
	hdomain = ani_domain_current();
	isrccred = NODEOS_INTERN_CRED(hdomain, srccred);
	assert(isrccred != 0);
	idstcred = NODEOS_INTERN_CRED(hdomain, dstcred);
	assert(idstcred != 0);

	err = ani_cred_resource_xfer(isrccred, idstcred, nthreads,
				     cpurate, memtotal, netrate);
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

/* an_cred_current is inlined in <nodeos/impl/an_cred_inline.h> */

/*
 * Include the inlined functions in non-inlined form
 * XXX must be last in the file
 */
#undef NODEOS_FUNC_INLINE
#define NODEOS_FUNC_INLINE

#include <nodeos/impl/an_cred_inline.h>
