/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * The Packet API
 */
#include <nodeos/an_error.h>
#include <nodeos/an_demuxkey.h>
#include <nodeos/ani/ani_nodeos.h>

#include <nodeos/ani/ani_domain.h>
#include <nodeos/ani/ani_mempool.h>
#include <nodeos/ani/ani_demuxkey.h>
#include <nodeos/ani/ani_error.h>

ani_demuxkey_t *
NODEOS_CREATE_DEMUXKEY(ani_domain_t *domain, an_demuxkey_t *xdemuxkey)
{
	if (xdemuxkey == 0) {
		xdemuxkey = ani_mempool_alloc(ani_domain_mempool(domain),
					      sizeof(an_demuxkey_t));
		if (xdemuxkey == 0)
			return 0;
		xdemuxkey->nodeosobj = xdemuxkey;
	} else
		xdemuxkey->nodeosobj = 0;

	return &xdemuxkey->demuxkey;
}

void
NODEOS_DESTROY_DEMUXKEY(ani_domain_t *domain, ani_demuxkey_t *idemuxkey)
{
	an_demuxkey_t *xdemuxkey = (an_demuxkey_t *)idemuxkey;

	assert(xdemuxkey != 0);
	if (xdemuxkey->nodeosobj != 0) {
		assert(xdemuxkey->nodeosobj == xdemuxkey);
		ani_mempool_free(ani_domain_mempool(domain), xdemuxkey,
				 sizeof(an_demuxkey_t));
	}
}

an_demuxkey_t *
an_demuxkey_create(void *mem, an_domain_t *domain)
{
	an_demuxkey_t *xkey;
	ani_demuxkey_t *ikey;
	ani_domain_t *idomain, *hdomain;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	/*
	 * Create a new object
	 */
	ikey = NODEOS_CREATE_DEMUXKEY(hdomain, mem);
	if (ikey == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "cannot create new demuxkey");
		goto done;
	}

	idomain = NODEOS_INTERN_DOMAIN(hdomain, domain);
	assert(idomain != 0);

	assert(idomain == hdomain);
	err = ani_demuxkey_init(ikey);

 done:
	if (ani_error_occured(err)) {
		if (ikey)
			NODEOS_DESTROY_DEMUXKEY(hdomain, ikey);
		xkey = 0;
	} else
		xkey = NODEOS_EXTERN_DEMUXKEY(hdomain, ikey);

	NODEOS_LEAVE;
	return xkey;
}

an_error_t
an_demuxkey_destroy(an_demuxkey_t *key)
{
	ani_demuxkey_t *ikey;
	ani_domain_t *hdomain;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	ikey = NODEOS_INTERN_DEMUXKEY(hdomain, key);
	assert(ikey != 0);

	err = ani_demuxkey_deinit(ikey);

	if (!ani_error_occured(err))
		NODEOS_DESTROY_DEMUXKEY(hdomain, ikey);

	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_demuxkey_t *
an_demuxkey_clone(void *mem, an_demuxkey_t *orig)
{
	an_demuxkey_t *xkey;
	ani_demuxkey_t *inew, *iorig;
	ani_domain_t *hdomain;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	/*
	 * Create a new object
	 */
	inew = NODEOS_CREATE_DEMUXKEY(hdomain, mem);
	if (inew == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "cannot create new an_demuxkey");
		goto done;
	}

	iorig = NODEOS_INTERN_DEMUXKEY(hdomain, orig);
	assert(iorig != 0);

	err = ani_demuxkey_clone(inew, iorig);

 done:
	if (ani_error_occured(err)) {
		if (inew)
			NODEOS_DESTROY_DEMUXKEY(hdomain, inew);
		xkey = 0;
	} else
		xkey = NODEOS_EXTERN_DEMUXKEY(hdomain, inew);

	NODEOS_LEAVE;
	return xkey;
}

/*
 * @allocation: demuxkey segment
 */
an_error_t
an_demuxkey_add(an_demuxkey_t *key, int len, void *datap, void *maskp)
{
	ani_error_t err;
	ani_demuxkey_t *ikey;
	ani_demuxkeyseg_t *seg;

	NODEOS_ENTER;

	ikey = NODEOS_INTERN_DEMUXKEY(ani_domain_current(), key);
	assert(ikey != 0);

	/*
	 * Create a tuple segment with indicated key/mask info.
	 *
	 * N.B. no copyin needed in our implementation.
	 */
	seg = ani_demuxkeyseg_tuple(0, len, datap, maskp);
	if (seg == 0) {
		err = ani_error_set(AN_ERR_NOMEMORY,
				    "no memory for demuxkey segment");
		goto done;
	}

	err = ani_demuxkey_append(ikey, seg);
	if (ani_error_occured(err))
		ani_demuxkeyseg_destroy(seg);

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

/*
 * @allocation: demuxkey segment
 */
an_error_t
an_demuxkey_addeth(an_demuxkey_t *key, int flags,
		   u_int8_t src[6], u_int8_t dst[6], u_int16_t etype)
{
	ani_error_t err;
	ani_demuxkey_t *ikey;
	ani_demuxkeyseg_t *seg;

	NODEOS_ENTER;

	ikey = NODEOS_INTERN_DEMUXKEY(ani_domain_current(), key);
	assert(ikey != 0);

	/*
	 * Create a tuple segment with indicated ethernet header info.
	 *
	 * N.B. no copyin needed in our implementation.
	 */
	seg = ani_demuxkeyseg_eth(0, flags, src, dst, etype);
	if (seg == 0) {
		err = ani_error_set(AN_ERR_NOMEMORY,
				    "no memory for demuxkey segment");
		goto done;
	}

	err = ani_demuxkey_append(ikey, seg);
	if (ani_error_occured(err))
		ani_demuxkeyseg_destroy(seg);

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_demuxkey_addipv4(an_demuxkey_t *key, int flags,
		    u_int32_t src, u_int32_t srcmask,
		    u_int32_t dst, u_int32_t dstmask,
		    u_int8_t proto)
{
	ani_demuxkey_t *ikey;
	ani_demuxkeyseg_t *seg;
	ani_error_t err;

	NODEOS_ENTER;

	ikey = NODEOS_INTERN_DEMUXKEY(ani_domain_current(), key);
	assert(ikey != 0);

	/*
	 * Create a segment with indicated IP header info.
	 */
	seg = ani_demuxkeyseg_ipv4(0, flags,
				   src, srcmask, dst, dstmask, proto);
	if (seg == 0) {
		err = ani_error_set(AN_ERR_NOMEMORY,
				    "no memory for demuxkey segment");
		goto done;
	}

	err = ani_demuxkey_append(ikey, seg);
	if (ani_error_occured(err))
		ani_demuxkeyseg_destroy(seg);

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_demuxkey_addudp(an_demuxkey_t *key, int flags, u_int16_t src, u_int16_t dst)
{
	ani_demuxkey_t *ikey;
	ani_demuxkeyseg_t *seg;
	ani_error_t err;

	NODEOS_ENTER;

	ikey = NODEOS_INTERN_DEMUXKEY(ani_domain_current(), key);
	assert(ikey != 0);

	/*
	 * Create a segment with indicated UDP header info.
	 */
	seg = ani_demuxkeyseg_udp(0, flags, src, dst);
	if (seg == 0) {
		err = ani_error_set(AN_ERR_NOMEMORY,
				    "no memory for demuxkey segment");
		goto done;
	}

	err = ani_demuxkey_append(ikey, seg);
	if (ani_error_occured(err))
		ani_demuxkeyseg_destroy(seg);

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_demuxkey_addtcp(an_demuxkey_t *key, int flags, u_int16_t src, u_int16_t dst)
{
	ani_demuxkey_t *ikey;
	ani_demuxkeyseg_t *seg;
	ani_error_t err;

	NODEOS_ENTER;

	ikey = NODEOS_INTERN_DEMUXKEY(ani_domain_current(), key);
	assert(ikey != 0);

	/*
	 * Create a segment with indicated TCP header info.
	 */
	seg = ani_demuxkeyseg_tcp(0, flags, src, dst);
	if (seg == 0) {
		err = ani_error_set(AN_ERR_NOMEMORY,
				    "no memory for demuxkey segment");
		goto done;
	}

	err = ani_demuxkey_append(ikey, seg);
	if (ani_error_occured(err))
		ani_demuxkeyseg_destroy(seg);

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_demuxkey_addanep(an_demuxkey_t *key, int flags,
		    u_int8_t version, u_int16_t protoid)
{
	ani_demuxkey_t *ikey;
	ani_demuxkeyseg_t *seg;
	ani_error_t err;

	NODEOS_ENTER;

	ikey = NODEOS_INTERN_DEMUXKEY(ani_domain_current(), key);
	assert(ikey != 0);

	/*
	 * Create a segment with indicated TCP header info.
	 */
	seg = ani_demuxkeyseg_anep(0, flags, version, protoid);
	if (seg == 0) {
		err = ani_error_set(AN_ERR_NOMEMORY,
				    "no memory for demuxkey segment");
		goto done;
	}

	err = ani_demuxkey_append(ikey, seg);
	if (ani_error_occured(err))
		ani_demuxkeyseg_destroy(seg);

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_demuxkey_tostring(an_demuxkey_t *key, char *buf, int buflen)
{
	ani_demuxkey_t *ikey;
	ani_error_t err;

	NODEOS_ENTER;
	
	ikey = NODEOS_INTERN_DEMUXKEY(ani_domain_current(), key);
	assert(ikey != 0);

	if (buf == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid char* parameter");
		goto done;
	}

	/* XXX no intern of buf, buflen */
	err = ani_demuxkey_tostring(ikey, buf, buflen);

 done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

int
an_demuxkey_matches(an_demuxkey_t *key, void *data, int datalength)
{
	ani_demuxkey_t *ikey;
	ani_error_t err;
	int offset;
	
	NODEOS_ENTER;

	ikey = NODEOS_INTERN_DEMUXKEY(ani_domain_current(), key);
	assert(ikey != 0);

	if (data == 0 || datalength < 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "bad data/length argument "
				    "in an_demuxkey_matches");
		offset = -1;
		goto done;
	}
	
	/*
	 * N.B. no copyin of data necessary in our implementation.
	 */
	offset = ani_demuxkey_match(ikey, data, datalength);
	err = ani_error_clear();

 done:
	NODEOS_LEAVE;
	return offset;
}

/*
 * Include the inlined functions in non-inlined form
 * XXX must be last in the file
 */
#undef NODEOS_FUNC_INLINE
#define NODEOS_FUNC_INLINE

#include <nodeos/impl/an_demuxkey_inline.h>
