/*
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <nodeos/an_event.h>

#include <nodeos/ani/ani_nodeos.h>
#include <nodeos/ani/ani_event.h>
#include <nodeos/ani/ani_domain.h>

ani_event_t *
NODEOS_CREATE_EVENT(ani_domain_t *domain, an_event_t *xevent)
{
	if (xevent == 0) {
		xevent = ani_mempool_alloc(ani_domain_mempool(domain),
					   sizeof(an_event_t));
		if (xevent == 0)
			return 0;
		xevent->nodeosobj = xevent;
	} else
		xevent->nodeosobj = 0;

	return &xevent->event;
}

void
NODEOS_DESTROY_EVENT(ani_domain_t *domain, ani_event_t *ievent)
{
	an_event_t *xevent = (an_event_t *)ievent;

	assert(xevent != 0);
	if (xevent->nodeosobj != 0) {
		assert(xevent->nodeosobj == xevent);
		ani_mempool_free(ani_domain_mempool(domain), xevent,
				 sizeof(an_event_t));
	}
}

an_event_t *
an_event_create(void *mem)
{
	an_event_t *xevent;
	ani_event_t *ievent;
	ani_domain_t *hdomain;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	/*
	 * Create a new object
	 */
	ievent = NODEOS_CREATE_EVENT(hdomain, mem);
	if (ievent == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "cannot create new event");
		goto done;
	}

	err = ani_event_init(ievent, ani_domain_threadpool(hdomain));

 done:
	if (ani_error_occured(err)) {
		if (ievent)
			NODEOS_DESTROY_EVENT(hdomain, ievent);
		xevent = 0;
	} else
		xevent = NODEOS_EXTERN_EVENT(hdomain, ievent);

	NODEOS_LEAVE;
	return xevent;
}

an_error_t
an_event_destroy(an_event_t *event,
		 an_eventtype_t *eventType, an_eventstate_t *eventState)
{
	ani_error_t err;
	ani_domain_t *hdomain;
	ani_event_t *ievent;
	int iflags;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	ievent = NODEOS_INTERN_EVENT(hdomain, event);
	assert(ievent != 0);

	err = ani_event_deinit(ievent, &iflags);
	if (!ani_error_occured(err)) {
		NODEOS_DESTROY_EVENT(hdomain, ievent);
		if (eventType)
			*eventType = ani_event_flagstotype(iflags);
		if (eventState)
			*eventState = ani_event_flagstostate(iflags);
	}

	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

/* an_event_schedule is in <nodeos/impl/an_event_inline.h> */
/* an_event_fire is in <nodeos/impl/an_event_inline.h> */
/* an_event_status is in <nodeos/impl/an_event_inline.h> */
/* an_event_deschedule is in <nodeos/impl/an_event_inline.h> */

an_error_t
an_event_redefine(an_event_t *event, an_timespec_t *oldtime,
		  an_eventtype_t *oldtype, an_eventstate_t *oldstate,
		  an_eventfunc_t *oldfunc, void **oldfarg,
		  an_timespec_t *newtime, an_eventtype_t newtype,
		  an_eventfunc_t newfunc, void *newfarg)
{
	ani_event_t *ievent;
	ani_error_t ierr;
	int oflags;

	NODEOS_ENTER;
	ievent = NODEOS_INTERN_EVENT(ani_domain_current(), event);
	ierr = ani_event_redefine(ievent, oldtime, &oflags, oldfunc, oldfarg,
				  newtime, newtype, newfunc, newfarg);
	NODEOS_LEAVE;

	if (!ani_error_occured(ierr)) {
		if (oldtype)
			*oldtype = ani_event_flagstotype(oflags);
		if (oldstate)
			*oldstate = ani_event_flagstostate(oflags);
	}

	return NODEOS_EXTERN_ERROR(ierr);
}

/* an_event_granularity is in <nodeos/impl/an_event_inline.h> */

/*
 * Include the inlined functions in non-inlined form
 * XXX must be last in the file
 */
#undef NODEOS_FUNC_INLINE
#define NODEOS_FUNC_INLINE

#include <nodeos/impl/an_event_inline.h>
