/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * NodeOS mutex entrypoints
 */

#include <nodeos/an_mutex.h>
#include <nodeos/an_error.h>
#include <nodeos/ani/ani_mutex.h>
#include <nodeos/ani/ani_domain.h>
#include <nodeos/ani/ani_mempool.h>

ani_mutex_t *
NODEOS_CREATE_MUTEX(ani_domain_t *domain, an_mutex_t *xmutex)
{
	if (xmutex == 0) {
		xmutex = ani_mempool_alloc(ani_domain_mempool(domain),
					   sizeof(an_mutex_t));
		if (xmutex == 0)
			return 0;
		xmutex->nodeosobj = xmutex;
	} else
		xmutex->nodeosobj = 0;

	return &xmutex->mutex;
}

void
NODEOS_DESTROY_MUTEX(ani_domain_t *domain, ani_mutex_t *imutex)
{
	an_mutex_t *xmutex = (an_mutex_t *)imutex;

	assert(xmutex != 0);
	if (xmutex->nodeosobj != 0) {
		assert(xmutex->nodeosobj == xmutex);
		ani_mempool_free(ani_domain_mempool(domain), xmutex,
				 sizeof(an_mutex_t));
	}
}

/*
 * Initalize an existing piece of memory to be a mutex object.
 */
an_mutex_t *
an_mutex_create(void *mem)
{
	an_mutex_t *xmux;
	ani_mutex_t *imux;
	ani_domain_t *hdomain;
	ani_error_t err;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	/*
	 * Create a new object
	 */
	imux = NODEOS_CREATE_MUTEX(hdomain, mem);
	if (imux == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "cannot create new mutex");
		goto done;
	}

	err = ani_mutex_init(imux);

 done:
	if (ani_error_occured(err)) {
		if (imux)
			NODEOS_DESTROY_MUTEX(hdomain, imux);
		xmux = 0;
	} else
		xmux = NODEOS_EXTERN_MUTEX(hdomain, imux);

	NODEOS_LEAVE;
	return xmux;
}

an_error_t
an_mutex_destroy(an_mutex_t *mutex)
{
	ani_error_t err;
	ani_domain_t *hdomain;
	ani_mutex_t *imux;

	NODEOS_ENTER;

	hdomain = ani_domain_current();

	imux = NODEOS_INTERN_MUTEX(hdomain, mutex);
	assert(imux != 0);

	err = ani_mutex_deinit(imux);
	if (!ani_error_occured(err))
		NODEOS_DESTROY_MUTEX(hdomain, imux);

	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

/* an_mutex_lock is in <nodeos/impl/an_mutex_inline.h> */
/* an_mutex_trylock is in <nodeos/impl/an_mutex_inline.h> */
/* an_mutex_unlock is in <nodeos/impl/an_mutex_inline.h> */

/*
 * Include the inlined functions in non-inlined form
 * XXX must be last in the file
 */
#undef NODEOS_FUNC_INLINE
#define NODEOS_FUNC_INLINE

#include <nodeos/impl/an_mutex_inline.h>
