/*
 * Copyright (c) 1999-2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * A "clock" channel driver.
 *
 * In-channel only.  Packets are delivered at a rate given by the address info.
 * (e.g., if address is "5", a packet is delivered every 5th clock tick).
 * The delivered packet contains the 4-byte internal time value (``myticks'').
 */
#include <nodeos/config.h>

#ifdef CHAN_CLOCK

#include <string.h>
#include <sys/time.h>

#include <nodeos/impl/an_error.h>
#include <nodeos/ani/ani_domain.h>
#include <nodeos/ani/ani_chan.h>
#include <nodeos/ani/ani_pbuf.h>
#include <nodeos/ani/ani_event.h>
#include <nodeos/ani/ani_mempool.h>

static void chan_clock_init(void);
static void chan_clock_deinit(void);
static int chan_clock_matches(an_chanproto_t *proto);
static ani_error_t chan_clock_create(ani_chan_t *chan);
static void chan_clock_destroy(ani_chan_t *chan);
static ani_error_t chan_clock_enable(ani_chan_t *chan);
static void chan_clock_disable(ani_chan_t *chan);

static char chan_clock_prefix[] = "clock";	/* XXX */

struct chanops chan_clock_ops = {
	chan_clock_init,	chan_clock_deinit,	chan_clock_matches,
	chan_clock_create,	chan_clock_destroy,	0,
	chan_clock_enable,	chan_clock_disable,
	0,			0
};

#define CBUFSIZ	sizeof(unsigned int)

struct chandata {
	ani_event_t event;
	ani_chan_t *chan;
	int freq;
};

static ani_event_t ticker;
static unsigned int mytime;
static void chan_clock_intr(struct an_event *event, void *arg);
static unsigned long nspertick;
static unsigned long tickspersec;

void
chan_clock_init(void)
{
	/*
	 * XXX chan_clock_init is called too early in the wrong context
	 * so we don't do anything.
	 */
}

void
chan_clock_deinit(void)
{
	ani_event_deinit(&ticker, 0);
}

void
chan_clock_start(void)
{
	an_timespec_t gap;
	ani_error_t err;
	static int done;

	if (done++)
		return;

	ani_event_setup();
	ani_event_info(&nspertick);
	tickspersec = (1000*1000*1000) / nspertick;

	gap.tv_sec = 0;
	gap.tv_nsec = nspertick;
	
	err = ani_event_init(&ticker,
			     ani_domain_threadpool(ani_domain_current()));
	assert(!ani_error_occured(err));
	err = ani_event_schedule(&ticker, &gap,
				 AN_EVTYPE_PERIODIC,
				 ANI_EV_INKERN|ANI_EV_DIRECT,
				 chan_clock_intr, 0);
	assert(!ani_error_occured(err));
}

static int
chan_clock_matches(an_chanproto_t *proto)
{
	int len;

	len = strlen(chan_clock_prefix);
	if (proto->len < len)
		return 0;

	return (strncmp(proto->data, chan_clock_prefix, len) == 0);
}

ani_error_t
chan_clock_create(ani_chan_t *chan)
{
	int freq;
	struct chandata *cdata;

	chan_clock_start();	/* XXX */

	freq = chan->addr.clock.freq;
	if (freq < 1)
		return ani_error_set(EINVAL, "invalid clockchan freqency");

	cdata = ani_mempool_alloc(ani_mempool_current(), sizeof *cdata);
	if (cdata == 0)
		return ani_error_set(ENOMEM, "no memory for clockchan data");

	cdata->chan = chan;
	cdata->freq = freq;
	chan->data = cdata;

	return ani_error_clear();
}

void
chan_clock_destroy(ani_chan_t *chan)
{
	struct chandata *cdata;

	assert(chan->data != 0);
	cdata = chan->data;

	/* XXX shouldn't be needed, but just in case */
	(void) ani_event_deinit(&cdata->event, 0);

	ani_mempool_free(ani_mempool_current(), cdata, sizeof *cdata);
	chan->data = 0;
}

ani_error_t
chan_clock_enable(ani_chan_t *chan)
{
	struct chandata *cdata;
	an_timespec_t gap;
	ani_error_t err;

	assert(chan->data != 0);
	cdata = chan->data;

	err = ani_event_init(&cdata->event, chan->tpool);
	if (!ani_error_occured(err)) {
		gap.tv_sec = cdata->freq / tickspersec;
		gap.tv_nsec = (cdata->freq % tickspersec) * nspertick;
		err = ani_event_schedule(&cdata->event, &gap,
					 AN_EVTYPE_PERIODIC,
					 ANI_EV_INKERN|ANI_EV_DIRECT,
					 chan_clock_intr, chan);
	}
	return err;
}

void
chan_clock_disable(ani_chan_t *chan)
{
	struct chandata *cdata;

	assert(chan->data != 0);
	cdata = chan->data;

	(void) ani_event_deinit(&cdata->event, 0);
}

#ifdef OSKIT
static void
chan_clock_intr(struct an_event *event, void *arg)
{
	ani_chan_t *chan = arg;
	ani_pbuf_t *pbuf;
	ani_error_t err;

	assert(event != 0);

	/*
	 * No channel means the ticker event.
	 */
	if (chan == 0) {
		mytime++;
		return;
	}

	err = ani_chan_pbuf_alloc(chan, CBUFSIZ, &pbuf);
	if (!ani_error_occured(err)) {
		*(unsigned int *)ani_pbuf_data(pbuf) = mytime;
		ani_chan_thread_startintr(chan, pbuf, 0);
	}
}
#else
static void
chan_clock_intr(struct an_event *event, void *arg)
{
	ani_chan_t *chan = arg;
	ani_pbuf_t *pbuf;
	ani_error_t err;

	assert(event != 0);

	/*
	 * No channel means the ticker event.
	 */
	if (chan == 0) {
		mytime++;
		return;
	}

	ani_chan_lock(chan);
	if ((chan->flags & CHAN_DYING) == 0) {
		err = ani_chan_pbuf_alloc(chan, CBUFSIZ, &pbuf);
		if (!ani_error_occured(err)) {
			*(unsigned int *)ani_pbuf_data(pbuf) = mytime;
			ani_chan_thread_start(chan, pbuf, 0);
		}
	}
	ani_chan_unlock(chan);
}
#endif

#endif
