/*
 * Copyright (c) 1999-2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * A raw ethernet channel driver.
 */
#include <nodeos/config.h>

#ifdef CHAN_ETH

#include <stdio.h>
#include <string.h>
#include <assert.h>

#ifdef OSKIT
#include <stdlib.h>
#include <oskit/startup.h>
#include <oskit/dev/dev.h>
#include <oskit/dev/ethernet.h>
#include <oskit/dev/net.h>
#include <oskit/net/ether.h>
#include <oskit/io/bufio.h>
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>

#ifndef ETHER_HDR_SIZE
#define ETHER_HDR_SIZE 14
#endif

#include "os/unix_eth.h"
#endif

#include <nodeos/an_info.h>

#include <nodeos/impl/an_impl.h>
#include <nodeos/impl/an_pbuf.h>

#include <nodeos/ani/ani_error.h>
#include <nodeos/ani/ani_nodeos.h>
#include <nodeos/ani/ani_ilock.h>
#include <nodeos/ani/ani_domain.h>
#include <nodeos/ani/ani_thread.h>
#include <nodeos/ani/ani_threadpool.h>
#include <nodeos/ani/ani_chan.h>
#include <nodeos/ani/ani_pbuf.h>
#include <nodeos/ani/ani_demuxkey.h>
#include <nodeos/ani/ani_mempool.h>
#include <nodeos/ani/ani_time.h>

#include "ani_os.h"
#include "pq/pq.h"

/*
 * Max packets to pull off of one interface during a poll call (0 for all)
 */
#define CHAN_ETH_MAXPOLL	0

static void chan_eth_init(void);
static void chan_eth_deinit(void);
static int chan_eth_matches(an_chanproto_t *proto);
static ani_error_t chan_eth_create(ani_chan_t *chan);
static void chan_eth_destroy(ani_chan_t *chan);
static ani_error_t chan_eth_send(ani_chan_t *chan, ani_pbuf_t *pbuf);
static ani_error_t chan_eth_enable(ani_chan_t *chan);
static void chan_eth_disable(ani_chan_t *chan);
#ifdef USE_POLLING
static int chan_eth_poll(void);
static void chan_eth_txflush(ani_chan_t *chan);
#else
#define chan_eth_poll	0
#define chan_eth_txflush 0
#endif

static char chan_eth_prefix[] = "eth";	/* XXX */

struct chanops chan_eth_ops = {
	chan_eth_init,		chan_eth_deinit,	chan_eth_matches,
	chan_eth_create,	chan_eth_destroy,	chan_eth_send,
	chan_eth_enable,	chan_eth_disable,
	chan_eth_poll,		chan_eth_txflush
};

/*
 * Per-device structure.
 * Most oskit drivers restrict to only a single open of a device,
 * we use a single send/recv netio shared by many channels.
 *
 * The etherdev lock is lower in the locking hierarchy than the channel
 * lock.
 */
struct etherdev {
	ani_ilock_t	 lock;
	char		 name[NODEOS_MAXIFNAME]; /* "eth0", etc. */
	unsigned char	 addr[6];	/* our address */
	int		 refcount;	/* # of channels using device */
#ifdef OSKIT
	oskit_etherdev_t *eth;		/* device handle */
#else
	ani_thread_t	 thread;	/* service thread */
	ani_icond_t	 cond;		/* synch with service thread */
	int		 flags;		/* more synch with service thread */
	char		 if_name[IFNAMSIZ]; /* system ethernet device name */
	int		 recv;		/* socket file descriptor */
#endif
	pq_scheduler_t	 *pq_scheduler;	/* packet scheduler */
	unsigned int	 bwavail;	/* total available bandwidth of IF */
	unsigned int	 bwinuse;	/* bandwidth allocated to channels */
	unsigned int	 bwscale;	/* scale to pq ticket values */

	/* accessed at interrupt time, must be protected with intr_disable */
	ani_demuxkeylist_t demux;	/* list of demuxers for this device */
#ifdef OSKIT
	oskit_netio_t	 *send;		/* handle for outgoing packets */
	oskit_netio_t	 *recv;		/* handle for incoming packets */
#endif
};

#define ANI_EDEV_DYING	1
#define ANI_EDEV_DEAD	2

/*
 * Per-channel structure.
 * You get one of these per channel.
 */
struct etherdata {
	struct etherdev	*edev;		/* Per device Per channel */
	pq_client_t	*pq;		/* Each channel talks to a PQ */
};

/* used for comparisons */
static unsigned char zaddr[6] = { 0 };

static struct etherdev *etherdevs;
static int netherdevs, nactivedevs;
static ani_ilock_t ethlock;

static ani_error_t chan_eth_demux_prepend(ani_chan_t *chan);
static struct etherdev *chan_edev_find(char *name);
static ani_error_t chan_edev_init(struct etherdev *edev);
static void chan_edev_deinit(struct etherdev *edev);
static ani_error_t chan_edev_allocbw(struct etherdev *edev, ani_chan_t *chan);
static void chan_edev_freebw(struct etherdev *edev, ani_chan_t *chan);
#ifdef OSKIT
#ifdef USE_POLLING
static int chan_edev_poll(struct etherdev *edev, int maxpackets);
static void chan_edev_txflush(struct etherdev *edev);
#else
static oskit_error_t chan_edev_intr(void *arg, oskit_bufio_t *b,
				    oskit_size_t psize);
#endif
#else
static void chan_edev_intr(void *arg1, void *arg2);
#endif

/*
 * Initialize support for ethernet devices.  Runs once at startup.
 */
static void
chan_eth_init(void)
{
	int ndev, i;
#ifdef OSKIT
	oskit_etherdev_t **edev;
#else
	struct if_info *edev;
#endif	      

	ani_ilock_init(&ethlock);
	ani_ilock_lock(&ethlock);

	/*
	 * Look up all the ethernet devices
	 */
#ifdef OSKIT
	ndev = osenv_device_lookup(&oskit_etherdev_iid, (void***)&edev);
#else 
	ndev = unix_eth_get_if_info(&edev);
#endif
	if (ndev <= 0) {
		ani_ilock_unlock(&ethlock);
		return;
	}

	/*
	 * Create a device to name mapping
	 */
	etherdevs = ani_mempool_alloc(ani_mempool_current(),
				      ndev * sizeof(struct etherdev));
	if (etherdevs == 0) {
		ani_ilock_unlock(&ethlock);
		return;
	}
	
	memset(etherdevs, 0, ndev * sizeof(struct etherdev));
	for (i = 0; i < ndev; i++) {
		snprintf(etherdevs[i].name, sizeof etherdevs[i].name,
			 "%s%d", chan_eth_prefix, i);
#ifdef OSKIT
		etherdevs[i].eth = edev[i];
		oskit_etherdev_getaddr(edev[i], etherdevs[i].addr);
#else 
		strncpy(etherdevs[i].if_name, edev[i].name, IFNAMSIZ);
		memcpy(etherdevs[i].addr, edev[i].addr, 6);
		ani_icond_init(&etherdevs[i].cond);
		etherdevs[i].recv = -1;
#endif
		etherdevs[i].bwavail = 100 * 1000 * 1000;	/* XXX */
		etherdevs[i].bwscale = 100 * 1000;		/* XXX */
		ani_demuxkeylist_init(&etherdevs[i].demux);

		ani_ilock_init(&etherdevs[i].lock);
	}
	netherdevs = ndev;
	
#ifdef OSKIT
	free(edev);  /* XXX should be osenv_devicearray_free() */
#else
	unix_eth_free_if_info(ndev, edev);
#endif

	ani_ilock_unlock(&ethlock);
}

static void
chan_eth_deinit(void)
{
	int i;

	ani_ilock_lock(&ethlock);
	for (i = 0; i < netherdevs; i++) {
		assert(etherdevs[i].refcount == 0);
#ifndef OSKIT
		ani_icond_deinit(&etherdevs[i].cond);
#endif
		ani_ilock_deinit(&etherdevs[i].lock);
	}
	ani_mempool_free(ani_mempool_current(),
			 etherdevs, netherdevs * sizeof(struct etherdev));
	ani_ilock_unlock(&ethlock);
	ani_ilock_deinit(&ethlock);
}

static int
chan_eth_matches(an_chanproto_t *proto)
{
	int len;

	len = strlen(chan_eth_prefix);
	if (proto->len < len)
		return 0;

	return (strncmp(proto->data, chan_eth_prefix, len) == 0);
}


#ifdef USE_POLLING
static int
chan_eth_poll(void)
{
	int i, count = 0;
	struct etherdev *edev;

	assert(netherdevs > 0);

	/*
	 * XXX quick check.
	 * We don't lock access to this variable but we manipulate it in
	 * such a way that the worst that will happen is that we may scan
	 * the list when nothing is actually active.
	 */
	if (nactivedevs == 0)
		return 0;

	for (i = 0; i < netherdevs; i++) {
		edev = &etherdevs[i];
		if (ani_ilock_trylock(&edev->lock) == 0) {
			if (edev->refcount > 0)
				count += chan_edev_poll(edev, CHAN_ETH_MAXPOLL);
			ani_ilock_unlock(&edev->lock);
		}
	}

	return count;
}

static void
chan_eth_txflush(ani_chan_t *chan)
{
	struct etherdev *edev = ((struct etherdata *)chan->data)->edev;

	assert(ani_chan_type(chan) == CHAN_OUT);

	if (ani_ilock_trylock(&edev->lock) == 0) {
		if (edev->refcount > 0)
			chan_edev_txflush(edev);
		ani_ilock_unlock(&edev->lock);
	}
}
#endif

/*
 * Hook up to the ethernet device.  Note that we do not allocate any actual
 * channel buffers for in-chans.  We use the buffer memory for incoming
 * packets.
 *
 * @allocation: ethernet specific channel data structure
 */
static ani_error_t
chan_eth_create(ani_chan_t *chan)
{
	struct etherdev *edev;
	struct etherdata *edata;
	ani_error_t err;

	assert(ani_chan_type(chan) != CHAN_CUT);

	/*
	 * Lookup the device corresponding to the name given
	 */
	edev = chan_edev_find(ani_chan_name(chan));
	if (edev == 0)
		return ani_error_set(EINVAL, "not a valid chaneth device");

	/*
	 * Special in-channel processing.
	 */
	if (ani_chan_type(chan) == CHAN_IN) {
		/*
		 * Augment the demux filter with the address and demux info
		 * as necessary.
		 */
		err = chan_eth_demux_prepend(chan);
		if (err)
			return err;
	}

	/*
	 * Special out-channel processing.
	 */
	if (ani_chan_type(chan) == CHAN_OUT) {
		/*
		 * Fill in the channel address source interface address with
		 * the interface provided info (if user didn't explicitly
		 * provide one).
		 */
		if (memcmp(chan->addr.eth.saddr, zaddr, sizeof zaddr) == 0)
			memcpy(chan->addr.eth.saddr, edev->addr,
			       sizeof edev->addr);
	}

	/*
	 * Allocate per-channel thingie.  XXX allocation.
	 */
	edata = ani_mempool_alloc(ani_mempool_current(),
				  sizeof(struct etherdata));
	if (edata == 0)
		return ani_error_set(ENOMEM, "no memory for chaneth data");
	memset(edata, 0, sizeof *edata);
	chan->data = edata;

	/*
	 * Now hook up to the ethernet device
	 */
	ani_ilock_lock(&edev->lock);

	/*
	 * If device hasn't been initialized, do so now
	 */
	edata->edev = edev;
	if (edev->refcount == 0) {
		err = chan_edev_init(edev);
		if (ani_error_occured(err)) {
			ani_ilock_unlock(&edev->lock);
			ani_mempool_free(ani_mempool_current(), edata,
					 sizeof(struct etherdata));
			return err;
		}
	}
	edev->refcount++;

	/*
	 * Ensure bandwidth requested can be allocated on outchans.
	 */
	if (ani_chan_type(chan) == CHAN_OUT) {
		err = chan_edev_allocbw(edev, chan);
		if (ani_error_occured(err)) {
			if (--edev->refcount == 0)
				chan_edev_deinit(edev);
			ani_ilock_unlock(&edev->lock);
			ani_mempool_free(ani_mempool_current(), edata,
					 sizeof(struct etherdata));
			return err;
		}
	}

	ani_ilock_unlock(&edev->lock);
	return ani_error_clear();
}

/*
 * Called with the channel locked
 */
static void
chan_eth_destroy(ani_chan_t *chan)
{
	struct etherdev *edev;

	edev = ((struct etherdata *)chan->data)->edev;
	ani_ilock_lock(&edev->lock);

	/*
	 * Release reserved bandwidth
	 */
	if (ani_chan_type(chan) == CHAN_OUT)
		chan_edev_freebw(edev, chan);

	/*
	 * Shutdown the device if this is the last channel using it.
	 */
	if (--edev->refcount == 0)
		chan_edev_deinit(edev);
	assert(edev->refcount >= 0);

	/* XXX kill service threads, free up packets/bufios */
	
	ani_ilock_unlock(&edev->lock);

	/*
	 * Finally, free up the per-channel structure.
	 */
	ani_mempool_free(ani_mempool_current(),	 chan->data,
			 sizeof(struct etherdata));
}

static ani_error_t
chan_eth_enable(ani_chan_t *chan)
{
	struct etherdev *edev;
	ani_error_t err;

	assert(ani_chan_type(chan) == CHAN_IN);

	nactivedevs++;	/* XXX not locked */

	edev = ((struct etherdata *)chan->data)->edev;
	ani_ilock_lock(&edev->lock);

	ani_os_intr_disable();
	err = ani_demuxkeylist_add(&edev->demux, chan->tdata.in.demux);
	ani_os_intr_enable();

	ani_ilock_unlock(&edev->lock);
	return err;
}

static void
chan_eth_disable(ani_chan_t *chan)
{
	struct etherdev *edev;

	assert(ani_chan_type(chan) == CHAN_IN);

	edev = ((struct etherdata *)chan->data)->edev;
	ani_ilock_lock(&edev->lock);
	ani_os_intr_disable();
	ani_demuxkeylist_remove(&edev->demux, chan->tdata.in.demux);
	ani_os_intr_enable();
	ani_ilock_unlock(&edev->lock);

	nactivedevs--;	/* XXX not locked */
}

/*
 * Send a packet out on a channel.
 * Called with the channel locked.
 */
static ani_error_t
chan_eth_send(ani_chan_t *chan, ani_pbuf_t *pbuf)
{
#ifdef OSKIT
	oskit_bufio_t *bio;
	size_t psize;
#else
	void *buf;
	size_t bufsize;
#endif
	struct etherdata *edata;
	struct etherdev *edev;
	int didcopy;
	size_t hsize;
	void *hdr;
	ani_error_t err;
	int ierr;

	assert(ani_chan_type(chan) == CHAN_OUT);

	edata = chan->data;
	edev = edata->edev;

	if ((chan->flags & CHAN_ISRAW) == 0) {
		hdr = &chan->addr.eth;
		hsize = ETHER_HDR_SIZE;
	} else {
		hdr = 0;
		hsize = 0;
	}

#ifdef OSKIT	    
	/*
	 * Create a bufio for the packet and header.
	 * XXX note we record the size of the pbuf before this call
	 * since it might change the size of the pbuf (to include the header).
	 */
	psize = ani_pbuf_size(pbuf);
	err = ani_pbuf_tobufio(pbuf, hdr, hsize, edev->send, &bio, &didcopy);
	if (ani_error_occured(err)) {
		CHANSTATS(chan->stats.dropfailedbuf++);
		return err;
	}

	/*
	 * Push the packet out. We should probably give our bufio reference
	 * to the pq code instead, since it does an addref and this does
	 * a release.
	 */
	ani_pbuf_setbusy(pbuf);
	ierr = pq_send_packet(edata->pq, bio, hsize+psize);
	if (ierr != 0) {
		ani_pbuf_clrbusy(pbuf);
		err = ani_error_set(ierr, "pq_send_packet failed");
		CHANSTATS(chan->stats.dropnobufs++);
	}
	oskit_bufio_release(bio);
	if (didcopy)
		CHANSTATS(chan->stats.outbufcopies++);
#else
	err = ani_pbuf_tomem(pbuf, hdr, hsize, &buf, &bufsize, &didcopy);
	if (ani_error_occured(err)) {
		CHANSTATS(chan->stats.dropfailedbuf++);
		return err;
	}
	ierr = ani_os_write_eth(edev->recv, buf, bufsize);
	if (ierr < 0) {
		err = ani_error_set(errno, "os_write_packet failed");
		CHANSTATS(chan->stats.dropnobufs++);
	}
	if (didcopy) {
		ani_pbuf_relmem(buf, bufsize);
		CHANSTATS(chan->stats.outbufcopies++);
	}
#endif /* OSKIT */	  
	return err;
}

/*
 * Prepend channel address info into the demux information.
 * Called with the channel locked.
 *
 * XXX this function changes the user-provided demux,
 * it may also screw it up royally on an error.
 */
static ani_error_t
chan_eth_demux_prepend(ani_chan_t *chan)
{
	an_ethaddrspec_t *eaddr;
	ani_error_t err;
	ani_demuxkeyseg_t *seg;
	int flags;

	assert(ani_chan_type(chan) == CHAN_IN);
	assert(chan->tdata.in.demux != 0);

	/*
	 * If RAW, nothing to do.
	 */
	if (chan->flags & CHAN_ISRAW)
		return ani_error_clear();

	eaddr = &chan->addr.eth;

	/*
	 * Determine what fields should be wildcarded.
	 * XXX is it safe to use 0 addresses as wildcards?
	 */
	flags = 0;
	if (memcmp(eaddr->daddr, &zaddr, sizeof zaddr) == 0)
		flags |= AN_DEMUXKEY_ETH_ANYDADDR;
	if (memcmp(eaddr->saddr, &zaddr, sizeof zaddr) == 0)
		flags |= AN_DEMUXKEY_ETH_ANYSADDR;
	if (eaddr->type == 0)
		flags |= AN_DEMUXKEY_ETH_ANYETYPE;

	seg = ani_demuxkeyseg_eth(0, flags, eaddr->saddr, eaddr->daddr,
				  eaddr->type);
	if (seg == 0)
		err = ani_error_set(ENOMEM,
				    "no memory for eth header demuxkey");
	else {
		err = ani_demuxkey_prepend(chan->tdata.in.demux, seg);
		if (ani_error_occured(err))
			ani_demuxkeyseg_destroy(seg);
	}

	return err;
}

static struct etherdev *
chan_edev_find(char *name)
{
	int i, len;

	len = strlen(name);

	ani_ilock_lock(&ethlock);
	for (i = 0; i < netherdevs; i++)
		if (strncmp(etherdevs[i].name, name, len) == 0) {
			ani_ilock_unlock(&ethlock);
			return &etherdevs[i];
		}

	ani_ilock_unlock(&ethlock);
	return 0;
}

#ifdef OSKIT
#ifdef USE_POLLING
/*
 * Flush completed transmissions on the device.
 * This can trigger bufio releases for packets that were sent earlier.
 */
static void
chan_edev_txflush(struct etherdev *edev)
{
	(void) oskit_etherdev_rxpoll(edev->eth, 0, 0);
}

/*
 * Poll the ethernet device for input packets.
 * Returns the number of packets read/processed.
 * Called with edev locked.
 */
static int
chan_edev_poll(struct etherdev *edev, int maxpackets)
{
	oskit_bufio_t *b = 0;
	oskit_size_t psize;
	oskit_off_t psize_offt;
	int count, num, ierr;
	ani_chan_t *chan;
	ani_pbuf_t *pbuf;
	size_t hsize;
	void *data;
	ani_error_t err;
	int msize;

	/*
	 * Repeatedly poll for a packet and process it.
	 */
	for (num = 0; maxpackets == 0 || num < maxpackets; num++) {
		/*
		 * Grab a packet
		 */
		if (b != 0) {
			oskit_bufio_release(b);
			b = 0;
		}
		count = 1;
		ierr = oskit_etherdev_rxpoll(edev->eth, &count, &b);
		assert(ierr != OSKIT_ENODEV);
		if (count == 0)
			break;

		/*
		 * Map the packet for demux examination
		 */
		oskit_bufio_getsize(b, &psize_offt);
		psize = (oskit_size_t)psize_offt;
		assert(psize >= ETHER_HDR_SIZE);
		if (oskit_bufio_map(b, &data, 0, psize) != 0)
			continue;

		/*
		 * Find a channel for this packet.
		 * If no channel matches, the packet is dropped.
		 *
		 * Note that the demux is applied to the entire packet
		 * including the ethernet header even for "if" (!CHAN_ISRAW)
		 * channels.  This works for "if" channels because the demux
		 * has been adjusted to account for the header.
		 */
		msize = (int)psize;
		chan = ani_demuxkeylist_apply(&edev->demux, data, &msize);
		oskit_bufio_unmap(b, data, 0, psize);
		if (chan == 0)
			continue;

		/*
		 * XXX dubious: if the channel is already locked, we don't
		 * want to block this (potentially unrelated) thread.
		 * So we drop the packet.
		 *
		 * Note that a trylock is also necessary because we are
		 * grabbing the chan and edev locks out of order.
		 */
		assert_ani_chan_valid(chan);
		if (ani_chan_trylock(chan)) {
			CHANSTATS(chan->stats.droppackets++);
			CHANSTATS(chan->stats.dropbytes +=
				  (psize - ((chan->flags & CHAN_ISRAW) ?
					    0 : ETHER_HDR_SIZE)));
			CHANSTATS(chan->stats.droplocked++);
			continue;
		}
		assert((chan->flags & CHAN_DYING) == 0);

		/*
		 * Check for a buffer as early as possible
		 */
		if (ani_pbuflist_isempty(&chan->tdata.in.rbufs)) {
			CHANSTATS(chan->stats.droppackets++);
			CHANSTATS(chan->stats.dropbytes +=
				  (psize - ((chan->flags & CHAN_ISRAW) ?
					    0 : ETHER_HDR_SIZE)));
			CHANSTATS(chan->stats.dropnobufs++);
			ani_chan_unlock(chan);
			continue;
		}

		/*
		 * XXX here we are again, needing another lock.
		 * Before we start up a thread in thread_startintr,
		 * we must make sure we have the channel's threadpool lock.
		 * We try to grab it here before going through the work
		 * of setting up the pbuf.
		 */
		if (ani_threadpool_trylock(chan->tpool)) {
			CHANSTATS(chan->stats.droppackets++);
			CHANSTATS(chan->stats.dropbytes +=
				  (psize - ((chan->flags & CHAN_ISRAW) ?
					    0 : ETHER_HDR_SIZE)));
			CHANSTATS(chan->stats.droplocked++);
			ani_chan_unlock(chan);
			continue;
		}

		/*
		 * Wrap the bufio with a NodeOS packet buffer
		 * This will strip the ethernet header for "if" channels.
		 */
		hsize = (chan->flags & CHAN_ISRAW) ? 0 : ETHER_HDR_SIZE;
		psize -= hsize;
		err = ani_pbuf_frombufio(chan->domain, chan, b,
					 hsize, psize, &pbuf);

		/*
		 * Finally, if we got a buffer, schedule the packet for
		 * delivery.  Note the adjustment of the demux offset to
		 * account for the header being stripped from "if" channels.
		 */
		if (!ani_error_occured(err))
			ani_chan_thread_startintr(chan, pbuf, msize-hsize);

		ani_threadpool_unlock(chan->tpool);
		ani_chan_unlock(chan);
	}
	if (b != 0)
		oskit_bufio_release(b);

	return num;
}
#else
/*
 * Receive netio push function, called from the interrupt handler
 * in an unknown pthread context.
 */
static oskit_error_t
chan_edev_intr(void *arg, oskit_bufio_t *b, oskit_size_t psize)
{
	struct etherdev *edev = arg;
	ani_chan_t *chan;
	ani_pbuf_t *pbuf;
	size_t hsize;
	void *data;
	ani_error_t err;
	int msize;

#ifndef OSKIT_UNIX
	/* XXX we are actually called at softint level */
	ani_os_intr_disable();
#endif

	/*
	 * Map the packet for demux examination
	 */
	assert(psize >= ETHER_HDR_SIZE);
	if (oskit_bufio_map(b, &data, 0, psize) != 0)
		goto done;

	/*
	 * Find a channel for this packet.
	 * If no channel matches, the packet is dropped.
	 *
	 * Note that the demux is applied to the entire packet including
	 * the ethernet header even for "if" (!CHAN_ISRAW) channels.
	 * This works for "if" channels because the demux has been adjusted
	 * to account for the header.
	 */
	msize = (int)psize;
	chan = ani_demuxkeylist_apply(&edev->demux, data, &msize);
	oskit_bufio_unmap(b, data, 0, psize);
	if (chan == 0)
		goto done;
	assert_ani_chan_valid(chan);
	assert((chan->flags & CHAN_DYING) == 0);

	/*
	 * Check for a buffer as early as possible
	 */
	if (ani_pbuflist_isempty(&chan->tdata.in.rbufs)) {
		CHANSTATS(chan->stats.droppackets++);
		CHANSTATS(chan->stats.dropbytes +=
			  (psize - ((chan->flags & CHAN_ISRAW) ?
				    0 : ETHER_HDR_SIZE)));
		CHANSTATS(chan->stats.dropnobufs++);
		goto done;
	}

	/*
	 * Wrap the buffio with a NodeOS packet buffer
	 * This will strip the ethernet header for "if" channels.
	 */
	hsize = (chan->flags & CHAN_ISRAW) ? 0 : ETHER_HDR_SIZE;
	psize -= hsize;
	err = ani_pbuf_frombufio(chan->domain, chan, b, hsize, psize, &pbuf);

	/*
	 * Finally, if we got a buffer, schedule the packet for delivery.
	 * Note the adjustment of the demux offset to account for the
	 * header being stripped from "if" channels.
	 */
	if (!ani_error_occured(err))
		ani_chan_thread_startintr(chan, pbuf, msize-hsize);

 done:
#ifndef OSKIT_UNIX
	ani_os_intr_enable();
#endif
	return 0;
}
#endif

#else

void
chan_edev_intr(void *arg, void *arg2)
{
	struct etherdev *edev = arg;
	int count, msize;
	ani_error_t err;
	char *buf;
	int fd, bufsize;
	ani_pbuf_t *rbuf, *pbuf;
	ani_chan_t *chan;

	/*
	 * Locking forces synchronization with our creator
	 */
	ani_ilock_lock(&edev->lock);
	fd = edev->recv;
	ani_ilock_unlock(&edev->lock);

	/*
	 * Create an initial read buffer.  We swap this buffer with
	 * empty ones from the devices' channels when we hand off packets.
	 *
	 * XXX should the caller allocate this?
	 */
	rbuf = NODEOS_CREATE_PBUF(ani_domain_current(), 0);
	assert(rbuf != 0);
	err = ani_pbuf_init(rbuf, ani_domain_current(), 0,
			    NODEOS_CHAN_BUFSIZE, 0, NODEOS_CHAN_BUFSIZE);
	assert(!ani_error_occured(err));
	buf = ani_pbuf_data(rbuf);
	bufsize = ani_pbuf_size(rbuf);

	while (1) {
		chan = 0;

		do {
			count = ani_os_read_eth(fd, buf, bufsize, 500);
		} while (count == 0 && (edev->flags & ANI_EDEV_DYING) == 0);
		ani_ilock_lock(&edev->lock);
		if (count <= 0) {
			err = ani_error_set(errno, "os_read_eth failed");
			break;
		}

		/*
		 * Find a channel for this packet.
		 * If no channel matches, the packet is dropped.
		 */
		msize = count;
		chan = ani_demuxkeylist_apply(&edev->demux, buf, &msize);
		ani_ilock_unlock(&edev->lock);
		if (chan == 0)
			continue;

		/*
		 * Allocate a packet buffer from the channel.
		 * If none are available, just drop the packet.
		 */
		ani_chan_lock(chan);
		err = ani_chan_pbuf_alloc(chan, NODEOS_CHAN_BUFSIZE, &pbuf);
		if (ani_error_occured(err)) {
			ani_chan_unlock(chan);
			continue;
		}

		/*
		 * Swap the buffer obtained from the channel with our buffer.
		 * The channel gets the full buffer in exchange for one of its
		 * empty ones that we will use for the next read.
		 */
		ani_pbuf_setsize(rbuf, count, 0);
		ani_pbuf_swap(rbuf, pbuf);

		/*
		 * If this is not a raw ethernet channel, strip off the
		 * ethernet header.
		 */
		if ((chan->flags & CHAN_ISRAW) == 0)
			ani_pbuf_setoffset(pbuf, ETHER_HDR_SIZE, 0);

		/*
		 * Finally, fire up a handler thread.
		 */
		ani_chan_thread_start(chan, pbuf, msize);

		ani_chan_unlock(chan);

		/*
		 * Reset the buffer pointer/size since we have a new
		 * physical buffer.
		 */
		buf = ani_pbuf_data(rbuf);
		bufsize = ani_pbuf_size(rbuf);
	}

	/*
	 * We have broken out of the recvfrom.	Most likely because the
	 * device was closed.  In that case, ANI_EDEV_DYING should be set
	 * and we free the read buffer and exit.  Otherwise we drop dead.
	 */
	if (edev->flags & ANI_EDEV_DYING) {
		ani_pbuf_deinit(rbuf);
		ani_mempool_free(ani_mempool_current(), rbuf, sizeof *rbuf);
		edev->flags |= ANI_EDEV_DEAD;
		ani_icond_broadcast(&edev->cond);
		ani_ilock_unlock(&edev->lock);
	} else {
		ani_ilock_unlock(&edev->lock);
		ani_os_panic("NodeOS: enet interface error (%d)", err);
	}
}
#endif /* OSKIT */

static ani_error_t
chan_edev_init(struct etherdev *edev)
{
#ifdef OSKIT	    
	oskit_error_t ierr;
	int flags = 0;

	assert(edev->recv == 0 && edev->send == 0);

#ifdef USE_POLLING
	/*
	 * Make sure the device is setup for polling.
	 * The receive netio is unused.
	 */
	flags |= OSKIT_ETHERDEV_RX_POLLING;
#ifdef OSKIT_ETHERDEV_TX_POLLING
	flags |= OSKIT_ETHERDEV_TX_POLLING;
#endif
#else
	/*
	 * Create the netio object we will use for receiving packets.
	 * We create a dedicated thread to service the incoming packets in a
	 * full thread context.
	 */
	edev->recv = oskit_netio_create(chan_edev_intr, edev);
	if (edev->recv == 0)
		return ani_error_set(ENOMEM, "oskit_netio_create failed");
#endif

	/*
	 * Open the device registering our receive netio and getting
	 * back the send netio.
	 */
	ierr = oskit_etherdev_open(edev->eth, flags, edev->recv, &edev->send);
	if (ierr) {
		if (edev->recv) {
			oskit_netio_release(edev->recv);
			edev->recv = 0;
		}
		return ani_error_set(ierr, "oskit_etherdev_open failed");
	}

	/*
	 * Create a packet scheduler for this device. Eventually, this might
	 * work under non-oskit situations.
	 */
	ierr = pq_create_scheduler(edev->send, 50, &edev->pq_scheduler);
	if (ierr) {
		if (edev->recv)
			oskit_netio_release(edev->recv);
		oskit_netio_release(edev->send);
		edev->recv = edev->send = 0;
		return ani_error_set(ierr, "pq_create_scheduler failed");
	}
#else
	ani_error_t err;

	assert(edev->recv < 0);

	edev->recv = ani_os_open_eth(edev->if_name);
	if (edev->recv < 0) 
		return ani_error_set(EINVAL, "cannot open ethernet device");
	
	/*
	 * Create the reader thread
	 */
	err = ani_rootdomain_startthread(&edev->thread,
				(ani_thread_func_t)chan_edev_intr, edev, 0, 0);
	if (ani_error_occured(err)) {
		ani_os_close_eth(edev->recv);
		edev->recv = -1;
		return err;
	}
#endif	      
	return ani_error_clear();
}

/*
 * Gotta release the ethernet devices or my PC hangs up in netboot reboot.
 * Called with device lock held.
 */
static void
chan_edev_deinit(struct etherdev *edev)
{
#ifdef OSKIT
	oskit_netio_release(edev->send);
	if (edev->recv)
		oskit_netio_release(edev->recv);
	pq_destroy_scheduler(edev->pq_scheduler);
	edev->send = edev->recv = 0;
#else
	edev->flags = ANI_EDEV_DYING;
	while ((edev->flags & ANI_EDEV_DEAD) == 0)
		ani_icond_wait(&edev->cond, &edev->lock);
	edev->flags = 0;
	ani_os_close_eth(edev->recv);
	edev->recv = -1;
#endif	      
}

/*
 * Take whatever steps necessary to reserve bandwidth on the device.
 * Device should be locked on call/return.
 */
static ani_error_t
chan_edev_allocbw(struct etherdev *edev, ani_chan_t *chan)
{
	assert(ani_chan_type(chan) == CHAN_OUT);

	if (edev->bwinuse + chan->tdata.out.mbs > edev->bwavail)
		return ani_error_set(ENOMEM, "inadequate bandwidth on device");
	edev->bwinuse += chan->tdata.out.mbs;

#ifdef OSKIT
	/*
	 * Create a packet scheduler client of the packet scheduler
	 * for the associated etherdev.
	 *
	 * XXX we arbitrarily pick a queue depth of 1000
	 */
	{
		int ierr, tickets;

		/*
		 * XXX scale a bit.  Note that if the caller specified
		 * no bandwidth (or a very small amount) we let them have
		 * uncontested bandwidth.
		 */
		tickets = chan->tdata.out.mbs / edev->bwscale;
		if (tickets == 0)
			tickets++;

		ierr = pq_create_client(edev->pq_scheduler, 1500, tickets,
					&((struct etherdata *)chan->data)->pq);
		if (ierr != 0)
			return ani_error_set(ierr, "pq_create_client failed");
	}
#endif

	return ani_error_clear();
}

static void
chan_edev_freebw(struct etherdev *edev, ani_chan_t *chan)
{
	assert(ani_chan_type(chan) == CHAN_OUT);

	edev->bwinuse -= chan->tdata.out.mbs;
	assert(edev->bwinuse >= 0);

#ifdef	OSKIT
	/*
	 * Kill the packet scheduler.
	 * Any queued packets will be killed.
	 */
	pq_destroy_client(((struct etherdata *)chan->data)->pq);
#endif
}
#endif /* CHAN_ETH */
