/*
 * Copyright (c) 1999-2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * A "socket" channel driver.
 *
 * Currently we recognize only "socket:if/ipv4/udp".
 */
#include <nodeos/config.h>

#ifdef CHAN_SOCKET

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#ifdef OSKIT
#include <oskit/startup.h>
#include <oskit/com/listener.h>
#include <oskit/dev/dev.h>
#include <oskit/net/socket.h>
#include <oskit/dev/timer.h>
#endif

#include <nodeos/impl/an_error.h>
#include <nodeos/impl/an_pbuf.h>

#include <nodeos/ani/ani_nodeos.h>
#include <nodeos/ani/ani_domain.h>
#include <nodeos/ani/ani_chan.h>
#include <nodeos/ani/ani_demuxkey.h>
#include <nodeos/ani/ani_thread.h>
#include <nodeos/ani/ani_mempool.h>
#include <nodeos/ani/ani_threadpool.h>
#include <nodeos/ani/ani_log.h>

#include "ani_os.h"
#include "ani_debug.h"

static void chan_socket_init(void);
static void chan_socket_deinit(void);
static int chan_socket_matches(an_chanproto_t *proto);
static ani_error_t chan_socket_create(ani_chan_t *chan);
static void chan_socket_destroy(ani_chan_t *chan);
static ani_error_t chan_socket_send(ani_chan_t *chan, ani_pbuf_t *pbuf);
static ani_error_t chan_socket_enable(ani_chan_t *chan);
static void chan_socket_disable(ani_chan_t *chan);
#ifdef USE_POLLING
#define chan_socket_poll	0
//static int chan_socket_poll(void);
#else
#define chan_socket_poll	0
#endif

static char chan_socket_prefix[] = "socket";	/* XXX */

struct chanops chan_socket_ops = {
	chan_socket_init,	chan_socket_deinit,	chan_socket_matches,
	chan_socket_create,	chan_socket_destroy,	chan_socket_send,
	chan_socket_enable,	chan_socket_disable,
	chan_socket_poll,	0,
};

/*
 * One structure per socket.
 * Sockets may be shared by multiple channels, with each channel having
 * a different demux for inChans.
 */
struct socketdata {
	struct socketdata  *next;
	short		   stype; /* SOCK_DGRAM or SOCK_STREAM */
	u_int16_t	   lport; /* local port (network order) */
	u_int32_t	   laddr; /* local address (network order) */
	int		   fd;	  /* socket file descriptor */
	ani_demuxkeylist_t demux; /* list of demuxers for this socket */
	int		   refs;  /* refcount */
	int		   inrefs;/* inchan refs */
	int		   flags; /* flags */
	ani_thread_t	   thread;/* service thread */
	ani_icond_t	   cond;  /* synch with service thread */
	ani_ilock_t	   lock;  /* protects the demux list */
};

#define ANI_SD_THREAD_RUNNING	0x01
#define ANI_SD_THREAD_DYING	0x02

static struct socketdata *socketlist;
static ani_ilock_t socketlock;

/*
 * Channel specific data (chan->data) is a pointer to a socketdata
 * structure (inChan) or a socket file descriptor (outChan).
 */

static ani_error_t chan_sock_demux_prepend(ani_chan_t *chan);
static ani_error_t chan_socketdata_find(int stype, int ctype,
					an_udp4addrspec_t *caddr,
					struct socketdata **out_sdata);
static void chan_socketdata_release(struct socketdata *sdata, int ctype);
static void chan_socketdata_intr(void *arg, void *arg2, void *arg3);

static void
chan_socket_init(void)
{
	ani_ilock_init(&socketlock);

#if defined(OSKIT) && !defined(OSKIT_UNIX)
	{
		int err;

		ani_ilock_lock(&socketlock);
		err = ani_os_init_chansock();
		assert(err == 0);
		ani_ilock_unlock(&socketlock);
	}
#endif
}

static void
chan_socket_deinit(void)
{
#if defined(OSKIT) && !defined(OSKIT_UNIX)
	ani_os_deinit_chansock();
#endif
	ani_ilock_deinit(&socketlock);
}

static int
chan_socket_matches(an_chanproto_t *proto)
{
	int len;

	len = strlen(chan_socket_prefix);
	if (proto->len < len)
		return 0;

	return (strncmp(proto->data, chan_socket_prefix, len) == 0);
}

static ani_error_t
chan_socket_create(ani_chan_t *chan)
{
	struct socketdata *sdata;
	ani_error_t err;
	int stype, len;
	char *cp;

	assert(ani_chan_type(chan) != CHAN_CUT);

	cp = ani_chan_name(chan);
	len = ani_chan_namelen(chan);
	if (strncmp(cp, "socket:if/ipv4/udp", len) == 0)
		stype = SOCK_DGRAM;
#if 0
	else if (strncmp(cp, "socket:if/ipv4/tcp", len) == 0)
		stype = SOCK_STREAM;
#endif
	else
		return ani_error_set(EINVAL,
				     "invalid proto spec for socket channel");

	/*
	 * Find/create a socket of the appropriate type.
	 * Returns the socket data locked.
	 */
	err = chan_socketdata_find(stype, ani_chan_type(chan),
				   &chan->addr.udp4, &sdata);
	if (ani_error_occured(err))
		return err;

	/*
	 * Special in-channel processing.
	 */
	if (ani_chan_type(chan) == CHAN_IN) {
		/*
		 * Augment the demux filter with the address info
		 * as necessary.
		 */
		err = chan_sock_demux_prepend(chan);
		if (err) {
			chan_socketdata_release(sdata, ani_chan_type(chan));
			return err;
		}
	}

	ani_ilock_unlock(&sdata->lock);

	chan->data = sdata;
	return ani_error_clear();
}

static void
chan_socket_destroy(ani_chan_t *chan)
{
	struct socketdata *sdata;

	assert(chan->data);
	sdata = chan->data;

	ani_ilock_lock(&sdata->lock);
	chan_socketdata_release(sdata, ani_chan_type(chan));
	/* lock got destroyed, no need to unlock. */
}

static ani_error_t
chan_socket_enable(ani_chan_t *chan)
{
	struct socketdata *sdata;
	ani_error_t err;

	assert(chan->data);

	if (ani_chan_type(chan) == CHAN_IN) {
		sdata = chan->data;
		ani_ilock_lock(&sdata->lock);
		err = ani_demuxkeylist_add(&sdata->demux, chan->tdata.in.demux);
		ani_ilock_unlock(&sdata->lock);
	} else
		err = ani_error_clear();

	return err;
}

static void
chan_socket_disable(ani_chan_t *chan)
{
	struct socketdata *sdata;

	assert(chan->data);

	if (ani_chan_type(chan) == CHAN_IN) {
		sdata = chan->data;
		ani_ilock_lock(&sdata->lock);
		ani_demuxkeylist_remove(&sdata->demux, chan->tdata.in.demux);
		ani_ilock_unlock(&sdata->lock);
	}
}

static ani_error_t
chan_socket_send(ani_chan_t *chan, ani_pbuf_t *pbuf)
{
	struct socketdata *sdata;
	struct sockaddr_in sa;
	int ierr, size;
	ani_error_t err;

#if defined(__FreeBSD__) || defined(OSKIT)
	sa.sin_len = sizeof sa;
#endif
	sa.sin_family = AF_INET;
	sa.sin_port = chan->addr.udp4.dport;
	sa.sin_addr.s_addr = chan->addr.udp4.daddr;

	sdata = chan->data;
	ani_ilock_lock(&sdata->lock);
	DBG(DBG_SOCKFD, {
		ani_dprintf("C%p: write fd=%d, pbuf=%p (data=%p, size=%d)\n",
			    chan, sdata->fd, pbuf,
			    ani_pbuf_data(pbuf), (int)ani_pbuf_size(pbuf));
	});
	ani_pbuf_setbusy(pbuf);
	size = ani_pbuf_size(pbuf);
	ierr = sendto(sdata->fd, ani_pbuf_data(pbuf), size,
		      0, (struct sockaddr *)&sa, sizeof sa);
	ani_pbuf_clrbusy(pbuf);
	ani_ilock_unlock(&sdata->lock);
	if (ierr == size)
		err = ani_error_clear();
	else if (ierr < 0) {
		CHANSTATS(chan->stats.dropnobufs++);	/* XXX */
		err = ani_error_set(errno, "socket write failed");
	} else {
		CHANSTATS(chan->stats.dropnobufs++);	/* XXX */
		err = ani_error_set(ENOSPC, "socket write incomplete");
	}

	return err;
}

/*
 * Lookup or create a socket corresponding to the desired local address.
 * Returns 0 on success with the socketdata struct locked.
 */
static ani_error_t
chan_socketdata_find(int stype, int ctype, an_udp4addrspec_t *caddr,
		     struct socketdata **out_sdata)
{
	struct socketdata *sdata;
	ani_error_t err;
	int s;
	u_int16_t lport;
	u_int32_t laddr;
	union {
		struct sockaddr gen;
		struct sockaddr_in ipv4;
	} saddr;

	assert(ctype != CHAN_CUT);
	if (ctype == CHAN_IN) {
		lport = caddr->dport;
		laddr = caddr->daddr;
	} else {
		lport = caddr->sport;
		laddr = caddr->saddr;
	}

	/*
	 * Look for an existing socket
	 */
	ani_ilock_lock(&socketlock);
	for (sdata = socketlist; sdata != 0; sdata = sdata->next)
		if (sdata->stype == stype &&
		    sdata->laddr == laddr && sdata->lport == lport) {
			ani_ilock_lock(&sdata->lock);
			ani_ilock_unlock(&socketlock);
			DBG(DBG_SOCKFD, {
				ani_dprintf("Sd%p: found socket "
					    "fd=%d, laddr=%#lx:%d\n",
					    sdata, sdata->fd,
					    ntohl(laddr), ntohs(lport));
			});
			if (ctype == CHAN_IN && ++sdata->inrefs == 1) {
				assert(!(sdata->flags & ANI_SD_THREAD_RUNNING));
				err = ani_rootdomain_startthread(&sdata->thread,
							chan_socketdata_intr,
							sdata, 0, 0);
				if (ani_error_occured(err)) {
					sdata->inrefs--;
					ani_ilock_unlock(&sdata->lock);
					return err;
				}
				sdata->flags |= ANI_SD_THREAD_RUNNING;
			}
			sdata->refs++;
			*out_sdata = sdata;
			return ani_error_clear();
		}
	ani_ilock_unlock(&socketlock);

	/*
	 * Otherwise, create a new one
	 * XXX who should we charge this to?
	 */
	sdata = ani_mempool_alloc(ani_mempool_rootpool(), sizeof *sdata);
	if (sdata == 0)
		return ani_error_set(ENOMEM,
				     "no memory for socket channel data");

	sdata->fd = socket(PF_INET, stype, 0);
	if (sdata->fd < 0) {
		ani_mempool_free(ani_mempool_rootpool(), sdata, sizeof *sdata);
		return ani_error_set(errno, "socket call failed");
	}
	DBG(DBG_SOCKFD, {
		ani_dprintf("Sd%p: created socket fd=%d, laddr=%#lx:%d\n",
			    sdata, sdata->fd, ntohl(laddr), ntohs(lport));
	});

	/*
	 * Make socket buffer sizes as large as possible.
	 */
	s = 32 * 1024;
	if (setsockopt(sdata->fd, SOL_SOCKET, SO_RCVBUF, &s, sizeof s) < 0 ||
	    setsockopt(sdata->fd, SOL_SOCKET, SO_SNDBUF, &s, sizeof s) < 0) {
		err = ani_error_set(errno, "setsockopt call failed");
		DBG(DBG_SOCKFD, {
			ani_dprintf("Sd%p: error, closing fd=%d\n",
				    sdata, sdata->fd);
		});
		close(sdata->fd);
		ani_mempool_free(ani_mempool_rootpool(), sdata, sizeof *sdata);
		return err;
	}

	/*
	 * We initialize and lock the sdata lock so we can use it to
	 * synchronize with any reader thread.  We set the ref count now
	 * so that we can use socketdata_release to cleanup on errors.
	 */
	ani_ilock_init(&sdata->lock);
	ani_ilock_lock(&sdata->lock);
	ani_icond_init(&sdata->cond);

	/*
	 * Convert NodeOS address info to BSD-style sockaddr
	 */
#if defined(__FreeBSD__) || defined(OSKIT)
	saddr.ipv4.sin_len = sizeof saddr.ipv4;
#endif
	saddr.ipv4.sin_family = AF_INET;
	saddr.ipv4.sin_port = lport;
	saddr.ipv4.sin_addr.s_addr = laddr;
	memset(&saddr.ipv4.sin_zero, 0, sizeof(saddr.ipv4.sin_zero));

	/*
	 * Bind to the local address
	 */
	DBG(DBG_SOCKFD, {
		ani_dprintf("Sd%p: bind fd=%d addr=%#lx:%d\n",
			    sdata, sdata->fd, ntohl(laddr), ntohs(lport));
	});
	if (bind(sdata->fd, &saddr.gen, sizeof saddr.gen) < 0) {
		err = ani_error_set(errno, "bind call failed");
		goto error;
	}

	sdata->refs = 1;
	sdata->inrefs = 0;
	sdata->flags = 0;

	/*
	 * Fire up a receiver thread if necessary
	 */
	if (ctype == CHAN_IN) {
		err = ani_rootdomain_startthread(&sdata->thread,
					       chan_socketdata_intr, sdata,
					       0, 0);
		if (ani_error_occured(err))
			goto error;
		sdata->flags |= ANI_SD_THREAD_RUNNING;
		sdata->inrefs++;
	}

	/*
	 * Initialize the rest of the socket data
	 *
	 * Note that we acquire the socketlist lock out of order here.
	 * This is safe since no one else could hold our lock yet and cause
	 * deadlock.
	 */
	sdata->stype = stype;
	sdata->laddr = laddr;
	sdata->lport = lport;
	ani_demuxkeylist_init(&sdata->demux);
	ani_ilock_lock(&socketlock);
	sdata->next = socketlist;
	socketlist = sdata;
	ani_ilock_unlock(&socketlock);

	*out_sdata = sdata;
	return ani_error_clear();

error:
	DBG(DBG_SOCKFD, {
		ani_dprintf("Sd%p: error, closing fd=%d\n",
			    sdata, sdata->fd);
	});
	close(sdata->fd);
	ani_icond_deinit(&sdata->cond);
	ani_ilock_unlock(&sdata->lock);
	ani_ilock_deinit(&sdata->lock);
	ani_mempool_free(ani_mempool_rootpool(), sdata, sizeof *sdata);
	return err;
}

/*
 * Drop a reference to a socketdata struct.
 * Called with the sdata locked.  If no more refs, the sdata is destroyed.
 * Otherwise, the lock is dropped.
 */
static void
chan_socketdata_release(struct socketdata *sdata, int ctype)
{
	struct socketdata **sdp;

	assert(sdata->refs >= 1);
	assert(ctype != CHAN_IN || sdata->inrefs >= 1);

	DBG(DBG_SOCKFD, {
		ani_dprintf("Sd%p: release, refs=%d, inrefs=%d\n",
			    sdata, sdata->refs, sdata->inrefs);
	});

	/*
	 * If no more inchans, wait for the service thread to terminate.
	 */
	if (ctype == CHAN_IN && --sdata->inrefs == 0) {
		assert((sdata->flags & ANI_SD_THREAD_RUNNING) != 0);

		DBG(DBG_SOCKFD, {
			ani_dprintf("Sd%p: release, "
				    "waiting for receive thread to exit\n",
				    sdata);
		});
		sdata->flags |= ANI_SD_THREAD_DYING;
		while ((sdata->flags & ANI_SD_THREAD_RUNNING) != 0)
			ani_icond_wait(&sdata->cond, &sdata->lock);
	}
	if (--sdata->refs == 0) {
		/*
		 * If we cannot grab the list lock, do the funky lock dance.
		 * Drop our lock to get the global lock then reaquire ours.
		 * Then check to make sure someone didn't latch on to our
		 * sdata in the meantime.
		 */
		if (ani_ilock_trylock(&socketlock)) {
			ani_ilock_unlock(&sdata->lock);
			ani_ilock_lock(&socketlock);
			ani_ilock_lock(&sdata->lock);
			if (sdata->refs != 0) {
				ani_ilock_unlock(&sdata->lock);
				ani_ilock_unlock(&socketlock);
				return;
			}
		}

		/*
		 * Locate and remove from the list
		 */
		for (sdp = &socketlist; *sdp != 0; sdp = &(*sdp)->next)
			if (*sdp == sdata)
				break;
		assert(*sdp != 0);
		*sdp = sdata->next;
		ani_ilock_unlock(&socketlock);

		DBG(DBG_SOCKFD, {
			ani_dprintf("Sd%p: release, closing fd=%d\n",
				    sdata, sdata->fd);
		});
		close(sdata->fd);
		ani_icond_deinit(&sdata->cond);
		ani_ilock_unlock(&sdata->lock);
		ani_ilock_deinit(&sdata->lock);
		ani_mempool_free(ani_mempool_rootpool(), sdata, sizeof *sdata);
	} else
		ani_ilock_unlock(&sdata->lock);
}

/*
 * We wake up the receiver thread peroidically to see if the socket needs
 * to be shutdown (i.e., all using channels have been destroyed).
 * As always, there seems to be no way to do this that works on all platforms.
 * For most systems we use select to wait and only do the receive if there
 * is something to read.  For the OSKit, the select implementation is dubious
 * so we set a socket timeout instead.  It would be nice (and the cleanest)
 * to use a socket timeout all around, but Linux and Solaris don't implement
 * it and it doesn't seem to work with the threads package in FreeBSD 3.4.
 */
#if !defined(OSKIT) || defined(OSKIT_UNIX)
#define USE_SELECT
#endif

static struct timeval timeo;
#ifdef USE_SELECT
static fd_set sofdset;
#endif

/*
 * Receiver thread for inChan sockets.
 * Applies the demux filters for the socket to determine a channel to
 * deliver to and fires up a channel thread to deal with it.
 *
 * XXX for TCP support, do listen and accept here.
 */
static void
chan_socketdata_intr(void *arg, void *arg2, void *arg3)
{
	struct socketdata *sdata = arg;
	void *bdata;
	ssize_t bsize, size;
	ani_error_t err;
	int s, msize;
	ani_pbuf_t *rbuf, *pbuf;
	ani_chan_t *chan;
	struct sockaddr_in hdr;
	int hsize;

	/*
	 * Locking forces synchronization with our creator
	 */
	ani_ilock_lock(&sdata->lock);

	/*
	 * Set a receive timeout so we break out of the recvfrom once in
	 * a while and can check for socket death.
	 */
	s = sdata->fd;
#if !defined(USE_SELECT)
	{
		int ierr;

		timeo.tv_sec = 0;
		timeo.tv_usec = 500000;	/* 500ms */
		ierr = setsockopt(s, SOL_SOCKET, SO_RCVTIMEO,
				  &timeo, sizeof timeo);
		assert(ierr == 0);
	}
#endif
	ani_ilock_unlock(&sdata->lock);

	/*
	 * Create an initial read buffer.  We swap this buffer with
	 * empty ones from the socket's channels when we hand off packets.
	 */
	rbuf = NODEOS_CREATE_PBUF(ani_domain_current(), 0);
	assert(rbuf != 0);
	err = ani_pbuf_init(rbuf, ani_domain_current(), 0,
			    NODEOS_CHAN_BUFSIZE, 0, NODEOS_CHAN_BUFSIZE);
	assert(!ani_error_occured(err));
	bdata = ani_pbuf_data(rbuf);
	bsize = ani_pbuf_size(rbuf);

	/*
	 * XXX run the loop at lowest priority.  This is counter to what a
	 * traditional interrupt handler does but its just too easy to
	 * livelock here otherwise.  This choice can lead to starvation
	 * in a system with other busy domains.  This is counter to our
	 * resource isolation tenet and needs to be addressed.  The quick
	 * and easy way would be to make sure that domains cannot share
	 * sockets and to make the service thread be a domain thread.
	 */
	err = ani_thread_setprio(ani_thread_current(), 0);
	assert(!ani_error_occured(err));

	while (1) {
		chan = 0;
		hsize = sizeof(hdr);
		
		DBG(DBG_SOCKFD, {
			ani_dprintf("St%p: intr, blocking on fd=%d\n",
				    ani_thread_current(), s);
		});
#ifdef USE_SELECT
		assert(s < sizeof(sofdset)*8);
		FD_ZERO(&sofdset); 
		FD_SET(s, &sofdset);
		timeo.tv_sec = 0;
		timeo.tv_usec = 500000;	/* 500ms */
		
		/*
		 * Only one fd in the set, so set the nfds parameter
		 * to +1 that fd.
		 */
		size = select(s+1, &sofdset, 0, 0, &timeo);

		assert(size == 0 || size == 1 || size == -1);
		if (size == 1)
			size = recvfrom(s, bdata, bsize, 0,
					(struct sockaddr *)&hdr, &hsize);
		else if (size == 0)
			errno = EAGAIN;
#else
		size = recvfrom(s, bdata, bsize, 0,
				(struct sockaddr *)&hdr, &hsize);
#endif
		ani_ilock_lock(&sdata->lock);

		DBG(DBG_SOCKFD, {
			ani_dprintf("St%p: intr, woke up on fd=%d "
				    "(dead: %s, timeout: %s, error: %s)\n",
				    ani_thread_current(), s,
				    (sdata->flags&ANI_SD_THREAD_DYING)?"yes":"no",
				    (size <= 0 && errno == EAGAIN)?"yes":"no",
				    (size <= 0 && errno != EAGAIN)?"yes":"no");
		});

		/*
		 * The socket is being destroyed, exit now.
		 */
		if (sdata->flags & ANI_SD_THREAD_DYING) {
			ani_ilock_unlock(&sdata->lock);
			errno = EAGAIN;
			break;
		}

		/*
		 * Socket read got an error or EOF.
		 * If error was a periodic timeout, ignore it and keep reading.
		 * Otherwise, exit now.
		 */
		if (size <= 0) {
			ani_ilock_unlock(&sdata->lock);
			if (errno == EAGAIN)
				continue;
			break;
		}

		DBG(DBG_SOCKFD, {
			ani_dprintf("St%p: demuxing %d byte packet "
				    "from %#lx/%d\n",
				    ani_thread_current(), size,
				    ntohl(hdr.sin_addr.s_addr),
				    ntohs(hdr.sin_port));
		});

		/*
		 * Valid packet.  Locate the target channel for this data.
		 * If none, just drop it and continue.
		 */
		msize = (int)size;
		chan = ani_demuxkeylist_apply_with_header(&sdata->demux,
							  &hdr, hsize,
							  bdata, &msize);
		ani_ilock_unlock(&sdata->lock);
		if (chan == 0)
			continue;

		msize -= hsize;
		assert(msize >= 0);

		/*
		 * Grab the channel lock for the remainder of the processing.
		 * Note that we dropped the sdata lock above (to avoid deadlock)
		 * so we have to recheck if the channel is still alive since
		 * someone could have run through the destroy code while our
		 * locks were down.
		 *
		 * XXX in theory we would deadlock here because the thread
		 * doing the destroy sleeps with the channel lock held
		 */
		ani_chan_lock(chan);
		if ((chan->flags & CHAN_DYING) != 0) {
			/* XXX figure out what to do here */
			ani_os_panic("chan_socket: dying and I can't cleanup!");
		}

		/*
		 * Allocate a packet buffer from the channel.
		 * If none are available, just drop the packet.
		 */
		err = ani_chan_pbuf_alloc(chan, NODEOS_CHAN_BUFSIZE, &pbuf);
		if (ani_error_occured(err)) {
			ani_chan_unlock(chan);
			continue;
		}

		/*
		 * Swap the buffer obtained from the channel with our buffer.
		 * The channel gets the full buffer in exchange for one of its
		 * empty ones that we will use for the next read.
		 */
		ani_pbuf_setsize(rbuf, size, 0);
		ani_pbuf_swap(rbuf, pbuf);

		/*
		 * Finally, fire up a handler thread.
		 */
		ani_chan_thread_start(chan, pbuf, msize);

		ani_chan_unlock(chan);

		/*
		 * Reset the buffer pointer/size since we have a new
		 * physical buffer.
		 */
		bdata = ani_pbuf_data(rbuf);
		bsize = ani_pbuf_size(rbuf);
	}

	/*
	 * We have broken out of the recvfrom.  Most likely because the
	 * socket was closed.  Free the read buffer, mark that we are dead,
	 * inform any interested party, and exit.
	 */
	ani_pbuf_deinit(rbuf);
	ani_ilock_lock(&sdata->lock);
	DBG(DBG_SOCKFD, {
		ani_dprintf("St%p: exiting, flags=0x%x\n",
			    ani_thread_current(), sdata->flags);
	});
	sdata->flags &= ~ANI_SD_THREAD_RUNNING;
	if (sdata->flags & ANI_SD_THREAD_DYING)
		ani_icond_broadcast(&sdata->cond);
	else
		/* unexpected exit */
		ani_log_printf(ANI_LOG_WARN,
			       "0x%x/%d: receive thread died: %s\n",
			       (unsigned int)ntohl(sdata->laddr),
			       ntohs(sdata->lport),
			       errno ? strerror(errno) : "EOF");
	ani_thread_unlock_exit(&sdata->lock);
}

/*
 * Prepend channel address info into the demux information.
 * Called with the channel locked.
 *
 * XXX this function changes the user-provided demux,
 * it may also screw it up royally on an error.
 */
static ani_error_t
chan_sock_demux_prepend(ani_chan_t *chan)
{
	an_udp4addrspec_t *uaddr;
	ani_error_t err;
	ani_demuxkeyseg_t *seg;
	struct sockaddr_in key, mask;

	assert(ani_chan_type(chan) == CHAN_IN);
	assert(chan->tdata.in.demux != 0);

	uaddr = &chan->addr.udp4;

	/*
	 * Even if we wildcard source info and don't need to add a segment,
	 * we need to adjust the offset of any existing key since a header
	 * is always included with the packet when the demux_match function
	 * is called.
	 */
	if (uaddr->sport == 0 && uaddr->saddr == 0) {
		(void) ani_demuxkey_adjust(chan->tdata.in.demux, sizeof(key));
		err = ani_error_clear();
		goto done;
	}

	memset(&key, 0, sizeof(key));
	memset(&mask, 0, sizeof(mask));
	if (uaddr->sport != 0) {
		key.sin_port = uaddr->sport;
		mask.sin_port = ~0;
	}
	if (uaddr->saddr != 0) {
		key.sin_addr.s_addr = uaddr->saddr;
		mask.sin_addr.s_addr = ~0;
	}

	seg = ani_demuxkeyseg_tuple(0, sizeof(key), &key, &mask);
	if (seg == 0)
		err = ani_error_set(ENOMEM,
				    "no memory for UDP header demuxkey");
	else {
		err = ani_demuxkey_prepend(chan->tdata.in.demux, seg);
		if (ani_error_occured(err))
			ani_demuxkeyseg_destroy(seg);
	}
 done:
	DBG(DBG_SOCKFD, {
		ani_dprintf("C%p: header demux:\n", chan);
		ani_demuxkey_dump(chan->tdata.in.demux);
	});

	return err;
}


#endif /* CHAN_SOCKET */
