/*
 * Copyright (c) 2000-2002 The University of Utah. 
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
/*
 * ani_debug.c
 */

#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <nodeos/ani/ani_mempool.h>
#include <nodeos/ani/ani_domain.h>
#include <nodeos/ani/ani_log.h>
#include <nodeos/ani/ani_error.h>

#include "ani_debug.h"
#include "ani_os.h"

#if !defined(NDEBUG) && defined(DEBUG)
/* --- Debugging is enabled --- */ 

/* The debugging mask all DBG() statements check. */
ani_debugmask_t ani_debug_mask = 0;

/*
 * Global variables related to the debugging buffer
 *
 * Note: these are not 'static' so that they're easy to find
 * in a debugger.
 */
char *ani_debug_buffer = NULL;
int ani_debug_bufferbegin = 0;
int ani_debug_buffersize = 0;
int ani_debug_bufferwrapped = 0;

/*
 * For each debugging option in ani_debug.h, provide a
 * string name, a mask (may be more than one option |'d),
 * and a description string.
 */
static struct {
	const char *name;
	ani_debugmask_t mask;
	const char *desc;
} ani_debug_opttable[] = {
#define D(name, str) { #name, DBG_ ## name, str }
	D(NONE,		"Nothing"),
	D(SOCKFD,	"Show socket-channel related file descriptor operations."),
	D(DEMUX_MATCH_DUMP,"Show all demuxkey matching of packets."),
	D(IP_MATCH_DUMP,"Show IPv4 demuxkey matching of packets."),
	D(DOMAINS,	"Show domain-related activity."),
	D(CRED,		"Show credential-related activity."),
	D(THREADS,	"Show thread-related operations."),
	D(THREADPOOLS,	"Show threadpool-related operations."),
	D(THREADPOOL_STATS,	"Show threadpool-related statistics."),
	D(DEVICE_INIT,	"Show device initialization debugging."),
	D(OSKIT_LOG_NONE, "Suppress all osenv_log messages."),
	D(OSKIT_LOG_NOTICE, "Suppress osenv_log messages w/pri < LOG_NOTICE."),
	D(OSKIT_LOG_ERR, "Suppress osenv_log messages w/pri < LOG_ERR."),
	
	/* Options handled special-like by ani_debug_setmaskstr() */
	{ "buffer", 0,	"Write all ani_dprintf() output into an internal buffer." },
	{ "dump", 0,	"Dump said internal buffer when moab exits." },
		
	/* These must be last, as they are markers for some GDB macros. */
	D(ALL,		"Everything under the sun...."),
	D(ANY,		"Ibid.")
};
	
#define ANI_DEBUG_DEFAULT_BUFFERSIZE (16 * 1024)
static void ani_debug_bufferinit(int size);
static void ani_debug_bufferprint(void);

#define NELEMS(a)	(sizeof (a) / sizeof(a[0]))

static int ani_debug_initialized = 0;

/*
 * initialize ani_debug system
 */
void
ani_debug_init(void)
{
	if (ani_debug_initialized)
		return;
	ani_debug_initialized = 1;
}

/*
 * Set the debug mask to the given mask
 */
void
ani_debug_setmask(ani_debugmask_t mask)
{
	ani_debug_init();  /* in case its done before moab initialization. */
	assert(ani_debug_initialized);

	ani_debug_mask = mask;
}


/*
 * Set the debug mask as defined by the entries in the given string.
 * Each entry should be a name in the ani_debug_opttable table.  Entries may
 * be separated by "|" or ",".  The special entry 'list' causes all
 * of the debugging entries to be listed.
 *
 * This function may NOT use any of Moab's run-time support.  It is usually
 * invoked before an_moab_setup().
 * 
 * Returns NULL if things went okay, otherwise returns a pointer to
 * the maskstr element that caused a problem.
 *
 * This function may be invoked before an_moab_setup() (and thus
 * before any of the OSKit initialization).  So it should avoid
 * using dynamic memory, threads, or any shutdown functions.
 */
const char *
ani_debug_setmaskstr(const char *mask_str)
{
	int i;
	int skipped;
	const char *separators = "|,";
	const char *curr_maskstr;
	
	ani_debug_init();  /* in case its done before moab initialization. */
	assert(ani_debug_initialized);
	
	if (mask_str == NULL)
		return NULL;

	/*
	 * Note that we don't use "strtok" because it modifies its argument
	 * and we *need* to be able to handle constant strings.  We also
	 * want to avoid allocating any memory this early in the game.
	 */

	curr_maskstr = mask_str;

	/* count leading separators */
	skipped = strspn(curr_maskstr, separators);
	if (skipped >= strlen(curr_maskstr))
	{
		/* String was nothing but separators, this is an error. */
		ani_debug_mask = 0;
		return "No valid options passed to ani_debug_setmaskstr";
	}

	/* skip leading separators */
	curr_maskstr+=skipped;
	
	/* Special target 'list' lists all the defined options, and returns 0 */
	if (!strcasecmp(curr_maskstr, "list")) {
		ani_dprintf("Available debug opts: \n");
		ani_dprintf("  %-15s\t%16s  %s\n",
			    "Option", "Mask (hex)", "Description");
		for (i = 0; i < NELEMS(ani_debug_opttable); i++)
		{
			const char* name;
			const char* desc;
			int hi, lo;

			name = ani_debug_opttable[i].name;
			desc = ani_debug_opttable[i].desc;
			hi = (int)(ani_debug_opttable[i].mask>>32);
			lo = (int)(ani_debug_opttable[i].mask);

			if (hi)
				ani_dprintf("  %-15s\t%8X%08X  %s\n",
					    name, hi, lo, desc);
			else
				ani_dprintf("  %-15s\t        %8X  %s\n", 
					    name, lo, desc);
		}
		return "'list' option passed to ani_debug_setmaskstr";
	}
	

	/*OPTSTRINGLOOP:*/
	while (curr_maskstr) {
		size_t optlen;
		/* Find out how big this option is */
		optlen = strcspn(curr_maskstr, separators);
		assert(optlen > 0);
		assert(optlen <= strlen(curr_maskstr));

		/* Handle special-case opts */
		if (!strncasecmp(curr_maskstr, "buffer", optlen))
			ani_debug_bufferinit(ANI_DEBUG_DEFAULT_BUFFERSIZE);
		else if (!strncasecmp(curr_maskstr, "dump", optlen))
			atexit(ani_debug_bufferprint);
		/* Handle run-of-the-mill opts */
		else
		{
			int invertedOpt = 0;

			/* Check for inverted opts */
			if (curr_maskstr[0] == '~') 
			{
				invertedOpt = 1;
				curr_maskstr++;

				/* ignore the degenerate case */
				if (curr_maskstr[0] == '\0')
					continue /*OPTSTRINGLOOP*/;
			}
			
			/*TABLECHECK:*/ 
			for (i = 0; i < NELEMS(ani_debug_opttable); i++)
				if (!strncasecmp(curr_maskstr, ani_debug_opttable[i].name, optlen))
				{
					if (invertedOpt)
						ani_debug_mask &= ~(ani_debug_opttable[i].mask);
					else 
						ani_debug_mask |= ani_debug_opttable[i].mask;
					break /*TABLECHECK*/;
				}
		
			/* Return bogon element if it wasn't found in table. */
			if (i == NELEMS(ani_debug_opttable)) 
			{
				return curr_maskstr;
			}
		}
		
		/* Skip the option just parsed. */
		curr_maskstr += optlen;

		/* End of options? */
		if (strlen(curr_maskstr) == 0)
			curr_maskstr = NULL;
		else
		{
			int sepct;
			/* Count the separators following the last option. */
			sepct = strspn(curr_maskstr, separators);
			assert(sepct > 0);
			assert(sepct <= strlen(curr_maskstr));

			/* Skip the separators */
			curr_maskstr+=sepct;

			/* Is that it? */
			if (strlen(curr_maskstr) == 0)
				curr_maskstr = NULL;
		}

		/*
		 * At this point curr_maskstr either points to the
		 * the next option name, or NULL (in which case the
		 * loop will end).
		 */
	}

	return NULL;
}

/*
 * Create a buffer into which debugging output will be written.
 * See the 'buffer' option.
 */
static void
ani_debug_bufferinit(int size)
{
	ani_debug_buffersize = size;
	ani_debug_buffer = ani_mempool_alloc(ani_mempool_current(), size);
	assert(ani_debug_buffer != 0);
}


/*
 * The following function is atexit'd if the 'dump' option was given.
 */
void
ani_debug_bufferprint(void)
{
	int soff, eoff, resid;

	if (ani_debug_bufferwrapped == 0) {
		soff = 0;
		eoff = ani_debug_bufferbegin;
		resid = 0;
	} else {
		soff = ani_debug_bufferbegin;
		eoff = ani_debug_buffersize;
		resid = ani_debug_bufferbegin;
	}
	ani_os_log_write(ANI_LOG_DEBUG, ani_debug_buffer + soff, eoff - soff);
	if (resid != 0) {
		ani_os_log_write(ANI_LOG_DEBUG, ani_debug_buffer, resid);
	}
}

void
ani_dprintf(const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	ani_log_vprintf(ANI_LOG_DEBUG, fmt, args);
	va_end(args);
}

void
ani_dvprintf(const char *fmt, void *args)
{
	ani_log_vprintf(ANI_LOG_DEBUG, fmt, args);
}

/*
 * XXX hack.  This is called by ani_log_write in place of ani_os_log_write
 * so that we can capture ANI_LOG_DEBUG output.  This function in turn calls
 * ani_os_log_write if we are not buffering such output.
 */
void
ani_debug_write(void *buf, int buflen)
{
	if (ani_debug_buffersize != 0) {
		int n;

		assert(ani_debug_bufferbegin >= 0);
		assert(ani_debug_bufferbegin < ani_debug_buffersize);

		/*
		 * Only save the last buffersize bytes
		 */
		if (buflen > ani_debug_buffersize) {
			buf += (buflen - ani_debug_buffersize);
			buflen = ani_debug_buffersize;
		}
				
		/*
		 * Write from current offset to end of data or debug buffer
		 */
		n = ani_debug_buffersize - ani_debug_bufferbegin;
		if (n > buflen)
			n = buflen;
		memcpy(ani_debug_buffer + ani_debug_bufferbegin, buf, n);
		buf += n;
		buflen -= n;

		/*
		 * If more to write, write from beginning of buffer to
		 * former current offset.
		 */
		if (buflen > 0) {
			memcpy(ani_debug_buffer, buf, buflen);
			ani_debug_bufferbegin = buflen;
		} else {
			ani_debug_bufferbegin += n;
		}

		assert(ani_debug_bufferbegin <= ani_debug_buffersize);
		if (ani_debug_bufferbegin == ani_debug_buffersize) {
			ani_debug_bufferbegin = 0;
			ani_debug_bufferwrapped = 1;
		}
	} else
		ani_os_log_write(ANI_LOG_DEBUG, buf, buflen);
}
#endif /* Debugging is enabled */

