/*
 * Copyright (c) 1999-2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <stdlib.h>
#include <assert.h>

#include <nodeos/an_resource.h>
#include <nodeos/an_thread.h>

#include <nodeos/impl/an_error.h>
#include <nodeos/ani/ani_thread.h>
#include <nodeos/ani/ani_threadpool.h>
#include <nodeos/ani/ani_nodeos.h>
#include <nodeos/ani/ani_cred.h>
#include <nodeos/ani/ani_resource.h>
#include <nodeos/ani/ani_domain.h>
#include <nodeos/ani/ani_mempool.h>
#include <nodeos/ani/ani_chan.h>

#include "ani_os.h"		/* XXX for ani_os_panic */
#include "ani_debug.h"

static int ani_domain_disable_internal(struct hnode *h, void *arg);
static int ani_domain_destroy_internal(struct hnode *h, void *arg);

/* flags to destroy_internal */
#define DI_DISABLE	1
#define DI_WAITFOR	2

extern int bootstrapped;
static ani_domain_t *root_domain;

/*
 * Boot time startup
 */
void
ani_domain_setup(ani_domain_t *rdomain)
{
	/*
	 * rdomain has not been initialized at this point, its still
	 * just an arbitrary hunk of memory.
	 */
	root_domain = rdomain;
}

/*
 * Initialize the root domain, firing off the first true NodeOS thread.
 */
ani_error_t
ani_rootdomain_init(ani_domain_t *rdomain, ani_cred_t *rcred,
		  ani_mempool_t *rmempool, int nthreads, int stacksize,
		  void **root_threads, void **root_threadstacks,
		  an_domaininit_func_t initfunc, void *initarg)
{
	an_cpuspec_t cspec;
	an_memspec_t mspec;
	ani_error_t err;

	/* Provide default stacksize if one isn't given. */
	if (stacksize <= 0) {
		assert(nthreads == 0); /* XXX */
		assert(root_threads == NULL); /* XXX */
		assert(root_threadstacks == NULL); /* XXX */
		stacksize = 64 * 1024;
	}
	
	/* Provide default number of implicit threads if not given. */
	if (nthreads <= 0) {
		assert(root_threads == NULL); /* XXX */
		assert(root_threadstacks == NULL); /* XXX */
		nthreads = 1;
	}

	if (root_threads != NULL) {
		int i;

		assert(root_threadstacks != NULL); /* XXX */

		/* Convert unitialized thread object memory into ani_thread_t */
		for (i = 0; i < nthreads; i++) {
			ani_thread_t *ith;

			assert(root_threads[i] != NULL); /* XXX */
			assert(root_threadstacks[i] != NULL); /* XXX */
			ith = NODEOS_INTERN_THREAD(rdomain, root_threads[i]);
			assert(ith != NULL);
			err = ani_thread_init(ith,
					      root_threadstacks[i], stacksize);
			if (ani_error_occured(err))
				return err;
		}
		
		cspec.ttype = AN_TH_EXPLICIT;
		cspec.threads.ex.nthreads = nthreads;
		cspec.threads.ex.threads = (struct an_thread **)root_threads;
	} else {
		cspec.ttype = AN_TH_IMPLICIT;
		cspec.threads.im.maxthreads = nthreads;
		cspec.threads.im.stacksize = stacksize;
	}

	/*
	 * Root domain threads are high priority and get whatever they need.
	 * The assumption being that they are short-running interrupt handlers.
	 */
	{
		extern ani_resource_t ani_moab_noderes;

		assert(nthreads <= 100);

		/* XXX bigass priority for our threads */
		cspec.rate = nthreads * ani_moab_noderes.cpurate;

		/* XXX ignored */
		cspec.slice.tv_sec = 0;
		cspec.slice.tv_nsec = 0;
		cspec.sched = 0;
	}

	/* XXX stack + 128k for every thread */
	mspec.nchunks = nthreads *
		(stacksize + 128*1024) / NODEOS_MEMPOOL_CHUNKSIZE;
	mspec.pool = NODEOS_EXTERN_MEMPOOL(rdomain, rmempool);

	err = ani_domain_init(rdomain, rcred, &cspec, &mspec,
			      initfunc, initarg, 0);
	return err;
}

/*
 * Deinit the root domain.
 * All of its children should already have been cleaned up via deinit_children.
 * We just need to clean up the root threadpool.
 * Called outside the context of any NodeOS domain.
 */
void
ani_rootdomain_deinit(ani_domain_t *domain)
{
	assert(domain == root_domain);

	ani_domain_lock(root_domain);
	assert(ani_domain_child(root_domain) == 0);
	assert(root_domain->chanlist == 0);

	root_domain->flags |= DOMAIN_DYING;

	/* Kill any remaining threads */
	ani_threadpool_disable(&root_domain->tpool, 1);
	ani_threadpool_deinit(&root_domain->tpool, 1);

	/* XXX need to ani_deinit_thread EXPLICIT threads here */

	ani_domain_unlock(root_domain);

	/*
	 * Release the reference we took above now that we have dropped the
	 * lock.  The NodeOS world is over at this point.
	 */
	ani_domain_release(root_domain);
}

/*
 * Deinit all the root domain children, but not the root domain itself
 */
void
ani_rootdomain_deinit_children(ani_domain_t *domain)
{
	ani_domain_t *cdomain;
	int err;

	assert(domain);
	assert(domain == root_domain);

	ani_domain_lock(root_domain);
	for (cdomain = ani_domain_child(domain); cdomain != 0;
	     cdomain = ani_domain_sibling(cdomain)) {
		err = hnode_traverse(&cdomain->me, ani_domain_destroy_internal,
				     (void *)(DI_DISABLE|DI_WAITFOR),
				     HNODE_PARENT, 1);
		assert(err == 0); /* XXX */
	}
	ani_domain_unlock(root_domain);
}

ani_error_t
ani_rootdomain_startthread(ani_thread_t *thread, ani_thread_func_t func,
			 void *farg1, void *farg2, void *farg3)
{
	ani_error_t err;

	assert(root_domain != 0);
	assert(thread != 0);

	err = ani_thread_init(thread, 0, 0);
	if (!ani_error_occured(err)) {
		err = ani_threadpool_osthread_start(&root_domain->tpool,
						    thread, 1, func,
						    farg1, farg2, farg3);
		if (ani_error_occured(err))
			ani_thread_deinit(thread);
	}
	DBG(DBG_THREADPOOLS, {
		if (ani_error_occured(err))
			ani_dprintf("TP%p: rootdomain: "
				    "startthread failed, err=%x\n",
				    ani_domain_threadpool(root_domain),
				    ani_error_code(err));
		else
			ani_dprintf("TP%p: rootdomain: "
				    "oneshot T%p started for %p(%p)\n",
				    ani_domain_threadpool(root_domain), thread,
				    func, farg1);
	});

	return err;
}



/* ani_domain_current is in ani_domain_inline.h */

ani_error_t
ani_domain_init(ani_domain_t *domain, ani_cred_t *cred,
	      an_cpuspec_t *cspec, an_memspec_t *mspec,
	      an_domaininit_func_t initfunc, void *initarg,
	      an_domainterm_func_t termfunc)
{
	ani_domain_t *pdomain;
	ani_mempool_t *mpool;
	ani_resource_t res;
	ani_error_t err;

	/* Sanity */
	if (domain == 0 || cred == 0 || cspec == 0 || mspec == 0)
		return ani_error_set(EINVAL,
				     "invalid argument to ani_domain_init");

	/* Can this domain create sub domains? */
	if (ani_cred_check(cred, CCHECK_DOMAIN_CREATE) == CCHECK_FAIL)
		return ani_error_set(EPERM,
				     "insufficient permission to create domain");

	/* Ensure requested resources are available and reserve them */
	memset(&res, 0, sizeof res);
	ani_resource_initcpu(&res, cspec);
	ani_resource_initmem(&res, mspec);
	err = ani_cred_resource_alloc(cred, &res);
	if (err)
		return err;

	pdomain = ani_domain_current();

	/*
	 * Extract the target mempool so we can start charging it.
	 *
	 * XXX this doesn't really belong here.  We may eventually
	 * just pass in a mempool here, with the current memspec
	 * params being associated with the pool when it is created.
	 */
	mpool = NODEOS_INTERN_MEMPOOL(pdomain, mspec->pool);
	if (mpool == 0) {
		ani_cred_resource_free(cred, &res);
		return ani_error_set(EINVAL,
				     "invalid mempool");
	}

	/* Initialize the domain */
	hnode_init(&domain->me);
	domain->cred = cred;
	domain->res = res;
	domain->chanlist = 0;
	domain->flags = 0;
	domain->termfunc = termfunc;
	domain->refs = 1;
	ani_ilock_init(&domain->lock);

	/* Link to the mempool now for threadpool memory allocations */
	domain->mpool = mpool;
	ani_mempool_adddomain(domain->mpool, domain);

	/* Create the threadpool */
	err = ani_threadpool_init(&domain->tpool, domain, cspec, &res);
	if (err) {
		ani_mempool_removedomain(mpool, domain);
		ani_ilock_deinit(&domain->lock);
		ani_cred_resource_free(cred, &res);
		return err;
	}

	/* Link into domain hierarchies (special handling for first domain) */
	if (domain == root_domain) {
		ani_domain_lock(domain);
		hnode_attach(&domain->me, 0, 0);
	} else {
		ani_domain_lock(root_domain);
		ani_domain_lock(domain);
		hnode_attach(&domain->me, &ani_domain_current()->me, 0);
		ani_domain_unlock(root_domain);
	}

	ani_domain_unlock(domain);

	/* All setup, fire up thread in the domain to call the init function */
	if (initfunc) {
		err = ani_domain_startthread(domain, 0,
					   (void (*)(void *, void *))initfunc,
					   initarg, 0);
		if (err) {
			/*
			 * Cannot deinit the root domain if something
                         *  goes wrong at startup.
			 */
			if (domain != root_domain)
				ani_domain_deinit(domain);

			return err;
		}
	}

	return ani_error_clear();
}

ani_error_t
ani_domain_deinit(ani_domain_t *domain)
{
	ani_domain_t *odomain;
	int me = 0, ierr;
	ani_error_t err;

	assert_ani_domain_valid(domain);

	if (domain == root_domain)
		return ani_error_set(EINVAL, "cannot deinit root domain");

	ani_domain_lock(domain);

	/*
	 * Domain must be us or one of our immediate children
	 */
	if (domain == ani_domain_current())
		me = 1;
	else {
#if 0
		for (odomain = ani_domain_child(ani_domain_current());
		     odomain != 0;
		     odomain = ani_domain_sibling(odomain))
			if (odomain == domain)
				break;
		if (odomain == 0) {
			ani_domain_unlock(domain);
			return ani_error_set(EINVAL,
					     "domain argument is not self or child domain");
		}
#endif
	}

	/*
	 * Prevent recursive upcalls and, effectively, attempts to destroy
	 * an already dying domain.  Note that the domain lock is not held while
	 * in the upcall so we need this other mechanism to ensure this.
	 */
	if (domain->flags & DOMAIN_INUPCALL) {
		ani_domain_unlock(domain);
		return ani_error_clear();
	}
	domain->flags |= DOMAIN_INUPCALL;
	ani_domain_unlock(domain);

	/*
	 * First disable all the threadpools so that no new domain activity
	 * will occur while we are shutting down.
	 */
	ani_domain_lock(root_domain);
	ierr = hnode_traverse(&domain->me, ani_domain_disable_internal, 0,
			      HNODE_PARENT, 1);
	assert(ierr == 0);
	ani_domain_unlock(root_domain);

	/*
	 * Now make the termination upcall.
	 */
	if (domain->termfunc != 0) {
		ani_threadwork_t work;

		odomain = ani_domain_parent(domain);
		assert(odomain != 0);
		DBG(DBG_DOMAINS, {
			ani_dprintf("F%p: making parent upcall to %p:%p\n",
				    domain, odomain, domain->termfunc);
		});
		ani_threadwork_init(&work);
		ani_threadwork_set(&work, 0, 0, 
				   (ani_thread_func_t)domain->termfunc,
				   domain, 0, 0);
		err = ani_threadpool_thread_run(&odomain->tpool, &work, 0);
		DBG(DBG_DOMAINS, {
			ani_dprintf("F%p: parent upcall returned, err=0x%x\n",
				    domain, err);
		});

		/*
		 * XXX no point in returning an error now, the domain
		 * is too far gone to recover.
		 */
		if (ani_error_occured(err))
			err = ani_error_clear();
	} else
		err = ani_error_clear();

	if ((domain->flags & DOMAIN_DYING) == 0) {
		ani_domain_lock(root_domain);
		ierr = hnode_traverse(&domain->me, ani_domain_destroy_internal,
				      (void *)(me ? 0 : DI_WAITFOR),
				      HNODE_PARENT, 1);
		assert(ierr == 0);
		ani_domain_unlock(root_domain);
	}

	/*
	 * If we just destroyed ourself, there is no point in returning.
	 * Our thread's reference is keeping the domain alive at this point,
	 * exiting will take care of that!
	 */
	if (me)
		ani_thread_exit();

	return err;
}

/*
 * Disable all future domain activity by removing all threads from the
 * domain's threadpool.
 */
static int
ani_domain_disable_internal(struct hnode *h, void *arg)
{
	ani_domain_t *domain = (ani_domain_t *)h;

	assert_ani_domain_valid(domain);

	ani_domain_lock(domain);
	ani_threadpool_disable(&domain->tpool, 1);
	ani_domain_unlock(domain);	

	return 0;
}

/*
 * The guts of domain destruction.  Called by the hnode_traverse in
 * ani_domain_destroy.
 */
static int
ani_domain_destroy_internal(struct hnode *h, void *arg)
{
	ani_domain_t *domain = (ani_domain_t *)h;
	int flags = (int)arg;
	ani_chan_t *chan, *next;

	assert_ani_domain_valid(domain);
	assert(domain != root_domain);

	ani_domain_lock(domain);
	domain->flags |= DOMAIN_DYING;

	/* Remove from domain hierarchy */
	hnode_detach(h, HNODE_PARENT);

	/* Kill any remaining threads */
	if (flags & DI_DISABLE)
		ani_threadpool_disable(&domain->tpool, 1);
	ani_threadpool_deinit(&domain->tpool, (flags & DI_WAITFOR));

	/* Destroy all the channels */
	for (chan = domain->chanlist; chan != 0; chan = next) {
		next = chan->next;
		ani_chan_deinit(chan);
	}

	/* Release the initial reference to the domain */
	ani_domain_unlock(domain);
	ani_domain_release(domain);

	return 0;
}

/*
 * Called when there are no more references to the domain.
 * At this point we can safely free resources.
 */
void
ani_domain_cleanup(ani_domain_t *domain)
{
	assert_ani_domain_valid(domain);

	ani_domain_lock(domain);

	/* Remove from mempool */
	assert(hnode_listhead(&domain->me) != 0);
	ani_mempool_removedomain(domain->mpool, domain);

	/* Return resources to the system */
	ani_cred_resource_free(domain->cred, &domain->res);

	/* Uninitialize and free all the data structures */
	ani_domain_unlock(domain);
	ani_ilock_deinit(&domain->lock);
}

void
ani_domain_addchannel(ani_domain_t *domain, ani_chan_t *chan)
{
	assert_ani_domain_valid(domain);

	ani_domain_lock(domain);
	ani_chan_lock(chan);

	/* XXX make sure requested channel resources are consistent with domain */

	chan->domain = domain;
	chan->next = domain->chanlist;
	domain->chanlist = chan;

	ani_chan_unlock(chan);
	ani_domain_unlock(domain);
}

void
ani_domain_removechannel(ani_domain_t *domain, ani_chan_t *chan)
{
	ani_chan_t *ch, **pch;

	assert_ani_domain_valid(domain);
	assert(chan->domain == domain);

	/*
	 * Domains never return from the dead so we don't lock this access
	 */
	if (domain->flags & DOMAIN_DYING)
		return;

	ani_domain_lock(domain);
	ani_chan_lock(chan);

	for (pch = &domain->chanlist; (ch = *pch) != 0; pch = &ch->next)
		if (ch == chan) {
			*pch = ch->next;
			break;
		}
	chan->domain = 0;

	ani_chan_unlock(chan);
	ani_domain_unlock(domain);
}

ani_error_t
ani_domain_startthread(ani_domain_t *domain, int inkern,
		     void (*func)(void *, void *), void *farg1, void *farg2)
{
	ani_threadwork_t work;

	ani_threadwork_init(&work);
	ani_threadwork_set(&work, inkern, 0,
			   (ani_thread_func_t)func, farg1, farg2, 0);
	return ani_threadpool_thread_start(&domain->tpool, &work, 0);
}

an_domainstats_t *
ani_domain_stats(ani_domain_t *domain)
{
	ani_threadstats_t tstats;

	assert_ani_domain_valid(domain);
	ani_domain_lock(domain);

	ani_threadpool_stats(&domain->tpool, &tstats);
	domain->stats.curthreads = tstats.active;
	domain->stats.maxthreads = tstats.maxactive;
	domain->stats.activations = tstats.starts;
	assert(tstats.pastcpu.units == 0 ||
	       tstats.pastcpu.units == tstats.curcpu.units);
	switch (tstats.curcpu.units) {
	case ANI_CPU_USECS:
		domain->stats.cpuunits = NODEOS_CPU_USECS;
		break;
	case ANI_CPU_CYCLES:
		domain->stats.cpuunits = NODEOS_CPU_CYCLES;
		break;
	default:
		ani_os_panic("bad threadpool stats units");
		break;
	}
	domain->stats.cpuused = tstats.pastcpu.amount + tstats.curcpu.amount;

	/* XXX NodeOS charges memory to mempools */
	domain->stats.memused = 0;

	ani_domain_unlock(domain);
	return &domain->stats;
}

#ifdef DEBUG
#include <stdio.h>
#include <nodeos/ani/ani_log.h>

static int
dump_domain(struct hnode *node, void *arg)
{
	ani_domain_t *domain = (ani_domain_t *)node;

	ani_log_printf(ANI_LOG_DEBUG,
		       "DOMAIN %p: parent=%p, child=%p, sibling=%p, "
		       "cred=%p, mpool=%p, tpool=%p\n",
		       domain, hnode_parent(node), hnode_child(node),
		       hnode_sibling(node),
		       domain->cred, domain->mpool, &domain->tpool);
	
	return 0;
}

void
ani_domain_dump(ani_domain_t *domain)
{
	if (domain == 0)
		domain = ani_domain_current();

	hnode_traverse(&domain->me, dump_domain, 0, HNODE_PARENT, 0);
}
#endif

/*
 * Include the inlined functions in non-inlined form
 * XXX must be last in the file
 */
#undef NODEOS_FUNC_INLINE
#define NODEOS_FUNC_INLINE

#include <nodeos/ani/ani_domain_inline.h>
