/*
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <assert.h>
#include <string.h> /* XXX memset */

#include <nodeos/an_event.h>
#include <nodeos/impl/an_impl.h>
#include <nodeos/ani/ani_nodeos.h>
#include <nodeos/ani/ani_ilock.h>
#include <nodeos/ani/ani_event.h>
#include <nodeos/ani/ani_threadpool.h>
#include <nodeos/ani/ani_threadwork.h>
#include <nodeos/ani/ani_domain.h>
#include <nodeos/ani/ani_time.h>

#include "ani_os.h"
#include "ani_debug.h"

static void *eventtimer;
static ani_event_t *eventlist;
static ani_ilock_t eventlock;

/* XXX move some of this time stuff into ani_time? */
/* XXX a lot of this time/ticks stuff is duplicated in chan_clock.c */

static unsigned long nspertick = 0;
static unsigned long tickspersec = 0;
static unsigned long myticks = 0;
static an_timespec_t starttime; /* abstime for myticks == 0 */
#define ABILLION (1000 * 1000 * 1000)

static void ani_event_intr(void *oskitthingee, void *arg);

static inline void
ani_ticks_to_reltime(ani_tick_t ticks, an_timespec_t *timespec)
{
	timespec->tv_sec = (ticks / tickspersec);
	timespec->tv_nsec = (ticks % tickspersec) * nspertick;
}

static inline ani_tick_t
ani_reltime_to_ticks(an_timespec_t *timespec)
{
	ani_tick_t ticks;

	ticks = timespec->tv_sec * tickspersec;
	ticks += timespec->tv_nsec / nspertick;

	return ticks;
}

static inline void
ani_ticks_to_abstime(ani_tick_t ticks, an_timespec_t *timespec)
{
	an_timespec_t reltime;

	/* Convert ticks into a relative time. */
	ani_ticks_to_reltime(ticks, &reltime);
	
	timespec->tv_sec = starttime.tv_sec + reltime.tv_sec;
	timespec->tv_nsec = starttime.tv_nsec + reltime.tv_nsec;
	if (reltime.tv_nsec >= ABILLION) {
		reltime.tv_nsec -= ABILLION;
		reltime.tv_sec++;
	}
}


static inline ani_tick_t
ani_abstime_to_ticks(an_timespec_t *timespec)
{
	an_timespec_t reltime;

	reltime.tv_sec = timespec->tv_sec - starttime.tv_sec;
	reltime.tv_nsec = timespec->tv_nsec - starttime.tv_nsec;
	if (reltime.tv_nsec < 0) {
		reltime.tv_sec--;
		reltime.tv_nsec += ABILLION;
	}

	return ani_reltime_to_ticks(&reltime);
}


void
ani_event_setup(void)
{
	static int called;

	if (called++)
		return;

	ani_ilock_init(&eventlock);

	eventtimer = ani_os_timer_init(ani_event_intr, 0, &nspertick);
	/* XXX ani_os_timer_init() actually returns uspertick, so upgrade: */
	nspertick *= 1000;
	tickspersec = ABILLION / nspertick;
	assert(eventtimer != 0);
	assert(nspertick > 0);
	assert(nspertick < ABILLION);

	/* XXX ani_os_timer is in us not ns: */
	myticks = 0;
	ani_os_timer_start(eventtimer, nspertick/1000);
	ani_time_current(&starttime);
}

void
ani_event_shutdown(void)
{
	assert(!ani_ilock_islocked(&eventlock));

	ani_os_timer_deinit(eventtimer);
	ani_ilock_deinit(&eventlock);
}

void
ani_event_info(unsigned long *timerres)
{
	assert(timerres);
	*timerres = nspertick;
}

static inline void
ani_event_add_locked(ani_event_t *event)
{
	ani_event_t *ev, **pevp;

	assert(event);
	assert(ani_event_state(event) == AN_EVSTATE_SCHEDULED);
	
	for (pevp = &eventlist; (ev = *pevp) != 0; pevp = &ev->next)
		if (event->ticks < ev->ticks)
			break;
	event->next = ev;
	*pevp = event;
}

static inline ani_error_t
ani_event_remove_locked(ani_event_t *event)
{
	ani_event_t *ev, **pevp;

	assert(ani_ilock_islocked(&eventlock));

	for (pevp = &eventlist; (ev = *pevp) != 0; pevp = &ev->next)
		if (event == ev) {
			*pevp = ev->next;
			return ani_error_clear();
		}

	return ani_error_set(ENOENT, "no such event");
}

ani_error_t
ani_event_init(ani_event_t *event, ani_threadpool_t *tpool)
{
	assert(event);
	assert(tpool);

	memset(event, 0, sizeof *event);

	/* tpool and threadwork are setup with valid values. */
	ani_threadwork_init(&event->work);
	event->tpool = tpool;
	event->flags = ANI_EV_USER|AN_EVTYPE_UNDEFINED|AN_EVSTATE_DESCHEDULED;
	/* leave all the other fields as 0/NULL */

	return ani_error_clear();
}

ani_error_t
ani_event_deinit(ani_event_t *event, int *oflags)
{
	ani_error_t err;

	assert(event);

	err = ani_event_deschedule(event, oflags);

	/* ani_threadwork_deinit(&event->work); */
	event->tpool = 0;

	return err;
}

ani_error_t 
ani_event_schedule(ani_event_t *event, an_timespec_t *itime,
		   an_eventtype_t type, int iflags,
		   an_eventfunc_t func, void *farg)
{
	assert(event);
	assert(itime);
	assert_valid_eventtype(type);
	assert(func);
	
	if (ani_event_state(event) != AN_EVSTATE_DESCHEDULED)
		return ani_error_set(AN_ERR_INVALIDARG,
				     "event is already scheduled");
	
	assert_ani_threadpool_valid(event->tpool);
	/* assert_ani_threadwork_valid(&event->work); */
	if (type == AN_EVTYPE_ONESHOT_ABSOLUTE) {
		/*
		 * It is okay if ticks < myticks, just means
		 * the event will fire on the next tick.
		 */
		event->freq = 0;
		event->ticks = ani_abstime_to_ticks(itime);
	} else {
		unsigned long ticks;
		ticks = ani_reltime_to_ticks(itime);

		/* AN_EVTYPE_ONESHOT_RELATIVE can have a time of 0. */
		if (ticks == 0 && type == AN_EVTYPE_PERIODIC) {
			/*
			 * Be sure returning here doesn't leave a half-changed
			 * event lying around.
			 */
			return ani_error_set(AN_ERR_INVALIDARG,
					     "invalid period for event");
		}
		
		event->ticks = myticks + ticks;
		event->freq = ticks; /* only needed by periodic, of course */
	}
	
	event->func = func;
	event->farg = farg;

	ani_ilock_lock(&eventlock);
	assert(event->flags == 0);
	event->flags = iflags;
	ani_event_settype(event, type);
	ani_event_setstate(event, AN_EVSTATE_SCHEDULED);

	ani_os_intr_disable();
	ani_event_add_locked(event);
	ani_os_intr_enable();
	ani_ilock_unlock(&eventlock);

	return ani_error_clear();
}

ani_error_t
ani_event_fire(ani_event_t *event, an_eventfunc_t func, void *farg)
{
	ani_error_t err;

	assert(event);
	assert(func);

	err = ani_error_clear();

	if (ani_event_state(event) != AN_EVSTATE_DESCHEDULED)
		return ani_error_set(AN_ERR_INVALIDARG,
				     "event is already scheduled");
	
	assert_ani_threadpool_valid(event->tpool);
	/* assert_ani_threadwork_valid(&event->work); */
	event->func = func;
	event->farg = farg;
	event->ticks = myticks; /* XXX */
	event->freq = 0;
	
	ani_ilock_lock(&eventlock);
	ani_event_settype(event, AN_EVTYPE_ONESHOT_RELATIVE); /* XXX AN_EVSTATE_ONESHOT_IMMEDIATE */
	ani_event_setstate(event, AN_EVSTATE_SCHEDULED); /* XXX AN_EVSTATE_FIRED_BUT_NOT_YET_RUNNING */
	ani_os_intr_disable();
	ani_event_add_locked(event);
	ani_os_intr_enable();
	ani_ilock_unlock(&eventlock);

	return err;
}

ani_error_t
ani_event_deschedule(ani_event_t *event, int *oflags)
{
	ani_error_t err;
	
	assert(event);

	err = ani_error_clear();
	
	/*
	 * Atomically fetch old flags and de-schedule the event.
	 */
	ani_ilock_lock(&eventlock);
	if (oflags)
		*oflags = event->flags;
	if (ani_event_state(event) == AN_EVSTATE_SCHEDULED) {
		ani_os_intr_disable();
		err = ani_event_remove_locked(event);
		ani_os_intr_enable();
	}
	/* XXX undefine event, regardless of error? */
	ani_event_setstate(event, AN_EVSTATE_DESCHEDULED);
	ani_event_settype(event, AN_EVTYPE_UNDEFINED);
	ani_ilock_unlock(&eventlock);

	return err;
}

ani_error_t
ani_event_redefine(ani_event_t *event,
		   an_timespec_t *otime, int *oflags,
		   an_eventfunc_t *ofunc, void **ofarg,
		   an_timespec_t *ntime, an_eventtype_t type,
		   an_eventfunc_t nfunc, void *nfarg)
{
	ani_tick_t ofiretime, nfiretime;
	ani_tick_t ofreq, nfreq;
	int myoflags;
	
	assert(event);
	assert(ntime);
	assert_valid_eventtype(type);
	assert(nfunc);
	
	/*
	 * Check validity of new state, compute new firetime/freq
	 */
	if (type == AN_EVTYPE_ONESHOT_ABSOLUTE) {
		nfreq = 0;
		nfiretime = ani_abstime_to_ticks(ntime);
	} else {
		unsigned long nticks;
		
		nticks = ani_reltime_to_ticks(ntime); /* absolute or relative */

		/* Only periodic events can have unreasonable tickers */
		if ((nticks == 0)
		    && (type == AN_EVTYPE_PERIODIC))
			return ani_error_set(AN_ERR_INVALIDARG,
					     "invalid period for event");

		nfiretime = myticks + nticks;
		nfreq = nticks;
	}
	
	assert_ani_threadpool_valid(event->tpool);
	/* assert_ani_threadwork_valid(&event->work); */

	/*
	 * Atomically save the old state, deschedule, set the new state, and
	 * schedule the event.
	 */
	ani_ilock_lock(&eventlock);
	ofiretime = event->ticks;
	ofreq = event->freq;
	myoflags = event->flags;
	if (oflags)
		*oflags = event->flags;
	if (ofunc)
		*ofunc = event->func;
	if (ofarg)
		*ofarg = event->farg;
	
	/* deschedule event if its scheduled */
	if (ani_event_state(event) == AN_EVSTATE_SCHEDULED) {
		ani_error_t err;
		ani_os_intr_disable();
		err = ani_event_remove_locked(event);
		/* Its okay if the event wasn't found. */
		ani_os_intr_enable();
	}

	/* Set the new state */
	event->func = nfunc;
	event->farg = nfarg;
	event->ticks = nfiretime;
	event->freq = nfreq;

	/* Schedule event */
	ani_event_settype(event, type);
	ani_event_setstate(event, AN_EVSTATE_SCHEDULED);
	ani_os_intr_disable();
	ani_event_add_locked(event);
	ani_os_intr_enable();
	ani_ilock_unlock(&eventlock);

	/*
	 * Convert ofiretime/ofreq into an_timespec_t, if they really
	 * want that.
	 */
	if (otime != NULL) {
		if (ani_event_flagstotype(myoflags) == AN_EVTYPE_ONESHOT_ABSOLUTE) {
			/*
			 * Convert firetime of the ONESHOT_ABSOLUTE to the
			 * time passed to user.
			 */
			ani_ticks_to_abstime(ofiretime, otime);
		} else {
			/*
			 * Convert freq to a relative time to pass
			 * back to user.
			 */
			ani_ticks_to_reltime(ofreq, otime);
		}
	}

	return ani_error_clear();
}

static inline ani_error_t
ani_event_dispatch(ani_event_t *ev, int flags);

static ani_error_t
ani_event_dispatch(ani_event_t *ev, int flags)
{
	int inkern;
	an_event_t *xev;
	ani_error_t err;

	assert(ev);
	assert_ani_threadpool_valid(ev->tpool);

	err = ani_error_clear();

	/*
	 * Direct (to kernel) calls are easy.
	 * Note that event argument is an internal ani_event not the
	 * external an_event.
	 */
	if (flags & ANI_EV_DIRECT) {
		assert((flags & ANI_EV_INKERN) != 0);
		ev->func((struct an_event*)ev, ev->farg);
		return err;
	}

	/*
	 * Event is already in the threadpool queue.
	 * This most likely happens for a periodic event where
	 * we missed the previous period.  We just let it go.
	 */
	if (ani_threadwork_busy(&ev->work))
		return err;

	inkern = (flags & ANI_EV_INKERN);
	if (inkern)
		xev = (an_event_t *)ev;
	else {
		xev = NODEOS_EXTERN_EVENT(ani_threadpool_domain(ev->tpool), ev);
		if (xev == 0)
			return ani_error_set(AN_ERR_INVALIDOBJECT,
					     "attempt to fire invalid event");
	}

	ani_threadwork_set(&ev->work, inkern, 0, (ani_thread_func_t)ev->func,
			   xev, ev->farg, 0);

#ifdef OSKIT
	ani_threadpool_thread_startintr(ev->tpool, &ev->work);
#else
	err = ani_threadpool_thread_schedule(ev->tpool, &ev->work, 0);
#endif
	return err;
}


/*
 * Called every nspertick nanoseconds.  Fires all the events that are due.
 *
 * OSKit version is special in that it cannot take a lock (actually called
 * at interrupt time), and OSKit version uses a different function to kick
 * a threadwork.
 *
 * XXX if we cannot schedule a thread to handle the event, we ignore the
 * event.  This should only happen if the domain (threadpool) is shutting down
 * (err == EINVAL).
 */
#ifdef OSKIT
static void
ani_event_intr(void *oskitthingee, void *arg)
{
	ani_event_t *ev;
	ani_error_t err;

	myticks++;
	assert(myticks != 0);

	while ((ev = eventlist) != 0 && ev->ticks <= myticks) {
		eventlist = ev->next;

		/*
		 * Change the state of the event to reflect that it
		 * is running/has run.
		 */
		assert(ani_event_state(ev) == AN_EVSTATE_SCHEDULED);
		if (ani_event_type(ev) != AN_EVTYPE_PERIODIC)
			/* one-shot event */
			ani_event_setstate(ev, AN_EVSTATE_DESCHEDULED);

		err = ani_event_dispatch(ev, ev->flags);

		/* Reset timer for periodic events */
		if (ani_error_code(err) != AN_ERR_INVALIDOBJECT &&
		    ani_event_flagstotype(ev->flags) == AN_EVTYPE_PERIODIC) {
			assert(ev->freq > 0);
			ev->ticks += ev->freq;
			ani_event_add_locked(ev);
		}
	}
}
#else
static void
ani_event_intr(void *oskitthingee, void *arg)
{
	ani_event_t *ev;
	ani_error_t err;
	int flags;

	myticks++;
	assert(myticks != 0);

	ani_ilock_lock(&eventlock);
	while ((ev = eventlist) != 0 && ev->ticks <= myticks) {
		eventlist = ev->next;

		/*
		 * Change the state of the event to reflect that it
		 * is running/has run.
		 */
		assert(ani_event_state(ev) == AN_EVSTATE_SCHEDULED);
		if (ani_event_type(ev) != AN_EVTYPE_PERIODIC)
			/* one-shot event */
			ani_event_setstate(ev, AN_EVSTATE_DESCHEDULED);

		flags = ev->flags;
		ani_ilock_unlock(&eventlock);

		err = ani_event_dispatch(ev, flags);
		
		ani_ilock_lock(&eventlock);
		/* Reset timer for periodic events */
		if (ani_error_code(err) != AN_ERR_INVALIDOBJECT &&
		    ani_event_flagstotype(flags) == AN_EVTYPE_PERIODIC) {
			assert(ev->freq > 0);
			ev->ticks += ev->freq;
			ani_event_add_locked(ev);
		}
	}
	ani_ilock_unlock(&eventlock);
}
#endif

/*
 * Remove all events associated with a particular threadpool.  Should be
 * called with the threadpool unlocked, otherwise we go against the locking
 * order used by event_intr above (eventlock then threadpool lock).
 */
void
ani_event_removeall(ani_threadpool_t *tpool)
{
	ani_event_t *ev, *next;

	assert(ani_threadpool_islocked(tpool) == 0);

	ani_ilock_lock(&eventlock);
	ani_os_intr_disable();
	for (ev = eventlist; ev != 0; ev = next) {
		next = ev->next;
		if (ev->tpool == tpool)
			ani_event_remove_locked(ev);
	}
	ani_os_intr_enable();
	ani_ilock_unlock(&eventlock);
}

void
ani_event_dump(char *str)
{
#ifdef DEBUG
	ani_event_t *ev;

	ani_ilock_lock(&eventlock);
	ani_os_intr_disable();
	ani_dprintf("%s:\n", str);
	for (ev = eventlist; ev != 0; ev = ev->next) {
		ani_dprintf("%p: ticks=%lu, freq=%lu, tpool=%p, flags=%#x, "
			    "func=%p, farg=%p\n",
			    ev, ev->ticks, ev->freq, ev->tpool, ev->flags,
			    ev->func, ev->farg);
	}
	ani_os_intr_enable();
	ani_ilock_unlock(&eventlock);
#endif
}

/*
 * Include the inlined functions in non-inlined form
 * XXX must be last in the file
 */
#undef NODEOS_FUNC_INLINE
#define NODEOS_FUNC_INLINE

#include <nodeos/ani/ani_event_inline.h>
