/*
 * Copyright (c) 1999, 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * Network interface information.
 */
#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include <nodeos/an_info.h>
#include <nodeos/an_error.h>

#include <nodeos/ani/ani_chan.h>
#include <nodeos/ani/ani_if.h>
#include <nodeos/config.h>

#include "ani_os.h"

#ifdef CHAN_ETH
extern struct chanops chan_eth_ops;
#endif

void
ani_if_setup(void)
{
}

void
ani_if_shutdown(void)
{
}

an_infoif_t *
ani_if_getifinfo(int *count)
{
	return ani_os_getifinfo(count);
}

an_infoipv4_t *
ani_if_getipv4info(int *count)
{
	return ani_os_getipv4info(count);
}

#ifdef CHAN_IF

/*
 * Hack "if" channel type support.
 *
 * "if" channels are "generic" network interface channels.  They return
 * packet payloads stripped of any interface headers; e.g., ethernet packets
 * stripped of the 14 byte ethernet header.  Mapping of physical interfaces
 * to logical "if" names is done via the ID contained in the ifinfo struct.
 */

static int chan_if_matches(an_chanproto_t *proto);
static ani_error_t chan_if_create(ani_chan_t *chan);
static void chan_if_destroy(ani_chan_t *chan);
static ani_error_t chan_if_send(ani_chan_t *chan, ani_pbuf_t *pbuf);
static ani_error_t chan_if_enable(ani_chan_t *chan);
static void chan_if_disable(ani_chan_t *chan);
static void chan_if_txflush(ani_chan_t *chan);

static char chan_if_prefix[] = "if";	/* XXX */

struct chanops chan_if_ops = {
	ani_if_setup,		ani_if_shutdown,	chan_if_matches,
	chan_if_create,		chan_if_destroy,	chan_if_send,
	chan_if_enable,		chan_if_disable,
	0,			chan_if_txflush
};

static int
chan_if_matches(an_chanproto_t *proto)
{
	int len;

	len = strlen(chan_if_prefix);
	if (proto->len < len)
		return 0;

	return (strncmp(proto->data, chan_if_prefix, len) == 0);
}

/*
 * Translate "if" name to underlying interface(s)
 * Returns 1 if found, 0 otherwise.
 */
static int
chan_if_find(char *name, an_infoif_t **infop, an_infoiftype_t *iftype)
{
	an_infoif_t *ifinfo, *info;
	int id, numifs;

#if 0
	if (strcmp(name, "if") == 0) {
		*iftype = ANIF_ANY;
		*infop = 0;
		return 1;
	}
	if (strcmp(name, "ifeth") == 0) {
		*iftype = ANIF_ETH;
		*infop = 0;
		return 1;
	}
#endif

	ifinfo = ani_os_getifinfo(&numifs);

	id = atoi(&name[2]);
	if (id < 0 || id >= numifs)
		return 0;

	for (info = ifinfo; info < &ifinfo[numifs]; info++)
		if (id == info->id)
			break;
	if (info == &ifinfo[numifs])
		return 0;

	*infop = info;
	*iftype = info->type;
	return 1;
}

static ani_error_t
chan_if_create(ani_chan_t *chan)
{
	an_infoiftype_t iftype;
	an_infoif_t *info;
	ani_error_t err;

	if (!chan_if_find(ani_chan_name(chan), &info, &iftype))
		return ani_error_set(EINVAL, "invalid channel name");

	/*
	 * XXX support multi-device channels some day
	 */
	if (info == 0)
		return ani_error_set(ENODEV, "no multi-IF channel support yet");

	switch (iftype) {
#ifdef CHAN_ETH
	case ANIF_ETH:
		chan->flags &= ~CHAN_ISRAW;

		/*
		 * XXX hackery.  Replace user-specified proto spec
		 * with one crafted to reflect the actual interface.
		 */
		strcpy(chan->proto.data, info->pname);
		chan->proto.len = strlen(info->pname);

		err = chan_eth_ops.create(chan);
		break;
#endif
	default:
		err = ani_error_set(EINVAL, "bad IF channel type");
		break;
	}

	return err;
}

static void
chan_if_destroy(ani_chan_t *chan)
{
	an_infoiftype_t iftype;
	an_infoif_t *info;
	int rv;

	rv = chan_if_find(ani_chan_name(chan), &info, &iftype);
	assert(rv != 0);
	assert(info != 0);

	switch (iftype) {
#ifdef CHAN_ETH
	case ANIF_ETH:
		chan_eth_ops.destroy(chan);
		break;
#endif
	default:
		ani_os_panic("destroy: bad IF channel type");
		break;
	}
}

static ani_error_t
chan_if_enable(ani_chan_t *chan)
{
	an_infoiftype_t iftype;
	an_infoif_t *info;
	ani_error_t err;

	if (!chan_if_find(ani_chan_name(chan), &info, &iftype))
		return ani_error_set(EINVAL, "invalid channel name");

	/*
	 * XXX support multi-device channels some day
	 */
	if (info == 0)
		return ani_error_set(ENODEV, "no multi-IF channel support yet");

	switch (iftype) {
#ifdef CHAN_ETH
	case ANIF_ETH:
		err = chan_eth_ops.enable(chan);
		break;
#endif
	default:
		err = ani_error_set(EINVAL, "bad IF channel type");
		break;
	}

	return err;
}

static void
chan_if_disable(ani_chan_t *chan)
{
	an_infoiftype_t iftype;
	an_infoif_t *info;
	int rv;

	rv = chan_if_find(ani_chan_name(chan), &info, &iftype);
	assert(rv != 0);
	assert(info != 0);

	switch (iftype) {
#ifdef CHAN_ETH
	case ANIF_ETH:
		chan_eth_ops.disable(chan);
		break;
#endif
	default:
		ani_os_panic("disable: bad IF channel type");
		break;
	}
}

static ani_error_t
chan_if_send(ani_chan_t *chan, ani_pbuf_t *pbuf)
{
	an_infoiftype_t iftype;
	an_infoif_t *info;
	ani_error_t err;
	int rv;

	rv = chan_if_find(ani_chan_name(chan), &info, &iftype);
	assert(rv != 0);
	assert(info != 0);

	switch (iftype) {
#ifdef CHAN_ETH
	case ANIF_ETH:
		err = chan_eth_ops.send(chan, pbuf);
		break;
#endif
	default:
		ani_os_panic("send: bad IF channel type");
		err = ani_error_set(EINVAL, "bad IF channel type");
		break;
	}

	return err;
}

static void
chan_if_txflush(ani_chan_t *chan)
{
#ifdef USE_POLLING
	an_infoiftype_t iftype;
	an_infoif_t *info;
	ani_error_t err;
	int rv;

	rv = chan_if_find(ani_chan_name(chan), &info, &iftype);
	assert(rv != 0);
	assert(info != 0);

	switch (iftype) {
#ifdef CHAN_ETH
	case ANIF_ETH:
		chan_eth_ops.txflush(chan);
		break;
#endif
	default:
		ani_os_panic("send: bad IF channel type");
		err = ani_error_set(EINVAL, "bad IF channel type");
		break;
	}
#endif
}

#endif
