/*
 * Copyright (c) 1999-2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h> /* XXX memset */

#include <nodeos/an_domain.h>
#include <nodeos/an_moab_setup.h>
#include <nodeos/an_info.h>
#include <nodeos/an_debug.h>
#include <nodeos/an_thread.h>

#include <nodeos/an_log.h>

static void load_ee(void *arg);
static void ee_died(an_domain_t *);
static const char *parsemoabargs(int argc, char **argv, an_initargs_t* initargs);

static an_cred_t root_cred;
static an_cred_t initee_cred;
static an_mempool_t root_mempool;
static an_domain_t root_domain;
static an_domain_t initee_domain;

#define ROOTDOMAIN_THREAD_COUNT      (2)
#define ROOTDOMAIN_THREAD_STACKSIZE  (20 * 1024) /* OSKit pthread minimum */


/*
 * Default main routine that is pulled in if an EE doesn't already define one.
 *
 * Note that this code does not use any internal (ani_*) NodeOS interfaces.
 * It uses just public interfaces and two functions from its host environment
 * (e.g., OSKit, Linux, BSD): printf and exit.  Thus it should be directly
 * usable (i.e., cut and paste) by EE's that want to do magic things at system
 * startup.
 * 
 * XXX We should probably get rid of this once the examples are all
 * converted over...
 */
int
main(int argc, char **argv)
{
	static an_initargs_t load_ee_args;
	const char *rc;

	/*
	 * Shut the overly-chatty OSKit up.  This keeps all messages
	 * below "error" level from clogging the airwaves.
	 * You can reverse this by passing '-moabdebug ~OSKIT_LOG_ERR'
	 * on the command line.
	 */
	an_debug_setmask("OSKIT_LOG_ERR");

	/*
	 * Parse arguments, pulling out Moab-specific ones.
	 * Push the remaining arguments into the global
	 * initargs struct, which will be passed to the
	 * initial EE.  This function must be careful
	 * to not invoke an an_functions other than the
	 * magical an_debug_foo() functions.
	 */
	rc = parsemoabargs(argc, argv, &load_ee_args);
	if (rc) {
		printf("Error parsing -moabdebug arg: %s\n", rc);
		exit(1);
	}

	/*
	 * This code all runs without a domain or NodeOS thread
	 * context.  There are few things that can be done
	 * in this context.
	 */
	an_moab_setup();
	
	printf("\nUtah NodeOS (version 0.%04d)\n", an_moab_version());
	printf("Options: %s\n", an_info_getoptlist());

	an_moab_start(&root_domain, &root_cred, &root_mempool,
		      ROOTDOMAIN_THREAD_COUNT, ROOTDOMAIN_THREAD_STACKSIZE,
		      /* use implicit threads: */ NULL, NULL, 
		      load_ee, &load_ee_args);

	/* XXX */
	printf("Root domain died, exiting...\n");

	exit(0);
}

/*
 * Domain initialization function for the root domain started by an_moab_start.
 * The root domain has a very limited set of resources since this function and
 * ee_died are the only things that ever run in its context.
 */

static char loadmsg[] = "Loading initial EE...\n";
static char diedmsg[] = "Initial EE domain died...\n";

static void
load_ee(void *arg)
{
	void (*func)(void *);
	an_cred_t *cred;
	an_cpuspec_t cspec;
	an_memspec_t mspec;
	an_error_t err;
	unsigned int nthreads, cpurate, memtotal, netrate;

	an_log_write(AN_LOG_INFO, loadmsg, strlen(loadmsg));

	/*
	 * Create a new credential from ours
	 */
	cred = an_cred_create(&initee_cred);
	assert(cred == &initee_cred);

	/*
	 * Give it all of the root resources.  This is ok since domains
	 * created with the root credential are not subject to admission
	 * checks.
	 *
	 * XXX since memory is not a renewable resource, we really do
	 * need to account for the root credential's use so we don't run
	 * out.  So we just siphon off a little bit.
	 */
	err = an_cred_resource_get(an_cred_current(), &nthreads, &cpurate,
				   &memtotal, &netrate);
	assert(memtotal > 1024*1024);
	memtotal -= 1024*1024;

	err = an_cred_resource_xfer(an_cred_current(), cred,
				    nthreads, cpurate, memtotal, netrate);
	assert(err == 0);

	{
		extern an_domaininit_func_t InitEE;

		func = InitEE;
		memset(&cspec, 0, sizeof cspec);
		cspec.ttype = AN_TH_IMPLICIT;
		cspec.threads.im.maxthreads = 3;
		cspec.threads.im.stacksize = 32 * 1024;
		cspec.rate = 0; /* XXX default rate */
		mspec.nchunks = cspec.threads.im.maxthreads *
			(cspec.threads.im.stacksize + 128*1024) /
			NODEOS_MEMPOOL_CHUNKSIZE;
		mspec.pool = an_mempool_current();
	}

	/* Create the domain, pass our void* arg on to func. */
	if (an_domain_create(&initee_domain, cred, &cspec, &mspec,
			     func, arg, ee_died) == 0)
		ee_died(0);

	an_thread_exit();
}

static void
ee_died(an_domain_t *domain)
{
	an_log_write(AN_LOG_INFO, diedmsg, strlen(diedmsg));
	an_moab_shutdown();
}

static const char *
parsemoabargs(int argc, char **argv, an_initargs_t *initargs)
{
	int i;
	static char *noargs[1] = { 0 };

	if (argc == 0) {
		argv = noargs;
		return 0;
	}

	/*
	 * To avoid allocating memory, we just stomp all over
	 * argv.  Since we go through argv linearly, and
	 * occasionally skip elements, there will be no
	 * problems.
	 */
	initargs->argv = argv;

	/* Assume that no args are given to initargs. */
	initargs->argc = 0;

	for (i = 0; i < argc; i++) {
		if (!strcmp("-moabdebug", argv[i])) {
			const char *rc;

			if (++i >= argc)
				return "<no args given>";
			rc = an_debug_setmask(argv[i]);
			if (rc)
				return rc;
		} else {
			/* Give this argument to initargs */
			initargs->argv[initargs->argc] = argv[i];
			initargs->argc++;
		}
	}

	return 0;
}
