/*
 * Copyright (c) 1999-2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <assert.h>
#include <string.h>
#include <stdlib.h>

#include <nodeos/impl/an_error.h>
#include <nodeos/ani/ani_hnode.h>
#include <nodeos/ani/ani_mempool.h>
#include <nodeos/ani/ani_domain.h>

#include "ani_debug.h"

extern int bootstrapped;

static struct hnode allpools;
static ani_mempool_t *root_mempool = NULL;

void
ani_mempool_setup(ani_mempool_t *pool)
{
	ani_error_t err;

	assert(pool);
	assert(root_mempool == 0);

	err = ani_mempool_init(pool, 0, 0);
	assert(!ani_error_occured(err));
	root_mempool = pool;
}

void
ani_mempool_shutdown(void)
{
	root_mempool = NULL;
}

/*
 * Get the mempool associated with the current thread's domain.
 * May return 0 if called before the root_mempool is setup.
 */
ani_mempool_t *
ani_mempool_current(void)
{
	ani_domain_t *domain;
	ani_mempool_t *pool;

	domain = ani_domain_current();
	pool = domain ? domain->mpool : root_mempool;
	assert(domain == 0 || pool != 0);

	return pool;
}

/*
 * Return the root mempool
 */
ani_mempool_t *
ani_mempool_rootpool(void)
{
	return root_mempool;
}

/*
 * Create a new mempool with no associated domains.  The parent
 * parameter defines this pool's position in the pool hierarchy.
 * The callback routine is invoked whenever the NodeOS detects that
 * the indicated pool is over its memory limit.
 */
ani_error_t
ani_mempool_init(ani_mempool_t *pool, ani_mempool_t *parent,
		 an_mempoolfull_func_t callback)
{
	assert(pool != 0);

	memset(pool, 0, sizeof *pool);

	hnode_init(&pool->me);
	hnode_attach(&pool->me, 0, &allpools);
	if (parent)
		hnode_attach(&pool->me, (struct hnode *)parent, 0);
	pool->physcurrent = pool->physlimit = 0;
	pool->callback = callback;

	return ani_error_clear();
}

/*
 * Destroy a mempool.  There should be no domains associatd with the pool.
 */
ani_error_t
ani_mempool_deinit(ani_mempool_t *pool)
{
	assert(pool);

	if (hnode_child(&pool->me) != 0)
		return ani_error_set(EBUSY,
				     "mempool has attached domains");

	hnode_detach(&pool->me, HNODE_ALL);

	return ani_error_clear();
}

/*
 * Donates the memory resources associated with a domain to the pool.
 */
void
ani_mempool_adddomain(ani_mempool_t *pool, ani_domain_t *domain)
{
	assert(pool);
	assert(domain);

	hnode_attach((struct hnode *)domain, 0, &pool->domainlist);

	pool->physlimit += (domain->res.memtotal / ANI_MEMPOOL_CHUNKSIZE);
	pool->stats.numdomains++;

#if 0
	/*
	 * XXX first attach to the root pool,
	 * debit for memory allocated before this point.
	 */
	if (pool == root_mempool && pool->physcurrent == 0) {
		extern unsigned long earlymempages;

		assert(domain == ani_domain_root_domain());
		assert(earlymempages <= pool->physlimit);

		pool->physcurrent += earlymempages;
	}
#endif
}

/*
 * Removes the memory resources associated with a domain from the pool.
 */
void
ani_mempool_removedomain(ani_mempool_t *pool, ani_domain_t *domain)
{
	hnode_detach((struct hnode *)domain, HNODE_GLOBAL);

	pool->physlimit -= (domain->res.memtotal / ANI_MEMPOOL_CHUNKSIZE);
	pool->stats.numdomains--;

#if 0
	/*
	 * If we are over the pool limit and it matters, invoke the callback
	 * to free up some memory.  If no callback, kill all the domains!
	 *
	 * XXX we need to establish the proper context for the callback to
	 *     run in (whatever that might be)
	 * XXX we need to do something reasonable if the callback doesn't
	 *     return or doesn't free memory.
	 */
	if (callback == 0)
		;
	while (pool->physcurrent > pool->physlimit)
		pool->callback(pool, pool->physcurrent, pool->physlimit);
#endif
}

/*
 * Simple malloc/free allocation interface
 *
 * XXX if pool == 0, root_mempool does not yet exist or has no resources yet,
 * remember allocations for later.
 *
 * XXX maintain seperate heaps for each pool, charge accordingly
 */

void *
ani_mempool_alloc(ani_mempool_t *pool, size_t size)
{
	void *ptr;

	assert(pool != 0);

	/* XXX fer now */
	ptr = malloc(size);

	if (ptr)
		pool->stats.exchunks += size;	/* XXX need locking */

	return ptr;
}

void
ani_mempool_free(ani_mempool_t *pool, void *ptr, size_t size)
{
	assert(pool != 0);

	/* XXX fer now */
	free(ptr);

	pool->stats.exchunks -= size;	/* XXX need locking */
}

ani_error_t
ani_mempool_stats(ani_mempool_t *pool, an_mempoolstats_t *stats)
{
	assert(pool != 0);

	stats->exchunks = pool->stats.exchunks;
	stats->imchunks = pool->stats.imchunks;
	stats->chunksize = 1; /* XXX */
	stats->numdomains = pool->stats.numdomains;
#if defined(OSKIT) && !defined(OSKIT_UNIX)
	/*
	 * For the OSKit, implicit allocations include explicit ones as well.
	 */
	assert(stats->imchunks >= stats->exchunks);
	stats->imchunks -= stats->exchunks;
#endif
	return ani_error_clear();
}

#ifdef DEBUG
#include <stdio.h>

static int
dump_mempool(struct hnode *node, void *arg)
{
	ani_mempool_t *pool = (ani_mempool_t *)node;

	ani_dprintf("POOL %p: parent=%p, child=%p, sibling=%p, "
		    "physcur=%d, physlimit=%d\n",
		    pool, hnode_parent(node), hnode_child(node),
		    hnode_sibling(node), pool->physcurrent, pool->physlimit);
	return 0;
}

void
ani_mempool_dump(ani_mempool_t *pool)
{
	if (pool == 0)
		hnode_traverse(&ani_mempool_current()->me, dump_mempool, 0,
			       HNODE_GLOBAL, 0);
	else
		hnode_traverse(&pool->me, dump_mempool, 0, HNODE_PARENT, 0);
}
#endif
