/*
 * Copyright (c) 1999-2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <nodeos/an_domain.h>

#include <nodeos/ani/ani_thread.h>
#include <nodeos/ani/ani_ilock.h>
#include <nodeos/ani/ani_cred.h>
#include <nodeos/ani/ani_mempool.h>
#include <nodeos/ani/ani_chan.h>
#include <nodeos/ani/ani_event.h>
#include <nodeos/ani/ani_moab_setup.h>
#include <nodeos/ani/ani_log.h>

#include "ani_os.h"
#include "ani_debug.h"

/*
 * Setup Moab.
 *
 * See main.c for example code that sets up Moab.
 * These functions should only be called once and should
 * be called immediately by the EE's main().
 */

#include "version.h" 
static int version = VERSION;

static ani_icond_t shutdownCond;
static ani_ilock_t shutdownLock;
static int childdead;

#define MOABSTATE_NUTHIN         0
#define MOABSTATE_BARESETUPSTART 1
#define MOABSTATE_BARESETUPDONE  2
#define MOABSTATE_INITNODEOS     3
#define MOABSTATE_ALLSETUPDONE   4
#define MOABSTATE_STARTSHUTDOWN  5
#define MOABSTATE_ROOTDOMAINGONE 6
#define MOABSTATE_COMPLETETOAST  7

int ani_moab_state = MOABSTATE_NUTHIN;

ani_resource_t ani_moab_noderes;

/*
 * Initialize the underlying device layers.
 */
void
ani_moab_setup(void)
{
	ani_moab_state = MOABSTATE_BARESETUPSTART;

	ani_debug_init();

	ani_os_init_basic();

	ani_os_init_mem();
	ani_os_init_devices();
	ani_os_init_net();
	ani_os_init_fs();
#ifdef GPROF
	/* after fs setup */
	ani_os_gprof_init();
#endif

	ani_moab_state = MOABSTATE_BARESETUPDONE;
}

/*
 * Return the compiled version of Moab, as an integer.
 */
int
ani_moab_version(void)
{
	return version;
}

/*
 * Return available resources for the node.
 */
void
ani_moab_resources(ani_resource_t *res)
{
	*res = ani_moab_noderes;
}

/*
 * Start a new thread in the root domain on its way.
 * 
 * Initialize various NodeOS components.  Register the (uninitialized)
 * global objects passed in.
 *
 * XXX initialization of credential will probably have to change.
 *
 * DOES NOT RETURN UNTIL MOAB IS SHUTDOWN.
 */
void
ani_moab_start(ani_domain_t *root_domain, ani_cred_t *root_cred,
	       ani_mempool_t *root_mempool,
	       int nthreads, int stacksize,
	       void **root_threads, void **root_threadstacks,
	       void (*initee_func)(void *), void *initee_arg)
{
	ani_error_t err;

	assert(root_domain);
	assert(root_mempool);
	assert(root_cred);
	assert(initee_func);

	ani_moab_state = MOABSTATE_INITNODEOS;

	/*
	 * Initialize NodeOS components, setup global objects.
	 */
	ani_resource_setup(&ani_moab_noderes);
	ani_cred_setup(root_cred, &ani_moab_noderes);
	if (ani_mempool_rootpool() == 0)
		ani_mempool_setup(root_mempool);
	assert(ani_mempool_rootpool() == root_mempool);
	ani_thread_setup();
	ani_domain_setup(root_domain);

	/* Initialize the root domain shutdown condition variable */
	ani_ilock_init(&shutdownLock);
	ani_icond_init(&shutdownCond);

	ani_moab_state = MOABSTATE_ALLSETUPDONE;

	childdead = 0;
	ani_ilock_lock(&shutdownLock);

#ifdef MEMLOGGING
	/* after one-time allocations */
	ani_os_memlog_start();
#endif

	/*
	 * Create the root domain.
	 * This will fire off a thread at initee_func which does
	 * real root domain initialization in the context of the root
	 * domain.  All memory allocations are charged to a domain from
	 * this point on.
	 *
	 * NOTE we don't pass the initialization function here.
	 * Doing so would create a race between that initialization thread
	 * and our finishing up our setup (e.g., ani_chan_setup).  Instead,
	 * we explicitly fire up a domain thread in that function when we are
	 * done.
	 *
	 * NOTE that there is no shutdown callback for the root
	 * domain.  It is never destroyed.  ani_moab_shutdown()
	 * should be called instead.
	 */
	err = ani_rootdomain_init(root_domain, root_cred, root_mempool,
				nthreads, stacksize,
				root_threads, root_threadstacks,
				0, 0);
	assert(!ani_error_occured(err)); /* XXX */

	/*
	 * Initialize subsystems that might need the root domain
	 * (i.e., to start a thread).
	 */
	ani_event_setup();
	ani_chan_setup(root_domain);

	/*
	 * We are done initializing so fire up a domain thread to run the
	 * initialization.
	 */
	err = ani_domain_startthread(root_domain, 0,
				   (void (*)(void *, void *))initee_func,
				   initee_arg, 0);
	assert(!ani_error_occured(err)); /* XXX */

	/*
	 * Wait until ani_moab_shutdown is called.
	 */
	while (childdead == 0)
		ani_icond_wait(&shutdownCond, &shutdownLock);
	ani_ilock_unlock(&shutdownLock);

#ifdef THREAD_CANCEL_BROKEN
	ani_log_printf(ANI_LOG_WARN, "WARNING: "
		       "abrupt shutdown due to broken pthread_cancel\n");
#else
	ani_moab_state = MOABSTATE_STARTSHUTDOWN;

	/*
	 * XXX prevent new threads from getting started.
	 * Makes life easier.
	 */
	ani_threadpool_shutdown();

	/*
	 * Shutdown all children of the root domain first.
	 *
	 * XXX this sucks.  We should just shutdown the root_domain but
	 * we run into problems if we deinit event/channel threads that
	 * were started in the root domain before we do shutdown on those
	 * objects.  For example if rootdomain_deinit cancels the event
	 * mechanism's timer under linux/solaris, the cancelled thread
	 * never returns from its select.  So we need to let the event
	 * system shutdown itself gracefully.  If we try to do event/channel
	 * shutdown first, then we are yanking infrastructure out from
	 * under running threads.  So, we shutdown most of the world,
	 * clean up events/channels, and then destroy the root domain proper.
	 */
	ani_rootdomain_deinit_children(root_domain);

	/*
	 * Gracefully shutdown services that might have spawned their own
	 * root_domain thread now that any users of that service have been
	 * killed.
	 */
	ani_chan_shutdown();
	ani_event_shutdown();

	/*
	 * Finally, we destroy the actual root domain
	 */
	ani_rootdomain_deinit(root_domain);
	ani_mempool_shutdown();
	ani_moab_state = MOABSTATE_ROOTDOMAINGONE;

	ani_os_deinit_fs();
	ani_os_deinit_net();

	/* Return to the Domain-free environment from whence we came. */
	ani_moab_state = MOABSTATE_COMPLETETOAST;

#ifdef MEMLOGGING
	ani_os_memlog_dump();
#endif
#endif
}

void
ani_moab_shutdown(void)
{
	int otype;

	ani_ilock_lock(&shutdownLock);
	if (childdead) {
		ani_ilock_unlock(&shutdownLock);
		return;
	}

	childdead = 1;
	ani_icond_signal(&shutdownCond);

	/*
	 * XXX wait til we get cancelled
	 *
	 * Note that we may or may not have a nodeos thread context
	 * when called and have not done a NODEOS_ENTER, so we must be
	 * careful what we do.  In particular, pthread_cond_wait's (and
	 * thus ani_icond_wait's), cancellation semantics (w.r.t. reacquiring
	 * the lock) depend on whether ASYNC or DEFERRED cancellation is set.
	 * Hence we first ensure that we are in the DEFERRED cancel mode so
	 * that ani_icond_wait works right.
	 */
	pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, &otype);
	while (1)
		ani_icond_wait(&shutdownCond, &shutdownLock);
}
