/*
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <oskit/dev/dev.h> /* osenv_log() */
#include <stdarg.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>

#include <nodeos/ani/ani_log.h>
#ifdef UTAHTESTBED
#include <oskit/tmcp.h>
#endif

#include "ani_os.h"
#include "ani_debug.h"

extern void sigcontext_dump(struct sigcontext *scp);

/*
 * NodeOS logging functions
 */

/*
 * Write a buffer to a logfile.
 *
 * Note: this function may be called very early in the initialization
 * or very late in shutdown of Moab and hence must work independent of the
 * state of the NodeOS.
 */
void
ani_os_log_write(int prio, void *buf, int buflen)
{
	int fd = 1;

#ifdef UTAHTESTBED
	/*
	 * Log EE output to the testbed log as well as console
	 */
	if (prio < ANI_LOG_LOW)
		tmcp_log("%.*s", buflen, buf);
#endif
	/*
	 * XXX use stdout (instead of stderr) for debug too.
	 * You really don't want to know (but if you insist,
	 * see posix/sys/write.c)
	 */
	if (prio == ANI_LOG_DEBUG)
		fd = 1;

	write(fd, buf, buflen);
}

/*
 * Emergency shutdown function
 * Basically, just the OSKit panic but removes a few levels of function
 * call by going direct.
 */
void
ani_os_panic(const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	ani_log_vprintf(ANI_LOG_PANIC, fmt, args);
	va_end(args);
	ani_log_printf(ANI_LOG_PANIC, "\n");

	exit(1);
}

/*
 * Dump thread state
 */
void
ani_os_dump_state(void *state)
{
	sigcontext_dump((struct sigcontext *)state);
}

/*
 * Replacement for the OSKit OSENV logging functions.  Puts all of it
 * under control of the Moab Logging/Debugging infrastructure.
 */

/*
 * All log-like messages generated by OSKit code come through
 * here.  They may (but usually don't) include newlines.
 *
 * Priority should be between 0 and 7.  '0' is for emergency messages,
 * 7 is for random crap. grep for OSENV_LOG_ in oskit/dev/dev.h
 */
void
osenv_vlog(int priority, const char *fmt, void *args)
{
	int doit;

#ifdef DEBUG
	/* Maximum verbosity is default when debugging. */
	doit = 1;
#else
	/* Non-debug builds, only show error messages. */
	doit = (priority < 3);
#endif

	DBG(DBG_OSKIT_LOG_NOTICE, doit = (priority < 5));
	DBG(DBG_OSKIT_LOG_ERR, doit = (priority < 3));
	DBG(DBG_OSKIT_LOG_NONE, doit = 0);

	if (doit) {
#ifdef DEBUG		
		ani_log_vprintf(ANI_LOG_DEBUG, fmt, args);
#else
		ani_log_vprintf(ANI_LOG_INFO, fmt, args);
#endif
	}
}

void
osenv_log(int priority, const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	osenv_vlog(priority, fmt, args);
	va_end(args);
}
