/*
 * Copyright (c) 1997-2002 University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include <oskit/dev/osenv.h>
#include <oskit/dev/osenv_mem.h>
#include <oskit/dev/ethernet.h>
#ifdef UTAHTESTBED
#include <oskit/tmcp.h>
#endif

#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>

/*
 * Set PROMPTME to 1 if you want the OSKit to prompt via the console
 * for IP/mask on interfaces that don't bootp or have info set via the
 * environment.
 */
#define PROMPTME	1

/*
 * XXX hack: we don't include oskit_libkern so we cannot get the expanded
 * versions of the atomic routines.  Hence we have to make sure they get
 * inlined.  Must be done before include of skbufio.h (which is included
 * via ani_os.h).
 */
#undef OSKIT_INLINE
#define OSKIT_INLINE static
#include <oskit/machine/atomic.h>

#include "ani_os.h"
#include "ani_debug.h"

#include <nodeos/an_info.h>

#include <nodeos/ani/ani_error.h>
#include <nodeos/ani/ani_pbuf.h>
#include <nodeos/impl/an_impl.h>
#include <nodeos/ani/ani_log.h>

static void oskit_skbufio_mem_flush(void);
static int oskit_skbufio_shutdown;

/*
 * Override portions of the OSKit skbuff infrastructure for both native and
 * unix versions of the OSKit:
 *
 *	1. Provides an skbufio wrapper implementation allowing us to catch
 *		the release of an skbufio when a device transmit is done
 *		(for an_pbuf_isbusy).
 *	2. Provides memory allocators for the OSKit linux driver glue
 *		and the OSkit/Unix skbufio implementation
 */

static an_infoif_t *iftab;
static an_infoipv4_t *ipv4tab;
static int numifs;

/*
 * For the OSKit, all actual device initialization has been done.
 * Here we just determine all the information we can about the available
 * interfaces.
 */
void
ani_os_init_net(void)
{
	oskit_etherdev_t **edev;
	int err, ndev, i;
	char *addr, *mask;

#ifdef UTAHTESTBED
	err = start_tmcp(1);
	DBG(DBG_DEVICE_INIT, {
		if (err) {
			ani_dprintf("No TMCP info (err=%x)\n", err);
		} else {
			ani_dprintf("TMCP info:\n");
			tmcp_dumpinfo();
		}
	});
#endif

	/*
	 * Look up all the ethernet devices
	 * XXX assumes only ethernet right now
	 */
	ndev = osenv_device_lookup(&oskit_etherdev_iid, (void***)&edev);
	if (ndev <= 0)
		return;

	iftab = osenv_mem_alloc(ndev * sizeof(an_infoif_t),
				OSENV_PHYS_WIRED, 0);
	ipv4tab = osenv_mem_alloc(ndev * sizeof(an_infoipv4_t),
				  OSENV_PHYS_WIRED, 0);
	if (iftab == 0 || ipv4tab == 0) {
		err = OSKIT_ENOMEM;
		goto error;
	}
	memset(iftab, 0, ndev * sizeof(an_infoif_t));
	memset(ipv4tab, 0, ndev * sizeof(an_infoipv4_t));

	for (i = 0; i < ndev; i++) {
		/*
		 * Raw interface info
		 * XXX beware of hardwired goo (name string, bandwidth)
		 */
		iftab[i].type = ANIF_ETH;
		iftab[i].id = i;
		snprintf(iftab[i].pname, sizeof iftab[i].pname, "eth%d", i);
		oskit_etherdev_getaddr(edev[i], iftab[i].addr.eth.daddr);
		iftab[i].bandwidth = 100 * 1000 * 1000;

		/*
		 * IPv4 info
		 */
		ipv4tab[i].ifid = i;
#ifdef UTAHTESTBED
		err = start_tmcp_getinfo_if(i, &addr, &mask);
#else
		err = start_getinfo_network_eif(i, &addr, &mask, PROMPTME);
#endif
		if (err == 0) {
			ipv4tab[i].addr = inet_addr(addr);
			free(addr);
			ipv4tab[i].mask = inet_addr(mask);
			free(mask);
		}
	}
	numifs = ndev;
	err = 0;

	/*
	 * Summarize
	 */
	DBG(DBG_DEVICE_INIT, {
		ani_dprintf("Interface configuration:\n");
		for (i = 0; i < ndev; i++) {
			ani_dprintf("%d: %s: mac=%02x:%02x:%02x:%02x:%02x:%02x",
				    iftab[i].id,
				    iftab[i].pname,
				    iftab[i].addr.eth.daddr[0],
				    iftab[i].addr.eth.daddr[1],
				    iftab[i].addr.eth.daddr[2],
				    iftab[i].addr.eth.daddr[3],
				    iftab[i].addr.eth.daddr[4],
				    iftab[i].addr.eth.daddr[5]);
			if (ipv4tab[i].addr != 0 && ipv4tab[i].mask != 0) {
				struct in_addr tmp;

				tmp.s_addr = ipv4tab[i].addr;
				ani_dprintf(", IPaddr=%s", inet_ntoa(tmp));
				tmp.s_addr = ipv4tab[i].mask;
				ani_dprintf(", IPmask=%s", inet_ntoa(tmp));
			}
			ani_dprintf("\n");
		}
	})

 error:
	if (err) {
		if (iftab)
			osenv_mem_free(iftab, OSENV_PHYS_WIRED,
				       ndev * sizeof(an_infoif_t));
		if (ipv4tab)
			osenv_mem_free(ipv4tab, OSENV_PHYS_WIRED,
				       ndev * sizeof(an_infoipv4_t));
	}
	for (i = 0; i < ndev; i++)
		oskit_etherdev_release(edev[i]);
	free(edev);
}

an_infoif_t *
ani_os_getifinfo(int *count)
{
	*count = numifs;
	return iftab;
}

an_infoipv4_t *
ani_os_getipv4info(int *count)
{
	*count = numifs;
	return ipv4tab;
}

void
ani_os_deinit_net(void)
{
	if (iftab)
		osenv_mem_free(iftab, OSENV_PHYS_WIRED,
			       numifs * sizeof(an_infoif_t));
	if (ipv4tab)
		osenv_mem_free(ipv4tab, OSENV_PHYS_WIRED,
			       numifs * sizeof(an_infoipv4_t));

	oskit_skbufio_shutdown = 1;
#ifdef OSKIT
	oskit_skbufio_mem_flush();
#endif
#ifdef UTAHTESTBED
	tmcp_shutdown(0, 1); 
#endif
}

/*
 * 1. NodeOS skbufio wrapper implemenation.
 *
 * We provide a wrapper skbufio implementation so that we can interpose on the
 * release method and thus know when a driver is done transmitting a packet
 * (indicating that the pbuf can be reused).
 *
 * ani_pbuf_wrap_skbufio is the one exported method, used to initialize a
 * wrapper skbufio.
 */

static OSKIT_COMDECL 
skbuff_io_query(oskit_skbufio_t *io, const oskit_iid_t *iid, void **out_ihandle) 
{
	ani_pbuf_skbufio_t *skbio = (ani_pbuf_skbufio_t *) io;

	if (memcmp(iid, &oskit_iunknown_iid, sizeof(*iid)) == 0 ||
	    memcmp(iid, &oskit_skbufio_iid, sizeof(*iid)) == 0 ||
	    memcmp(iid, &oskit_bufio_iid, sizeof(*iid)) == 0) {
		*out_ihandle = &skbio->ourif;
		atomic_inc(&skbio->refs);
		return 0;
	}
	*out_ihandle = 0;
	return OSKIT_E_NOINTERFACE;
}

static OSKIT_COMDECL_U 
skbuff_io_addref(oskit_skbufio_t *io)
{
	ani_pbuf_skbufio_t *skbio = (ani_pbuf_skbufio_t *) io;

	/*
	 * XXX hope no one cares about the actual value since atomic_inc
	 * doesn't return the new value so we have to read it non-atomically.
	 */
	atomic_inc(&skbio->refs);

	return atomic_read(&skbio->refs);
}

static OSKIT_COMDECL_U 
skbuff_io_release(oskit_skbufio_t *io) 
{
	ani_pbuf_skbufio_t *skbio = (ani_pbuf_skbufio_t *) io;
	unsigned iszero;

	/*
	 * Decrement and read reference count atomically.
	 * Needed since this routine will be called by the network transmit
	 * done interrupt handler.
	 */
	iszero = atomic_dec_and_test(&skbio->refs);
	if (iszero) {
		skbio->func(skbio->farg);
		oskit_skbufio_release(skbio->realif);
		return 0;
	}

	/* XXX again, hope the caller doesn't care about the actual value */
	return 1;
}

static OSKIT_COMDECL
skbuff_io_getsize(oskit_skbufio_t *io, oskit_off_t *out_size)
{
	ani_pbuf_skbufio_t *skbio = (ani_pbuf_skbufio_t *) io;
	return oskit_skbufio_getsize(skbio->realif, out_size);
}

static OSKIT_COMDECL
skbuff_io_setsize(oskit_skbufio_t *io, oskit_off_t new_size)
{
	ani_pbuf_skbufio_t *skbio = (ani_pbuf_skbufio_t *) io;
	return oskit_skbufio_setsize(skbio->realif, new_size);
}

static OSKIT_COMDECL 
skbuff_io_read(oskit_skbufio_t *io, void *dest, oskit_off_t offset,
	       oskit_size_t count, oskit_size_t *out_size) 
{
	ani_pbuf_skbufio_t *skbio = (ani_pbuf_skbufio_t *) io;
	return oskit_skbufio_read(skbio->realif, dest, offset, count, out_size);
}

static OSKIT_COMDECL 
skbuff_io_write(oskit_skbufio_t *io, const void *src, oskit_off_t offset,
		oskit_size_t count, oskit_size_t *out_size)
{
	ani_pbuf_skbufio_t *skbio = (ani_pbuf_skbufio_t *) io;
	return oskit_skbufio_write(skbio->realif, src, offset, count, out_size);
}

static OSKIT_COMDECL 
skbuff_io_map(oskit_skbufio_t *io, void **out_addr, oskit_off_t offset,
	      oskit_size_t count)
{
	ani_pbuf_skbufio_t *skbio = (ani_pbuf_skbufio_t *) io;
	return oskit_skbufio_map(skbio->realif, out_addr, offset, count);
}

static OSKIT_COMDECL 
skbuff_io_unmap(oskit_skbufio_t *io, void *addr, oskit_off_t offset,
		oskit_size_t count) 
{
	ani_pbuf_skbufio_t *skbio = (ani_pbuf_skbufio_t *) io;
	return oskit_skbufio_unmap(skbio->realif, addr, offset, count);
}

static OSKIT_COMDECL 
skbuff_io_wire(oskit_skbufio_t *io, oskit_addr_t *out_phys_addr,
	       oskit_off_t offset, oskit_size_t count)
{
	ani_pbuf_skbufio_t *skbio = (ani_pbuf_skbufio_t *) io;
	return oskit_skbufio_wire(skbio->realif, out_phys_addr, offset, count);
}

static OSKIT_COMDECL 
skbuff_io_unwire(oskit_skbufio_t *io, oskit_addr_t phys_addr,
		 oskit_off_t offset, oskit_size_t count)
{
	ani_pbuf_skbufio_t *skbio = (ani_pbuf_skbufio_t *) io;
	return oskit_skbufio_unwire(skbio->realif, phys_addr, offset, count);
}

static OSKIT_COMDECL
skbuff_io_copy(oskit_skbufio_t *io, oskit_off_t offset, oskit_size_t count,
	       oskit_skbufio_t **out_io)
{
	return OSKIT_E_NOTIMPL;
}

/*
 * metadata_map:
 */
static OSKIT_COMDECL
skbuff_io_metadata_map(oskit_skbufio_t *io, oskit_u32_t version,
		       struct sk_buff **out_addr)
{
	ani_pbuf_skbufio_t *skbio = (ani_pbuf_skbufio_t *) io;
	return oskit_skbufio_metadata_map(skbio->realif, version, out_addr);
}

/*
 * metadata_unmap:
 */
static OSKIT_COMDECL
skbuff_io_metadata_unmap(oskit_skbufio_t *io, struct sk_buff *addr)
{
	ani_pbuf_skbufio_t *skbio = (ani_pbuf_skbufio_t *) io;
	return oskit_skbufio_metadata_unmap(skbio->realif, addr);
}

static OSKIT_COMDECL
skbuff_io_clone(oskit_skbufio_t *io, oskit_skbufio_t **out_io)
{
	return OSKIT_E_NOTIMPL;
}

/*
 * Cloned:
 */
static OSKIT_COMDECL
skbuff_io_cloned(oskit_skbufio_t *io, int *cloned)
{
	return OSKIT_E_NOTIMPL;
}

static struct oskit_skbufio_ops nodeos_skbuff_io_ops = {
	skbuff_io_query, 
	skbuff_io_addref, 
	skbuff_io_release,
	(void *) 0,
	skbuff_io_read, 
	skbuff_io_write, 
	skbuff_io_getsize,
	skbuff_io_setsize,
	skbuff_io_map, 
	skbuff_io_unmap, 
	skbuff_io_wire, 
	skbuff_io_unwire,
	skbuff_io_copy,
	skbuff_io_metadata_map,
	skbuff_io_metadata_unmap,
	skbuff_io_clone,
	skbuff_io_cloned,
};

/*
 * Wrap an skbufio in one of our own so that we can intercept the release call.
 */
void
ani_os_wrap_skbufio(ani_pbuf_skbufio_t *skbio, oskit_skbufio_t *io,
		    void (*func)(void *), void *arg)
{
	assert(io->ops != &nodeos_skbuff_io_ops);

	skbio->ourif.ops = &nodeos_skbuff_io_ops;
	atomic_set(&skbio->refs, 1);
	skbio->realif = io;
	oskit_skbufio_addref(io);
	skbio->func = func;
	skbio->farg = arg;
}

/*
 * 2. skbuff memory allocation routines.
 *
 * We return fixed-size (NODEOS_CHAN_BUFSIZE) buffers for all requested
 * skbuffs.
 */

#define MAXFREEBUFS	64

static int freecount;
static struct sk_buff *freelist;
static struct {
	unsigned int allocs;
	unsigned int alloc_empty;
	unsigned int frees;
	unsigned int free_full;
	unsigned int max_free;
} skb_stats;

void *
oskit_skbufio_mem_alloc(oskit_size_t size, int flags, oskit_size_t *out_size)
{
	struct sk_buff *skb;
	size_t bufsize;
	int enabled;

	assert(flags == OSENV_NONBLOCKING);

	/*
	 * XXX hack deinit_net should shutdown the network interfaces
	 * but the OSKit network stack code has no interface for that.
	 */
	if (oskit_skbufio_shutdown)
		return 0;

	/*
	 * Allow for a maximum-sized buffer plus room for an skbuff header.
	 * This is also the maximum size we should ever have passed in to us. 
	 */
	assert(size <= NODEOS_CHAN_BUFSIZE+SKB_HDRSIZE);
	bufsize = NODEOS_CHAN_BUFSIZE + SKB_HDRSIZE;

	enabled = ani_os_intr_save_disable();
	skb_stats.allocs++;

	skb = freelist;
	if (skb == NULL) {
		assert(freecount == 0);
		skb = osenv_mem_alloc(bufsize, OSENV_PHYS_WIRED|flags, 0);
		skb_stats.alloc_empty++;
	} else {
		assert(freecount > 0);
		freelist = skb->next;
		skb->next = 0;
		freecount--;
	}
	if (enabled)
		ani_os_intr_enable();

	if (skb)
		*out_size = bufsize;
	return skb;
}

void
oskit_skbufio_mem_free(void *ptr, oskit_size_t size)
{
	struct sk_buff *skb = ptr;
	int enabled;

	assert(size == NODEOS_CHAN_BUFSIZE + SKB_HDRSIZE);

	/*
	 * XXX hack deinit_net should shutdown the network interfaces
	 * but the OSKit network stack code has no interface for that.
	 */
	if (oskit_skbufio_shutdown)
		return;


	enabled = ani_os_intr_save_disable();
	skb_stats.frees++;

	if (freecount < MAXFREEBUFS) {
		skb->next = freelist;
		freelist = skb;
		freecount++;
		if (freecount > skb_stats.max_free)
			skb_stats.max_free = freecount;
	} else {
		osenv_mem_free(ptr, OSENV_PHYS_WIRED, size);
		skb_stats.free_full++;
	}
	if (enabled)
		ani_os_intr_enable();
}

/*
 * There should be no activity at this point so no locking/int-disable.
 */
static void
oskit_skbufio_mem_flush(void)
{
	struct sk_buff *skb, *next;
	oskit_size_t size = NODEOS_CHAN_BUFSIZE + SKB_HDRSIZE;

	for (skb = freelist; skb != NULL; skb = next) {
		next = skb->next;
		osenv_mem_free(skb, OSENV_PHYS_WIRED, size);
	}
	freecount = 0;
	freelist = NULL;
}

#if 1
#include <stdio.h>

void
oskit_linux_skbmem_dump(void)
{
	ani_log_printf(ANI_LOG_INFO,
		       "sk_buff freelist: %d allocations, %d frees\n",
		       skb_stats.allocs, skb_stats.frees);
	ani_log_printf(ANI_LOG_INFO,
		       "  %d empty on alloc, %d full on free, %d max entries\n",
		       skb_stats.alloc_empty, skb_stats.free_full,
		       skb_stats.max_free);
}
#endif
